
header {
  #include <string.h>
  #include "oswald.h"
  #include "core-classes.h"
  #include "threads.h"
  #include "j-spot.h"
  #include "node.h"
  #include "mimic-stack.h"
  #include "register-allocator.h"
  #include "code-generator.h"
  #include "assembler.h"
  #include "liveness.h"
  #include "rt-hooks.h"
  #include "ir.h"
  #include "is.h"

  void cg_patch_jump(j_patch patch);
  void cg_patch_exit(j_patch patch);

  /* 
  ** Emit code that compares two operands.
  */
  static void i386_cmp_rr(j_node node, j_node node1, j_node node2) {
    /*
    ** Implementation guidelines:
    **
    **   mov mem reg  :  4 clock cycles, 3 or 6 bytes depending on displacement
    **   cmp mem reg  :  5 clock cycles, 3 or 6 bytes depending on displacement
    **   cmp reg mem  :  6 clock cycles, 3 or 6 bytes depending on displacement
    **   cmp reg reg  :  2 clock cycles, 2 bytes
    */

    /*
    ** Compare memory operand with register operand.
    */

    if (ra_on_stack(node1) && !ra_in_register(node1)) {
      ra_allocate(node, node2);

      // check whether we can use an 8-bit displacement or not:
      if (node1->stack > -128 && node1->stack < 127) {  
        asm_emit(assembler->body, x86_cmp_m8r(node1->stack, EBP, node2->reg));
      }
      else {
        asm_emit(assembler->body, x86_cmp_m32r(node1->stack, EBP, node2->reg));
      }
    }

    /*
    ** Compare register operand with memory operand.
    */

    else if (ra_on_stack(node2) && !ra_in_register(node2)) {
      ra_allocate(node, node1);

      // check whether we can use an 8-bit displacement or not:
      if (node2->stack > -128 && node2->stack < 127) {  
        asm_emit(assembler->body, x86_cmp_rm8(node1->reg, node2->stack, EBP));
      }
      else {
        asm_emit(assembler->body, x86_cmp_rm32(node1->reg, node2->stack, EBP));
      }
    }

    /*
    ** Compare register operand with register operand.
    */

    else {
      ra_allocate(node, node1);
      ra_allocate(node, node2);
      asm_emit(assembler->body, x86_cmp_rr(node1->reg, node2->reg));
    }
  }

  static void i386_cmp_ir(j_node node, j_node node1, int value) {

    /*
    ** Implementation guidelines:
    **
    **   mov mem reg  :  4 clock cycles, 3 or 6 bytes depending on displacement
    **   cmp imm reg  :  2 clock cycles, 6 bytes
    **   cmp imm mem  :  6 clock cycles, 4 bytes
    */

    int offset = node1->stack;
   
    /*
    ** On x86 we can compare an 8-bit signed immediate value with an operand
    ** in memory addressable using an 8-bit signed displacement.
    */
    
    if (ra_on_stack(node1) && (offset > -128 && offset < 127) && (value > -128 && value < 127)) {
      asm_emit(assembler->body, x86_cmp_i8m8(EBP, offset, value));
    }
    else {
      ra_allocate(node, node1);
      asm_emit(assembler->body, x86_cmp_i32r(node1->reg, value));
    }
  }

  static void emit_stack_map_entries(j_node instruction, j_node operand, va_list list) {

    int id = va_arg(list, int);
    j_node call = (j_node)id;    
    j_node node;

    node = instruction;

    if (operand) {
      node = operand;
    }
    else {
      node = instruction;
    }

    if (node != call && node->visited != id && node_is_collectable(node)) {
      // Load the node into register EAX:
      cg_node2register(node, EAX);

      // Push the parameter onto the stack:
      asm_emit(assembler->body, x86_push_r(EAX));

      // Invalidate EAX:
      jm->registers[EAX] = node;

      // Mark the node as being "saved":
      node->visited = id;
    }
  }

  static void emit_node_to_stack(j_node node, int reg) {
    if (jm->registers[node->reg] == node) {
      // Push the paramater onto the stack:
      asm_emit(assembler->body, x86_push_r(node->reg));
    }
    else if (node->operator == j_constant) {
      // Push the constant on the stack:
      asm_emit(assembler->body, x86_push_i32((j_word)(node->data.constant.value)));
    }
    else {
      // Load the parameter into the register:
      cg_node2register(node, reg);

      // Push the paramater onto the stack:
      asm_emit(assembler->body, x86_push_r(reg));

      // Invalidate the register:
      jm->registers[reg] = node;
    }
  }
}

terminal j_assign, j_store, j_neg, j_iseq, j_isne, j_sub, j_add, j_mul, j_and, j_or, j_xor, j_native_call, j_java_call, j_null_pointer, j_array_bounds, j_jump, j_lt, j_gt, j_le, j_ge, j_eq, j_ne, j_return, j_constant, j_variable;

statement: register;

register : j_assign(register, register)
  code {
    ra_allocate(node, node->node1);
    ra_assign(node, node->node2, node->node1->reg);

    // TODO: if the use counter is 2 and if the target node is the source operand of a phi-node, move the value directly to memory?
  }

register : j_store(register, register)
  code {
    int offset;

    if (node->node2->node2->operator == j_constant) {
      // load the value in a register:
      ra_allocate(node, node->node1);

      // load the base address in a register:
      ra_allocate(node, node->node2->node1);

      // compute the constant offset:
      offset = node->node2->node2->data.constant.value * 4;

      if (offset > -128 && offset < 127) {
        /*
        ** Emit a move instruction that uses a 8-bit displacement.
        */

        asm_emit(assembler->body, x86_mov_rm8(node->node2->node1->reg, offset, node->node1->reg));
      }
      else {
        /*
        ** Emit a move instruction that uses a 32-bit displacement.
        */

        asm_emit(assembler->body, x86_mov_rm32(node->node2->node1->reg, offset, node->node1->reg));
      }
    }
    else {
      jabort("instruction not supported yet");
    }
  }


register : j_neg(register)
  code {
    ra_allocate(node, node->node1);
    ra_assign(node, node, node->node1->reg);

    asm_emit(assembler->body, x86_neg_r(node->reg));
  }
  cost {
    return 2;
  }

register: j_neg(constant) 
  code {
    jabort("perform constant propagation\n");
  }

register : j_iseq(register)
  code {
    ra_allocate(node, node->node1);
    ra_assign(node, node, node->node1->reg);
   
    asm_emit(assembler->body, x86_cmp_i32r(node->reg, 0));
    asm_emit(assembler->body, x86_mov_ri32(node->reg, 0));
    asm_emit(assembler->body, x86_sete(node->reg));

    /*
    ** The "and" instruction below has been removed and replaced by
    ** a "mov $reg, 0" because it triggers partial register stalls 
    ** on Pentium III and above causing significant performance 
    ** degradation.
    ** 
    ** asm_emit(assembler->body, x86_and_i32r(node->reg, 0x000000ff));
    */ 
  }
  cost {
    return 2;
  }

register: j_iseq(constant) 
  code {
    jabort("perform constant propagation\n");
  }

register : j_isne(register)
  code {
    ra_allocate(node, node->node1);
    ra_assign(node, node, node->node1->reg);
   
    asm_emit(assembler->body, x86_cmp_i32r(node->reg, 0));
    asm_emit(assembler->body, x86_mov_ri32(node->reg, 0));
    asm_emit(assembler->body, x86_setne(node->reg));

    /*
    ** The "and" instruction below has been removed and replaced by
    ** a "mov $reg, 0" because it triggers partial register stalls 
    ** on Pentium III and above causing significant performance 
    ** degradation.
    ** 
    ** asm_emit(assembler->body, x86_and_i32r(node->reg, 0x000000ff));
    */ 
  }
  cost {
    return 2;
  }

register: j_isne(constant) 
  code {
    jabort("perform constant propagation\n");
  }

register: j_sub(register, register)
  code {
    ra_allocate(node, node->node1);
    ra_allocate(node, node->node2);
    ra_assign(node, node, node->node1->reg);

    asm_emit(assembler->body, x86_sub_rr(node->reg, node->node2->reg));
  }
  cost {
    return 3;
  }

register: j_sub(register, constant)
  code {
    int value = node->node2->data.constant.value;
    
    ra_allocate(node, node->node1);
    ra_assign(node, node, node->node1->reg);

    if (value == 0) {
      // do nothing
    }
    else if (value == 1) {
      asm_emit(assembler->body, x86_dec_r(node->reg));
    }
    else {
      asm_emit(assembler->body, x86_sub_ri32(node->reg, value));
    }
  }
  cost {
    return 2;
  }

register: j_sub(constant, constant) 
  code {
    jabort("perform constant propagation\n");
  }

register: j_add(register, register)
  code {
    ra_allocate(node, node->node1);
    ra_allocate(node, node->node2);
    ra_assign(node, node, node->node1->reg);

    asm_emit(assembler->body, x86_add_rr(node->reg, node->node2->reg));
  }
  cost {
    return 3;
  }

register: j_add(register, constant)
  code {
    int value = node->node2->data.constant.value;

    ra_allocate(node, node->node1);
    ra_assign(node, node, node->node1->reg);

    if (value == 0) {
      // do nothing
    }
    else if (value == 1) {
      asm_emit(assembler->body, x86_inc_r(node->reg));
    }
    else {
      asm_emit(assembler->body, x86_add_ri32(node->reg, value));
    }
  }
  cost {
    return 2;
  }

register: j_add(constant, register)
  code {
    int value = node->node1->data.constant.value;
    
    ra_allocate(node, node->node2);
    ra_assign(node, node, node->node2->reg);

    if (value == 0) {
      // do nothing
    }
    else if (value == 1) {
      asm_emit(assembler->body, x86_inc_r(node->reg));
    }
    else {
      asm_emit(assembler->body, x86_add_ri32(node->reg, value));
    }
  }
  cost {
    return 2;
  }

register: j_add(constant, constant) 
  code {
    jabort("perform constant propagation\n");
  }


register: j_mul(register, register)
  code {
    ra_allocate(node, node->node1);
    ra_allocate(node, node->node2);
    ra_assign(node, node, node->node1->reg);
    
    asm_emit(assembler->body, x86_mul_rr(node->reg, node->node2->reg));
  }
  cost {
    return 2;
  }

register: j_mul(constant, constant) 
  code {
    jabort("perform constant propagation\n");
  }

register: j_and(register, register)
  code {
    ra_allocate(node, node->node1);
    ra_allocate(node, node->node2);
    ra_assign(node, node, node->node1->reg);
    
    asm_emit(assembler->body, x86_and_rr(node->node2->reg, node->reg));
  }
  cost {
    return 2;
  }

register: j_and(constant, constant) 
  code {
    jabort("perform constant propagation\n");
  }

register: j_or(register, register)
  code {
    ra_allocate(node, node->node1);
    ra_allocate(node, node->node2);
    ra_assign(node, node, node->node1->reg);
    
    asm_emit(assembler->body, x86_or_rr(node->node2->reg, node->reg));
  }
  cost {
    return 2;
  }

register: j_or(constant, constant) 
  code {
    jabort("perform constant propagation\n");
  }


register: j_xor(register, register)
  code {
    ra_allocate(node, node->node1);
    ra_allocate(node, node->node2);
    ra_assign(node, node, node->node1->reg);
    
    asm_emit(assembler->body, x86_xor_rr(node->node2->reg, node->reg));
  }
  cost {
    return 2;
  }

register: j_xor(constant, constant) 
  code {
    jabort("perform constant propagation\n");
  }

register: j_native_call
  code {
    j_node temp;
    int i;

    /*
    ** When we are dealing with a function call, we have to do some
    ** extra housekeeping (1) to deal with those function arguments
    ** that are kept in registers and (2) to take care of the nodes
    ** that live in any of the caller-saved registers. 
    **
    ** Note that we update the method's 'reg'- field after the node
    ** has been initialized.  That is because it's only _after_ the
    ** call that some of the registers might have become available,
    ** with the exception of EAX.
    */
  
    /*
    ** Push EAX onto the stack: we have (a) to store its value (when in
    ** use) and (b) we want to use it to push the callee's arguments on
    ** the stack.
    */

    if (!ra_is_free(EAX)) {
      ra_spill(node, jm->registers[EAX]);
    }

    /*
    ** Save the caller saved registers on stack:
    */

    if (!ra_is_free(ECX)) {
      ra_spill(node, jm->registers[ECX]);
    }

    if (!ra_is_free(EDX)) {
      ra_spill(node, jm->registers[EDX]);
    } 

    /*
    ** Function arguments:
    */
    
    i = node->data.call.argc;
    
    while (i > 0) {
      i = i - 1;

      temp = *(j_node *)(node->data.call.args + i);
      emit_node_to_stack(temp, EAX); 
      ra_consume(temp);
    }

    /*
    ** Call far, absolute indirect address given in register:
    */

    asm_emit(assembler->body, x86_mov_ri32(EAX, (j_word)node->data.call.value));
    asm_emit(assembler->body, x86_call_r(EAX));

    /*
    ** Decrement the stack size to clean up the arguments:
    */

    if (node->data.call.argc) {
      asm_emit(assembler->body, x86_add_ri8(ESP, 4 * node->data.call.argc));
    }

    /*
    ** Setup the register to return control.
    */
  
    if (node->type == j_void) {
      jm->registers[EAX] = NULL;
      jm->registers[ECX] = NULL;
      jm->registers[EDX] = NULL;
      node->reg = 0;
    }
    else {
      jm->registers[EAX] = node;
      jm->registers[ECX] = NULL;
      jm->registers[EDX] = NULL;
      node->reg = EAX;
    }
  }

register: j_java_call
  code {
    j_node temp;
    int i;
    int l;
    int s;

    /*
    ** When we are dealing with a function call, we have to do some
    ** extra housekeeping (1) to deal with those function arguments
    ** that are kept in registers and (2) to take care of the nodes
    ** that live in any of the caller-saved registers. 
    **
    ** Note that we update the method's 'reg'- field after the node
    ** has been initialized.  That is because it's only _after_ the
    ** call that some of the registers might have become available,
    ** with the exception of EAX.
    */
  
    /*
    ** Push EAX onto the stack: we have (a) to store its value (when in
    ** use) and (b) we want to use it to push the callee's arguments on
    ** the stack.
    */

    if (!ra_is_free(EAX)) {
      ra_spill(node, jm->registers[EAX]);
    }

    /*
    ** Push the caller saved registers on stack:
    */

    if (!ra_is_free(ECX)) {
      ra_spill(node, jm->registers[ECX]);
    }

    if (!ra_is_free(EDX)) {
      ra_spill(node, jm->registers[EDX]);
    } 

    /*
    ** Store the current stack pointer in EBX:
    */
    
    if (!ra_is_free(EBX)) {
      ra_spill(node, jm->registers[EBX]);
    } 

    jm->registers[EBX] = NULL;
    jm->usage = jm->usage | (1 << EBX);

    asm_emit(assembler->body, x86_mov_rr(EBX, ESP));
  
    /*
    ** Protect the live references by padding them to the stack frame,
    ** terminated by '42'.
    */

    // push the terminator:
    asm_emit(assembler->body, x86_push_i32((j_word)(42)));

    l = method_size_locals();
    s = method_size_stack();

    // push the references:
    for (i = 0; i < l + s; i++) {
      temp = ms_node_state(node->pc, j_post_execution, i);
      if (temp) {
        node_operands(temp, emit_stack_map_entries, node);
      }
    }

    /*
    ** Push the slots for the return value of the callee(s): 
    */

    // asm_emit(assembler->body, x86_push_i32(0));  // disabled for now; we don't support 64-bit values yet
    // asm_emit(assembler->body, x86_push_i32(0));  // disabled for now; we don't support 64-bit values yet
    asm_emit(assembler->body, x86_push_i32(0));
    asm_emit(assembler->body, x86_push_i32(0));

    /*
    ** Push the arguments:
    */
   
    i = node->data.call.argc;
    
    while (i > 1) {
      i = i - 1;

      temp = *(j_node *)(node->data.call.args + i);

      /*
      ** Push the slot's identifier:
      */
      
      if (temp->type == j_unknown) {
        jabort("we can't protect parameters of type j_unknown\n");
      }
      else if (temp->type == j_address) {
        asm_emit(assembler->body, x86_push_i32(stack_trace));
      }
      else {
        asm_emit(assembler->body, x86_push_i32(0));
      }
 
      /*
      ** Push the slot's value:
      */
  
      emit_node_to_stack(temp, EAX);   
     
      ra_consume(temp);
    }

    /*
    ** Reserve place for the stack frame:
    */
    
    asm_emit(assembler->body, x86_sub_ri8(ESP, sizeof(x_Frame)));
 
    /*
    ** Push a pointer to the frame and the method on stack:
    */
    
    temp = *(j_node *)(node->data.call.args);

    asm_emit(assembler->body, x86_push_r(ESP));
    asm_emit(assembler->body, x86_push_i32((j_word)(temp->data.constant.value)));
        
    /*
    ** Call far, absolute indirect address given in register:
    */

    asm_emit(assembler->body, x86_mov_ri32(EAX, (j_word)node->data.call.value));
    asm_emit(assembler->body, x86_call_r(EAX));

    /*
    ** Clean up the arguments and the stack frame:
    */

    asm_emit(assembler->body, x86_mov_rr(ESP, EBX));

    /*
    ** Setup the register to return control.
    */
    if (node->type == j_void) {
      jm->registers[EAX] = NULL;
      jm->registers[ECX] = NULL;
      jm->registers[EDX] = NULL;
      node->reg = 0;
    }
    else {
      jm->registers[EAX] = node;
      jm->registers[ECX] = NULL;
      jm->registers[EDX] = NULL;
      node->reg = EAX;
    }
  } 

register: j_null_pointer(register)
  code {
    // compare the instance with NULL:
    i386_cmp_ir(node, node->node1, 0);
    asm_emit(assembler->body, x86_ne_32(0));

    // if the instance is non-NULL, jump over the function call:
    asm_add_patch(asm_size_body(), 0, ir_get_branch(node), &cg_patch_jump);

    // push the exception clazz on the stack:
    asm_emit(assembler->body, x86_push_i32((j_word)(clazzNullPointerException)));
   
    // move the address of the calle into EAX: 
    asm_emit(assembler->body, x86_mov_ri32(EAX, (j_word)rt_exception_throw));

    // call the exception handling code: 
    asm_emit(assembler->body, x86_call_r(EAX));
  }

register: j_array_bounds(register, register)
  code {
    /*
    ** Save the caller saved registers on stack:
    */
 
    if (!ra_is_free(EAX)) {
      ra_spill(node, jm->registers[EAX]);
    }

    if (!ra_is_free(ECX)) {
      ra_spill(node, jm->registers[ECX]);
    }

    if (!ra_is_free(EDX)) {
      ra_spill(node, jm->registers[EDX]);
    } 

    /*
    ** Prepare the function arguments:
    */

    emit_node_to_stack(node->node2, EAX);  // push index
    emit_node_to_stack(node->node1, EAX);  // push array

    /*
    ** Call far, absolute indirect address given in register:
    */

    asm_emit(assembler->body, x86_mov_ri32(EAX, (j_word)rt_check_array_bounds));
    asm_emit(assembler->body, x86_call_r(EAX));

    /*
    ** Decrement the stack size to clean up the arguments:
    */

    if (node->data.call.argc) {
      asm_emit(assembler->body, x86_add_ri8(ESP, 8));
    }

    /*
    ** Setup the register to return control.
    */
  
    jm->registers[EAX] = NULL;
    jm->registers[ECX] = NULL;
    jm->registers[EDX] = NULL;
  }

register: j_jump
  code {
    asm_emit(assembler->body, x86_jmp_32(0));
    asm_add_patch(asm_size_body(), 0, ir_get_branch(node), &cg_patch_jump);
  }

register: j_lt(register, register)
  code {
    i386_cmp_rr(node, node->node1, node->node2);
    asm_emit(assembler->body, x86_lt_32(0));
    asm_add_patch(asm_size_body(), 0, ir_get_branch(node), &cg_patch_jump);
  }
  cost {
    return 2;
  }

register: j_lt(register, constant)
  code {
    i386_cmp_ir(node, node->node1, node->node2->data.constant.value);
    asm_emit(assembler->body, x86_lt_32(0));
    asm_add_patch(asm_size_body(), 0, ir_get_branch(node), &cg_patch_jump);
  }
 
register: j_gt(register, register)
  code {
    i386_cmp_rr(node, node->node1, node->node2);
    asm_emit(assembler->body, x86_gt_32(0));
    asm_add_patch(asm_size_body(), 0, ir_get_branch(node), &cg_patch_jump);
  }
  cost {
    return 2;
  }

register: j_gt(register, constant)
  code {
    i386_cmp_ir(node, node->node1, node->node2->data.constant.value);
    asm_emit(assembler->body, x86_gt_32(0));
    asm_add_patch(asm_size_body(), 0, ir_get_branch(node), &cg_patch_jump);
  }

register: j_le(register, register)
  code {
    i386_cmp_rr(node, node->node1, node->node2);
    asm_emit(assembler->body, x86_le_32(0));
    asm_add_patch(asm_size_body(), 0, ir_get_branch(node), &cg_patch_jump);
  }
  cost {
    return 2;
  }

register: j_le(register, constant)
  code {
    i386_cmp_ir(node, node->node1, node->node2->data.constant.value);
    asm_emit(assembler->body, x86_le_32(0));
    asm_add_patch(asm_size_body(), 0, ir_get_branch(node), &cg_patch_jump);
  }

register: j_ge(register, register)
  code {
    i386_cmp_rr(node, node->node1, node->node2);
    asm_emit(assembler->body, x86_ge_32(0));
    asm_add_patch(asm_size_body(), 0, ir_get_branch(node), &cg_patch_jump);
  }
  cost {
    return 2;
  }

register: j_ge(register, constant)
  code {
    i386_cmp_ir(node, node->node1, node->node2->data.constant.value);
    asm_emit(assembler->body, x86_ge_32(0));
    asm_add_patch(asm_size_body(), 0, ir_get_branch(node), &cg_patch_jump);
  }

register: j_eq(register, register)
  code {
    i386_cmp_rr(node, node->node1, node->node2);
    asm_emit(assembler->body, x86_eq_32(0));
    asm_add_patch(asm_size_body(), 0, ir_get_branch(node), &cg_patch_jump);
  }
  cost {
    return 2;
  }

register: j_eq(constant, register)
  code {
    ra_allocate(node, node->node2);

    asm_emit(assembler->body, x86_cmp_i32r(node->node2->reg, node->node1->data.constant.value));
    asm_emit(assembler->body, x86_eq_32(0));
    asm_add_patch(asm_size_body(), 0, ir_get_branch(node), &cg_patch_jump);
  }

register: j_eq(register, constant)
  code {
    i386_cmp_ir(node, node->node1, node->node2->data.constant.value);
    asm_emit(assembler->body, x86_eq_32(0));
    asm_add_patch(asm_size_body(), 0, ir_get_branch(node), &cg_patch_jump);
  }

register: j_ne(register, register)
  code {
    i386_cmp_rr(node, node->node1, node->node2);
    asm_emit(assembler->body, x86_ne_32(0));
    asm_add_patch(asm_size_body(), 0, ir_get_branch(node), &cg_patch_jump);
  }
  cost {
    return 2;
  }

register: j_ne(register, constant)
  code {
    i386_cmp_ir(node, node->node1, node->node2->data.constant.value);
    asm_emit(assembler->body, x86_ne_32(0));
    asm_add_patch(asm_size_body(), 0, ir_get_branch(node), &cg_patch_jump);
  }

register: j_return(register)
  code {
    /*
    ** On x86, %eax is used to hold the result, or %edx:%eax for 64-bit 
    ** results.
    */
   
    cg_node2register(node->node1, EAX);
   
    /*
    ** If the given node is not the last node in the linearized DAG, we 
    ** are compiling a method with multiple returns.  If that is the case,
    ** we make all nodes but the last node jump to a single exit point.
    */
 
    if (node->next) {
      asm_emit(assembler->body, x86_jmp_32(0));
      asm_add_patch(asm_size_body(), 0, NULL, &cg_patch_exit);
    }
  }

register: j_return
  code {
    /*
    ** If the given node is not the last node in the linearized DAG, we 
    ** are compiling a method with multiple returns.  If that is the case,
    ** we make all nodes but the last node jump to a single exit point.
    */
 
    if (node->next) {
      asm_emit(assembler->body, x86_jmp_32(0));
      asm_add_patch(asm_size_body(), 0, NULL, &cg_patch_exit);
    }
  }

constant : j_constant
  code {
    /*
    ** Do nothing - we encountered a literal, not an instruction.  Every
    ** instruction is held responsible for "preparing" its own operands.
    ** This is required because (a) some x86 instructions can keep one or
    ** more operands in memory and (b) to get the addresses of the join 
    ** points (forward and backward references) right.
    */
  }

register : j_constant
  code {
    /*
    ** Do nothing - we encountered a literal, not an instruction.  Every
    ** instruction is held responsible for "preparing" its own operands.
    ** This is required because (a) some x86 instructions can keep one or
    ** more operands in memory and (b) to get the addresses of the join 
    ** points (forward and backward references) right.
    */
  }

register : j_variable
  code {
    /*
    ** Do nothing - we encountered a literal, not an instruction.  Every
    ** instruction is held responsible for "preparing" its own operands.
    ** This is required because (a) some x86 instructions can keep one or
    ** more operands in memory and (b) to get the addresses of the join 
    ** points (forward and backward references) right.
    */
  }

register : j_variable(register, register)
  code {
    /*
    ** Do nothing - we encountered a literal, not an instruction.  Every
    ** instruction is held responsible for "preparing" its own operands.
    ** This is required because (a) some x86 instructions can keep one or
    ** more operands in memory and (b) to get the addresses of the join 
    ** points (forward and backward references) right.
    */
  }


footer {
  void burg_walk(j_node node, int goal, int depth) {

    j_node kids[3];
    short *nts;
    int rule;

    if (node->emitted == 0) {
 
      node->offset = asm_size_body();
      rule = burg_rule(node->state, goal);

      nts = burg_nts[rule];

      // woempa(9, "%snode %s (%p), goal %d, rule '%s' (%d), nts %p, kids[0] = %d, kids[1] = %d\n", burg_spaces(depth), node_get_operator(node), node, goal, burg_rule_name[rule], rule, nts, (kids[0] ? kids[0]->operator : 0), (kids[1] ? kids[1]->operator : 0));

      if (rule && nts) {
        burg_kids(node, rule, kids);

        if (kids[0] && (depth == 0 || !node_is_call(kids[0]))) {
          burg_walk(kids[0], nts[0], depth + 1);
        }

        if (kids[1] && (depth == 0 || !node_is_call(kids[1]))) {
          burg_walk(kids[1], nts[1], depth + 1);
        }

      }
      else {
        woempa(9, "%snode '%s', goal %d, rule '%s' (%d), nts %p\n", burg_spaces(depth), node_get_operator(node), goal, burg_rule_name[rule], rule, nts);
        fprintf(stderr, "failing operator '%s'\n", node_get_operator(node));
        exit(0);
      }

      burg_code(rule, node);
      node->emitted = 1;
    }
  }
}

