#ifndef _J_NODE_H
#define _J_NODE_H

/**************************************************************************
* Copyright (c) 2001, 2002, 2003 by Acunia N.V. All rights reserved.      *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Philips-site 5, box 3       info@acunia.com                           *
*   3001 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: node.h 3985 2003-06-04 12:36:10Z dbuytaert $ 
*/

#include "j-spot.h"

#define j_max_edges 10

typedef void (*node_callback)(j_node instruction, j_node operand, va_list list);

typedef enum {
  j_none          = 0,
  j_constant      = 1,
  j_variable      = 2,
  j_add           = 3,
  j_sub           = 4,
  j_mul           = 5,
  j_div           = 6,
  j_rem           = 7,
  j_neg           = 8,
  j_lsh           = 9,
  j_rsh           = 10,
  j_and           = 11,
  j_or            = 12,
  j_xor           = 13,
  j_le            = 14,
  j_gt            = 15,
  j_eq            = 16,
  j_ne            = 17,
  j_lt            = 18,
  j_ge            = 19,
  j_iseq          = 20,
  j_isne          = 21,
  j_native_call   = 22,
  j_java_call     = 23,
  j_assign        = 24,
  j_return        = 25,
  j_jump          = 26,
  j_null_pointer  = 27,
  j_array_bounds  = 28,
  j_store         = 29
} j_operator;

typedef enum {
  j_parameter     = 1,
  j_local         = 2,
  j_memory        = 3
} j_location;

typedef enum {
  j_unknown       = 0,
  j_void          = 1,
  j_address       = 2,
  j_boolean       = 3,
  j_byte          = 4, 
  j_character     = 5, 
  j_short         = 6, 
  j_double        = 7,
  j_integer       = 8,
  j_float         = 9,
  j_long          = 10
} j_type;

char * operator2string(j_operator operator);

typedef struct j_Node {
  char *name;                     // the name of this node
  
  j_operator operator;            // operator
  j_node node1;                   // operand node #1
  j_node node2;                   // operand node #2
  j_type type;
 
  // Data flow:
  j_node definition;              
  
  // Control flow:
  j_node incoming[j_max_edges];   // a list of outgoing nodes
  j_node outgoing[j_max_edges];   // a list of incoming nodes
  j_node next;                    // the next sequential node
  j_node list;                    // the next worklist node

  int visited;
  int lattice;

  int use;                        // use counter
  int pc;                         // program counter
 
  // Code generation:
  j_state state;

  // Register allocation: 
  int stack;                      // offset relative to base pointer
  int spill;
  int reg;
  int dfs;

  // Assembler, debugger:
  int offset;                     // the offset in the generated machine code
  int emitted;
   
  // Optimizations:
  int array_size;                 // the maximum array size
  int null_check;
  int critical;

  // Detailed information:
  union {
    struct {
      int value;
    } constant;
    struct {
      int offset;
      j_location location;
    } variable;
    struct {
      void * value;
      int size;
      int argc;
      j_node * args;
    } call;
  } data;

  j_node src[j_max_edges];
  j_node dst[j_max_edges];
} j_Node;

j_node node_create_constant(j_type type, int value);
  /* Create a new constant with the specified value. */

j_node node_create_parameter(j_type type, int value);
  /* Create a new variable with the specified location. */

j_node node_create_local_variable(j_type type, int value);
  /* Create a new variable with the specified location. */

j_node node_create_memory_variable(j_type type, j_node base, j_node offset);
  /* Create a new variable with the specified location. */

j_node node_create_insn(j_operator operator, j_node n1, j_node n2);
  /* Create a new instruction. */

j_node node_create_call(char * name, j_operator, j_type type, void * function, int argc);
  /* Create a new function call. */

void node_add_param(j_node function, j_node parameter);
  /* Add a parameter to a function-node. */

void node_add_link(j_node n1, j_node n2, int label);
  /* Add a link-relation between node #1 and node #2. */

void node_add_clone(j_node src, j_node dst);
  /* Add a clone-relation between node #1 and node #2. */

int node_is_memory_variable(j_node node);
  /* Returns true when the specified node is a memory variable. */

int node_is_local_variable(j_node node);
  /* Returns true when the specified node is a local variable. */

int node_is_parameter(j_node node);
  /* Returns true when the specified node is a function parameter. */

int node_is_instruction(j_node node);
  /* Returns true when the specified node is an instruction node. */

int node_is_call(j_node node);
  /* Returns true when the specified node is a call node. */

int node_is_escapable(j_node node);
  /* Returns true when the specified node or a parameter thereof can escape. */

int node_is_collectable(j_node node);
  /* Returns true when the node might not be referenced. */
 
int node_check_null_pointer(j_node node);
  /* Returns true when the node needs a null-pointer check. */
 
int node_check_array_bounds(j_node array, j_node idx);
  /* Returns true when the node needs an array bound check. */
   
int node_are_compatible(j_type type1, j_type type2);
  /* Returns whether the nodes have compatible types. */

char * node_get_operator(j_node node);
  /* Give the node's operator. */

void node_inherit(j_node parent, j_node child);
  /* Inherit data from parent node. */

void node_operands(j_node node, node_callback function, ...);
  /* Call the specified function for each operand. */
   
char * node_print(char * buffer, int * remain, void * data, int width, int precision, unsigned int flags);
  /* Write a node description in the buffer. */

#endif /* _J_NODE_H */
