
#include <stdarg.h>
#include "oswald.h"

#include "j-spot.h"
#include "arena.h"
#include "ir.h"
#include "assembler.h"
#include "code-generator.h"

j_assembler assembler = NULL;

void asm_init(void) {

  if (assembler == NULL) {
    /*
    ** The first time this function is called, we initialize all the 
    ** components used by the assembler, resp. a prologue, body, and
    ** epilogue buffer.  For each of them, we allocate and initialize
    ** a chunk of memory that we'll use to emit the machine code to.
    */
    
    assembler = (j_assembler)x_mem_alloc(sizeof(j_Assembler));

    assembler->prologue = (j_buffer)x_mem_alloc(sizeof(j_Buffer));
    assembler->prologue->size = 1024;
    assembler->prologue->base = (void *)x_mem_alloc(assembler->prologue->size);
    
    assembler->epilogue = (j_buffer)x_mem_alloc(sizeof(j_Buffer));
    assembler->epilogue->size = 1024;
    assembler->epilogue->base = (void *)x_mem_alloc(assembler->epilogue->size);
    
    assembler->constant = (j_buffer)x_mem_alloc(sizeof(j_Buffer));
    assembler->constant->size = 5 * 1024;
    assembler->constant->base = (void *)x_mem_alloc(assembler->constant->size);
        
    assembler->body = (j_buffer)x_mem_alloc(sizeof(j_Buffer));
    assembler->body->size = 50 * 1024;
    assembler->body->base = (void *)x_mem_alloc(assembler->body->size);

    // TODO: consider using one big memory chunk, allocated at once and to memmov the different parts together as it rumoured to be better cache-wise.
  }
    
  assembler->body->patches = NULL;
 
  /*
  ** Reset or initialize the buffers.
  */
 
  assembler->prologue->step = assembler->prologue->base;
  assembler->epilogue->step = assembler->epilogue->base;
  assembler->constant->step = assembler->constant->base;
  assembler->body->step = assembler->body->base;
}

int asm_emit(j_buffer buffer, int endian, int count, ...) {

  va_list ap;

  int shift = 0;
  int bytes = 0;
  int data = 0;
  int insn = 0;
  int bits = 0;
  int loop = 0;
  int size = 0;

  /*
  ** This function hosts a generic code emitter that, given an instruction
  ** description, emits binary machine code in the specified buffer.  Can
  ** be used for both big and little endian targets.
  */
  
  size = count / 8; 
  loop = count;
  
  if (endian) {
    buffer->step = buffer->step + size - 1;
  }
  
  va_start(ap, count);

  while (loop > 0) {
    bits = va_arg(ap, int);
    data = va_arg(ap, int);

    if (bits < 0 || bits > 32) {
      jabort("invalid instruction: bits = %d\n", bits);
    }
   
    if (buffer->step - 32 > buffer->base + buffer->size) {
      jabort("code buffer is full: size = %d\n", buffer->size);
    }
    
    shift = shift + bits;
    insn |= data << (32 - shift);

    if (endian) {
      // NOTE: this function has to be as fast as possible, so we put the check on the endianess outside the loop.
      while (shift > 7) {
        *buffer->step-- = (char)(insn >> 24);
        bytes = bytes + 1;
        shift = shift - 8;
        insn = insn << 8;
      }
    }
    else {
      while (shift > 7) {
        *buffer->step++ = (char)(insn >> 24);
        bytes = bytes + 1;
        shift = shift - 8;
        insn = insn << 8;
      }
    }
     
    loop = loop - bits;
  }
  
  va_end(ap);
 
  if (shift != 0 || bytes != size) {
    jabort("invalid instruction: shift = %d, size = %d, bytes = %d\n", shift, size, bytes);
  }

  if (endian) {
    buffer->step = buffer->step + size + 1;
  }
 
  return size;
}

void asm_add_patch(int offset, int value, j_node dst, void (*patcher)(j_patch patch)) {

  j_patch patch = arena_allocate(sizeof(j_Patch));

  /*
  ** Add a patch, typically an unresolved forward reference, to the
  ** internal list of patches.  These patches will be "applied" just
  ** before the code is exported through "asm_export()".
  */

  patch->patcher = patcher;
  patch->offset = offset;
  patch->value = value;
  patch->dst = dst;

  if (assembler->body->patches) {
    patch->next = assembler->body->patches;
    assembler->body->patches = patch;
  }
  else {
    assembler->body->patches = patch;
  }

  // NOTE: we don't care about duplicate entries.
}

void asm_patch(void) {

  /*
  ** Patch the unresolved references in the body.
  */
  
  j_patch patch = assembler->body->patches;
  
  while (jm->failed == 0 && patch) {
    patch->patcher(patch);
    patch = patch->next;
  }
}

char * asm_export(void) {
 
  int offset;
  char *code;

  asm_patch();
  
  code = x_mem_alloc((x_size)asm_size_total());
   // NOTE: we assume that the pointer pointing to this memory chunk is word aligned.
  
  offset = 0;
  
  if (assembler->prologue->step - assembler->prologue->base) {
    w_memcpy(code + offset, assembler->prologue->base, (size_t)(assembler->prologue->step - assembler->prologue->base));
    offset = offset + assembler->prologue->step - assembler->prologue->base;
  }
  
  if (assembler->body->step - assembler->body->base) {
    w_memcpy(code + offset, assembler->body->base, (size_t)(assembler->body->step - assembler->body->base));
    offset = offset + assembler->body->step - assembler->body->base;
  }
  
  if (assembler->epilogue->step - assembler->epilogue->base) {
    w_memcpy(code + offset, assembler->epilogue->base, (size_t)(assembler->epilogue->step - assembler->epilogue->base));
    offset = offset + assembler->epilogue->step - assembler->epilogue->base;
  }
  
  if (assembler->constant->step - assembler->constant->base) {
    w_memcpy(code + offset, assembler->constant->base, (size_t)(assembler->constant->step - assembler->constant->base));
  }

  cg_clear_cache(code, code + asm_size_total());
  
  return code; 
}

void asm_dump() {
  char * i;
  char * j;

  // TODO: the behavior of this function is temporary "broken" as only a function's body is being printed
   
  j_buffer buffer = assembler->body;

  fprintf(stderr, "----------------\n");
  fprintf(stderr, "raw compiled method:\n");
  
  for (i = buffer->base; i < buffer->step; i = i + 4) {
    fprintf(stderr, "   %10p  |  %s  (%8x hex)\n", i, asm_insn(*(j_word *)i), *(j_word *)i);
  }

  fprintf(stderr, "disassembled method:\n");

  for (j = buffer->base; j < buffer->step; j = j) {
    fprintf(stderr, "   %10p  |  ", j);
    j = (char *)cg_disassemble((j_word *)j);
    fprintf(stderr, "\n");
  }
  
  fprintf(stderr, "----------------\n");
}

char * asm_insn(j_word inst) {

  /*
  ** Return a string with the bit-wise representation of the specified 
  ** word/instruction.
  */
 
  unsigned int size = 31; 
  char * buffer = (char *)arena_allocate(size);
  unsigned int mask = (unsigned)(1 << (size));
  int i = 0;

  while (mask) {
    if (i % 9 == 8) {
      *(buffer + i++) = ' ';
    }
    if (inst & mask) {
      *(buffer + i++) = '1';
    }
    else {
      *(buffer + i++)  = '0';
    }
    mask >>= 1;
  }
  
  *(buffer + i++) = '\0';
                                                      
  return (char *)buffer;
}

int asm_size_prologue(void) {
  return (int)(assembler->prologue->step - assembler->prologue->base);
}

int asm_size_epilogue(void) {
  return (int)(assembler->epilogue->step - assembler->epilogue->base);
}

int asm_size_constant(void) {
  return (int)(assembler->constant->step - assembler->constant->base);
}

int asm_size_body(void) {
  return (int)(assembler->body->step - assembler->body->base);
}

int asm_size_total(void) {
  return asm_size_prologue() + asm_size_body() + asm_size_epilogue() + asm_size_constant();
}

