/**************************************************************************
* Copyright (c) 2001, 2002, 2003 by Acunia N.V. All rights reserved.      *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Philips-site 5, box 3       info@acunia.com                           *
*   3001 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: dominance.c 3863 2003-04-25 09:23:25Z dbuytaert $
*/

#include "j-spot.h"
#include "method.h"
#include "arena.h"
#include "node.h"

/*
** Some terminology and definitions:
**
**  + Dom: a node "x" dominates a node "y" if and only if every path 
**  from the start node "s" to the node "y", passes through "x".  By
**  definition, every node dominates itself.
**
**  + Sdom: a dominator "x" of "y" is said to be "strict" if x != y.  
**
**  + Idom: an "immidiate" dominator "i" of a node "y" is a strict 
**  dominator when all strict dominators of "y" also dominate "i".  
**  Every node but the start node has a unique immediate dominator.
**
** To compute dominators we use the algorithm developed by T. Lengauer
** and R. E. Tarjan as described in their paper: "A fast algorithm for
** finding dominators in a flowgraph".  The book "Modern compiler 
** implementation in C" by A. W. Appel explains the algorithm and its
** variants rather well.
*/

/*
** TODO:
**   - We implemented and tested the algorithm using nodes, but we should
**     use basic blocks really ...
*/

/*
** Create a Depth First Spanning (DFS) tree of the current method.
*/

void dfs_recursive(j_node node, int * number, int parent) {

  int i;

  /*
  ** We traverse the flow-graph in depth first order to give each node
  ** a depth-first index (we number the nodes in pre-oder) and to link
  ** them to their most recently visited predecessor.
  */

  node->dfs = *number;
  jm->dfs[*number] = node;
  jm->parent[*number] = parent;
  *number += 1; 

  for (i = 0; i < j_max_edges; i++) {
    if (node->outgoing[i] && node->outgoing[i]->dfs == 0) {
      // NOTE: we call this only if the node is not visited already.
      dfs_recursive(node->outgoing[i], number, node->dfs);
    }
  }
}

void dfs(void) {

  int number;
  
  /*
  ** Allocate memory for the DFS tree related datastructures:
  */

  jm->dfs = arena_allocate(sizeof(j_node) * (jm->instructions + 1));
  jm->parent = arena_allocate(sizeof(int) * (jm->instructions + 1));
    // NOTE: + 1 because the initial DFS index is 1.
  
  /*
  ** The DFS numbers start with 1; 0 means "undefined".
  */
  
  number = 1;

  /*
  ** Currently we compute the DFS tree using recursion, but it would
  ** probably more efficient to use a non-recursive variant.  Eg. we
  ** could keep partially traversed paths a stack.
  */

  dfs_recursive(jm->start, &number, 0);

  /*
  ** The assert() below validates the correctness of the input DAGs
  ** and the DFS algorithm.
  */ 
  
  assert(number - jm->instructions == 1);
}

int dom_lowest_ancestor(int k) {
  
  int u = k;
 
  /*
  ** We use the "ancestor" array to search the DFS tree upwards starting
  ** from "k" looking for the node with the smallest semi-dominator.
  */
  
  while (jm->ancestor[k]) {
    if (jm->sdom[k] < jm->sdom[u]) {
      u = k;
    }

    if (jm->ancestor[k]) {
      k = jm->ancestor[k];
    }
  }

  return u;
}

int dom_link(j_node parent, j_node node) {

  return 1;
}

void idom(void) {

  j_node node;
  int i;
  int m;  // m dominates n
  int n;
  int p;  // parent
  int k;
  int w;

  /*
  ** Allocate memory for the dominator related datastructures:
  */

  jm->ancestor = arena_allocate(sizeof(int) * (jm->instructions + 1));
  jm->fbucket = arena_allocate(sizeof(int) * (jm->instructions + 1));
  jm->nbucket = arena_allocate(sizeof(int) * (jm->instructions + 1));
  jm->sdom = arena_allocate(sizeof(int) * (jm->instructions + 1));
  jm->idom = arena_allocate(sizeof(int) * (jm->instructions + 1));
    // NOTE: + 1 because the initial DFS index is 1.
 
  m = jm->instructions;
  p = 0;

  // We iterate in reverse depth-first order. 
  while (m > 1) {
   
    node = jm->dfs[m];
    p = jm->parent[m];

    assert(node);
    
    woempa(5, "%q has dfs value %d and parent %d\n", node, m, p);
      
    n = m;

    /*
    ** To find the semi-dominator of a node, we search for the smallest
    ** numbered node reachable from the given node by following the CFG
    ** branches backward, stopping at nodes with a smaller depth-first
    ** index.
    */
      
    for (i = 0; i < j_max_edges; i++) {
      if (node->incoming[i]) {
        k = node->incoming[i]->dfs;

        if (k > m) {
          k = jm->sdom[dom_lowest_ancestor(k)]; 
        }
          
        if (k < n) {
          n = k;
        }
      }
    }

    jm->sdom[m] = n;  // store "m"'s the semi-dominator "n"
    
    woempa(5, "semidominator of %d [%q] is %d\n", m, jm->dfs[m], n);
    
    /*
    ** We can now use these semi-dominators to compute the immidiate 
    ** dominators during the same pass.  If they can't be calculated
    ** yet, they are deferred until the required dominators are known.
    **
    ** The implementation of the deferred dominator calculations is 
    ** inspired by GCC's implementation.
    */
    
    jm->ancestor[m] = p;
    jm->nbucket[m] = jm->fbucket[n];
    jm->fbucket[n] = m;

    w = jm->fbucket[p];
   
    while (w) {
      k = dom_lowest_ancestor(w);
      
      if (jm->sdom[k] < jm->sdom[w]) {
        jm->idom[w] = k;
      }
      else {
        jm->idom[w] = p;
      }
      
      w = jm->nbucket[w];
    }   
     
    jm->fbucket[p] = 0;
    
    m = m - 1;
  }

  jm->idom[1] = 0;
  
  for (m = 2; m < jm->instructions; m++) {
    if (jm->idom[m] != jm->sdom[m]) {
      jm->idom[m] = jm->idom[jm->idom[m]];
    }
  }
  
}

void dominance(void) {

  dfs();

  idom();
}

