/**************************************************************************
* Copyright (c) 2001, 2002, 2003 by Acunia N.V. All rights reserved.      *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Philips-site 5, box 3       info@acunia.com                           *
*   3001 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: mimic-stack.c 3982 2003-06-03 14:17:13Z dbuytaert $ 
*/

#include <string.h>

#include "arena.h"
#include "bytecode.h"
#include "mimic-stack.h"

#undef swap  // because of the define in "bytecode.h"

static j_node ms_get(int offset, j_type type) {

  j_node node;

  if (offset < 0) {
    jabort("mimic stack: stack underflow: stack pointer is %d!\n", offset);
  }

  if (offset > jm->slots) {
    jabort("mimic stack: stack overflow: stack pointer is %d!\n", offset);
  }

  if (jm->stack[offset] == NULL) {
    jm->stack[offset] = node_create_local_variable(type, offset + method_size_locals());
  }

  // get the node from the stack:  
  node = jm->stack[offset];

  #ifdef DEBUG 
    /*
    ** We use the "type" parameter to verify the correctness of the
    ** propagated type information.  These checks are invaluable as
    ** they helped isolating many bugs.
    */

    if (node_are_compatible(node->type, type) == 0) {
      jabort("mimic stack: type mismatch: node is of type %d but expected type is %d!\n", node->type, type);
    }
  #endif
  
  return node;
}

void ms_init(void) {

  jm->sp = 0;

  /*
  ** TODO: the translation process uses a worklist-style algorithm and
  **       uses quite a bit of *temporary* allocated memory.  Look into
  **       means to reduce the number of allocations and the ammount of
  **       allocated memory.
  */
  
  // Allocate memory to store the parameters in:
  jm->parameters = arena_allocate((method_size_parameters()) * sizeof(j_node));

  // Allocate memory to store the local variables in:
  jm->locals = arena_allocate((method_size_locals() + 1) * sizeof(j_node));
    // NOTE: "+ 1" because with the exception of "aload_0" addressing of local variables starts at index 0.
  
  // Allocate memory to store the stack variables in:
  jm->stack = arena_allocate((method_size_stack() + 1) * sizeof(j_node));
    // NOTE: "+ 1" because we sometimes use an extra stack variable when translating bytecode instructions.
  
  // Allocate memory to store the stack variables in:
  jm->states = arena_allocate(jm->length * 2 * sizeof(void *));
}

void ms_save_state(int pc, j_timestamp timestamp) {
  
  j_node * state;
  int i;
  int l;
  int s;

  /*
  ** Save the stack variables.
  */

  if (jm->states[pc + timestamp * jm->length] == NULL) {
    l = method_size_locals();
    s = method_size_stack();

    // Allocate memory to store the state backups:
    state = arena_allocate(jm->slots * sizeof(j_node));

    for (i = 0; i < l + s; i++) {
      if (i < l) {
        state[i] = jm->locals[i];
      }
      else {
        state[i] = jm->stack[i - l];
      }
    }

    jm->states[pc + timestamp * jm->length] = state;
    
    woempa(9, "saved state information at pc %d\n", pc);
  }
}

void ms_load_state(int pc, j_timestamp timestamp) {

  int i;
  int l;
  int s;

  if (jm->states[pc + timestamp * jm->length]) { 

    l = method_size_locals();
    s = method_size_stack();

    /*
    ** Load the stack variables.
    */
 
    for (i = 0; i < l + s; i++) {
      if (i < l) {
        jm->locals[i] = jm->states[pc + timestamp * jm->length][i];
      }
      else {
        jm->stack[i - l] = jm->states[pc + timestamp * jm->length][i];
      }
    }
  }
  else {
    jabort("%q has no state information\n");
  }
}

j_node ms_node_state(int pc, j_timestamp timestamp, int slot) {

  if (jm->states[pc + timestamp * jm->length]) {
    return jm->states[pc + timestamp * jm->length][slot];
  }
  else {
    return NULL;
  }
}

void ms_clone(int inherit) {

  j_node node = NULL;
  int i;
  int l;
  int s;
  int p;
  
  l = method_size_locals();
  s = method_size_stack();
  p = method_size_parameters();

  for (i = 0; i < l + s; i++) {
    if (i < p) {
      // do nothing
    }
    else if (i < l) {
      if (jm->locals[i]) {
        if (node_is_memory_variable(jm->locals[i])) {
          jabort("something is wrong!\n");
        }
        else {
          node = node_create_local_variable(jm->locals[i]->type, i);
        }

        if (inherit) {
          node_inherit(jm->locals[i], node);
        }

        jm->locals[i] = node; 
      }
    }
    else {
      if (jm->stack[i - l]) {
        if (node_is_memory_variable(jm->stack[i - l])) {
          jabort("something is wrong!\n");
        }
        else {
          node = node_create_local_variable(jm->stack[i - l]->type, i);
        }

        if (inherit) {
          node_inherit(jm->stack[i - l], node);
        }

        jm->stack[i - l] = node; 
      }
    }
  }
}

void ms_push(j_node node) {

  j_node temp;
  
  /*
  ** Every push instruction gets translated into an assignment (move)
  ** instruction that moves the node being pushed to a temporary node
  ** on the stack.  
  **
  ** The downside is that a lot of move instructions are inserted, 
  ** however they can easily be optimized away later on.  The major
  ** benefit of this mechanism is that no control flow analysis, and
  ** the detection of join points in specific, needs to be performed
  ** and that we don't have to bother merging the operand stacks at 
  ** each control join point.  Instead, we use a one-pass translator
  ** that has to compute less.  As a result, the translation time is
  ** heavily reduced. 
  ** 
  ** Note that originally, we had a two-pass translator: one pass to
  ** pre-resolve basic information about control join points and one 
  ** pass to translate the bytecodes to the IR, using the join point
  ** information to determine when the elements on the operand stack
  ** had to be spilled to temporaries.  Unfortunately, this approach
  ** didn't really worked correctly (it did in 99% of the cases) and
  ** getting it to work correctly turned out to be both complex and 
  ** error-prone.  (Reminder: one problem was that of merging multiple
  ** join points of different bytecode instructions that map onto a 
  ** single IR instruction.)
  */

  if (bytecode_is_join(jm->inst_pc)) {
      // NOTE: join points require an explicit copy instruction
    woempa(9, "insert copy instruction at pc %d\n", jm->inst_pc);
    
    // stack variables can only be defined once:
    jm->stack[jm->sp] = NULL;

    // create a new node:
    temp = node_create_insn(j_assign, node, ms_get(jm->sp, node->type));

    // mark the node as a top-level node:
    ir_add_node(temp, j_link_next, j_link_none);
  }
  else if (bytecode_tos_consumed(jm->inst_pc) > 0) {
    woempa(9, "copy coalesce instruction at pc %d\n", jm->inst_pc);
    
    // update the mimic stack:
    jm->stack[jm->sp] = node;
  }
  else {
    woempa(9, "insert copy instruction at pc %d\n", jm->inst_pc);
    
    // stack variables can only be defined once:
    jm->stack[jm->sp] = NULL;

    // create a new node:
    temp = node_create_insn(j_assign, node, ms_get(jm->sp, node->type));

    // mark the node as a top-level node:
    ir_add_node(temp, j_link_next, j_link_none);
  }

  jm->sp = jm->sp + 1;
}

j_node ms_pop(j_type type) {

  j_node node;
 
  jm->sp = jm->sp - 1;
  
  node = ms_get(jm->sp, type);
  
  // stack variables can only be popped once:
  jm->stack[jm->sp] = NULL;

  return node;
}

j_node ms_top(j_type type) {

  j_node node;
  
  node = ms_get(jm->sp - 1, type);
  
  return node;
}

j_node ms_next(j_type type) {

  jfail("depricated method - we shouldn't use this for it violates the stack-based properties of the translation process\n");
  return ms_get(jm->sp, type);
}

void ms_dup1(void) {

  ms_push(ms_get(jm->sp - 1, j_unknown));
}

void ms_dup1_x1(void) {
  jabort("unsupported bytecode\n");
}

void ms_dup1_x2(void) {
  jabort("unsupported bytecode\n");
}

void ms_dup2(void) {

  ms_push(ms_get(jm->sp - 2, j_unknown));  // ms_push() will increment the stack pointer
  ms_push(ms_get(jm->sp - 2, j_unknown));
}

void ms_dup2_x1(void) {
  jabort("unsupported bytecode\n");
}

void ms_dup2_x2(void) {
  jabort("unsupported bytecode\n");
}

void ms_swap(void) {

  j_node temp;  
  j_node swap  = ms_get(jm->sp, j_unknown);
  j_node node1 = ms_get(jm->sp - 1, j_unknown);
  j_node node2 = ms_get(jm->sp - 2, j_unknown);
  
  temp = node_create_insn(j_assign, node1, swap);
  ir_add_node(temp, j_link_next, j_link_none);
  
  temp = node_create_insn(j_assign, node2, node1);
  ir_add_node(temp, j_link_next, j_link_none);
  
  temp = node_create_insn(j_assign, swap, node2);
  ir_add_node(temp, j_link_next, j_link_none);

  /*
  ** Clear the temporary slot that we have used to swap the other two
  ** slots.  Experience learned that leaving it populated might cause
  ** funny things to happen.
  */
  
  jm->stack[jm->sp] = NULL;
}

void ms_dump_stack(void) {
 
  j_node node;
  int i;

  for (i = jm->sp - 1; i >= 0; i--) {
    node = jm->stack[i];
    if (node) {
      woempa(9, "%2d  |  type: %2d  |  node: %q (%p)\n", i, node->type, node, node);
    }
  }
}

void ms_dump_state(int pc, j_timestamp timestamp) {

  j_node * state;
  j_node node; 
  int i;

  state = jm->states[pc + timestamp * jm->length];

  if (state) {
   for (i = 0; i < jm->slots; i++) {
     node = state[i];
     if (node) {
       woempa(9, "%2d  |  type: %2d  |  node: %q (%p)\n", i, node->type, node, node);
     }
   }
  }
}
   



int ms_is_stacked(j_node node) {
 
  int i;
  int s;

  /*
  ** TODO: came to realize this function might be incorrect as it does
  **       not inspect sub-expressions of instructions pushed on stack.
  **       Then again, this might not be a problem, as normally the 
  **       value would be "consumed" and thus lost.
  */
    
  s = method_size_stack();

  for (i = 0; i < s; i++) {
    if (jm->stack[i] == node) {
      return 1;
    }
  }

  return 0;
}

j_node ms_set_local(j_type type, int offset) {

  j_node node;

  #ifdef DEBUG
    if ((offset < 0) || (offset > method_size_locals())) {
      jabort("invalid local variable offset: the local variable at offset %d does not exist.", offset);
    }
  #endif

 
  if (offset >= method_size_parameters()) {
    /*
    ** When a local variable is pushed onto the stack, we have to
    ** push a copy of that local variable.  That is, when a local
    ** variable is changed or updated, copies on the stack are not.
    **
    ** This only triggers in rare cases.
    */

    if (jm->locals[offset] && ms_is_stacked(jm->locals[offset])) {
        // TODO: insert an assign instruction instead of bailing out.
      jfail("anti-dependence: attempt to change local variable when there is a copy on stack\n");
    }
    
    // local variables can only be defined once:
    jm->locals[offset] = node_create_local_variable(type, offset);
  }
  else {
    jfail("paramater handling is obnoxious due to control flow issues\n");
      // NOTE: we can't allow parameters being overwritten as this will cause the register allocator on ARM to go fubar.
  }

  // retrieve the node: 
  node = jm->locals[offset];

  #ifdef DEBUG
    if (jm->failed == 0 && node == NULL) {
      jabort("the requested local variable hasn't been initialized yet");
    }
  #endif
  
  woempa(9, "set local variable '%q' (%d) from slot %d\n", node, type, offset);
  
  return node;
}

j_node ms_get_local(j_type type, int offset) {

  j_node node;
  
  #ifdef DEBUG
    if ((offset < 0) || (offset > method_size_locals())) {
      jabort("invalid local variable offset: the local variable at offset %d does not exist.", offset);
    }
  #endif
    
  if (offset < method_size_parameters()) {
    if (jm->parameters[offset] == NULL) {
      jm->parameters[offset] = node_create_parameter(type, offset);
    }
    
    jm->locals[offset] = jm->parameters[offset];
  }
 
  // retrieve the node: 
  node = jm->locals[offset];

  #ifdef DEBUG
    if (jm->failed == 0 && node == NULL) {
      jabort("the requested local variable hasn't been initialized yet");
    }
  #endif
  
  woempa(9, "get local variable '%q' (%d) from slot %d\n", node, type, offset);

  return node;
}

