/**************************************************************************
* Copyright (c) 2001, 2002, 2003 by Acunia N.V. All rights reserved.      *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Philips-site 5, box 3       info@acunia.com                           *
*   3001 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: node.c 3985 2003-06-04 12:36:10Z dbuytaert $
*/

#include <string.h>
#include <stdarg.h>

#include "j-spot.h"
#include "arena.h"
#include "node.h"
#include "statistics.h"
#include "mimic-stack.h"
#include "register-allocator.h"
#include "ir.h"
#include "is.h"

char * node_operators[] = {
  "none",
  "constant",
  "variable",
  "add",
  "sub",
  "mul",
  "div",
  "rem",
  "neg",
  "lsh",
  "rsh",
  "and",
  "or",
  "xor",
  "le",
  "gt",
  "eq",
  "ne",
  "lt",
  "ge",
  "iseq", 
  "isne", 
  "call",
  "invoke",
  "assign",
  "return",
  "jump",
  "null_pointer",
  "array_bounds",
  "store"
};

j_node node_create_constant(j_type type, int value) {

  j_node node = arena_allocate(sizeof(j_Node));

  /*
  ** Create and initialize the constant's content:
  */
  
  node->type = type;
  node->pc = jm->inst_pc;
  node->operator = j_constant;
  node->data.constant.value = value;

  return node;
}

static inline j_node create_variable(j_type type, j_location location) {
  j_node node = arena_allocate(sizeof(j_Node));

  /*
  ** Create and initialize the variable's content:
  */
  
  node->type = type;
  node->pc = jm->inst_pc;
  node->operator = j_variable;
  node->data.variable.location = location;
  
  return node;
}

j_node node_create_parameter(j_type type, int value) {

  j_node node = create_variable(type, j_parameter);

  node->data.variable.offset = value;
  
  return node;
}

j_node node_create_local_variable(j_type type, int value) {

  j_node node = create_variable(type, j_local);

  node->data.variable.offset = value;
  
  return node;
}


j_node node_create_memory_variable(j_type type, j_node base, j_node offset) {

  j_node node = create_variable(type, j_memory);

  node->node1 = base;
  node->node2 = offset;
 
  // Increate the use counters by one:
  base->use += 1;
  offset->use += 1;
 
  return node;
}

j_node node_create_insn(j_operator operator, j_node n1, j_node n2) {

  j_node node;

  /*
  ** Allocate memory for the node:
  */

  node = arena_allocate(sizeof(j_Node));
    
  /*
  ** Create and initialize the node's content:
  */
  
  node->pc = jm->inst_pc;
  node->operator = operator;
  
  if (n1) {
    // Increate the node's use count by one:
    n1->use += 1;

    node->node1 = n1;
    node->type = n1->type;
  }

  if (n2) {    
    /*
    ** The right-hand side of an assignment defines a node; it does
    ** not consume a node so the use counter is not incremented.
    */
    
    if (operator != j_assign) {
      // Increate the node's use count by one:
      n2->use += 1;
    }

    node->node2 = n2;
    node->type = n2->type;
  }
 

  if (n1 && n2 && operator == j_assign) {
    /*
    ** If the instruction is an assignment instruction, we record 
    ** that in the assignee such that we can easily do def(node).
    */
    
    n2->definition = node;

    /*
    ** If the instruction is an assignment instruction, there might
    ** be some information to inherit.
    */
 
    node_inherit(n1, n2);
  }
  
  return node;
}

j_node node_create_call(char * name, j_operator operator, j_type type, void * function, int size) {

  j_node node;

  /*
  ** Allocate memory for the node:
  */

  node = arena_allocate(sizeof(j_Node));
    
  /*
  ** Create and initialize the function's content:
  */
  
  node->type = type;
  node->pc = jm->inst_pc;
  node->operator = operator;

  node->data.call.value = function;
  node->data.call.args = arena_allocate(size * sizeof(j_node));
  node->data.call.size = size;
  node->data.call.argc = 0;
 
  /*
  ** Copy the new name into the node's name field:
  */
 
  #ifdef DEBUG
    node->name = arena_allocate(strlen(name) + 1); // + 1 because of '\0'
    strcpy(node->name, name);
  #endif

  return node; 
}

int node_is_memory_variable(j_node node) {

  if (node->operator == j_variable && node->data.variable.location == j_memory) {
    return 1;
  }
  else {
    return 0;
  }
}

int node_is_local_variable(j_node node) {
  if (node->operator == j_variable && node->data.variable.location != j_local) {
    return 1;
  }
  else {
    return 0;
  }
}

int node_is_parameter(j_node node) {
  if (node->operator == j_variable && node->data.variable.location == j_parameter) {
    return 1;
  }
  else {
    return 0;
  }
}

int node_is_instruction(j_node node) {
  if (node->operator == j_constant || node->operator == j_variable) {
    return 0;
  }
  else {
    return 1;
  }
}


int node_is_call(j_node node) {
  if (node->operator == j_native_call || node->operator == j_java_call) {
    return 1;
  }
  else {
    return 0;
  }
}

int node_is_escapable(j_node node) {
  if (node->operator == j_native_call || node->operator == j_java_call || node->operator == j_store || node->operator == j_null_pointer || node->operator == j_array_bounds) {
    return 1;
  }
  else {
    return 0;
  }
}
int node_is_collectable(j_node node) {

  if (node->type == j_unknown) {
    jabort("%q might be a reference\n", node);
    return 0;
  }
  else if (node->operator == j_constant) {
    /*
    ** There are no constant references, yet there are constants 
    ** of type 'j_address' such as class and method pointers.
    */
    return 0;
  }
  else if (node->type == j_address) {
    /*
    ** If the reference is a parameter then it is already protected
    ** on the caller's stack frame.
    */
    return !node_is_parameter(node);
  }
  else {
    return 0;
  }
}

int node_check_null_pointer(j_node node) {

  /*
  ** If the variable "null_check" is set, the instance has already
  ** been checked so no null pointer check is required.
  */
  
  if (node->null_check) {
    statistics_null_pointer(1);
    return 0;
  }
  else { 
    /*
    ** Mark the node as being checked.
    */ 

    node->null_check = 1;
    
    /*
    ** In case of an instance method, parameter 0 is the this pointer,
    ** which is non-null by definition. 
    */
  
    if (!method_is_static() && node_is_parameter(node) && node->data.variable.offset == 0) {
      statistics_null_pointer(1);
      return 0;
    }
    else {
      statistics_null_pointer(0);
      return 1;
    }
  }
}

int node_check_array_bounds(j_node array, j_node index) {

  /*
  ** If we have a constant index whose value is positive and
  ** less than the maximum known array size, no array bound
  ** check is needed.
  */

  if (index->operator == j_constant && index->data.constant.value >= 0) {
    if (index->data.constant.value < array->array_size) {
      statistics_array_bounds(1);
      return 0;
    }
    else {
      array->array_size = index->data.constant.value;
      statistics_array_bounds(0);
      return 1; 
    }
  }
  else {
    statistics_array_bounds(0);
    return 1;
  }
}

int node_are_compatible(j_type type1, j_type type2) {

  if (type1 == type2) {
    return 1;
  }

  if (type1 == j_unknown || type2 == j_unknown) {
    return 1;
  }

  if ((type1 == j_boolean || type1 == j_byte || type1 == j_character || type1 == j_integer || type1 == j_short) && (type2 == j_boolean || type2 == j_byte || type2 == j_character || type2 == j_integer || type2 == j_short)) { 
    return 1;
  }

  return 0;
}

j_node node_state(j_node node, int slot) {

  if (jm->states[node->pc]) {
    return jm->states[node->pc][slot];
  }
  else {
    return NULL;
  }
}

char * node_get_operator(j_node node) {
  return (char *)node_operators[node->operator];
}

void node_add_param(j_node node, j_node parameter) {

  // Perform a run-time sanity check:
  if (node_is_call(node) == 0) {
    jabort("the specified node is not a valid function-node\n");
  }
  
  /*
  ** Promote function arguments that are subexpressions (instructions)
  ** to become root nodes.  As a result DAGs will get split in smaller
  ** DAGs that pre-calculate subexpressions, and that store the values
  ** in temporaries.  This is required for both code generation and 
  ** register allocation: in case of the statement 'foo(bar());', we
  ** would have to call 'bar()' after having started to setup 'foo()'s
  ** stack frame.  In essence, we split the DAG to read: 'x = bar();
  ** foo(x);'.
  */
  
  if (node_is_instruction(parameter)) {
    ir_add_node(parameter, j_link_next, j_link_none);
  }
  
  // Increase the function's argument count by one:
  node->data.call.argc += 1;
 
  // Increate the parameters's use count by one:
  parameter->use += 1;

  // Add the parameter to the end of the parameter list:
  *(node->data.call.args + node->data.call.size - node->data.call.argc) = parameter;
    
  // Perform a run-time sanity check: 
  if (node->data.call.argc > node->data.call.size) {
    jabort("the specified node's argument count is exceeded\n");
  }
}

void node_add_link(j_node node1, j_node node2, int label) {

  int i;

  /*
  ** Take care of the outgoing edges:
  */
  
  woempa(9, "adding edge with label %d between %p at pc %d and %p at pc %d\n", label, node1, node1->pc, node2, node2->pc);
  
  if (label) {
    node1->outgoing[0] = node2;   // "true" edge
  }
  else {
    node1->outgoing[1] = node2;   // "false" edge
  }
  
  /*
  ** Take care of the incoming edges:
  */
  
  for (i = 0; i < j_max_edges; i++) {
    if (node2->incoming[i] == NULL) {
      node2->incoming[i] = node1;
      break;
    }
  }
}

void node_add_clone(j_node src, j_node dst) {

  int i;


  if (node_are_compatible(src->type, dst->type)) {
    for (i = 0; i < j_max_edges; i++) {
      if (src->dst[i] == dst) {
        break;
      }

      if (src->dst[i] == NULL) {
        src->dst[i] = dst;
        break;
      }
    }

    for (i = 0; i < j_max_edges; i++) {
      if (dst->src[i] == src) {
        break;
      }

      if (dst->src[i] == NULL) {
        dst->src[i] = src;
        break;
      }
    }

    if (i == j_max_edges) {
      jfail("maximum number of edges exceeded\n");
    }

    if (src->dst[1] == NULL) { 
      // Increate the node's use count by one:
      src->use += 1;
    }
  }
  else {
     /*
     ** This triggers in rare occasions only and does not 
     ** necessarily indicate a IR problem.  For an example,
     ** compile ProblemTest.com_test_24 with Jikes 1.18.
     */

    joempa("warning: attempt to clone two nodes %q (%p) and %q (%p) that are not assignment compatible\n", src, src, dst, dst);
  }
}

void node_inherit(j_node parent, j_node child) {

   child->null_check = parent->null_check;
   child->array_size = parent->array_size;
}

void node_operands(j_node node, node_callback function, ...) {

  j_node temp;
  va_list list;
  int i;
 
  va_start(list, function);
  
  if (node_is_call(node)) {
    for (i = 0; i < node->data.call.argc; i++) {
      temp = *(j_node *)(node->data.call.args + i);
      function(node, temp, list);
    }
  }
  
  if (node->node1) {
    function(node, node->node1, list);
  }
 
  if (node->node2) {
    function(node, node->node2, list);
  }

  function(node, NULL, list);

  va_end(list);  
}

char * node_print(char * buffer, int * remain, void * data, int width, int precision, unsigned int flags) {

  j_node node = data;
  int pos = 0;
 
  if (node) {

    /*
    ** Print the node's string representation/information.
    */
  
    if (node->operator == j_constant && node->type == j_address) { 
      pos = sprintf(buffer, "constant '%p'", (j_word *)node->data.constant.value);
    }
    else if (node->operator == j_constant && node->type != j_address) { 
       pos = sprintf(buffer, "constant '%d'", node->data.constant.value);
    }
    else if (node->operator == j_variable && node->data.variable.location == j_parameter) {
      pos = sprintf(buffer, "parameter %d (%p)", node->data.variable.offset, node);
    }
    else if (node->operator == j_variable && node->data.variable.offset < method_size_locals()) {
      pos = sprintf(buffer, "temporary %d (local)", node->data.variable.offset);
    }
    else if (node->operator == j_variable && node->data.variable.offset >= method_size_locals()) {
      pos = sprintf(buffer, "temporary %d (stack)", node->data.variable.offset);
    }
    else if (node_is_call(node)) { 
      if (node->name) {
        pos = sprintf(buffer, "call '%s'", node->name);
      }
      else {
        pos = sprintf(buffer, "call %p", (j_word *)node->data.call.value);
      }
    }
    else {
      if (node->name) {
        pos = sprintf(buffer, "%s '%s'", node_get_operator(node), node->name);
      }
      else {
        pos = sprintf(buffer, "%s", node_get_operator(node));
      }
    }
  }
  else {
    pos = sprintf(buffer, "node 'NULL'");
  }

  *remain -= pos;
  buffer  += pos;

  return buffer;
}

