/**************************************************************************
* Copyright (c) 2001, 2002, 2003 by Acunia N.V. All rights reserved.      *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Philips-site 5, box 3       info@acunia.com                           *
*   3001 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: register-allocator.c 4032 2003-07-11 13:19:09Z dbuytaert $
*/

#include "j-spot.h"
#include "ir.h"
#include "arena.h"
#include "register-allocator.h"
#include "code-generator.h"
#include "liveness.h"
#include "mimic-stack.h"
#include "statistics.h"
#include "is.h"

int last;

/*
** Return the number of the first free register.  If no free register
** is available, return -1.
*/
static int select_free_register(j_node node, int exclude) {
  
  int i;

  for (i = 0; i < J_NUM_REGISTERS; i++) {
    if (exclude != i && ra_is_free(i)) {
      return i;
    }
  }

  return -1; /* We get here if no free register is found. */
}

/*
** Spill the node that lives longest (probably not 100% accurate).
*/
static int spill_longest(j_node instruction, int exclude) {
  
  int i;
  int ttl = 0;
  int ttl_spillee = 0;
  j_node node;
  j_node spillee = NULL;

  /*
  ** Search for the node with the longest time to live (ttl):
  */
  for (i = 0; i < J_NUM_REGISTERS; i++) {
    node = jm->registers[i];
    if (node && node->reg == i && exclude != i) {
      ttl = la_is_live(node);
      if (spillee == NULL) {
        spillee = node;
        ttl_spillee = ttl;
      }
      else if (ttl == 0) {
        spillee = jm->registers[i];
        ttl_spillee = jm->slots;
          // NOTE: we use 'jm->slots' to ensure that we won't overwrite dead spillees in the next if-clause when handling the remaining iterations. 
      }
      else if (ttl > ttl_spillee) {
        spillee = jm->registers[i];
        ttl_spillee = ttl;
      }
    }
  }

  // Spill the spillee:
  ra_spill(instruction, spillee);

  return spillee->reg;
}

/*
** Spill the node.
*/
void ra_spill(j_node instruction, j_node variable) {
 
  if (ra_in_register(variable) && !ra_on_stack(variable)) {
    /*
    ** 1. Only spill a node when it is kept in a register.
    ** 2. Only spill a node when it is not already on the stack.  This
    **    is correct because in SSA, each node is defined only once.
    */
    
    // Record the total number of spills:
    statistics_spill();

    // Delegate the actual spilling to the platform-specific backend:
    cg_spill(instruction, variable);
    
    // Mark the node as being spilled:
    variable->spill = 1;
 
    woempa(9, "spilled node %q (%p) from register %d to stack location %d\n", variable, variable, variable->reg, variable->stack);
  }
}

/*
** Check whether the specified node is on stack.
*/
int ra_on_stack(j_node node) {

  if (node->stack && node->spill) {
    return 1;
  }

  return 0;
}

/*
** Check whether the specified node is in memory.
*/
int ra_in_register(j_node node) {
  
  if (jm->registers[node->reg] == node) {
    return 1;
  }
    
  return 0;
}

/*
** Check whether the specified register is in use on the specified moment.
*/
int ra_is_free(int reg) {

  j_node node = jm->registers[reg];

  if (node) {
    woempa(9, "%q: reg = %d, node->reg = %d, node->spill = %d, node->stack = %d, live = %d ==> %d\n", node, reg, node->reg, node->spill, node->stack, la_is_live(node), (node->reg == reg && node->spill == 0 && la_is_live(node)));
  }

  if (node && ra_in_register(node) && la_is_live(node) && !ra_on_stack(node)) {
    /*
    ** NOTES:
    **
    **   1. We can't check against "node_is_parameter()" because depending
    **      on the calling conventions, some parameters might be passed in
    **      registers; on ARM for example, parameters 1 to 4 are passed in
    **      registers and might need spilling for later use.  If parameter
    **      spilling can or needs to be avoided (eg. on x86 parameters are
    **      always passed using the stack), make sure "node->spill" is set
    **      to true.  We typically do so in "cg_init()".
    **
    **   2. We can't check against "node->operator != j_constant" because
    **      we might have to spill contants that are merged by a phi-node
    **      at points where control flow joins.
    */
    return 0;
  }
  else {
    return 1;
  }
}

/*
** Replace the node currently in register reg with the new node.
** Spill if necessary.
*/
void ra_assign(j_node instruction, j_node variable, int reg) {

  if (!ra_is_free(reg)) {
    ra_spill(instruction, jm->registers[reg]);
  }
 
  /*
  ** Tell the method that it uses register "reg".
  */
  jm->usage = jm->usage | (1 << reg);
  
  /*
  ** Tell the method that register "reg" is currenlty occupied
  ** by the specified node.
  */
  jm->registers[reg] = variable;
  
  /*
  ** Update the node's register and stack information.
  */
  variable->reg = reg;
  variable->spill = 0;  // TODO: shouldn't this be redundant now we have SSI?  If you remove this j-spot, the generated machine code is incorrect.

  last = reg;
}

/*
** Consume a use of the node.
*/
void ra_consume(j_node node) {
  node->use = node->use - 1;
}

/*
** Allocate a register for the node.
*/
void ra_allocate(j_node instruction, j_node variable) {

  int reg = -1;
 
  /*
  ** Check whether the specified node hasn't been alocated already.
  */
  if (!ra_in_register(variable)) {

    reg = select_free_register(variable, last);
    if (reg == -1) {
      reg = spill_longest(instruction, last);
    }

    /*
    ** Perform a run-time sanity check.
    */
    if (reg < 0 || reg > J_NUM_REGISTERS) {
      jabort("allocated %q (%p) to invalid register %d\n", variable, variable, reg);
    }
    else {
      woempa(9, "allocated %q to register %d\n", variable, reg);
    }
   
    /*
    ** Ask the code generator to emit machine code that moves the
    ** node to the specified register.
    */ 
    cg_node2register(variable, reg);
 
    /*
    ** Tell the method that it uses register "reg".
    */
    jm->usage = jm->usage | (1 << reg);
  
    /*
    ** Tell the method that register "reg" is currenlty occupied
    ** by the specified node.
    */
    jm->registers[reg] = variable;
  
    /*
    ** Update the node's register information.
    */
    variable->reg = reg;
      // NOTE: we don't invalidate the node's copy on stack as we do in "ra_assign()".
  }

  ra_consume(variable);
  
  last = variable->reg;
}

/*
** Return the number of the stack slot this node has been allocated to.
*/
int ra_stack_slot(j_node instruction, j_node variable) {

  j_node node = NULL;
  int slot = 0;
  int size;
  int i;

  if (variable->stack == 0) {
    /*
    ** Try to identify the frame slot the variable resides in
    ** if it were executed as Java bytecode.  This information
    ** acts as liveness information which allows us to reuse
    ** slots.
    */  
    if (instruction) {
      size = method_size_locals() + method_size_stack();
      for (i = 0; i < size; i++) {
        node = ms_node_state(instruction->pc, j_pre_execution, i);
        if (node == variable) {
          return i + 1;
          // '+ 1' because stack offsets start at '0'.
        } 
      }
    }
    
    /*
    ** TODO: if we got here we typically ran out of registers
    ** while computing a sub-expression.  We may want to push 
    ** these nodes to a limited set of special puporpose stack 
    ** slots (eg. one stack slot for each available register)?
    */
      
    jm->slots = jm->slots + 1;
    slot = jm->slots;
  }

  return slot;
}

/*
** Called by the compiler when an basic block entry is encountered.
*/
void ra_basic_block_entry(j_node node) {
  
  int i;
  
  for (i = 0; i < J_NUM_REGISTERS; i++) {
    jm->registers[i] = NULL;
  }

  woempa(9, "cleared all registers at basic block entry %q (%p)\n", node, node);
}

/*
** Called by the compiler when an basic block exit is encountered.
*/
void ra_basic_block_exit(j_node node) {
 
  int i;

  /*
  ** If we reached an exit point of the method, there is nothing left
  ** to do.
  */ 
  if (ir_is_exit(node)) {
    return;
  }

  /*
  ** Spill all live variables that are live accross this basic block's 
  ** boundary.
  */ 
  for (i = 0; i < J_NUM_REGISTERS; i++) {
    /*
    ** The statement below is more accurate then ra_is_free() because
    ** ra_is_free() would spill the operands of the current branch 
    ** instruction.  For example, given the instruction 'eq a b', it
    ** would attempt to spill 'a' and 'b'.  This is the case bacause
    ** no code has been generated yet to prepare 'a' and 'b', and thus
    ** their use-counters have not been decremented yet.
    */
    
    if (jm->registers[i]) {
      if (node_is_parameter(jm->registers[i])) {
        ra_spill(node, jm->registers[i]);
      }
      else if (jm->registers[i]->dst[0]) {
        ra_spill(node, jm->registers[i]);
      }
    }
  }

  woempa(9, "spilled all registers at basic block exit %q (%p)\n", node, node);
}

/*
** Print the contents of all registers to screen.
*/
void ra_dump_registers(void) {

  int i;

  for (i = 0; i < J_NUM_REGISTERS; i++) {
    if (jm->registers[i]) {
      woempa(9, " %d. %q --> spill = %d, stack = %d\n", i, jm->registers[i], jm->registers[i]->spill, jm->registers[i]->stack);
    }
  }
}

