/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

#include "oswald.h"  /* x_thread_... */
#include "wonka.h"   /* the w_ types */
#include "ts-mem.h"  /* allocMem_system */
#include "realms.h"  /* allocMem_system */
#include "jdwp.h"
#include "jdwp_events.h"
#include "jdwp-debug.h"

/*
** Set : Process the event request sent by the debugger. When events occur and they match
**       one of the previous received event request, an event will be send to the debugger.
*/

struct set_header {
  w_ubyte  event_kind;
  w_ubyte  suspend_policy;
  w_short  mod1;
  w_short  mod2;
};

w_void jdwp_evtreq_set(jdwp_command_packet cmd) {
  struct set_header   *hdr = (struct set_header *)cmd->data;
  w_int               modifiers = swap_int(((hdr->mod2 << 16) + hdr->mod1));
  w_int               position;
  w_byte              mod_kind;
  w_int               i;
  w_int               strlength;
  jdwp_event          event;
  jdwp_event_modifier mod;

  woempa(9, "Event Request - Set\n");

  event = allocate_event(hdr->event_kind, hdr->suspend_policy);
  
  woempa(9, "  eventKind:     %4d (%s)\n", 
          event->event_kind, jdwp_event_kind_str(event->event_kind));
  woempa(9, "  suspendPolicy: %4d (%s)\n", 
          event->suspend_policy, jdwp_suspend_policy_str(event->suspend_policy));
  woempa(9, "  modifiers (%d):\n", modifiers);

  /* 
  ** Set the current position to the 7'th byte. This is right behind the header and puts us at the 
  ** first modifier. (if there are any at all). 
  */

  position = 6;

  /*
  ** Next we go over all the modifiers in the block and store them in the event.
  */
  
  for(i = 0; i < modifiers; i++) {
    
    memcpy(&mod_kind, &cmd->data[position], 1);      /* Get the kind of modifier    */
    position += 1;                                   /* Update the current position */

    mod = x_mem_calloc(sizeof(jdwp_Event_Modifier));
    mod->mod_kind = mod_kind;
    
    woempa(9, "    %2d %s\n", mod_kind, jdwp_mod_kind_str(mod_kind));

    switch(mod_kind) {
      
      /* 
      ** Count : Suppress the first (count - 1) occurrences of the event. Each occurrence will 
      **         decrement count. When it reaches 0, the other modifiers are checked to see if 
      **         the event should be blocked. If none of the modifiers suppress the event, it will 
      **         be reported. This modifier works on all event kinds.
      **
      **         NOTE: It's possible that once count has reached 0, no further events are 
      **         reported. I have to check it sometime. This means that a count of 1 will report 
      **         the next occurrence and then completely ignore further occurrences.
      */

      case  1: memcpy(&mod->condition.count, &cmd->data[position], 4);
               position += 4;
               break;
               
      /* 
      ** Conditional : This modifier is not used but reserved for future additions.
      */
      
      case  2: memcpy(&mod->condition.exprID, &cmd->data[position], 4);
               position += 4;
               break;
               
      /* 
      ** Thread only : Only report the event if it happened in the given thread. This works on all 
      **               event kinds except on the 'Class unload' event.
      */
      
      case  3: memcpy(&mod->condition.threadID, &cmd->data[position], 4);
               position += 4;
               break;
               
      /* 
      ** Class only : Restrict the 'Class Prepare' events to classes of the the given reference type 
      **              or subtype. For other events, restrict the events to classes whose location is 
      **              in the given reference type of subtype. If the class can be safely casted to 
      **              the given reference type, then an event will be passed on. This modifier 
      **              doesnt't work on 'Class unload', 'Thread start' and 'Thread end' events.
      */
      
      case  4: memcpy(&mod->condition.clazz, &cmd->data[position], 4);
               position += 4;
               break;
               
      /* 
      ** Class match : Block events from classes whose name doesn't match the given expression. 
      **               For 'Class load' and 'Class unload' events, the class name is checked. For
      **               all the other events, the name of the class where the event occurred is 
      **               checked. This modifier can not be used with 'Thread start' and 'Thread end' 
      **               events.
      **               The expression is limited to exact matches and patterns that have an '*' at
      **               the beginning or the end. e.g. 'java.awt.Button', '*.Button' & 'java.awt.*'
      */

      case  5: memcpy(&strlength, &cmd->data[position], 4);
               position += 4;
               strlength = swap_int(strlength);
               mod->condition.match_pattern = x_mem_calloc((w_word)(strlength + 1));
               memcpy(mod->condition.match_pattern, &cmd->data[position], (w_word)strlength);
               position += strlength;

               woempa(9, "       -> %s\n", mod->condition.match_pattern);
       
               break;
               
      /* 
      ** Class exclude : This modifier has the opposite effects as the 'Class match' modifier. When
      **                 a certain class name is blocked with 'Class match', it will be reported with
      **                 'Class exclude'. When a certain class name gets through with 'Class match', 
      **                 it will be blocked by 'Class exclude'. The same rules for the patterns and
      **                 events on which it can be used applies here aswell.
      */
      
      case  6: memcpy(&strlength, &cmd->data[position], 4);
               position += 4;
               strlength = swap_int(strlength);
               mod->condition.exclude_pattern = x_mem_calloc((w_word)(strlength + 1));
               memcpy(mod->condition.exclude_pattern, &cmd->data[position], (w_word)strlength);
               position += strlength;

               woempa(9, "       -> %s\n", mod->condition.exclude_pattern);
       
               break;

      /* 
      ** Location only : Only report events if they occurred at the given location. This modifier
      **                 only works with 'Breakpoint', 'Field access', 'Field modification', 'Step'
      **                 and exception events.
      */

      case  7: memcpy(&mod->condition.location, &cmd->data[position], 17);
               position += 17;
               break;
               
      /* 
      ** Exception only : This modifier works on 'Exception' events only and restricts exception to
      **                  the given type (or all types if the given type is null) and if they are 
      **                  caught or uncaught (depending on the 2 booleans that are given).
      **
      **                  NOTE: Sometimes it's not possible to determine if an exception is caught
      **                        or not at the time it's thrown...
      */

      case  8: memcpy(&mod->condition.ex.exception, &cmd->data[position], 4);
               position += 4;
               woempa(9, "          exception: 0x%x\n", mod->condition.ex.exception);
               memcpy(&mod->condition.ex.caught, &cmd->data[position], 1);
               position += 1;
               woempa(9, "          caught:    %d\n", mod->condition.ex.caught);
               memcpy(&mod->condition.ex.uncaught, &cmd->data[position], 1);
               position += 1;
               woempa(9, "          uncaught:  %d\n", mod->condition.ex.uncaught);
               break;
               
      /* 
      ** Field only : This modifier can only be used with 'Field access' and 'Field modification'
      **              events. Only events that occurred for a given field are reported.
      */
      
      case  9: memcpy(&mod->condition.field.declaring, &cmd->data[position], 4);
               position += 4;
               memcpy(&mod->condition.field.ID, &cmd->data[position], 4);
               position += 4;
               break;
               
      /* 
      ** Step : This modifier can only be used with 'Step' events. Only events that match the given
      **        depth and size are reported.
      */
     
      case 10: memcpy(&mod->condition.step.thread, &cmd->data[position], 4);
               position += 4;
               memcpy(&mod->condition.step.size, &cmd->data[position], 4);
               position += 4;
               memcpy(&mod->condition.step.depth, &cmd->data[position], 4);
               position += 4;
               break;
               
      /* 
      ** Instance only : Only report events in the given instance. This doens't work with 'Class 
      **                 prepare', 'Class unload', 'Thread start' and 'Thread end' events.
      **                 (introduced in JDWP 1.4)
      */

      case 11: memcpy(&mod->condition.instance, &cmd->data[position], 4);
               position += 4;
               break;
               
      /* 
      ** What's this ?  We don't know this kind of modifier...
      ** Let the debugger know about this little incident.
      */
      
      default: jdwp_send_reply(cmd->id, NULL, 0, jdwp_err_not_implemented);

               /* 
               ** Clean up the event and also all the modifiers that were added so far.
               */

               /* TODO */

               return; 
    }

    /*
    ** Add the modifier to the event.
    */

    add_modifier(event, mod);
  }

  /*
  ** If the event is a breakpoint, we need to do some more stuff.
  */

  if(event->event_kind == jdwp_evt_breakpoint) {
    jdwp_set_breakpoint(event);
  }

  /*
  ** When we're done parsing, add this event to the list.
  */  

  add_event(event);

  /*
  ** Let the debugger know that everything worked out fine.
  */

  jdwp_send_reply(cmd->id, (w_void *)&event->eventID, sizeof(w_int), jdwp_err_none);

  return;
}


/*
** Clear : Clear an event request. Events that match the event in question will no longer be send
**         to the debugger.
*/

w_void jdwp_evtreq_clear(jdwp_command_packet cmd) {
  w_ubyte             event_kind;
  w_int               eventID;
  jdwp_event          event;
  jdwp_event_modifier mod;
  jdwp_event_modifier tmp;

  /* 
  ** The data for this command has the following structure :
  **   byte  event_kind
  **   int   eventID
  */

  memcpy(&event_kind, &cmd->data[0], 1);   /* Get the event_kind */
  memcpy(&eventID,    &cmd->data[1], 4);   /* Get the eventID */

  woempa(9, "Event Request - Clear\n");
  woempa(9, "  eventKind: %4d (%s)\n", 
      event_kind, jdwp_event_kind_str(event_kind));
  woempa(9, "  eventID:   %4d\n", eventID); 

  /*
  ** Look for the event.
  */

  event = get_event_ID(eventID);

  if(event) {

    if(event->event_kind == jdwp_evt_breakpoint) {
      jdwp_clear_breakpoint(event);
    }

    /*
    ** Clear the modifiers and release memory.
    */

    mod = event->modifiers;
    if(mod != NULL) mod->prev->next = NULL;

    while(mod != NULL) {
      tmp = mod;
      mod = mod->next;

      /* TODO: Some modifiers have extra allocated memory. Deal with it. */
      
      x_mem_free(tmp);
    }

    /* 
    ** Clear the event and release memory.
    */

    remove_event(event->eventID);

    x_mem_free(event);

  }
  
  jdwp_send_reply(cmd->id, NULL, 0, jdwp_err_none);

  return;
}
    
/*
** Clear breakpoints : Clear all breakpoint event requests. All the breakpoints will be deleted and
**                     no more breakpoint events will be send.
*/

w_void jdwp_evtreq_clearbreak(jdwp_command_packet cmd) {

  woempa(9, "Event Request - Clear all breakpoints\n");

  /*
  ** TODO: Everything....
  */
  
  jdwp_send_reply(cmd->id, NULL, 0, jdwp_err_none);

  return;
}

/*
** The dispatcher for the 'Event request' command set.
*/

w_void dispatch_eventreq(jdwp_command_packet cmd) {

  switch((jdwp_eventreq_cmd)cmd->command) {
    case jdwp_eventreq_set:                  jdwp_evtreq_set(cmd);    break;
    case jdwp_eventreq_clear:                jdwp_evtreq_clear(cmd);  break;
    case jdwp_eventreq_clearAllBreakpoints:
    default:  
      jdwp_send_reply(cmd->id, NULL, 0, jdwp_err_not_implemented);
  }
}


