/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

#include "oswald.h"      /* x_thread_... */
#include "wonka.h"       /* the w_ types */
#include "ts-mem.h"      /* allocMem_system */
#include "realms.h"      /* allocMem_system */
#include "jdwp.h"
#include "clazz.h"       /* w_clazz struct */
#include "methods.h"     /* w_method struct */
#include "descriptor.h"  /* w_desc struct */

/*
** Returns the line table of a given method in a given reference.
*/

w_void jdwp_method_lines(jdwp_command_packet cmd) {
  w_clazz   clazz;
  w_method  method;
  w_int     i;
  w_long    codeindex;
  w_ubyte   *reply;
  w_int     position = 0;
  w_int     line;

  /*
  ** Get the clazz and method ID out of the packet.
  ** In our case, a methodID is enough because it's already quite unique and 
  ** the method holds a pointer to the class. The specification however says 
  ** that methodID's don't have to be unique if they are in different classes.
  */

  memcpy(&clazz,  &cmd->data[0], 4);
  memcpy(&method, &cmd->data[4], 4);

  /*
  ** Get the number of methods and allocate memory accordingly. A reply looks like this :
  **    w_long  start    the lowest possible code index for this method.
  **    w_long  end      the highest possible code index for this method.
  **    w_int   lines    the number of entries.
  **  Repeated 'lines' times :
  **    w_long  index    the code index for the current line.
  **    w_int   line     the linenumber that matches with this code index.
  */

  reply = x_mem_alloc((w_word)(8 + 8 + 4 + method->numLineNums * (8 + 4)));

  /*
  ** Put 'start' and 'end' into the reply. 'start' is always 0 (the first bytecode),
  ** 'end' is always the number of bytecodes (the code length).
  */

  codeindex = swap_int(0);
  memcpy(&reply[position], &codeindex, 8);
  position += 8;
  
  codeindex = method->code_length;
  codeindex = swap_long(codeindex);
  memcpy(&reply[position], &codeindex, 8);
  position += 8;

  /*
  ** Next in line is the number of entries in the line table.
  */

  i = swap_int(method->numLineNums);
  memcpy(&reply[position], &i, 4);
  position += 4;

  /*
  ** Go through the entire line table and add the entries to the reply.
  */

  for(i = 0; i < method->numLineNums; i++) {

    /*
    ** Get the line number and code index.
    */

    line = method->lineNums[i].line_number;
    codeindex = method->lineNums[i].start_pc;

    /*
    ** Swap 'em.
    */

    line = swap_int(line);
    codeindex = swap_long(codeindex);

    /*
    ** Add them to the reply packet.
    */

    memcpy(&reply[position], &codeindex, 8);
    position += 8;
    memcpy(&reply[position], &line, 4);
    position += 4;
    
  }

  /*
  ** Send the reply to the debugger.
  */ 
  
  jdwp_send_reply(cmd->id, reply, position, jdwp_err_none);

  /*
  ** Clean up.
  */ 
  
  x_mem_free(reply);
  
  return;
}


/*
** Get the variable table for the given method. This table includes the arguments to
** the method as well as all the locals used during execution of the method. In case 
** the given class is an instance, 'this' is also included.
*/

w_void jdwp_method_var_table(jdwp_command_packet cmd) {
  w_clazz     clazz;
  w_method    method;
  w_ubyte     *reply;
  w_int       length;
  w_int       words = 0;
  w_int       position = 0;
  w_int       i;
  w_long      pc_start;
  w_ubyte     *string;

  /*
  ** Get the class and method from the command packet.
  */

  memcpy(&clazz,  &cmd->data[0], 4);
  memcpy(&method, &cmd->data[4], 4);

  /*
  ** Allocate memory for the reply.
  */
  
  reply = x_mem_calloc((w_word)(4 + 4 + method->numLocalVars * (8 + 100 + 100 + 4 + 4)));

  /*
  ** Add the number of local variables to the reply. We start at position 4, because the total number of words
  ** used in the frame should be at position 0 (everything uses 1 word except doubles & longs). Since we don't 
  ** know this number up front, we leave the spot open and count the words as we add the variables. Later on
  ** this value will be filled in.
  */

  position = 4;
  length = swap_int(method->numLocalVars);
  memcpy(&reply[position], &length, 4);
  position += 4;

  /*
  ** Every variable takes at least one word. For longs and doubles we'll add an extra one.
  */

  words = method->numLocalVars;

  /*
  ** Go over the variables and add them to the reply.
  */

  for(i = 0; i < method->numLocalVars; i++) {
    
    /*
    ** Get the start code index at which this variable exists and store
    ** it in the reply packet.
    */
    
    pc_start = swap_long((w_long)method->localVars[i].start_pc);
    memcpy(&reply[position], &pc_start, 8);
    position += 8;

    /*
    ** Store the name of variable.
    */

    string = jdwp_string2UTF8(method->localVars[i].nat.resolved.name, &length);
    memcpy(&reply[position], string, (w_word)(length + 4));
    position += length + 4;
    x_mem_free(string);

    /*
    ** Store the signature of the variable.
    */

    string = jdwp_string2UTF8(method->localVars[i].nat.resolved.desc->name, &length);
    memcpy(&reply[position], string, (w_word)(length + 4));
    position += length + 4;
    x_mem_free(string);

    /*
    ** Store the code length during which this variable exists.
    */

    length = swap_int((method->localVars[i].end_pc -  method->localVars[i].start_pc));
    memcpy(&reply[position], &length, 4);
    position += 4;

    /*
    ** Store the index of the variable in the frame.
    */

    length = swap_int((w_int)method->localVars[i].slot);
    memcpy(&reply[position], &length, 4);
    position += 4;

    /*
    ** Add an extra word if this variable is a long or a double.
    */

    if(method->localVars[i].nat.resolved.desc == desc_long) words++;
    if(method->localVars[i].nat.resolved.desc == desc_double) words++;
    
  }

  /*
  ** Store the number of words at position 0 in the reply.
  */

  words = swap_int(words);
  memcpy(&reply[0], &words, 4);

  /*
  ** Send the reply to the debugger.
  */

  jdwp_send_reply(cmd->id, reply, position, jdwp_err_none);

  /*
  ** Clean up.
  */

  x_mem_free(reply);
  
  return;
}


/*
** Returns the bytecode of a given method in a given reference.
*/

w_void jdwp_method_bytecode(jdwp_command_packet cmd) {
  w_clazz   clazz;
  w_method  method;
  w_ubyte   *reply;
  w_int     length;
  w_int     position = 0;

  /*
  ** Get the clazz and method ID out of the packet.
  ** In our case, a methodID is enough because it's already quite unique and 
  ** the method holds a pointer to the class. The specification however says 
  ** that methodID's don't have to be unique if they are in different classes.
  */

  memcpy(&clazz,  &cmd->data[0], 4);
  memcpy(&method, &cmd->data[4], 4);

  /*
  ** Get the number of bytecodes and allocate memory for the reply. First we have
  ** to put in the number of bytecodes, so we need 4 bytes extra.
  */

  length = method->code_length;
  reply = x_mem_alloc((w_word)(4 + length));

  /*
  ** Store the number of bytecodes in the reply packet.
  */

  length = swap_int(length);
  memcpy(&reply[position], &length, 4);
  length = swap_int(length);
  position += 4;

  /*
  ** Put the actual bytecode into the packet.
  */

  memcpy(&reply[position], method->code, (w_word)length);
  position += length;
  
  /*
  ** Send the reply to the debugger.
  */ 
  
  jdwp_send_reply(cmd->id, reply, position, jdwp_err_none);

  /*
  ** Clean up.
  */ 
  
  x_mem_free(reply);
  
  return;
}


/*
** The dispatcher for the 'Method' command set.
*/

w_void dispatch_method(jdwp_command_packet cmd) {

  switch((jdwp_method_cmd)cmd->command) {
    case jdwp_method_lineTable:       jdwp_method_lines(cmd);       break;
    case jdwp_method_variableTable:   jdwp_method_var_table(cmd);   break;
    case jdwp_method_bytecodes:       jdwp_method_bytecode(cmd);    break;
    default:
      jdwp_send_reply(cmd->id, NULL, 0, jdwp_err_not_implemented);
  }
  
}

