/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

#include "oswald.h"  /* x_thread_... */
#include "wonka.h"   /* the w_ types */
#include "ts-mem.h"  /* allocMem_system */
#include "realms.h"  /* allocMem_system */
#include "jdwp.h"
#include "jdwp_events.h"
#include "clazz.h"       /* w_clazz */
#include "methods.h"     /* w_method */
#include "threads.h"     /* w_thread */
#include "descriptor.h"  /* w_desc */
#include "hashtable.h"   /* hashtables */
#include "fifo.h"        /* w_desc */

jdwp_event event_table[255]; 
jdwp_event event_hashtable[EVENT_HASHTABLE_SIZE]; 

#define BREAKPOINT_HASHTABLE_SIZE  97

jdwp_breakpoint breakpoint_hashtable[BREAKPOINT_HASHTABLE_SIZE];

/*
** Add a new event. This happens in 2 big steps. First the event is
** added to a cicular linked list of events of the same kind. Then 
** the event is also added to a hashtable with it's event ID as key
** so it can be recovered fast and easy (Keeping in mind that the 
** debugger refers to it by it's ID). 
*/

w_void add_event(jdwp_event event) {
  w_int hash_key = event->eventID % EVENT_HASHTABLE_SIZE;
  
  /*
  ** First add the event to the list
  */
  
  if(event_table[event->event_kind] == NULL) {      /* First one in the list */
    event_table[event->event_kind] = event;
    event->next = event;
    event->prev = event;
  } else {                                          /* Put it on the end of the list */
    event->next = event_table[event->event_kind];
    event->prev = event_table[event->event_kind]->prev;
    event->prev->next = event;
    event_table[event->event_kind]->prev = event;
  }

  /* 
  ** Then put it in the hashtable
  */

  if(event_hashtable[hash_key] == NULL) {           /* first one in the list */
    event_hashtable[hash_key] = event;
    event->hash_next = event;
    event->hash_prev = event;
  } else {                                          /* Put it on the end of the list */
    event->hash_next = event_hashtable[hash_key];
    event->hash_prev = event_hashtable[hash_key]->hash_prev;
    event->hash_prev->hash_next = event;
    event_hashtable[hash_key]->hash_prev = event;
  }
}


/*
** Throw away an event. Events have a unique ID, so we look for the event with this 
** specific ID. Removing an event also takes place in 2 steps. First it gets taken 
** out of the linked list of events of the same kind. Then it gets deleted in the 
** hashtable.
*/

w_void remove_event(w_int ID) {
  jdwp_event event = get_event_ID(ID);
  w_int      hash_key = event->eventID % EVENT_HASHTABLE_SIZE;

  /* 
  ** First remove it from the list 
  */

  if(event == event_table[event->event_kind]) {     /* First one on the list */
    if(event == event->next) {                      /* And also the last one */
      event_table[event->event_kind] = NULL;
    } else {
      event_table[event->event_kind] = event->next;
    }
  }
  
  event->next->prev = event->prev;
  event->prev->next = event->next;

  /*
  ** Then remove it from the hashtable
  */
  
  if(event == event_hashtable[hash_key]) {          /* First one on the list */
    if(event == event->hash_next) {                 /* And also the last one */
      event_hashtable[hash_key] = NULL;
    } else {
      event_hashtable[hash_key] = event->hash_next;
    }
  }
  
  event->hash_next->hash_prev = event->hash_prev;
  event->hash_prev->hash_next = event->hash_next;
}


/*
** Lookup an event. Every event has a unique ID and we already have a hashtable 
** with all events and hashed with there ID, so this is fairly straight forward.
*/

jdwp_event get_event_ID(w_int ID) {
  w_int      hash_key = ID % EVENT_HASHTABLE_SIZE;
  jdwp_event result = NULL;  
  jdwp_event current = event_hashtable[hash_key];  
 
  if(current == NULL) return NULL;                  /* Nothing here... */
  if(current->eventID == ID) return current;        /* First in line */  
    
  while((current->hash_next != event_hashtable[hash_key]) && (result == NULL)) {
    if(current->eventID == ID) result = current;
    current = current->hash_next;
  }
    
  return result;
}


/*
** Add an event modifier to a given event. These modifiers are stored in a 
** circular linked list in the event. Adding elements is easy.
*/

w_void add_modifier(jdwp_event event, jdwp_event_modifier modifier) {
  
  if(event->modifiers == NULL) {                    /* First one in the list */
    event->modifiers = modifier;
    event->modifiers->next = event->modifiers;
    event->modifiers->prev = event->modifiers;
  } else {                                          /* Put it on the end of the list */
    modifier->next = event->modifiers;
    modifier->prev = event->modifiers->prev;
    modifier->prev->next = modifier;
    event->modifiers->prev = modifier;
  }
}


/*
** The current event ID. Every event has it's own unique ID, so whenever a new event is 
** created, this counter will be incremented.
*/

static w_int current_eventID = 0;


/*
** Allocate a new event structure, give it an ID, store the eventkind and suspendpolicy.
*/

jdwp_event allocate_event(w_int event_kind, w_int suspend_policy) {
  jdwp_event event = x_mem_alloc(sizeof(jdwp_Event));
  
  current_eventID++;
  
  event->eventID = current_eventID;
  event->event_kind = event_kind;
  event->suspend_policy = suspend_policy;
  event->modifiers = NULL;
  
  return event;
}


/*
** Clear all events.
*/

w_void jdwp_clear_all_events() {
  w_int i;
  jdwp_event event;
  jdwp_event temp;

  for(i=0; i<255; i++) {
    if(event_table[i]) {
      event = event_table[i];
      event->prev->next = NULL;
      
      while(event) {
        temp = event->next;

        if(event->event_kind == jdwp_evt_breakpoint) {
          jdwp_clear_breakpoint(event);
        }

        /* TODO: modifiers ! */
        
        x_mem_free(event);
        
        event = temp;
      }
    
      event_table[i] = NULL;
    }
  }
  
  for(i=0; i<EVENT_HASHTABLE_SIZE; i++) {
    event_hashtable[i] = NULL;
  }

}


/*
** Add a breakpoint to the breakpoint hashtable.
*/

w_void add_breakpoint(jdwp_breakpoint breakpoint) {
  w_word hash_key = (w_word)breakpoint->code % BREAKPOINT_HASHTABLE_SIZE;
  
  if(breakpoint_hashtable[hash_key] == NULL) {      /* first one in the list */
    breakpoint_hashtable[hash_key] = breakpoint;
    breakpoint->next = breakpoint;
    breakpoint->prev = breakpoint;
  } else {                                          /* Put it on the end of the list */
    breakpoint->next = breakpoint_hashtable[hash_key];
    breakpoint->prev = breakpoint_hashtable[hash_key]->prev;
    breakpoint->prev->next = breakpoint;
    breakpoint_hashtable[hash_key]->prev = breakpoint;
  }
}


/*
** Delete a breakpoint from the breakpoint hashtable.
*/

w_void remove_breakpoint(jdwp_breakpoint breakpoint) {
  w_word hash_key = (w_word)breakpoint->code % BREAKPOINT_HASHTABLE_SIZE;

  if(breakpoint == breakpoint_hashtable[hash_key]) {   /* First one on the list */
    if(breakpoint == breakpoint->next) {              /* And also the last one */
      breakpoint_hashtable[hash_key] = NULL;
    } else {
      breakpoint_hashtable[hash_key] = breakpoint->next;
    }
  }
  
  breakpoint->next->prev = breakpoint->prev;
  breakpoint->prev->next = breakpoint->next;
}


/*
** Search for a breakpoint in the breakpoint hashtable.
*/

jdwp_breakpoint get_breakpoint(w_code code) {
  w_word          hash_key = (w_word)code % BREAKPOINT_HASHTABLE_SIZE;
  jdwp_breakpoint result = NULL;  
  jdwp_breakpoint current = breakpoint_hashtable[hash_key];  
 
  if(current == NULL) return NULL;                  /* Nothing here... */
  if(current->code == code) return current;         /* First in line */  
    
  while((current->next != breakpoint_hashtable[hash_key]) && (result == NULL)) {
    if(current->code == code) result = current;
    current = current->next;
  }
    
  return result;
}


/*
** Send an event to the debugger.
*/

w_void jdwp_send_event(jdwp_event event, w_void *data, w_int size) {
  w_ubyte  *reply;
  w_int    position = 0;
  w_int    count = 1;

  /*
  ** Allocate memory for the event data.
  **
  ** Normally JDWP uses composite events. This means that certain event types will be 
  ** grouped together and send to the debugger in one packet. Since this behaviour is 
  ** rather hard, we send events one by one.
  **
  ** The overal event packet looks like this :
  **
  **   byte   suspend_policy   The suspend policy used for this event.
  **   int    events           The number of events in the packet. In our case this is
                               always 1.
  **   byte   event_kind       The event kind of the event.
  **   int    eventID          The ID of the event.
  **   ...    ...              Event specific data.
  */

  reply = x_mem_alloc((w_word)(1 + 4 + 1 + 4 + size + 1));

  /*
  ** Store the suspend policy.
  */

  reply[position] = event->suspend_policy;
  position += 1;

  /*
  ** Store the number of events. (Always 1).
  */

  count = swap_int(count);
  memcpy(&reply[position], &count, 4);
  position += 4;

  /*
  ** Store the event kind.
  */

  reply[position] = event->event_kind;
  position += 1;

  /*
  ** Store the eventID.
  */

  memcpy(&reply[position], &(event->eventID), 4);
  position += 4;

  /*
  ** Add the event specific data to the packet.
  */

  memcpy(&reply[position], data, (w_word)size);
  position += size;

  /*
  ** Send the whole bunch to the debugger in the form of a command packet.
  */
  
  jdwp_send_command(reply, position, jdwp_cmdset_event, jdwp_event_composite);

  /*
  ** Clean up.
  */

  x_mem_free(reply);

  return;
}


/*
** Suspend all threads
*/

w_void jdwp_suspend_all_threads(void) {
  w_fifo    fifo;
  w_thread  thread = NULL;
  w_int     length;

  /*
  ** Use a fifo to get all the threads out of the thread_hashtable.
  */

  fifo = ht_list_values(thread_hashtable);
  length = fifo->numElements;

  /*
  ** Loop through the list and suspend them one by one.
  */

  while((thread = getFifo(fifo)) != NULL) {
    if(thread->kthread != jdwp_thread) {
      x_thread_suspend(thread->kthread);
    }
  }

  /*
  ** Release memory used by the fifo.
  */

  releaseFifo(fifo);

  return;
}


/*
** Check the suspend policy and act accordingly.
** There are 3 different suspend policies :
**   1. none           No threads will be suspended when the event occurs.
**   2. event_thread   Only the thread in which the event occured will be suspended.
**   3. all            All threads will be suspended.
*/

w_void jdwp_check_suspend(jdwp_event event) {

  switch(event->suspend_policy) {
    case jdwp_sp_none:          break;
    case jdwp_sp_event_thread:  x_thread_suspend(currentWonkaThread->kthread);
                                break;
    case jdwp_sp_all:           jdwp_suspend_all_threads();
                                break;
  }

  return;
}


/*
** Set a breakpoint.
*/

w_void jdwp_set_breakpoint(jdwp_event event) {
  jdwp_event_modifier  modifier = event->modifiers;
  w_long               pc;
  w_method             method;
  w_clazz              clazz;
  jdwp_breakpoint      point;
  
  woempa(9, "--== Setting a breakpoint ==--\n");

  /*
  ** Loop through the modifiers of the event to find the location_only modifier.
  ** This modifiers contains the clazz, method and pc where the breakpoint should be set.
  */
  
  while(modifier->mod_kind != 7) modifier = modifier->next;

  /*
  ** Get clazz, method and pc out of the location modifier.
  */
 
  memcpy(&clazz, &modifier->condition.location[1], 4);
  memcpy(&method, &modifier->condition.location[5], 4);
  memcpy(&pc, &modifier->condition.location[9], 8);
  pc = swap_long(pc);

  woempa(9, "    clazz: %w,  method: %w,  pc: %Ld,  original opcode: 0x%x\n", clazz->dotified, method->name, pc, method->code[pc]);

  /*
  ** Allocate memory for a breakpoint structure.
  */

  point = x_mem_alloc(sizeof(jdwp_Breakpoint));

  /*
  ** Fill in the breakpoint structure.
  */

  point->code = &method->code[pc];       /* The address of the replaced opcode */
  point->original = method->code[pc];    /* The original opcode that was replaced by a breakpoint */
  point->event = event;                  /* The event that requested this breakpoint to be set */
  memcpy(&point->location[0], &modifier->condition.location[0], 17);

  /*
  ** Store the breakpoint in the event as well.
  */
  
  event->break_point = point;

  /*
  ** Replace the original opcode with a breakpoint opcode (0xca).
  */

  method->code[pc] = 0xca;

  /*
  ** Add this breakpoint structure to the hashtable of breakpoints.
  */

  add_breakpoint(point);

}


/*
** Clear a breakpoint.
*/

w_void jdwp_clear_breakpoint(jdwp_event event) {
  w_method             method;
  w_clazz              clazz;
  
  woempa(9, "--== Clearing a breakpoint ==--\n");

  /*
  ** Get the clazz and method out of the event->breakpoint.
  */

  memcpy(&clazz, &event->break_point->location[1], 4);
  memcpy(&method, &event->break_point->location[5], 4);

  woempa(9, "    clazz: %w,  method: %w\n", clazz->dotified, method->name);

  /*
  ** Put the original opcode back where it belongs.
  */

  *event->break_point->code = event->break_point->original;

  /*
  ** Remove the breakpoint from the hashtable.
  */
  
  remove_breakpoint(event->break_point);
  
  /* 
  ** Clean up.
  */
  
  x_mem_free(event->break_point);
}


/*
** From this point on are the event handlers. Wonka calls these to let us know 
** something happened. We then need to look in our tables for a matching event 
** request. If one is found, the event gets through to the debuggger. If there 
** aren't any matching event requests stored, this particular event will be 
** ignored.
*/


/*
** A class has been prepared
*/

w_void jdwp_event_class_prepare(w_clazz clazz) {
  jdwp_event           events = event_table[jdwp_evt_class_prepare];
  jdwp_event_modifier  modifier;
  w_int                go_ahead;
  w_thread             thread = currentWonkaThread;
  w_ubyte              *reply;
  w_int                position = 0;
  w_int                ID;
  w_int                length;
  w_ubyte              reftype;
  w_ubyte              *sig;

  /*
  ** Only go through with this if the events are enabled.
  */

  if(!jdwp_events_enabled) return;

  /*
  ** Do some locking for sanity.
  */

  x_mutex_lock(jdwp_mutex, x_eternal);
  
  /*
  ** First check if there are any events requested for this event kind. 
  ** If there are any events, go through them one by one.
  */

  if(events) {
    
    do {

      go_ahead = 1;

      /*
      ** Check the modifiers.
      */

      modifier = events->modifiers;
/*
      if(modifier) {

        do {

          switch(modifier->mod_kind) {
            case 5:   sig = jdwp_stringcstring(clazz->desc->name, &length);
                      x_mem_free(sig);
                      
                      break;
            default:  go_ahead = 0;
          }
        
          modifier = modifier->next;

        } while(modifier != events->modifiers);
        
      }
*/
      if(go_ahead) {

        /*
        ** The modifiers didn't block us from sending this event, so we go ahead
        ** and send it to the debugger.
        */

        /*
        ** Allocate some memory for the event data. The data for this event looks like this :
        **   threadID   thread        The thread in which the class was prepared.
        **   byte       reftype_tag   The kind of reference type.
        **   reftypeID  typeID        The reference type.
        **   string     signature     The signature of the type.
        **   int        status        The classstatus.
        */

        reply = x_mem_alloc((w_word)(4 + 1 + 4 + 100 + 4));
        position = 0;

        /*
        ** Store the thread ID.
        */

        ID = (w_int)(thread->Thread);
        memcpy(&reply[position], &ID, 4);
        position += 4;

        /*
        ** Store the reference type tag.
        */

        if (isSet(clazz->flags, ACC_INTERFACE)) 
          reftype = jdwp_tt_interface;
        else 
          reftype = jdwp_tt_class;

        memcpy(&reply[position], &reftype, 1);
        position += 1;

        /*
        ** Store the type ID.
        */

        ID = (w_int)(clazz);
        memcpy(&reply[position], &ID, 4);
        position += 4;

        /*
        ** Store the signature.
        */

        sig = jdwp_string2UTF8(clazz->desc->name, &length);
        memcpy(&reply[position], sig, (w_word)(length + 4));
        position += length + 4;
        x_mem_free(sig);

        /*
        ** Store the status of the class.
        */

        length = swap_int((jdwp_cs_verified | jdwp_cs_prepared | jdwp_cs_initialized));
        memcpy(&reply[position], &length, 4);
        position += 4;

        /*
        ** Send the event to the debugger.
        */
        
        jdwp_send_event(events, reply, position);

        /*
        ** Clean up.
        */

        x_mem_free(reply);

        /*
        ** Check the suspend status of the event and act accordingly.
        */

        jdwp_check_suspend(events);
        
      }

      /*
      ** Go to the next event in the list.
      */
      
      events = events->next;
      
    } while(events != event_table[jdwp_evt_class_prepare]);
    
  }

  /*
  ** Unlock.
  */
  
  x_mutex_unlock(jdwp_mutex);
  
  return;
}


/*
** A thread has started.
*/

w_void jdwp_event_thread_start(w_thread thread) {
  jdwp_event           events = event_table[jdwp_evt_thread_start];
  jdwp_event_modifier  modifier;
  w_int                go_ahead;
  
  /*
  ** Only go through with this if the events are enabled.
  */

  if(!jdwp_events_enabled) return;

  /*
  ** Do some locking for sanity.
  */

  x_mutex_lock(jdwp_mutex, x_eternal);
  
  /* 
  ** Check if there are any events of this type requested. If there are, go over them
  ** one by one and check if we need to send an event.
  */

  if(events) {
    
    do {

      go_ahead = 1;

      /*
      ** Get the modifiers and go over them one by one to check if we need to send
      ** this event.
      */
      
      modifier = events->modifiers;

      if(modifier) {

        do {
        
          modifier = modifier->next;

        } while(modifier != events->modifiers);
        
      }

      if(go_ahead) {

        /*
        ** None of the modifiers blocked this event from being sent, so we can go ahead 
        ** and send an event to the debugger.
        */
        
        jdwp_send_event(events, &(thread->Thread), 4);

        /*
        ** Check the suspend status and act accordingly.
        */
        
        jdwp_check_suspend(events);
        
      }

      /*
      ** Go to the next event in the list.
      */

      events = events->next;

    } while(events != event_table[jdwp_evt_thread_start]);

  }

  /*
  ** Unlock.
  */

  x_mutex_unlock(jdwp_mutex);
  
  return;
}


/*
** A thread has ended.
*/

w_void jdwp_event_thread_end(w_thread thread) {
  jdwp_event           events = event_table[jdwp_evt_thread_end];
  jdwp_event_modifier  modifier;
  w_int                go_ahead;
  
  /*
  ** Only go through with this if the events are enabled.
  */

  if(!jdwp_events_enabled) return;

  /*
  ** Do some locking for sanity.
  */

  x_mutex_lock(jdwp_mutex, x_eternal);
  
  /* 
  ** Check if there are any events of this type requested. If there are, go over them
  ** one by one and check if we need to send an event.
  */

  if(events) {
    
    do {

      go_ahead = 1;

      /*
      ** Get the modifiers and go over them one by one to check if we need to send
      ** this event.
      */
      
      modifier = events->modifiers;

      if(modifier) {

        do {
        
          modifier = modifier->next;

        } while(modifier != events->modifiers);
        
      }

      if(go_ahead) {

        /*
        ** None of the modifiers blocked this event from being sent, so we can go ahead 
        ** and send an event to the debugger.
        */
        
        jdwp_send_event(events, &(thread->Thread), 4);

        /*
        ** Check the suspend status and act accordingly.
        */
        
        jdwp_check_suspend(events);
        
      }

      /*
      ** Go to the next event in the list.
      */

      events = events->next;

    } while(events != event_table[jdwp_evt_thread_end]);

  }

  /*
  ** Unlock.
  */

  x_mutex_unlock(jdwp_mutex);
  
  return;
}


/*
** We've hit a breakpoint !
*/

w_ubyte jdwp_event_breakpoint(w_ubyte *code) {
  jdwp_event           event;
  jdwp_breakpoint      point;
  w_instance           instance = currentWonkaThread->Thread;
  w_int                thread = (w_int)instance;
  w_ubyte              *reply;
  
  /*
  ** Only go through with this if the events are enabled.
  */

  if(!jdwp_events_enabled) return 0;

  /*
  ** Do some locking for sanity.
  */

  x_mutex_lock(jdwp_mutex, x_eternal);
  
  woempa(9, " -= EVENT =-    B R E A K P O I N T\n");

  /*
  ** Lookup the breakpoint in the hashtable and also get the event.
  */
 
  point = get_breakpoint(code);
  event = point->event;

  /* 
  ** Allocate memory for the event data. The data is organized like this:
  **   threadID  thread     The thread in which the breakpoint was hit.
  **   location  location   The location of the breakpoint.
  */

  reply = x_mem_alloc(4 + 17); 

  /*
  ** Store the thread ID.
  */

  memcpy(&reply[0], &thread, 4);

  /*
  ** Store the location.
  */
  
  memcpy(&reply[4], &point->location[0], 17);

  /*
  ** Send the event to the debugger.
  */
  
  jdwp_send_event(event, reply, 4 + 17);

  /*
  ** Check the suspend policy and act accordingly.
  */
        
  jdwp_check_suspend(event);

  /*
  ** Clean up.
  */

  x_mem_free(reply);

  /*
  ** Unlock.
  */

  x_mutex_unlock(jdwp_mutex);

  /*
  ** Return the original opcode. This is used in the interpreter loop where the breakpoint
  ** opcode will be overwritten with the original opcode. Then the pc is turned back by one
  ** so the interpreter will go over the original opcode. Then the breakpoint will be written
  ** over the original opcode once again.
  */

  return point->original;
}

