/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/*
** $Id: sem.c 3684 2003-03-19 14:20:16Z gray $
*/    
 
#include "oswald.h"

/*
 * Prototype:
 *   x_status x_sem_create(x_sem semaphore, 
 *                            w_size initial_count);
 * Description (of counting semaphores):
 *   Creates a counting semaphore for inter-thread synchronization.
 *   The initial semaphore count is specified as an input parameter.
 *   A semaphore can be used to guard accesses to a resource, or,
 *   alternatively, just to let processes wait for something to
 *   happen.  The 'x_sem_get' and 'x_sem_put' operations
 *   basicaly decrement and increment an integer count associated with
 *   the semaphore (hence the name).  When a process waits on a
 *   semaphore whose count is positive, then the process just 
 *   continues, having decremented the semaphore count possibly to
 *   zero.  If the semaphore count is zero, then the waiting process
 *   is blocked (suspended) until another process posts the semaphore
 *   by incrementing the semaphore count.
 * Implementation:
 *   Although POSIX.4 provides counting semaphores not all operating systems
 *   provide them and they are optimised for communication between processes
 *   Using a mutex and a condition variable is more likely to be supported
 *   and more efficient.
 */ 
 
x_status x_sem_create(x_sem semaphore, w_size initial_count) {

  x_status rval = xs_success;
  pthread_mutexattr_t attr;

  semaphore->deleted = 0;
  semaphore->count = initial_count;
  semaphore->waiters = 0;
  semaphore->owner = NULL;

  pthread_mutexattr_init(&attr);
  pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_NORMAL);

  pthread_mutex_init(&semaphore->sem_mutex , &attr);
  pthread_cond_init(&semaphore->sem_cond, NULL);

  return rval;

}

/*
 * Prototype:
 *   x_status x_sem_delete(x_sem semaphore);
 * Description:
 *   The 'x_sem_delete' operation deletes the specified 
 *   counting semaphore.
 * Implementation:
 *   Although POSIX.4 provides counting semaphores not all operating systems
 *   provide them and they are optimised for communication between processes
 *   Using a mutex and a condition variable is more likely to be supported
 *   and more efficient.
 */
 
x_status x_sem_delete(x_sem semaphore) {

  x_status rval = xs_success;
  
  if (pthread_mutex_lock(&semaphore->sem_mutex) !=0)
    abort();
  
  semaphore->deleted = 1;
  if (semaphore->waiters > 0) {
    rval = xs_competing;
    pthread_cond_broadcast(&semaphore->sem_cond);
  }

  if (pthread_mutex_unlock(&semaphore->sem_mutex) !=0)
    abort();

  return rval;

}

/*
 * Prototype:
 *   x_status x_sem_get(x_sem semaphore, 
 *                         x_sleep wait_option);
 * Description: 
 *   The get operation decreases the semaphore by one.  If the 
 *   semaphore is 0, the get operation will not be succesful.
 *   Selecting x_no_wait as wait option results in an immidiate
 *   return from this call regardless of whether or not it was
 *   succesful, while selecting x_eternal causes the calling
 *   thread to suspend until a semaphore instance is available.
 * Implementation:
 *   Although POSIX.4 provides counting semaphores not all operating systems
 *   provide them and they are optimised for communication between processes
 *   Using a mutex and a condition variable is more likely to be supported
 *   and more efficient.
 */

x_status x_sem_get(x_sem semaphore, x_sleep owait) {

  x_thread current = x_thread_current();
  int have_it = 0;

  if (pthread_mutex_lock(&semaphore->sem_mutex) !=0)
    abort();
  
  if (semaphore->count > 0) {
    semaphore->count -= 1;
    semaphore->owner = current;
    if (pthread_mutex_unlock(&semaphore->sem_mutex) != 0) 
      abort();
    return xs_success;
  }
  else {

    /*
    ** If we have to wait for it, indefinitely or for a while,
    ** we unlock the mutex and wait for it, otherwise
    ** xs_no_instance is returned.
    */

    if (owait == x_eternal || owait != x_no_wait) {
      semaphore->waiters += 1;
      while (! have_it) {
        if (owait == x_eternal) {
          if (pthread_cond_wait(&semaphore->sem_cond, &semaphore->sem_mutex) != 0)
            abort();
        }
        else {
          struct timespec ts;
          int ret;

          x_now_plus_ticks(owait, &ts);
          
          ret = pthread_cond_timedwait(&semaphore->sem_cond, &semaphore->sem_mutex, &ts);
          if (ret == ETIMEDOUT)
            return xs_no_instance;
          else if (ret != 0)
            abort();
        }

        if (semaphore->deleted) {
          pthread_mutex_unlock(&semaphore->sem_mutex);
          return xs_deleted;
        }

        if (semaphore->count > 0) {
          semaphore->count -= 1;
          semaphore->owner = current;
          have_it = 1;
        }
      }
      semaphore->waiters -= 1;
    }
    if (pthread_mutex_unlock(&semaphore->sem_mutex) != 0)
      abort();
  }
  
  return xs_success;

}

/*
 * Prototype:
 *   x_status x_sem_put(x_sem semaphore);
 * Description:
 *   This service, which is the inverse of the 'x_sem_get'
 *   operation, increases the semaphore by one.  
 * Implementation:
 *   Although POSIX.4 provides counting semaphores not all operating systems
 *   provide them and they are optimised for communication between processes
 *   Using a mutex and a condition variable is more likely to be supported
 *   and more efficient.
 */
 
x_status x_sem_put(x_sem semaphore) {
  x_status rval = xs_success;

  pthread_mutex_lock(&semaphore->sem_mutex);
  semaphore->count += 1;
  if (semaphore->waiters) {
    pthread_cond_signal(&semaphore->sem_cond);
  }
  pthread_mutex_unlock(&semaphore->sem_mutex);

  return rval;
}

/*
 * Prototype:
 *   x_status x_sem_wait(x_sem semaphore);
 * Description:
 * Implementation:
 *   Although POSIX.4 provides counting semaphores not all operating systems
 *   provide them and they are optimised for communication between processes
 *   Using a mutex and a condition variable is more likely to be supported
 *   and more efficient.
 */
 
x_status x_sem_wait(x_sem semaphore) {
  x_status rval = xs_success;
  x_thread current = x_thread_current();

  pthread_mutex_lock(&semaphore->sem_mutex);

  while (semaphore->count == 0) {
    semaphore->waiters++;
    pthread_cond_wait(&semaphore->sem_cond, &semaphore->sem_mutex);
    if (semaphore->deleted) {
      pthread_mutex_unlock(&semaphore->sem_mutex);
      return xs_deleted;
    }
    semaphore->waiters--;
  }
  semaphore->count--;
  semaphore->owner = current;

  pthread_mutex_unlock(&semaphore->sem_mutex);

  return rval;

}

#if 0
/*
 * Prototype:
 *   x_status x_sem_post(x_sem semaphore);
 * Description:
 * Implementation:
 *   Although POSIX.4 provides counting semaphores not all operating systems
 *   provide them and they are optimised for communication between processes
 *   Using a mutex and a condition variable is more likely to be supported
 *   and more efficient.
 */
 
x_status x_sem_post(x_sem semaphore) {
  x_status rval = xs_success;

  pthread_mutex_lock(&semaphore->sem_mutex);

  semaphore->count++;
  if (semaphore->waiters > 0) {
    pthread_cond_broadcast(&semaphore->sem_cond);
  }

  pthread_mutex_unlock(&semaphore->sem_mutex);

  return rval;
}
#endif
