/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: host.c 2908 2002-10-01 12:00:12Z gray $
*/

#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <sys/mman.h>

#include <oswald.h>

#define HOST_TIMER_SIGNAL  SIGALRM 
#define HOST_TIMER_TIMER   ITIMER_REAL 

/* 
** The default number of microseconds per timer interrupt. For NetBSD, don't make
** it too small since the kernel would not be able to service it. Note that the
** following value was experimentally found to be a correct number for most systems. Your mileage
** may vary...
*/

static x_size usecs_per_tick = 50000;

static struct sigaction segfault;
static struct sigaction tick;
static struct sigaction oldtick;
static struct itimerval ticker;
static struct itimerval oldtimer;
static sigset_t unblock_timersig;
static unsigned int pagesize;

sigset_t unblocked_signals;

static x_ubyte * host_memory;

x_ubyte * host_memory_end;

x_size max_heap_bytes = 0;

extern void stack_segfault_handler(int signum);

volatile x_word preemptive_switch = 0;

void start_host_timer(void);

static void host_timer_tick(int sig) {

  /*
  ** Avoid timer tick when we are booting or too excessive irq depth.
  */
  
  if (irq_depth > 10) {
    return;
  }
  
  irq_depth++;
  (*irq_handlers[IRQ_TICK]->top)(irq_handlers[IRQ_TICK]);
  irq_depth--;

  /*
  ** An interrupt handler can change thread_next, so we call xi_thread_reschedule 
  ** that will do the correct thing. We reset preemptive switch in case that no thread 
  ** switch took place so that the next non preempted switch will not do go through the
  ** if clause in x_host_post.
  */

  if (critical_status == 0) {
    preemptive_switch = 1;
    xi_thread_reschedule();
    preemptive_switch = 0;
  }

}
 
void x_host_post(void) {

  if (preemptive_switch) {

    /*
    ** Unblock the timer signal! At entry of the above handler, the kernel has disabled the fired signal
    ** and will only unblock it when we leave the handler. Since we leave through a thread
    ** switch, we would block the next delivery of the timer signal. The thread switch routine calls this
    ** post handler to unblock the signal again.
    */

    sigprocmask(SIG_UNBLOCK, &unblock_timersig, NULL); 
    //sigprocmask(SIG_SETMASK, &unblocked_signals, NULL);
    preemptive_switch = 0;

  }

  if (runtime_checks && thread_current->trigger && *thread_current->trigger != 0xaaaaaaaa) {
    loempa(9, "Trigger value is 0x%0x for thread %d.\n", *thread_current->trigger, thread_current->id);
    abort();
  }

//  x_thread_switched(thread_current);

}

void start_host_timer () {

  x_int result;

  ticker.it_interval.tv_sec = usecs_per_tick / 1000000;
  ticker.it_interval.tv_usec = usecs_per_tick % 1000000;
  ticker.it_value.tv_sec = usecs_per_tick / 1000000;
  ticker.it_value.tv_usec = usecs_per_tick % 1000000;
  result = setitimer(HOST_TIMER_TIMER, &ticker, &oldtimer);
  if (result == -1) {
    loempa(9, "%s %d: setitimer failed.\n", __FILE__, __LINE__);
    abort();
  }

}

inline static unsigned int round_up(unsigned int value, unsigned int rounding) {
  return (value + (rounding - 1)) & ~(rounding - 1);
}

inline static unsigned int round_down(unsigned int value, unsigned int rounding) {  
  return (value) & (unsigned int)(0 - rounding);
}

x_status x_stack_delete(x_stack stack) {
#if defined(__MACHINE_STACK_GROWS_UP)
  munmap(stack->first, (stack->capacity + 1) * pagesize);
#else
  munmap(stack->final, (stack->capacity + 1) * pagesize);
#endif
  return xs_success;
}

x_status x_stack_create(x_stack stack, unsigned int size, x_thread thread) {

  unsigned char * addr;
  unsigned char * top;
  unsigned char * bottom;
  
  size = round_up(size, pagesize);
  stack->thread = thread;
  stack->capacity = size / pagesize;
  loempa(9, "Stack capacity is %d\n", stack->capacity);

  x_preemption_disable;
  
  /*
  ** Let NetBSD decide where she wants to allocate this virtual range, in one swoop. We
  ** will add guarding pages later. Allocate 1 extra page for the final guard page in
  ** the same way the Linux code does for now. This is apparently a workaround in case
  ** somebody wantsto use pthreads in some native code.
  **
  ** The mapping we do first, for the complete area, will have its protection updated so
  ** that its one large read/write area with read only guard pages surrounding it.
  */
  
  addr = mmap(NULL, (stack->capacity + 1) * pagesize, PROT_NONE, MAP_ANON | MAP_PRIVATE, -1, 0);
  loempa(9, "would get %d bytes starting from range %p\n", (stack->capacity + 1) * pagesize, addr);

  /*
  ** Make the entire stack read/write; the top pointer of the page is rounded down/up to a word pointer
  ** and is the starting point for the stack, on which elements are pushed down/up to lower/higher
  ** addresses.
  */
  
  if (mprotect(addr, (stack->capacity) * pagesize, PROT_READ | PROT_WRITE) < 0) {
    x_preemption_enable;
    return xs_no_mem;
  }
  bottom = addr;
  top = bottom + stack->capacity * pagesize;

#if defined(__MACHINE_STACK_GROWS_UP)
  stack->first = bottom
  stack->final = stack->first + (pagesize * stack->capacity);
#else
  stack->first = top;
  stack->final = stack->first - (pagesize * stack->capacity);
#endif
  loempa(9, "start would then be %p\n", stack->first);

  stack->used = stack->capacity;
  // XXX NAH stack->top = (unsigned int *) round_down((unsigned int)(stack->first + (pagesize) - 1), 4);

  /*
  ** Allocate a page that acts as the unmoveable barrier. This page signals the
  ** end of the road in terms of stack...
  */
  
  stack->guard = stack->final;
  if (mprotect(stack->final, pagesize, PROT_READ) < 0) {
    x_preemption_enable;
    return xs_no_mem;
  }

  loempa(9, "OK, top %p, first %p, guard %p, final %p, capacity %d\n", stack->top, stack->first, stack->guard, stack->final, stack->capacity);

  x_preemption_enable;

  return xs_success;

}

void install_stack_handler(void) {

  x_int result;

  /*
  ** Set up our segmentation violation handler.
  */

  segfault.sa_handler = stack_segfault_handler;
  sigemptyset(&segfault.sa_mask);
  result = sigaction(SIGSEGV, &segfault, NULL);
  if (result == -1) {
    loempa(9, "%s %d: sigaction failed.\n", __FILE__, __LINE__);
    abort();
  }

  loempa(9, "Installed new segfault handler new, %p\n", segfault.sa_handler);

}

void NEWstack_segfault_handler(int signum) {
  write(1, "BBBB", 4);
}
  
void stack_segfault_handler(int signum) {

  x_stack stack = NULL;

  if (thread_current == NULL) {
    return;
  }

  loempa(9, "Segfault caught\n");

  //stack = thread_current->stack.v_stack;

  x_preemption_disable;
  
  if (stack->guard == stack->final) {
    loempa(9, "Givin up...\n");
    abort();
  }
  else {
    loempa(9, "Something has gone wrong...\n");
    abort();
  }

  loempa(9, "Used %d pages\n", stack->used);

  x_preemption_enable;
}

x_ubyte * x_irqs_setup(x_ubyte *mem) {

  x_size i;
  
  irq_handlers = x_alloc_static_mem(mem, sizeof(x_irq) * IRQ_MAX);

  for (i = 0; i < IRQ_MAX; i++) {
    irq_handlers[i] = irq_default;
  }
  
  irq_handlers[IRQ_TICK] = irq_tick;

  return mem;
  
}

x_ubyte * x_host_setup(x_ubyte *memory) {

  int result;

  /*
  ** Set up our timer tick interval timer and signal handler.
  */

  tick.sa_handler = host_timer_tick;
  sigemptyset(&tick.sa_mask);
  tick.sa_flags = SA_RESTART;

  result = sigaction(HOST_TIMER_SIGNAL, &tick, &oldtick);
  if (result == -1) {
    loempa(9, "%s %d: sigaction failed.\n", __FILE__, __LINE__);
    abort();
  }

  start_host_timer();

  /*
  ** Get currently unblocked signals to unblock them again in timer handler...
  */
    
  sigemptyset(&unblock_timersig);
  sigaddset(&unblock_timersig, HOST_TIMER_SIGNAL);

  sigprocmask(SIG_UNBLOCK, NULL, &unblocked_signals);

  return memory;
  
}

#define MIN_TRY (1024 * 1024 * 4)
#define TRIM_FACTOR (6)

#define  ext_malloc  malloc
#define  ext_free    free
#define  ext_realloc realloc

x_status x_oswald_init(x_size max_heap, x_size millis) {

  x_size heap = max_heap;
  const x_size decrement = (1024 * 100);

  loempa(1, "max_heap = %d bytes.\n", heap);
  critical_status = 1;
  max_heap_bytes = max_heap;

  usecs_per_tick = millis * 1000;
  
  /*
  ** Let's try to find out what memory we can take from this box. We try with max_heap and work our way
  ** down when that doesn't work. The minimum is MIN_TRY.
  */
  
  do {
    host_memory = ext_malloc(heap);
    heap -= decrement;
  } while (host_memory == NULL && heap > MIN_TRY);
  heap += decrement;

  if (host_memory == NULL) {
    loempa(9, "Could not even get %d bytes from host.\n", heap);
    return xs_no_mem;
  }

  /*
  ** We now trim down 10% since on a PDA
  ** like the Yopi, we get killed if the system gets to thight...
  */
  
  loempa(9, "Trimming available memory down from %d to %d bytes.\n", heap, heap - (heap / TRIM_FACTOR));
  heap -= (heap / TRIM_FACTOR);
  ext_free(host_memory);
  host_memory = ext_malloc(heap);
  if (host_memory == NULL) {
    loempa(9, "No host memory available.\n");
    return xs_no_mem;
  }

  loempa(9, "Host memory starts at 0x%08x, %d bytes available.\n", host_memory, heap);  
  host_memory_end = host_memory + heap - 1;

  pagesize = getpagesize();

  x_kernel_setup(host_memory);

  loempa(9, "Should not get here!\n");

  return xs_bad_argument;
  
}

/*
** Functions to convert seconds, microseconds to ticks and vice versa.
*/

static const x_size ticks_compensation = 2;

/*
** Return a compensated number of ticks for a number of seconds.
*/

x_size x_seconds2ticks(x_size seconds) {
  return (x_size) ((seconds * 1000) / (x_size) (usecs_per_tick / 1000) - ticks_compensation);
}

x_size x_millis2ticks(x_size millis) {

  x_size ticks = (millis / (x_size) (usecs_per_tick / 1000));
  
  return ticks ? ticks : 1;
  
}

x_size x_ticks2usecs(x_size ticks) {
  return (usecs_per_tick * ticks);
}

x_size x_usecs2ticks(x_size usecs) {
  return (usecs <= usecs_per_tick) ? 1 : ((x_size)((usecs-1) / usecs_per_tick)+1);
}

int my_error;

int * __errno_location(void) {
  if (thread_current){
    return (int *) & thread_current->tsd_errno;
  }
  else {
    return (int *) & my_error;
  }
}

void showhandler(void) {
  struct sigaction sf;
  sigaction(SIGSEGV, NULL, &sf);
  loempa(9, "handler = %p\n", sf.sa_handler);
}

x_ubyte * x_host_sbrk(x_int bytes) {
 return sbrk(bytes);
}

void x_host_break(x_ubyte * memory) {

  loempa(9, "Reallocing host static memory down to %d bytes.\n", memory - host_memory);
  ext_realloc(host_memory, memory - host_memory);

}
