/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/



/******************************************************************************************************************************************
 This file contains the helper classes needed to run TCP/IP tests in a mauve framework.
 In brief, it contains the classes:
 ThreadServer: runs a server in a separate thread. The server awaits ONE socket call and handles the connection
 ReplyHandler: the socket handler for the thread server
 MultithreadServer: as above, but this server runns multiple connections in separate threads, so it can handle multiple calls at one time
 ThreadReplayHandler: the socket handler class for the multi-threadserver
*/

package gnu.testlet.wonka.net;

import java.io.*;
import java.net.*;
import java.util.*;

import gnu.testlet.Testlet;
import gnu.testlet.TestHarness;

/**
****************************************************************************************************************************************
*
*  Basic external socket handler:
* This is a class to be called from the main tests handling testlet, or from an external server, implementing socket behavior
* it provides the following functionality:
*  <p> => constructor for a host and port, and constructor for a given socket
*  <p> => some help functions to retrieve handler name and to retrieve a string as would be sent by the class socket
*  <p> => a function to send a series of connection data messages
*  <p> => a function to send a string directly,
*  <p> => a function to send a string a certain time from now (this function runs in a thread as to not block the class)
*  <p> => a function to receive a message together with a timeout functionality (this function runs in a thread as to not block the class)
*  <p> => a function to wait for one message to be received and mirror it back to the sender (in a thread as to not block the class)
*  <p> => a function that mirrors all messages back to the sender untill it receives a quit-message (in a thread as to not block the class)
*  <p> => a close-down function
*
****************************************************************************************************************************************
**/

public class ExternalMulticastHandler  extends ExternalDatagramHandler
{
  //main definitions
  protected MulticastSocket mailgroupSocket;
  protected ArrayList       mailgroupsList;

  public ExternalMulticastHandler(MulticastSocket socket, String name, TestHarness ourharness)
  {
    super(name, ourharness);

    mailgroupSocket = socket;
    socketActive = true;
    // mailgroups
    mailgroupsList = new ArrayList();
  }

  public ExternalMulticastHandler(int port, String name, TestHarness ourharness)
  {
    super(name, ourharness);
    try
    {
      mailgroupSocket = new MulticastSocket(port);
      socketActive = true;
    }
    catch (Exception e)
    {
       ourharness.fail( e.toString()+ "( building socket for "+name+")");
       socketActive = false;
    }

     // mailgroups
    mailgroupsList = new ArrayList();
  }

/****************************************************************************************************************************************/
  /** overwrites ExternalDatagramsocket.getSocket: returns the multicast socket **/
  public DatagramSocket getSocket()  {  return (DatagramSocket)mailgroupSocket;  }
  /** return the handler port **/
  public int getPort(){return mailgroupSocket.getLocalPort();}
/****************************************************************************************************************************************/
  /**
   * forwards a 'fail'-message to the handlers TestHarness
   * @param the message string
   */
  public void fail(String errormessage) {harness.fail(errormessage + " (from ExternalMulticastHandler <"+handlerName+"> )");}

  /**
   * forwards a 'verbose'-message to the handlers TestHarness
   * @param the message string
   */
  public void verbose(String errormessage) {harness.verbose(errormessage+ " (from ExternalMulticastHandler <"+handlerName+"> )");}

/****************************************************************************************************************************************/
/*
*   set the default datagram address
*
*/
  /** constructor with default datagram address **/
  public ExternalMulticastHandler(MulticastSocket socket, String name, TestHarness ourharness, DatagramAddress p)
  {
    this(socket, name, ourharness);
    setDatagramAddress(p);
  }
  /** constructor with default datagram address **/
  public ExternalMulticastHandler(int port, String name, TestHarness ourharness, DatagramAddress p)
  {
    this(port, name, ourharness);
    setDatagramAddress(p);
  }
  /** constructor with default datagram address **/
  public ExternalMulticastHandler(MulticastSocket socket, String name, TestHarness ourharness, InetAddress i ,int  p)
  {
    this(socket, name, ourharness);
    setDatagramAddress(i,p);
  }
  /** constructor with default datagram address **/
  public ExternalMulticastHandler(int port, String name, TestHarness ourharness, InetAddress i ,int  p)
  {
    this(port, name, ourharness);
    setDatagramAddress(i,p);
  }
  /** constructor with default datagram address **/
  public ExternalMulticastHandler(MulticastSocket socket, String name, TestHarness ourharness, String s ,int  p)
  {
    this(socket, name, ourharness);
    setDatagramAddress(s,p);
  }
  /** constructor with default datagram address **/
  public ExternalMulticastHandler(int port, String name, TestHarness ourharness, String s ,int  p)
  {
    this(port, name, ourharness);
    setDatagramAddress(s,p);
  }
  /** constructor with default datagram address **/
  public ExternalMulticastHandler(MulticastSocket socket, String name, TestHarness ourharness, DatagramPacket d)
  {
    this(socket, name, ourharness);
    setDatagramAddress(d);
  }
  /** constructor with default datagram address **/
  public ExternalMulticastHandler(int port, String name, TestHarness ourharness, DatagramPacket d)
  {
    this(port, name, ourharness);
    setDatagramAddress(d);
  }
  /** constructor with default datagram address **/
  public ExternalMulticastHandler(MulticastSocket socket, String name, TestHarness ourharness, DatagramSocket s)
  {
    this(socket, name, ourharness);
    setDatagramAddress(s);
  }
  /** constructor with default datagram address from a datagram socket to connent to **/
  public ExternalMulticastHandler(int port, String name, TestHarness ourharness, DatagramSocket s)
  {
    this(port, name, ourharness);
    setDatagramAddress(s);
  }
  /** constructor with default datagram address **/
  public ExternalMulticastHandler(MulticastSocket socket, String name, TestHarness ourharness, ExternalDatagramHandler target)
  {
    this(socket, name, ourharness);
    setDatagramAddress(target);
  }
  /** constructor with default datagram address from a datagram socket to connent to **/
  public ExternalMulticastHandler(int port, String name, TestHarness ourharness, ExternalDatagramHandler target)
  {
    this(port, name, ourharness);
    setDatagramAddress(target);
  }

/****************************************************************************************************************************************/

  /**
   * Overrides Object.toString()
   * @ return the string representation of the current ExternalSocketHandler instance
   */
  public String toString()    {    return buildStringRepresentation(handlerName, mailgroupSocket);    }


  /**
   * as static ExternalDatagramHandler.buildStringRepresentation(name, socket), but for a multicast handler and multicast socket
   */
  static String buildStringRepresentation(String handlername, MulticastSocket handlersocket)
  {  return buildStringRepresentation(handlername, (handlersocket.getLocalAddress()).toString(), handlersocket.getLocalPort() );  }

  /**
   * as static ExternalDatagramHandler.buildStringRepresentation(name, adress), but for a multicast handler
   */
  static String buildStringRepresentation(String handlername, DatagramAddress address)
    {return buildStringRepresentation(handlername, (address.getAddress()).toString(), address.getPort() );}

  /**
   * as static ExternalDatagramHandler.buildStringRepresentation(name, Inet Address, port), but for a multicast handler
   */
  static String buildStringRepresentation(String handlername, InetAddress socketIPaddress,int localport)
    {return buildStringRepresentation(handlername, socketIPaddress.toString(), localport);}

  /**
   * as static ExternalDatagramHandler.buildStringRepresentation(name, adress-string, port), but for a multicast handler
   */
  static String buildStringRepresentation(String name, String addressstring, int port)
  {    return ( "MulticastHandler[name="+name+",addr="+addressstring+",port="+port+"]");  }

/****************************************************************************************************************************************
  /**
   * Overrides ExternalDatagramHandler.getConnectionData(handler) to return MulticastHandler.buildStringRepresentation()
   */
  static String getConnectionData(ExternalMulticastHandler handler)  {  return ("Connected to <"+handler.toString()+">");  }
/****************************************************************************************************************************************/
/**
*
* Specific multicast functions:
*   => InetAdress mailgroup list : get list,size, first element, last element, find element
*   => Subscribe/unsubscribe to an inetaddress
*   => Send to an Inetaddress mailgroup
*
****************************************************************************************************************************************/

/****************************************************************************************************************************************/
/* Get mailgroup list data and members
*/
  /** get mailgroup list **/
  public ArrayList getMailgroupsList()  {return mailgroupsList;}
  /** get mailgroup list size**/
  public int getMailgroupsSize()  {return mailgroupsList.size();}
  /** get last InetAddress element of mailgroup list **/
  public InetAddress getLastMailgroupAddress() {return (InetAddress)(mailgroupsList.get(mailgroupsList.size()-1));}
  /** get specific InetAddress element of mailgroup list **/
  public InetAddress getMailgroupAddress(int i){return(i>=0 && i<mailgroupsList.size())? (InetAddress)(mailgroupsList.get(i)): null;}
  /** get last DatagramAddress element of mailgroup list **/
  public DatagramAddress getLastMailgroup() {return new DatagramAddress(getLastMailgroupAddress(),getPort());}
  /** get specific DatagramAddress element of mailgroup list **/
  public DatagramAddress getMailgroup(int i)
  {return (i>=0 && i<mailgroupsList.size())?(new DatagramAddress(getMailgroupAddress(i),getPort())):null;}


/****************************************************************************************************************************************/
  /** find specific element of mailgroup list
  * @return :the number of that Inet Address in the list, or -1 if not found
  **/
  public int findMailgroup(InetAddress a)
  {
    int result = -1;
    for(int i=0;i<mailgroupsList.size() && result<0; i++)
    {
      if(a.equals(mailgroupsList.get(i)) )
        result=i;
    }
    return result;
  }
  /** check if given adress is in mailgroup list **/
  public boolean isSubscribed(InetAddress a) { return(findMailgroup(a) >= 0); }

  /** find a specific DatagramAddress in the mailgroup list
  * @return :the number of that address in the list, or -1 if not found (InetAddress not in list, or wrong port)
  **/
  public int findMailgroup(DatagramAddress a)  { return(a.getPort() == getPort())? findMailgroup(a.getTarget()):-1; }
  /** check if a specific DatagramAddress is in the mailgroup list
  * @return :true if the address is in the list and both adress and handler have the same port
  **/
  public boolean isSubscribed(DatagramAddress a) { return(a.getPort() == getPort() && findMailgroup(a.getTarget()) >= 0); }
  /** find specific element of mailgroup list **/

/****************************************************************************************************************************************/
/* Subscribe/unsubscribe mailgroup list
*/
  /**
   * Subscribes to a specific mailgroup. Keep the name of that mailgroup in the mailgroup list
   * @param the given mailgroup inet address
   */
  public int subscribe(InetAddress a)
  {
    int result = -1;
    try
    {
      if(a.isMulticastAddress())
      {
        if(!isSubscribed(a))
        {
          mailgroupSocket.joinGroup(a);
//System.out.println(toString()+" joins group "+a);
          mailgroupsList.add(a);
          result = mailgroupsList.size();
        }
        else
          verbose(" Already subscribed : "+a);
      }
      else
        verbose(" No multicastAddress : "+a);
    }
    catch(IOException iox)
    {
      fail(iox.toString());
    }
    return result;
  }

  public int subscribe(DatagramAddress a)
  {
    int result = -1;
    if(a.getPort() == getPort() )
      result = subscribe(a.getTarget() );
    else
        verbose(" Different port : "+a);
    return result;
  }
  /**
   * Unsubscribes from a specific mailgroup. and remove that mailgroup from the adress list
   * @param index of the mailgroup in the list
   * @return true if successfully removed, false if no mailgroup with that number exists
   */
  public int unSubscribe(int i)
  {
    int result = -1;
    if(i>=0 && i<mailgroupsList.size())
    {
      InetAddress a = (InetAddress)(mailgroupsList.get(i));
      try
      {
        mailgroupSocket.leaveGroup((InetAddress)(mailgroupsList.get(i)) );
        mailgroupsList.remove(i);
        result = mailgroupsList.size();
      }
      catch(IOException iox)
      {
        fail(iox.toString());
      }
    }
    else
      verbose(" No element "+i);
    return result;
  }

  /**
   * Unsubscribes from a specific mailgroup. and remove that mailgroup from the adress list
   * @param a the Inet adress of the mailgroup
   * @return true if successfully removed, false if adress not in mailgroup list
   */
  public int unSubscribe(InetAddress a)
  {
    int result = -1;
    int i= findMailgroup(a);
    if(i>=0 && i<mailgroupsList.size())
    {
      try
      {
        mailgroupSocket.leaveGroup(a);
        mailgroupsList.remove(i);
        result = mailgroupsList.size();
      }
      catch(IOException iox)
      {
        fail(iox.toString());
      }
    }
    else
      verbose(" No element "+a);
    return result;
  }

  /**
   * Unsubscribes from a specific mailgroup. and remove that mailgroup from the adress list
   * @param a the datagram adress of the mailgroup
   * @return true if successfully removed, false if different port or adress not in mailgroup list
   */
  public int unSubscribe(DatagramAddress a)
  {
    int result = -1;
    if(a.getPort() == getPort() )
      result = unSubscribe(a.getTarget() );
    else
        verbose(" Different port : "+a);
    return result;
  }

  /**
   * Unsubscribes from all mailgroups
   */
  public void unSubscribeAll()
  {
    InetAddress a;
    try
    {
      for(int i=0;i<mailgroupsList.size();i++)
        mailgroupSocket.leaveGroup((InetAddress)(mailgroupsList.get(0)) );
      mailgroupsList.clear();
    }
    catch(IOException iox)
    {
      fail(iox.toString());
    }
  }

//ExternalMulticastHandler ends here
}



