/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/******************************************************************************************************************************************
 This file contains the helper classes needed to run TCP/IP tests in a mauve framework.
 In brief, it contains the classes:
 ThreadServer: runs a server in a separate thread. The server awaits ONE socket call and handles the connection
 ReplyHandler: the socket handler for the thread server
 MultithreadServer: as above, but this server runns multiple connections in separate threads, so it can handle multiple calls at one time
 ThreadReplayHandler: the socket handler class for the multi-threadserver
*/

package gnu.testlet.wonka.net;

import java.io.*;
import java.net.*;

import gnu.testlet.Testlet;
import gnu.testlet.TestHarness;

/**
****************************************************************************************************************************************
*
*  Basic external socket handler:
* This is a class to be called from the main tests handling testlet, or from an external server, implementing socket behavior
* it provides the following functionality:
*  <p> => constructor for a host and port, and constructor for a given socket
*  <p> => some help functions to retrieve handler name and to retrieve a string as would be sent by the class socket
*  <p> => a function to send a series of connection data messages
*  <p> => a function to send a string directly,
*  <p> => a function to send a string a certain time from now (this function runs in a thread as to not block the class)
*  <p> => a function to receive a message together with a timeout functionality (this function runs in a thread as to not block the class)
*  <p> => a function to wait for one message to be received and mirror it back to the sender (in a thread as to not block the class)
*  <p> => a function that mirrors all messages back to the sender untill it receives a quit-message (in a thread as to not block the class)
*  <p> => a close-down function
*
****************************************************************************************************************************************
**/

public class ExternalSocketHandler
{
/**
  final variables for close-socket-on-exit questions: CLOSE_ON_EXIT and LEAVE_OPEN
**/
final static boolean CLOSE_ON_EXIT = true;
final static boolean LEAVE_OPEN = false;

  //main definitions
  private Socket          handlerSocket;
  private String          handlerName;
  private TestHarness     harness;

  private PrintWriter     toClient;
  private BufferedReader  fromClient;
  private int             threadCount; // counter to give each thread a different name/threadgroup
  private boolean         socketActive;

/****************************************************************************************************************************************/
  /**
   * general constructor for a socket connecting to a host on a certain port
   * @param  host: the name of the host to connect to
   * @param  port: the port of this server
   * @param  name: This will be used as an identifier for the socket handler on all send() and reply() calls
   * @param  ourharness: reference to the test harness
   */
  public ExternalSocketHandler(String host, int port, String name, TestHarness ourharness)
  {
    //name
    handlerName = name;
    //socket and socket data streams
    try
    {
      handlerSocket = new Socket(host, port);
      toClient = new PrintWriter(handlerSocket.getOutputStream(), true);
      fromClient = new BufferedReader( new InputStreamReader(handlerSocket.getInputStream() ) );
      socketActive = true;
    }
    catch (Exception e)
    {
       ourharness.fail( e.toString()+ "( building socket for "+name+")");
       socketActive = false;
    }

    //harness
    harness = ourharness;
    // thread counter
    threadCount = 0;
  }

  /**
   * constructor with given socket. This form is designed for calling the handler from within a server application,
   * when the socket is obtained by a call to ServerSocket.accept()
   * @param oursocket: the socket (for instance obtained through a call to ServerSocket.accept() )
   * @param name: This will be used as an identifier for the socket handler on all send() and reply() calls
   * @param ourharness: test harness to dispatch fail-messages to
   */
  public ExternalSocketHandler(Socket oursocket, String name, TestHarness ourharness)
  {
    //name and socket
    handlerName = name;
    handlerSocket = oursocket;
    //socket data streams
    try
    {
      toClient = new PrintWriter(handlerSocket.getOutputStream(), true);
      fromClient = new BufferedReader( new InputStreamReader(handlerSocket.getInputStream() ) );
      //always oen the writer first, otherwise there is a danger of deathlock
    }
    catch (Exception e)
    {
       ourharness.fail( e.toString()+ "( building socket for "+name+")");
    }

    //harness
    harness = ourharness;
    // thread counter
    threadCount = 0;
    socketActive = true;
  }

  /** return the handler name **/
  public String getName(){return handlerName;}

  /** return the handler socket **/
  public Socket getSocket(){return handlerSocket;}



/****************************************************************************************************************************************/
  /**
   * forwards a 'fail'-message to the handlers TestHarness
   * @param the message string
   */
  public void fail(String errormessage) {harness.fail(errormessage + " (from ExternalSocketHandler <"+handlerName+"> )");}

  /**
   * forwards a 'verbose'-message to the handlers TestHarness
   * @param the message string
   */
  public void verbose(String errormessage) {harness.verbose(errormessage+ " (from ExternalSocketHandler <"+handlerName+"> )");}



/****************************************************************************************************************************************/

  /**
   * Overrides Object.toString()
   * @ return the string representation of the current ExternalSocketHandler instance
   */
  public String toString() {return buildStringRepresentation(handlerName, handlerSocket);}

  /**
   * returns a string equal to the toString() representation for the given ExternalSocketHandler
   * @param handler : the name of the ExternalSocketHandler
   * @return the toString() representation of an ExternalSocketHandler with the specified name, adress connection port and local port
   */
  static String buildStringRepresentation(ExternalSocketHandler handler)
  {return buildStringRepresentation(handler.getName(),handler.getSocket() );}
  /**
   * returns a string equal to ExternalSocketHandler.toString() for a handler with the specified name,and socket
   * @param name : the name the ExternalSocketHandler socket would have
   * @param handlersocket : the handler socket the ExternalSocketHandler would use
   * @return the toString() representation of an ExternalSocketHandler with the specified name, adress connection port and local port
   */
  static String buildStringRepresentation(String handlername, Socket handlersocket)
  {return buildStringRepresentation(handlername, handlersocket.getInetAddress(), handlersocket.getPort(), handlersocket.getLocalPort() );}
  /**
   * returns a string equal to the toString() representation of an ExternalSocketHandler with the specified handler name,
   * socket address and socket ports
   * @param name : the name of the ExternalSocketHandler socket
   * @param socketIPaddress : the socket's IP adress
   * @param connectionport : connection port as per Socket.getPort()
   * @param localport : local port as per Socket.getLocalPort()
   * @return the toString() representation of an ExternalSocketHandler with the specified name, adress connection port and local port
   */
  static String buildStringRepresentation(String handlername, InetAddress socketIPaddress, int connectionport, int localport)
  {return buildStringRepresentation(handlername, socketIPaddress.toString(), connectionport, localport);}
  /**
   * returns a string equal to the toString() representation of an ExternalSocketHandler with the specified handler name,
   * socket name , socket host address and socket ports
   * @param name : the name of the ExternalSocketHandler socket
   * @param hostname : the host name for the socket as per IPAddress.getHostName()
   * @param hostaddress : the host IP adress for the socket as per IPAddress.getHostAddress()
   * @param connectionport : connection port as per Socket.getPort()
   * @param localport : local port as per Socket.getLocalPort()
   * @return the toString() representation of an ExternalSocketHandler with the specified name, adress connection port and local port
   */
  static String buildStringRepresentation(String handlername, String hostname, String hostaddress, int connectionport, int localport)
  {return buildStringRepresentation(handlername, hostname+"/"+hostaddress, connectionport, localport);}
  /**
   * returns a string equal to the toString() representation of an ExternalSocketHandler with the specified handler name,
   * socket address and socket ports
   * returns a string equal to the toString() representation of an ExternalSocketHandler with the specified handler name,
   * host name, host adress and ports
   * @param name : the name of the ExternalSocketHandler socket
   * @param address : string representation of the IP adress as per IPAddress.toString()
   * @param connectionport : connection port as per Socket.getPort()
   * @param localport : local port as per Socket.getLocalPort()
   * @return the toString() representation of an ExternalSocketHandler with the specified name, adress connection port and local port
   */
  static String buildStringRepresentation(String name, String addressstring, int connectionport, int localport)
  {return ( "SocketHandler[name="+name+",addr="+addressstring+",port="+connectionport+",localport="+localport+"]");}

/****************************************************************************************************************************************/
  /**
   * sends a line ExternalSocketHandler.toString() that defines name, address, local and connected port
   * @return the function is of type void, but calling it sends a string ExternalSocketHandler.getConnectionData() to the connected socket
   */
  public void sendConnectionData()
  {
    try
    {
       // send connection data
      toClient.println(getConnectionData());
    }
    catch (Exception e)
    {
       harness.fail(e.toString() );
    }
  }

  /**
   * returns a connection data message like ExternalSocketHandler  would send on a sendConnectionData() command
   *<p> this function is public to allow to check the message received by the connected socket against the expected result
   * @return the connection data as the handler would send it (hence the message that the receiver socket would expect)
   */
  public String getConnectionData() {return ("Connected to <"+buildStringRepresentation(handlerName, handlerSocket)+">");}
  /**
   * as getConnectionData(), but a public function that can be called specifying the desired ExternalsocketHandler name
   * @param handler : the ExternalSocketHandler of which the data should be retrieved
   * @return the connection data as the handler would send it (hence the message that the receiver socket would expect)
   */
  static String getConnectionData(ExternalSocketHandler handler)
    {return ("Connected to <"+buildStringRepresentation(handler.getName(), handler.getSocket())+">");}


/****************************************************************************************************************************************/
/**
* wraps the given message into a form buildSendMessage(message) and sends this
* @param message: the data to (wrap into <buildSentMessage(data)> and )send
* @return the function is of type void, but calling it sends a line ExternalSocketHandler.buildSentMessage(data) to the connected socket
*/
  public void sendMessage(String message)
  {
    /**sends the desired data**/
    try
    {
      //send the desired data
      toClient.println(buildSendMessage(message));
    }
    catch (Exception e)
    {
       harness.fail(e.toString() );
    }
  }

/**
* calls a PostThread timer that will send the given message after waiting <delay> milliseconds
* @param message: the data to (wrap into <buildSentMessage(data)> and )send
* @param delay: the delay in milliseconds
* @return the function is of type void, but calling it will send a line ExternalSocketHandler.buildSentMessage(data)
* to the connected socket after the specified time has elapsed
*/
  public void sendMessage(String message, long delay)
  {
    sendMessage(message,delay,false);
  }

/**
* as sendmessage(message, delay), but with the option to let the listener close the handler after the message is received
* @param message: the data to (wrap into <buildSentMessage(data)> and )send
* @param delay: the delay in milliseconds
* @param closedown: if true, this function will call ExternalSocketHandler.close() to clean up the handler after it finished sending
* @return the function is of type void, but calling it will send a line ExternalSocketHandler.buildSentMessage(data)
* to the connected socket after the specified time has elapsed
*/
  public void sendMessage(String message, long delay, boolean closedown)
  {
    DelayedTCPIPSender postman = new DelayedTCPIPSender(handlerName, threadCount++, toClient, message, delay, this, closedown);
    postman.start();
  }
/**
* sends a <quit> message
* <p> call this function when connected to another ExternalSocketHandler to break out a continuous listening loop
* @return the function  itself is of type void, but calling it will send a <quit> message to the socket it is connected with
*/
  public void sendQuitMessage()
  {
    try
    {
      //send quit message
      toClient.println(NETDEFS.QUITSTRING);
    }
    catch (Exception e)
    {
       harness.fail(e.toString() );
    }
  }

  /**
   * Wraps the string into a 'message'  like the ExternalsocketHandler would send on a sendMessage(message) command
   *<p> this function is public to allow to check a message received from this ExternalsocketHandler against the result expected
   * @param message : the message that would be sendt
   * @return the string as the current ExternalsocketHandler would send it
   */
  public String buildSendMessage (String message)  {return( buildSendMessage (handlerName, message) );}
  /**
   * as buildSendMessage(message), but a public function that can be called specifying a desired ExternalsocketHandler
   * @param handler : the handler specified
   * @param message : the message to send
   * @return the string as an ExternalsocketHandler would send it
   */
  static String buildSendMessage (ExternalSocketHandler handler,String message){return( buildSendMessage(handler.getName(), message));}
  /**
   * as buildSendMessage(message), but a public function that can be called specifying the desired ExternalsocketHandler name
   * @param name : the name of the handler that is asked to perform the send
   * @param message : the message to send
   * @return the string as an ExternalsocketHandler would send it
   */
  static String buildSendMessage (String name,String message){return("Sent by <"+name+"> : ["+message+"]");}

/****************************************************************************************************************************************/
  /**
   * waits for a message to arrive(and blocks while doing so)
   * @return the received data string
   * <p><b> THIS FUNCTION BLOCKS THE WHOLE APPLICATION UNTIL THE EXPECTED MESSAGE HAS ARIVED. THEREFORE ONLY CALL THIS FUNCTION
   * WHEN YOU DEFINITELY KNOW THAT A MESSAGE IS ON ITS WAY </b>
   */
  public String awaitMessage()
  {
    return receiveMessage(0,false);
  }
  /**
   * as receiveMessage() but with the option to verbose the received message to the test harness
   * @return the received data string
   * @param verbose: if true, this function will verbose the message directly after it received it
   * <p><b> THIS FUNCTION BLOCKS THE WHOLE APPLICATION UNTIL THE EXPECTED MESSAGE HAS ARIVED. THEREFORE ONLY CALL THIS FUNCTION
   * WHEN YOU DEFINITELY KNOW THAT A MESSAGE IS ON ITS WAY </b>
  */

  public String awaitMessage(boolean verbose)
  {
    return receiveMessage(0,verbose);
  }
  /**
   * waits for a datagram message to arrive(and blocks while doing so), but defines a maximum timeout time
   * If the message hasn't arrived within that specified time, the function nevertheless closes,
   * so the maximum blocking time is known, and can be bargained with
   * @param timeout, the maximum time to wait for the message (ergo the maximum time this function wil hang)
   * @return the received data or a NULL string on timeout
   */
  public String receiveMessage(long timeout)
  {
   return receiveMessage(timeout, false);
  }
  /**
   * as receiveMessage(timeout) but with the option to verbose received message to the test harness
   * @param timeout, the maximum time to wait for the message (ergo the maximum time this function wil hang)
   * @param verbose: if true, this function will verbose the message directly after it received it
   * @return the received data or a NULL string on timeout
   */
  //public synchronized String receiveMessage(long timeout, boolean verbose)
  public String receiveMessage(long timeout, boolean verbose)
  {
   String message = null;
   try
   {
      handlerSocket.setSoTimeout((int)timeout);//wait for message to arrive
      message = fromClient.readLine();
      if(verbose)
        verbose("Received message [" +message+"]");
    }
    catch (InterruptedIOException e)
    {
      if(verbose)
        verbose(e.toString() );
    }
    catch (Exception e)
    {
       harness.fail(e.toString() );
    }
    return message;
  }


/****************************************************************************************************************************************/
  /**
   * calls a ReplyThread listener to wait untill ONE message is received by the socket, and then mirrors this message back to the sender
   * <p>(actually this function will wrap the received data into a String <buildReplyMessage(data)> and send this)
   * @return the function  itself returns void ,but mirrors back the received data to sender as soon as it arrives
   */
  public void replyMessage()
  {
    replyMessage(false,false);
  }
  /**
   * calls a ReplyThread listener to wait untill ONE message is received by the socket, and then mirrors this message back to the sender
   * <p>(actually this function will wrap the received data into a String <buildReplyMessage(data)> and send this)
   * @return the function  itself returns void ,but mirrors back all received data to sender as soon as it arrives
   */
  public void replyMessagesContinuously()
  {
    replyMessagesContinuously(false, false);
  }

  /**
   * as replyMessage() but with the option to verbose the received message to the test harness immedialtely after mirroring it back
   * @param verbose: if true, this function will verbose every message it received and mirrorred back to the test harness
   * @return the function  itself returns void ,but mirrors back the received data to sender as soon as it arrives
   */
  public void replyMessage(boolean verbose)
  {
    replyMessage(verbose,false);
  }
  /**
   * as replyMessagesContinuously(), but with the option to verbose all mirrored strings to the test harness
   * @param verbose: if true, this function will verbose every message it received and mirrorred back to the test harness
   * @return the function  itself returns void ,but mirrors back all received data to sender as soon as it arrives
   */
  public void replyMessagesContinuously(boolean verbose)
  {
    replyMessagesContinuously(verbose, false);
  }

  /**
   * as replyMessage(verbose) but with the option to close down the handler after the last message is received and sent back
   * @param verbose: if true, this function will verbose every message it received and mirrorred back to the test harness
   * @param closedown: if true, this function will call ExternalSocketHandler.close() to clean up the handler as soon as it has finished
   * @return the function  itself returns void ,but mirrors back the received data to sender as soon as it arrives
   */
  public void replyMessage(boolean verbose, boolean closedown)
  {
    ReplyThread mirror = new ReplyThread(handlerName, threadCount++, toClient, fromClient, this, verbose, closedown);
    mirror.start();
  }

  /**
   * as replyMessagesContinuously(verbose), but with the option to close the handler when the last message is received and sent back
   * @param verbose: if true, this function will verbose every message it received and mirrorred back to the test harness
   * @param closedown: if true, this function will call ExternalSocketHandler.close() to clean up the handler as soon as it has finished
   * @return the function  itself returns void ,but mirrors back all received data to sender as soon as it arrives
   */
  public void replyMessagesContinuously(boolean verbose, boolean closedown)
  {
    ReplyThreadLoop mirror = new ReplyThreadLoop(handlerName, threadCount++, toClient, fromClient, this, verbose, closedown);
    mirror.start();
  }

  /**
   * Wraps the string into a 'message' like ExternalsocketHandler woulds reply back to sender on a ReplyMessage(message) command
   *<p> this function is public to allow to check a message received from this ExternalsocketHandler against the result expected
   * @param message : the message that would be replied
   * @return the string as the current ExternalsocketHandler would send it
   */
  public String buildReplyMessage (String message)  {return( buildReplyMessage (handlerName, message) );}
  /**
   * as buildReplyMessage(message), but a public function for a given ExternalSocketHandler
   * @param handler : the handler specified
   * @param message : the message that would be replied
   * @return the string as an ExternalsocketHandler would send it
   */
  static String buildReplyMessage (ExternalSocketHandler handler, String message){return( buildReplyMessage (handler.getName(), message));}
  /**
   * as buildReplyMessage(message), but a public function for a given ExternalSocketHandler name
   * @param name : the name of the handler would perform the reply
   * @param message : the message that would be replied
   * @return the string as an ExternalsocketHandler would send it
   */
  static String buildReplyMessage (String name, String message){return("replied to sender by <"+name+"> : ["+message+"]");}


/****************************************************************************************************************************************/
  /**
   * closes the socket
   * @return <code>true</code>if closed successfully <code>false</code>if either error while closing, or socket already closed
   */
  public boolean close()
  {
    if(socketActive)
    {
      try
      {
        fromClient.close();
        toClient.close();
        handlerSocket.close();
        socketActive = false;
        verbose(" Handler threads and Sockets closed");
      }
      catch (Exception e)
      {
         fail(e.toString() );
      }
    }
    return (!socketActive);
  }


  /** return true if socket active (not yet closed) **/
  public boolean isActive() {return socketActive;}


//end of class ExternalSocketHandler
}














/******************************************************************************************************************************************
*   externalSocketHandler  helper thread classes. These are designed to 'decouple' hanging functions to run in a separate thread
* and alow the main socket handler to continue its job
*/

/**
****************************************************************************************************************************************
*  posts a thread <delay> milliseconds in the future
**/
class DelayedTCPIPSender extends Thread
{
  // basic definitions
  private ExternalSocketHandler parent;
  private int                   threadCount;
  private PrintWriter           toClient;
  private String                messageData;
  private long                  messageDelay;
  private boolean               closeOnExit;

  public DelayedTCPIPSender(String name, int count, PrintWriter writer, String message, long delay, ExternalSocketHandler handler, boolean close)
  {
    super(name+count);

    threadCount   = count;
    toClient      = writer;
    messageData   = message;
    messageDelay  = delay;
    parent        = handler;
    closeOnExit   = close;
  }

  public void run()
  {
    try
    {
      //wait until the desired time is elapsed
      sleep(messageDelay);
      // send the data
      toClient.println(parent.buildSendMessage(messageData) );
    }
    catch(Exception e)
    {
      parent.fail("PostThread ("+threadCount+"): " + e.toString() );
    }
    //close down if desired
    if(closeOnExit)
      parent.close();
  }
}


/**
****************************************************************************************************************************************
*  waits for ONE message to be sent to the handler, then mirrors back this message to the sender in a reply string
*  "<"+parentName+"> received and mirrored ["+message+"]"
**/
class ReplyThread extends Thread
{
  // basic definitions
  private int             threadCount;
  private PrintWriter     toClient;
  private BufferedReader  fromClient;

  private ExternalSocketHandler parent;

  private boolean verboseReply;
  private boolean closeOnExit;


  public ReplyThread(String name, int count, PrintWriter writer, BufferedReader reader,
                                          ExternalSocketHandler handler, boolean verbose, boolean close)
  {
    super(name+count);

    threadCount   = count;
    toClient      = writer;
    fromClient    = reader;
    parent        = handler;
    verboseReply  = verbose;
    closeOnExit   = close;
  }

  public void run()
  {
    String message;
    try
    {
      // wait for message to arrive
      message = fromClient.readLine();
      // reply to client
      toClient.println(parent.buildReplyMessage(message) );
      //if wanted, verbose
      if(verboseReply)
        parent.verbose("ReplyThread ("+threadCount+") sent [" + message + "]");
    }
    catch (Exception e)
    {
       parent.fail("ReplyThread ("+threadCount+"): " +e.toString() );
    }

    if(closeOnExit)
       parent.close();
  }
}

/**
****************************************************************************************************************************************
*  waits for messages to be sent to the handler in a continuous loop. Upon receiving a message ,mirrors back it back in a reply string
*  "<"+parentName+"> received and mirrored ["+message+"]"
* this function will continue running untill it receives a message 'quit' : when calling this function in a program, you should always
* send a quit-message before closing down the program to avoid leaving threads open
**/
class ReplyThreadLoop extends Thread
{
  // basic definitions
  private int             threadCount;
  private PrintWriter     toClient;
  private BufferedReader  fromClient;

  private ExternalSocketHandler parent;

  private boolean verboseReply;
  private boolean closeOnExit;


  public ReplyThreadLoop(String name, int count, PrintWriter writer, BufferedReader reader,
                                          ExternalSocketHandler handler, boolean verbose, boolean close)
  {
    super(name+count);

    threadCount   = count;
    toClient      = writer;
    fromClient    = reader;
    parent        = handler;
    verboseReply  = verbose;
    closeOnExit   = close;
  }

  public void run()
  {
    String message;
    try
    {
      do
      {
      // wait for message to arrive
      message = fromClient.readLine();
      // reply to client
      toClient.println(parent.buildReplyMessage(message) );
      //if wanted, verbose
      if(verboseReply)
        parent.verbose("ContinuousReplyThread ("+threadCount+") sent [" + message + "]");
      }
      while(!message.equalsIgnoreCase(NETDEFS.QUITSTRING));
    }
    catch (Exception e)
    {
       parent.fail("ContinuousReplyThread ("+threadCount+"): " +e.toString() );
    }

    if(closeOnExit)
       parent.close();
  }
}
