/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/******************************************************************************************************************************************
 This file contains the helper classes needed to run TCP/IP tests in a mauve framework.
 In brief, it contains the classes:
 ThreadServer: runs a server in a separate thread. The server awaits ONE socket call and handles the connection
 ReplyHandler: the socket handler for the thread server
 MultithreadServer: as above, but this server runns multiple connections in separate threads, so it can handle multiple calls at one time
 ThreadReplayHandler: the socket handler class for the multi-threadserver
*/

package gnu.testlet.wonka.net;

import java.util.*;
import java.io.*;
import java.net.*;

import gnu.testlet.Testlet;
import gnu.testlet.TestHarness;

/**
****************************************************************************************************************************************
*
*  This application simulates a distant server : it runs in a separate thread 'independent' of the main program and dispatches all
*  incomming socket calls from clients to a propper external socket. THe behavior of this socket can be set by a number of boolean
*  flags, set when calling the constructor
*
* THIS CLASS WILL RUN FOR EVER : BE SURE TO CALL ThreadServer.interrupt()  FOLLOWED  by ThreadServer.close()
* BEFORE TERMINATING IN ORDER TO AVOID OPEN SOCKETS
****************************************************************************************************************************************
**/
public class ExternalSocketServer extends Thread
{
  //base definitions:
  String      serverName;
  TestHarness harness;
  boolean sendConnectionDesired;

  //server socket
  ServerSocket dispatcher  = null;

  // host handler sockets
  ArrayList handlerList = null;

/****************************************************************************************************************************************/
  /**
   * constructs an independent server in a separate thread, and tells the server which options to activate on handling a client request
   * @param  name: This will be used as an identifier for the socket server and forwarded to the assigned socket handlers
   * @param  port: the port for this server
   * @param  ourharness: reference to the test harness
   * @param  sendconnection: if true, every socket assigned by the server wil call  ExternalSocketHandler.sendConnectionData()
   * as soon as it establishes connection with the client socket
   */
  public ExternalSocketServer(int port, String name, TestHarness ourharness, boolean sendconnection)
  {
   /**
   * <b>ATTENTION, ASSIGNING A SERVER SOCKET CAN TAKE SOME TIME, SO WIAT UNTIL YOU ARE SURE CONNECTION IS ESTABLISHED BEFORE TRYNG
   * TO ACCESS THE ASSIGNED HANDLER :THis is best done BY SETTING *SENDCONNECTION* == TRUE AND WAITING FOR AN INCOMMING MESSAGE.</b>
   **/
    super(name + Integer.toString(port) );

    //variables
    serverName = name;
    harness = ourharness;
    sendConnectionDesired  = sendconnection;

    //server socket
    try
    {
      dispatcher = new ServerSocket(port);
    }
    catch (Exception e)
    {
       harness.fail("Error loading server socket: "+e );
       dispatcher = null;
    }

    //socket list
    handlerList = new ArrayList();
  }
/****************************************************************************************************************************************/
  /**
   * forwards a 'fail'-message to the handlers TestHarness
   * @param the message string
   */
  public void fail(String errormessage) {harness.fail(errormessage + " (from ThreadServer <"+serverName+"> )");}

  /**
   * forwards a 'verbose'-message to the handlers TestHarness
   * @param the message string
   */
  public void verbose(String errormessage) {harness.verbose(errormessage+ " (from ThreadServer <"+serverName+"> )");}
/****************************************************************************************************************************************/
  /**
   * the running thread: upon receipt of a client request, generate a new ExternalSocketHandler and dispatch it to deal with the client.
   * <p>(the behavior of this ExternalSocketHandler is determined by the behavior flags set at startup)
   */
  public void run()
  {
    try
    {
      while(true)
      {
        Socket dispatched = dispatcher.accept();
//   System.out.println(" =>   server accepted socket, dispatched to "+dispatched);
        String handlername = serverName + "_" +dispatched.getPort();
        ExternalSocketHandler newhandler = new ExternalSocketHandler(dispatched, handlername, harness);
//   System.out.println(" =>   dispatched to "+newhandler);
        if(sendConnectionDesired)
          newhandler.sendConnectionData();
        handlerList.add(newhandler);
//   System.out.println(" =>   added to list : ");
//   System.out.println(handlerList.toString());
      }
    }
    catch (InterruptedIOException e)
    {
      verbose("Accept() timed out");
    }
    catch (Exception e)
    {
      harness.fail("ThreadServer threadloop error : "+e.toString() );
    }
  }

  /** set accept() timeout**/
  public void setTimeOut(int timeout)
  {
    try
    {
      dispatcher.setSoTimeout(timeout);
    }
    catch (Exception e)
    {
       harness.fail("ThreadServer setting timeout : "+e.toString() );
    }
  }
/****************************************************************************************************************************************
* Socket handler access functions
*/

  /** return number of sockets **/
  public int getNumberOfHandlers() {return handlerList.size(); }
  /** return number of the last assigned socket **/
  public int getLastHandlerNumber() {return (handlerList.size()-1); }

  /** return the last assigned socket or <code>null</code> on exception (like no socket initialised yet) Also throws Fail if necessary **/
  public ExternalSocketHandler getLastHandler()  {return getHandler(handlerList.size() - 1);}
  /** return the desired socket or <code>null</code> on exception (like index out of bounds) Also throws Fail if necessary **/
  public ExternalSocketHandler getHandler(int index)
  {
    ExternalSocketHandler handler = null;
    try
    {
      handler = (ExternalSocketHandler)handlerList.get(index);
    }
    catch(Exception e)
    {
      fail(e.toString());
    }
    return handler;
  }

  /** return <code>true</code> if desired socket is active, <code>false</code> if not or on error **/
  public boolean isHandlertActive(int index)
  {
    ExternalSocketHandler handler = getHandler(index);
    return(handler != null)? handler.isActive():false;
  }
  /** Closes down the socket handler and returns ExternalSocketHandler.close() or <code>false</code> on error  **/
  public boolean closeHandler(int index)
  {
    ExternalSocketHandler handler = getHandler(index);
    return(handler != null)? handler.close():false;
  }
  /** Closes down all the handlers in the list **/
  public void closeAllHandlers()
  {
    for(int i=0; i<handlerList.size(); i++)
      getHandler(i).close();
  }
/****************************************************************************************************************************************/
  /**
   * closes down the ThreadServer and all its sockets
   */
  public void close()
  {
    try
    {
      //close all socket handlers
      for(int i=0; i<handlerList.size(); i++)
      {
        ExternalSocketHandler handler = (ExternalSocketHandler)handlerList.get(i);
        handler.close();
      }
      //then close server socket
      dispatcher.close();
    }
    catch (Exception e)
    {
       harness.fail("ThreadServer closing error : "+e.toString() );
    }
  }




  //ThreadServer class code ends here
}

