/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/



/******************************************************************************************************************************************
 This file contains the helper classes needed to run TCP/IP tests in a mauve framework.
 In brief, it contains the classes:
 ThreadServer: runs a server in a separate thread. The server awaits ONE socket call and handles the connection
 ReplyHandler: the socket handler for the thread server
 MultithreadServer: as above, but this server runns multiple connections in separate threads, so it can handle multiple calls at one time
 ThreadReplayHandler: the socket handler class for the multi-threadserver
*/

package gnu.testlet.wonka.net;

import java.io.*;
import java.net.*;

import gnu.testlet.Testlet;
import gnu.testlet.TestHarness;

/**
****************************************************************************************************************************************
*
* a datagram handler running in a remote thread allowing to send and receive datagrams
* While the Wonka classes still have difficulties with 'real' net commections, and the wonka's InetAddress class does not
* support InetAddress.getByName(), this is a special version only using the local host InetAddress.getLocalHost(). Different
* destinations are still possible using different ports on the local address.
*
* In this version, it reflects that where the ExternalDatagramHandler uses a
* DatagramAddress struct(InetAddress targetAddress, int targetPort) , LocalDatagram Handler knows that the destination Address is
* the local address, and can therefore replace the DatagramAddress member by a simple integer targetPort.
****************************************************************************************************************************************
**/

public class LocalDatagramHandler
{
/**
  final variables for close-socket-on-exit questions: CLOSE_ON_EXIT and LEAVE_OPEN
**/
final static int DEFAULTBUFSIZE = 0xff;

  //main definitions
  protected DatagramSocket  handlerSocket;
  protected String          handlerName;
  protected TestHarness     harness;

  //other variables
  protected int             targetPort = -1;
  protected int             threadCount; // counter to give each thread a different name/threadgroup
  protected boolean         socketActive;

/****************************************************************************************************************************************/
  /**
   * default constructor for a derived ExternalMulticastHandler
   * @param  name: This will be used as an identifier for the socket handler on all send() and reply() calls
   * @param  ourharness: reference to the test harness
   */
  public LocalDatagramHandler(String name, TestHarness ourharness)
  {
    handlerName = name;
    handlerSocket = null;
    socketActive = false;
    //harness
    harness = ourharness;
    // thread counter
    threadCount = 0;
  }

  /**
   * general constructor for a datagram socket handler tool
   * @param  socket: the DatagramSocket the handler uses
   * @param  name: This will be used as an identifier for the socket handler on all send() and reply() calls
   * @param  ourharness: reference to the test harness
   */
  public LocalDatagramHandler(DatagramSocket socket, String name, TestHarness ourharness)
  {
    handlerName = name;
    handlerSocket = socket;
    socketActive = true;

    //harness
    harness = ourharness;
    // thread counter
    threadCount = 0;
  }
  /**
   * general constructor for a datagram socket handler tool
   * @param  port: the DatagramSocket port
   * @param  name: This will be used as an identifier for the socket handler on all send() and reply() calls
   * @param  ourharness: reference to the test harness
   */
  public LocalDatagramHandler(int handlerport, String name, TestHarness ourharness)
  {
    handlerName = name;
    try
    {
      handlerSocket = new DatagramSocket(handlerport);
      socketActive = true;
    }
    catch (Exception e)
    {
       ourharness.fail( e.toString()+ "( building socket for "+name+")");
       socketActive = false;
    }

    //harness
    harness = ourharness;
    // thread counter
    threadCount = 0;
  }

/**************************************************************************************************************************************
handler name
**/
  /** return the handler name **/
  public String getName(){return handlerName;}


/**************************************************************************************************************************************
handler socket and socket port
**/
  /** set the handler socket **/
  public void setSocket(DatagramSocket socket)
  {
    handlerSocket = socket;
    socketActive = true;
  }

  /** set the handler socket **/
  public void setSocket(int handlerport)
  {
    try
    {
      handlerSocket = new DatagramSocket(handlerport);
      socketActive = true;
    }
    catch (Exception e)
    {
       fail( e.toString()+ "( building socket for "+handlerName+")");
       socketActive = false;
    }
  }

  /** return the handler socket **/
  public DatagramSocket getSocket(){return handlerSocket;}
  /** return the handler port **/
  public int getSocketPort(){return handlerSocket.getLocalPort();}

/**************************************************************************************************************************************
Target port
**/

  /** specify a default datagram address for the send-functions **/
  public void setTargetPort(int port){targetPort = port;}
  /** return the handler port **/
  public int getTargetPort(){return targetPort;}


/****************************************************************************************************************************************/
  /**
   * forwards a 'fail'-message to the handlers TestHarness
   * @param the message string
   */
  public void fail(String errormessage) {harness.fail(errormessage + " (from LocalDatagramHandler <"+handlerName+"> )");}

  /**
   * forwards a 'verbose'-message to the handlers TestHarness
   * @param the message string
   */
  public void verbose(String errormessage) {harness.verbose(errormessage+ " (from LocalDatagramHandler <"+handlerName+"> )");}

/****************************************************************************************************************************************/

  /**
   * Overrides Object.toString()
   * @ return the string representation of the current ExternalSocketHandler instance
   */
  public String toString()  {  return buildStringRepresentation(handlerName, handlerSocket);  }

  /** String representation of a certain Datagram handler with given name and DatagramSocket**/
  static String buildStringRepresentation(String handlername, DatagramSocket handlersocket)
  {  return buildStringRepresentation(handlername, handlersocket.getLocalPort() );  }

  /** String representation of a certain Datagram handler with given name, adress string and port**/
  static String buildStringRepresentation(String name, int port)
  {  return ( "LocalDatagramHandler[name="+name+",port="+port+"]");  }



/****************************************************************************************************************************************/
  /**
   * Help function to send a datagram containing the desired data to the desired address
   * @param data: data string to be sent
   * @param target: the target adressee's port (As this is a local simulation, the addressee's IP address is ALWAYS the local host)
   * @return the function is of type void, but calling it sends a string ExternalSocketHandler.getConnectionData() to the connected socket
   */
  protected void sendData(String data,int targetport)
  {
    //security on <null>-type (invalid) datagram address
    if(targetport <0 )
       fail("sendConnectionData ("+targetport+") invalid port or no port specified : "+targetport );
    else
    {
      try
      {
        byte[] outbuffer = NETDEFS.stringToBytes(data);
        //send back received message
        DatagramPacket outPacket = new DatagramPacket(outbuffer, outbuffer.length, InetAddress.getLocalHost(), targetport);
        handlerSocket.send(outPacket);
      }
      catch (Exception e)
      {
         fail(toString() );
      }
    }
  }

  /**
   * Help function to send a datagram containing the desired data to the desired address <delay> milliseconds in the future
   * (the sending is done in a separate thread, so the handler itself doesn't blok while waiting to send this message)
   * @param data: data string to be sent
   * @param targetPort: the target port (it's InetAddress is the local host)
   * @param delay: the delay on sending this address
   * (the sending is done in a separate thread, so while waiting, the handler itself doesn't blok)
   * @param closedown: If true, the socket closes down after sending this message
   * @return the function is of type void, but calling it sends a string ExternalSocketHandler.getConnectionData() to the connected socket
   */
  protected void sendDataDelayed(String data, int targetport, long delay, boolean closedown)
  {
    //security on <null>-type (invalid) datagram address
    if(targetport <0 )
       fail("sendConnectionData ("+targetport+") invalid port or no port specified : "+targetport );
    else
    {
      DelayedDatagramSender postman = new DelayedDatagramSender(handlerName, threadCount++, data, targetport, delay, this, closedown);
      postman.start();
    }
  }
/****************************************************************************************************************************************/
  /**
   * sends a line ExternalSocketHandler.toString() that defines name, address, local and connected port
   * @param targetport: the target addressee's port (it's InetAddress is the local host)
   * @return the function is of type void, but calling it sends a string ExternalSocketHandler.getConnectionData() to the connected socket
   */
  public void sendConnectionData(int targetport) {sendData(getConnectionData(),targetport);}
  /** SendConnectinData(target) with target address specified by a datagramPacket's destination **/
  public void sendConnectionData(DatagramPacket d){sendData(getConnectionData(), d.getPort() );}
  /** SendConnectinData(target) with target address specified by a target datagramsocket **/
  public void sendConnectionData(DatagramSocket s){sendData(getConnectionData(), s.getLocalPort() );}
  /** SendConnectinData(target) with target address specified by a target LocalDatagramHandler **/
  public void sendConnectionData(LocalDatagramHandler h){sendData(getConnectionData(), h.getSocketPort() );}
  /** sends the connection data to the handlers default address**/
  public void sendConnectionData(){sendData(getConnectionData(),targetPort);}

  /**
   * returns a connection data message like ExternalSocketHandler  woulds send on a sendConnectionData() command
   *<p> this function is public to allow to check the message received by the connected socket against the expected result
   * @return the connection data as the handler would send it (hence the message that the receiver socket would expect)
   */
  public String getConnectionData() {return ("Connected to <"+toString()+">");  }
  /**
   * as getConnectionData(), but a public function that can be called for a specific LocalDatagramHandler
   * @param handler : the datagramhandler
   * @return the connection data as the handler would send it (hence the message that the receiver socket would expect)
   */
  static String getConnectionData(LocalDatagramHandler handler)  {  return ("Connected to <"+handler.toString()+">");  }
/****************************************************************************************************************************************/
/**
* wraps the given message into a form buildSendMessage(message) and sends this to the target datagram adress
* @param targetport: the target addressee's port (it's InetAddress is the local host)
* @param message: the data to (wrap into <buildSentMessage(data)> and )send
* @return the function is of type void, but calling it sends a line ExternalSocketHandler.buildSentMessage(data) to the connected socket
*/
  public void sendMessage(String message, int targetport) {sendData(buildSendMessage(message),targetport);}
  /** SendConnectinData(target) with target address specified by a datagramPacket's destination **/
  public void sendMessage(String message, DatagramPacket d){sendData(buildSendMessage(message), d.getPort() );}
  /** SendConnectinData(target) with target address specified by a target datagramsocket **/
  public void sendMessage(String message, DatagramSocket s){sendData(buildSendMessage(message), s.getLocalPort() );}
  /** SendConnectinData(target) with target address specified by a target LocalDatagramHandler **/
  public void sendMessage(String message, LocalDatagramHandler h){sendData(buildSendMessage(message), h.getSocketPort() );}
  /** sends the connection data to the handlers default address**/
  public void sendMessage(String message){sendData(buildSendMessage(message),targetPort);}

/**
* calls a PostThread timer that will send the given message to the desired datagram address after waiting <delay> milliseconds
* @param message: the data to (wrap into <buildSentMessage(data)> and )send
* @param targetport: the target addressee's port (it's InetAddress is the local host)
* @param delay: the delay in milliseconds
* @return the function is of type void, but calling it will send a line ExternalSocketHandler.buildSentMessage(data)
* to the connected socket after the specified time has elapsed
*/
  public void sendMessageDelayed(String message, int targetport, long delay)
    {sendDataDelayed(buildSendMessage(message),targetport, delay, false);}
  /** SendConnectinData(target) with target address specified by a datagramPacket's destination **/
  public void sendMessageDelayed(String message, DatagramPacket d, long delay)
    {sendDataDelayed(buildSendMessage(message), d.getPort(), delay, false);}
  /** SendConnectinData(target) with target address specified by a target datagramsocket **/
  public void sendMessageDelayed(String message, DatagramSocket s, long delay)
    {sendDataDelayed(buildSendMessage(message), s.getLocalPort(), delay, false);}
  /** SendConnectinData(target) with target address specified by a target LocalDatagramHandler **/
  public void sendMessageDelayed(String message, LocalDatagramHandler h, long delay)
    {sendDataDelayed(buildSendMessage(message), h.getSocketPort(), delay, false);}
  /** sends the connection data to the handlers default address**/
  public void sendMessageDelayed(String message, long delay)
    {sendDataDelayed(buildSendMessage(message),targetPort, delay, false);}

/**
* as sendmessage(message, delay), but with the option to let the listener close the handler after the message is received
* @param message: the data to (wrap into <buildSentMessage(data)> and )send
* @param targetport: the target addressee's port (it's InetAddress is the local host)
* @param delay: the delay in milliseconds
* @param closedown: if true, this function will call ExternalSocketHandler.close() to clean up the handler after it finished sending
* @return the function is of type void, but calling it will send a line ExternalSocketHandler.buildSentMessage(data)
* to the connected socket after the specified time has elapsed
*/
  public void sendMessageDelayed(String message, int targetport, long delay, boolean closedown)
    {sendDataDelayed(buildSendMessage(message),targetport, delay, closedown);}
  /** SendConnectinData(target) with target address specified by a datagramPacket's destination **/
  public void sendMessageDelayed(String message, DatagramPacket d, long delay, boolean closedown)
    {sendDataDelayed(buildSendMessage(message), d.getPort(), delay, closedown);}
  /** SendConnectinData(target) with target address specified by a target datagramsocket **/
  public void sendMessageDelayed(String message, DatagramSocket s, long delay, boolean closedown)
    {sendDataDelayed(buildSendMessage(message), s.getLocalPort(), delay, closedown);}
  /** SendConnectinData(target) with target address specified by a target LocalDatagramHandler **/
  public void sendMessageDelayed(String message, LocalDatagramHandler h, long delay, boolean closedown)
    {sendDataDelayed(buildSendMessage(message), h.getSocketPort(), delay, closedown);}
  /** sends the connection data to the handlers default address**/
  public void sendMessageDelayed(String message, long delay, boolean closedown)
    {sendDataDelayed(buildSendMessage(message), targetPort, delay, closedown);}


/**
* sends a <quit> message
* <p> call this function when connected to another ExternalSocketHandler to break out a continuous listening loop
* @return the function  itself is of type void, but calling it will send a <quit> message to the socket it is connected with
*/

  public void sendQuitMessage(int targetport) {sendData(NETDEFS.QUITSTRING,targetport);}
  /** SendConnectinData(target) with target address specified by a datagramPacket's destination **/
  public void sendQuitMessage(DatagramPacket d){sendData(NETDEFS.QUITSTRING, d.getPort() );}
  /** SendConnectinData(target) with target address specified by a target datagramsocket **/
  public void sendQuitMessage(DatagramSocket s){sendData(NETDEFS.QUITSTRING, s.getLocalPort() );}
  /** SendConnectinData(target) with target address specified by a target LocalDatagramHandler **/
  public void sendQuitMessage(LocalDatagramHandler h){sendData(NETDEFS.QUITSTRING, h.getSocketPort() );}
  /** sends the connection data to the handlers default address**/
  public void sendQuitMessage(){sendData(NETDEFS.QUITSTRING,targetPort);}


  /**
   * Wraps the string into a 'message'  like the ExternalsocketHandler would send on a sendMessage(message) command
   *<p> this function is public to allow to check a message received from this ExternalsocketHandler against the result expected
   * @param message : the message that would be sendt
   * @return the string as the current ExternalsocketHandler would send it
   */
  public String buildSendMessage (String message)  {return( buildSendMessage (handlerName, message) );}
  /**
   * as buildSendMessage(message), but a public function that can be called specifying a desired ExternalsocketHandler
   * @param handler : the handler specified
   * @param message : the message to send
   * @return the string as an ExternalsocketHandler would send it
   */
  static String buildSendMessage (LocalDatagramHandler handler,String message){return( buildSendMessage(handler.getName(), message));}
  /**
   * as buildSendMessage(message), but a public function that can be called specifying the desired ExternalsocketHandler name
   * @param name : the name of the handler that is asked to perform the send
   * @param message : the message to send
   * @return the string as an ExternalsocketHandler would send it
   */
  static String buildSendMessage (String name,String message){return("Sent by <"+name+"> : ["+message+"]");}

/****************************************************************************************************************************************/
  /**
   * waits for a datagram message to arrive(and blocks while doing so)
   * @return the received data string
   * <p> note that the message storing buffer is of size DEFAULTBUFSIZE == 255 bytes,
   * A message longer then 255 bytes (127 chars) will be truncated) </p>
   * <p><b> THIS FUNCTION BLOCKS THE WHOLE APPLICATION UNTIL THE EXPECTED MESSAGE HAS ARIVED. THEREFORE ONLY CALL THIS FUNCTION
   * WHEN YOU DEFINITELY KNOW THAT A MESSAGE IS ON ITS WAY </b>
   */
  public String awaitMessage()
  {
    return receiveMessage(DEFAULTBUFSIZE, 0, false);
  }
  /**
   * waits for a datagram message to arrive(and blocks while doing so)
   * @return the received data string
   * @param bufersize the size in characters (NOT IN BYTES) of the message buffer. A message bigger then this size will be truncated
   * <p><b> THIS FUNCTION BLOCKS THE WHOLE APPLICATION UNTIL THE EXPECTED MESSAGE HAS ARIVED. THEREFORE ONLY CALL THIS FUNCTION
   * WHEN YOU DEFINITELY KNOW THAT A MESSAGE IS ON ITS WAY </b>
   */
  public String awaitMessage(int buffersize)
  {
    return receiveMessage(buffersize,  0, false);
  }
  /**
   * as receiveMessage() but with the option to verbose the received message to the test harness
   * @return the received data string
   * @param bufersize the size in characters (NOT IN BYTES) of the message buffer. A message bigger then this size will be truncated
   * @param verbose: if true, this function will verbose the message directly after it received it
   * <p><b> THIS FUNCTION BLOCKS THE WHOLE APPLICATION UNTIL THE EXPECTED MESSAGE HAS ARIVED. THEREFORE ONLY CALL THIS FUNCTION
   * WHEN YOU DEFINITELY KNOW THAT A MESSAGE IS ON ITS WAY </b>
  */

  public String awaitMessage(boolean verbose)
  {
    return receiveMessage(DEFAULTBUFSIZE, 0, verbose);
  }
  /**
   * as receiveMessage() but with the option to verbose the received message to the test harness
   * @return the received data string
   * @param bufersize the size in characters (NOT IN BYTES) of the message buffer. A message bigger then this size will be truncated
   * @param verbose: if true, this function will verbose the message directly after it received it
   * <p><b> THIS FUNCTION BLOCKS THE WHOLE APPLICATION UNTIL THE EXPECTED MESSAGE HAS ARIVED. THEREFORE ONLY CALL THIS FUNCTION
   * WHEN YOU DEFINITELY KNOW THAT A MESSAGE IS ON ITS WAY </b>
  */

  public String awaitMessage(int buffersize, boolean verbose)
  {
    return receiveMessage(buffersize, 0, verbose);
  }
  /**
   * waits for a datagram message to arrive(and blocks while doing so), but defines a maximum timeout time
   * If the message hasn't arrived within that specified time, the function nevertheless closes,
   * so the maximum blocking time is known, and can be bargained with
   * @param bufersize the size in characters (NOT IN BYTES) of the message buffer. A message bigger then this size will be truncated
   * @param timeout, the maximum time to wait for the message (ergo the maximum time this function wil hang)
   * @return the received data or a NULL string on timeout
   */
  public String receiveMessage(long timeout)
  {
   return receiveMessage(DEFAULTBUFSIZE, timeout, false);
  }
  /**
   * waits for a datagram message to arrive(and blocks while doing so), but defines a maximum timeout time
   * If the message hasn't arrived within that specified time, the function nevertheless closes,
   * so the maximum blocking time is known, and can be bargained with
   * @param bufersize the size in characters (NOT IN BYTES) of the message buffer. A message bigger then this size will be truncated
   * @param timeout, the maximum time to wait for the message (ergo the maximum time this function wil hang)
   * @return the received data or a NULL string on timeout
   */
  public String receiveMessage(int buffersize, long timeout)
  {
   return receiveMessage(buffersize, timeout, false);
  }
  /**
   * as receiveMessage(timeout) but with the option to verbose received message to the test harness
   * @param bufersize the size in characters (NOT IN BYTES) of the message buffer. A message bigger then this size will be truncated
   * @param timeout, the maximum time to wait for the message (ergo the maximum time this function wil hang)
   * @param verbose: if true, this function will verbose the message directly after it received it
   * @return the received data or a NULL string on timeout
   */
  //public synchronized String receiveMessage(long timeout, boolean verbose)
  public String receiveMessage(long timeout, boolean verbose)
  {
   return receiveMessage(DEFAULTBUFSIZE, timeout, verbose);
  }
  /**
   * as receiveMessage(timeout) but with the option to verbose received message to the test harness
   * @param bufersize the size in characters (NOT IN BYTES) of the message buffer. A message bigger then this size will be truncated
   * @param timeout, the maximum time to wait for the message (ergo the maximum time this function wil hang)
   * @param verbose: if true, this function will verbose the message directly after it received it
   * @return the received data or a NULL string on timeout
   */
  //public synchronized String receiveMessage(long timeout, boolean verbose)
  public String receiveMessage(int buffersize, long timeout, boolean verbose)
  {
   String message = null;
   if(verbose)
     verbose("Waiting for data, buffersize = "+buffersize+" timeout =  "+timeout);

   try
   {
      //set timeout
      getSocket().setSoTimeout((int)timeout);
      // build receiving datagram
      byte[] inbuffer = new byte[buffersize*2]; //buffersize in char, not in byte
      DatagramPacket inpacket = new DatagramPacket(inbuffer, buffersize*2);
      //wait for message to arrive
      getSocket().receive(inpacket);
      //get the data out of the datagram
      message = NETDEFS.bytesToString(inpacket.getData(),inpacket.getLength() );

      if(verbose)
        verbose("Received message [" +message+"]");
    }
    catch (InterruptedIOException e) //timed out
    {
      if(verbose)
        verbose(e.toString() );
    }
    catch (Exception e)
    {
       harness.fail(e.toString() );
    }
    return message;
  }

/****************************************************************************************************************************************/
  /**
   * calls a ReplyThread listener to wait untill ONE message is received by the socket, and then mirrors this message back to the sender
   * <p>(actually this function will wrap the received data into a String <buildReplyMessage(data)> and send this)
   * @return the function  itself returns void ,but mirrors back the received data to sender as soon as it arrives
   * <p> note that the message storing buffer is of size DEFAULTBUFSIZE == 255 bytes,
   * A message longer then 255 bytes (127 chars) will be truncated) </p>
   */
  public void replyMessage()
  {
    replyMessage(DEFAULTBUFSIZE, false,false);
   /** <p> note that the message storing buffer is of size DEFAULTBUFSIZE == 255 bytes,
   A message longer then 255 bytes (127 chars) will be truncated) </p>  **/
  }
  /**
   * calls a ReplyThread listener to wait untill ONE message is received by the socket, and then mirrors this message back to the sender
   * <p>(actually this function will wrap the received data into a String <buildReplyMessage(data)> and send this)
   * @return the function  itself returns void ,but mirrors back all received data to sender as soon as it arrives
   * <p> note that the message storing buffer is of size DEFAULTBUFSIZE == 255 bytes,
   * A message longer then 255 bytes (127 chars) will be truncated) </p>
   */
  public void replyMessagesContinuously()
  {
    replyMessagesContinuously(DEFAULTBUFSIZE, false, false);
   /** <p> note that the message storing buffer is of size DEFAULTBUFSIZE == 255 bytes,
   A message longer then 255 bytes (127 chars) will be truncated) </p>  **/
  }

  /**
   * as replyMessage() but with the option to verbose the received message to the test harness immedialtely after mirroring it back
   * @param verbose: if true, this function will verbose every message it received and mirrorred back to the test harness
   * @return the function  itself returns void ,but mirrors back the received data to sender as soon as it arrives
   */
  public void replyMessage(boolean verbose)
  {
    replyMessage(DEFAULTBUFSIZE, verbose, false);
   /** <p> note that the message storing buffer is of size DEFAULTBUFSIZE == 255 bytes,
   A message longer then 255 bytes (127 chars) will be truncated)   **/
  }
  /**
   * as replyMessagesContinuously(), but with the option to verbose all mirrored strings to the test harness
   * @param verbose: if true, this function will verbose every message it received and mirrorred back to the test harness
   * @return the function  itself returns void ,but mirrors back all received data to sender as soon as it arrives
   */
  public void replyMessagesContinuously(boolean verbose)
  {
    replyMessagesContinuously(DEFAULTBUFSIZE, verbose, false);
   /** <p> note that the message storing buffer is of size DEFAULTBUFSIZE == 255 bytes,
   A message longer then 255 bytes (127 chars) will be truncated)   **/
  }

  /**
   * as replyMessage() but with the option to verbose the received message to the test harness immedialtely after mirroring it back
   * @param verbose: if true, this function will verbose every message it received and mirrorred back to the test harness
   * @return the function  itself returns void ,but mirrors back the received data to sender as soon as it arrives
   */
  public void replyMessage(boolean verbose, boolean closedown)
  {
    replyMessage(DEFAULTBUFSIZE, verbose, closedown);
   /** <p> note that the message storing buffer is of size DEFAULTBUFSIZE == 255 bytes,
   A message longer then 255 bytes (127 chars) will be truncated)   **/
  }
  /**
   * as replyMessagesContinuously(), but with the option to verbose all mirrored strings to the test harness
   * @param verbose: if true, this function will verbose every message it received and mirrorred back to the test harness
   * @return the function  itself returns void ,but mirrors back all received data to sender as soon as it arrives
   */
  public void replyMessagesContinuously(boolean verbose, boolean closedown)
  {
    replyMessagesContinuously(DEFAULTBUFSIZE, verbose, closedown);
   /** <p> note that the message storing buffer is of size DEFAULTBUFSIZE == 255 bytes,
   A message longer then 255 bytes (127 chars) will be truncated)   **/
  }

  /**
   * calls a ReplyThread listener to wait untill ONE message is received by the socket, and then mirrors this message back to the sender
   * This functin allows you to specify the size (in characters) of the receiving message buffer. A message bigger then this buffer size
   * will be truncated
   * @param bufersize the size in characters (NOT IN BYTES) of the message buffer. A message bigger then this size will be truncated
   * @return the function  itself returns void ,but mirrors back the received data to sender as soon as it arrives
   */
  public void replyMessage(int buffersize)
  {
    replyMessage(buffersize, false,false);
  }
  /**
   * calls a ReplyThread listener to wait untill ONE message is received by the socket, and then mirrors this message back to the sender
   * This functin allows you to specify the size (in characters) of the receiving message buffer. A message bigger then this buffer size
   * will be truncated
   * @param bufersize the size in characters (NOT IN BYTES) of the message buffer. A message bigger then this size will be truncated
   * @return the function  itself returns void ,but mirrors back all received data to sender as soon as it arrives
   */
  public void replyMessagesContinuously(int buffersize)
  {
    replyMessagesContinuously(buffersize, false, false);
  }


  /**
   * as replyMessage(buffersize) but with the option to verbose the received message to the test harness immedialtely after mirroring it back
   * @param verbose: if true, this function will verbose every message it received and mirrorred back to the test harness
   * @param bufersize the size in characters (NOT IN BYTES) of the message buffer. A message bigger then this size will be truncated
   * @return the function  itself returns void ,but mirrors back the received data to sender as soon as it arrives
   */
  public void replyMessage(int buffersize, boolean verbose)
  {
    replyMessage(buffersize, verbose,false);
  }
  /**
   * as replyMessagesContinuously(buffersize), but with the option to verbose all mirrored strings to the test harness
   * @param verbose: if true, this function will verbose every message it received and mirrorred back to the test harness
   * @param bufersize the size in characters (NOT IN BYTES) of the message buffer. A message bigger then this size will be truncated
   * @return the function  itself returns void ,but mirrors back all received data to sender as soon as it arrives
   */
  public void replyMessagesContinuously(int buffersize, boolean verbose)
  {
    replyMessagesContinuously(buffersize, verbose, false);
  }
  /**
   * as replyMessage(buffersize, verbose) but with the option to close down the handler after the last message is received and sent back
   * @param verbose: if true, this function will verbose every message it received and mirrorred back to the test harness
   * @param closedown: if true, this function will call ExternalSocketHandler.close() to clean up the handler as soon as it has finished
   * @param bufersize the size in characters (NOT IN BYTES) of the message buffer. A message bigger then this size will be truncated
   * @return the function  itself returns void ,but mirrors back the received data to sender as soon as it arrives
   */
  public void replyMessage(int buffersize, boolean verbose, boolean closedown)
  {
    DatagramReplyThread mirror = new DatagramReplyThread(handlerName, threadCount++, buffersize * 2, this, verbose, closedown);
    mirror.start();
  }
  /**
   * as replyMessagesContinuously(buffersize, verbose), but with the option to close the handler when the last message is received and sent back
   * @param verbose: if true, this function will verbose every message it received and mirrorred back to the test harness
   * @param closedown: if true, this function will call ExternalSocketHandler.close() to clean up the handler as soon as it has finished
   * @param bufersize the size in characters (NOT IN BYTES) of the message buffer. A message bigger then this size will be truncated
   * @return the function  itself returns void ,but mirrors back all received data to sender as soon as it arrives
   */
  public void replyMessagesContinuously(int buffersize, boolean verbose, boolean closedown)
  {
    DatagramReplyThreadLoop mirror = new DatagramReplyThreadLoop(handlerName, threadCount++, buffersize * 2, this, verbose, closedown);
    mirror.start();
  }

  /**
   * Wraps the string into a 'message' like ExternalsocketHandler woulds reply back to sender on a ReplyMessage(message) command
   *<p> this function is public to allow to check a message received from this ExternalsocketHandler against the result expected
   * @param message : the message to send
   * @return the string as the current ExternalsocketHandler would send it
   */
  public String buildReplyMessage (String message)  {return( buildReplyMessage (handlerName, message) );}
  /**
   * as buildReplyMessage(message), but a public function for a given LocalDatagramHandler
   * @param handler : the specified handler
   * @param message : the message that would be replyed
   * @return the string as an ExternalsocketHandler would send it
   */
  static String buildReplyMessage (LocalDatagramHandler handler,String message){return( buildReplyMessage(handler.getName(), message));}
  /**
   * as buildReplyMessage(message), but a public function for a given LocalDatagramHandler name
   * @param handlername : the name of the handler that would to perform the reply
   * @param message : the message that would be replyed
   * @return the string as an ExternalsocketHandler would send it
   */
  static String buildReplyMessage (String name, String message){return("replied to sender by <"+name+"> : ["+message+"]");}


/****************************************************************************************************************************************/
  /**
   * closes the socket
   * @return <code>true</code>if closed successfully <code>false</code>if either error while closing, or socket already closed
   */
  public boolean close()
  {
    if(socketActive)
    {
      try
      {
        getSocket().close();
        socketActive = false;
        verbose(" Handler Socket closed");
      }
      catch (Exception e)
      {
         fail(e.toString() );
      }
    }
    return (!socketActive);
  }


  /** return true if socket active (not yet closed) **/
  public boolean isActive() {return socketActive;}


//end of class ExternalSocketHandler
}














/******************************************************************************************************************************************
*   externalSocketHandler  helper thread classes. These are designed to 'decouple' hanging functions to run in a separate thread
* and alow the main socket handler to continue its job
*/

/**
****************************************************************************************************************************************
*  posts a datagram <delay> milliseconds in the future
**/
class DelayedDatagramSender extends Thread
{
  // basic definitions
  private LocalDatagramHandler  parent;
  private int                   threadCount;

  private String                messageData;
  private int                   messagePort;
  private long                  messageDelay;

  private boolean               closeOnExit;

  public DelayedDatagramSender(String name, int count, String data, int targetport, long delay, LocalDatagramHandler handler, boolean close)
  {
    super(name+count);

    threadCount   = count;
    messageData   = data;
    messagePort   = targetport;
    messageDelay  = delay;
    parent        = handler;
    closeOnExit   = close;
  }

  public void run()
  {
    try
    {
      //wait until the desired time is elapsed
      sleep(messageDelay);

      // send the data
      byte[] outbuffer = NETDEFS.stringToBytes(messageData); //messageData.getBytes();
      DatagramPacket outPacket = new DatagramPacket(outbuffer, outbuffer.length, InetAddress.getLocalHost(), messagePort);
      DatagramSocket socket = parent.getSocket();
      socket.send(outPacket);
    }
    catch(Exception e)
    {
      parent.fail("PostThread ("+threadCount+"): " + e.toString() );
    }
    //close down if desired
    if(closeOnExit)
      parent.close();
  }
}


/**
****************************************************************************************************************************************
*  waits for ONE datagram to be sent to the handler, then mirrors back its data to the sender in a reply datagram
*  "<"+parentName+"> received and mirrored ["+message+"]"
**/
class DatagramReplyThread extends Thread
{
  // basic definitions
  private int                  threadCount;
  private LocalDatagramHandler parent;

  private boolean verboseReply;
  private boolean closeOnExit;

  private DatagramPacket inPacket;
  private DatagramPacket outPacket;


  public DatagramReplyThread(String name, int count, int expectedbytes, LocalDatagramHandler handler, boolean verbose, boolean close)
  {
    super(name+count);

    threadCount   = count;
    parent        = handler;
    verboseReply  = verbose;
    closeOnExit   = close;

    byte[]inbuffer = new byte[expectedbytes];
    inPacket = new DatagramPacket(inbuffer, expectedbytes);
  }

  public void run()
  {
    try
    {
      DatagramSocket socket = parent.getSocket();

      //wait for datagram to arrive
      socket.receive(inPacket);

      // handle received message
      String reply = parent.buildReplyMessage(NETDEFS.bytesToString(inPacket.getData(),inPacket.getLength()) );
      byte[] outbuffer = NETDEFS.stringToBytes(reply); //reply.getBytes();

      //verbose if wanted
      if(verboseReply)
        parent.verbose("DatagramReplyThread ("+threadCount+") sent ["+reply+"] to "+inPacket.getAddress()+" (port "+inPacket.getPort()+")");

      //send back received message
      outPacket = new DatagramPacket(outbuffer, outbuffer.length, inPacket.getAddress(), inPacket.getPort());
      socket.send(outPacket);
      /*
      alternate method using the same packet:
      inpacket.setAddress(inpacket.getAddress());
      inpacket.setPort(inPacket.getPort());
      inPacket.setData(outbuffer);
      inPacket.setLength(outbuffer.length);
      */
    }
    catch (Exception e)
    {
       parent.fail("ReplyThread ("+threadCount+"): " +e.toString() );
    }

    if(closeOnExit)
       parent.close();
  }
}

/**
****************************************************************************************************************************************
*  waits for messages to be sent to the handler in a continuous loop. Upon receiving a message ,mirrors back it back in a reply string
*  "<"+parentName+"> received and mirrored ["+message+"]"
* this function will continue running untill it receives a message 'quit' : when calling this function in a program, you should always
* send a quit-message before closing down the program to avoid leaving threads open
**/
class DatagramReplyThreadLoop extends Thread
{
  // basic definitions
//  private String          parentName;
  private int                     threadCount;
  private LocalDatagramHandler parent;

  private boolean verboseReply;
  private boolean closeOnExit;

  private DatagramPacket inPacket;
  private DatagramPacket outPacket;


  public DatagramReplyThreadLoop(String name, int count, int expectedbytes, LocalDatagramHandler handler, boolean verbose, boolean close)
  {
    super(name+count);

    threadCount   = count;
    parent        = handler;
    verboseReply  = verbose;
    closeOnExit   = close;

    byte[]inbuffer = new byte[expectedbytes];
    inPacket = new DatagramPacket(inbuffer, expectedbytes);
  }

  public void run()
  {
    try
    {
      DatagramSocket socket = parent.getSocket();
      String message;
      String reply;
      byte[] outbuffer;

      do
      {
        //wait for datagram to arrive
        socket.receive(inPacket);

        // handle received message
        message = NETDEFS.bytesToString(inPacket.getData(),inPacket.getLength());
        reply = parent.buildReplyMessage(message );
        outbuffer = NETDEFS.stringToBytes(reply); //reply.getBytes();

        //verbose if wanted
        if(verboseReply)
        parent.verbose("DatagramReplyThread ("+threadCount+") sent ["+reply+"] to "+inPacket.getAddress()+" (port "+inPacket.getPort()+")");

        //send back received message
        outPacket = new DatagramPacket(outbuffer, outbuffer.length, inPacket.getAddress(), inPacket.getPort());
        socket.send(outPacket);
      /*
      alternate method using the same packet:
      inpacket.setAddress(inpacket.getAddress());
      inpacket.setPort(inPacket.getPort());
      inPacket.setData(outbuffer);
      inPacket.setLength(outbuffer.length);
      */
      }
      while(!message.equals(NETDEFS.QUITSTRING));
    }
    catch (Exception e)
    {
       parent.fail("ReplyThread ("+threadCount+"): " +e.toString() );
    }

    if(closeOnExit)
       parent.close();
  }
}
