/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
*                                                                         *
* Modifications copyright (c) 2005 by Chris Gray, /k/ Embedded Java       *
* Solutions. Permission is hereby granted to distribute under the terms   *
* of the Wonka Public Licence.                                            *
*                                                                         *
**************************************************************************/

/*
** $Id: exec.c 4294 2006-05-10 11:35:19Z chris $
*/

#include <signal.h>
#include <errno.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <sys/socket.h>
#include <sys/signal.h>
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include "hashtable.h"
#include "wonka.h"
#include "oswald.h"
#include "exec.h"
#include "network.h"

extern char *current_working_dir;

/*
** Spawn a child process used to execute 'cmd', using path 'path'. Note that
** currently 'env' is not used (bug???). The child process's stdin, stdout,
** and stderr are redirected to new fd's, which are returned via fd_in, fd_out,
** and fd_err. Returns the pid of the child process, or <0 if fork/exec failed.
*/
w_int host_exec(char **cmd, char **env, char *path, w_int *fd_in, w_int *fd_out, w_int *fd_err) {
  w_int pid;
  w_int socks_inout[2];
  w_int socks_err[2];

  socketpair(PF_UNIX, SOCK_STREAM, 0, socks_inout);
  socketpair(PF_UNIX, SOCK_STREAM, 0, socks_err);

  *fd_in = socks_inout[0];
  *fd_out = socks_inout[0];
  *fd_err = socks_err[0];
  
  fcntl(*fd_in, F_SETFL, O_NONBLOCK);
  fcntl(*fd_err, F_SETFL, O_NONBLOCK);
  
  pid = vfork();

  if (pid < 0) {
    perror("mika: Runtime.exec(): vfork");
  }
  else if (pid == 0) {

    /*
    ** Close stdin, stdout and stderr.
    */
    
    close(0);
    close(1);
    close(2);

    /*
    ** Reassign stdin, stdout and stderr to the socketpairs.
    */

    dup2(socks_inout[1], 0);
    dup2(socks_inout[1], 1);
    dup2(socks_err[1], 2);

    /*
    ** Change path 
    */

    chdir(current_working_dir);

    if(path) chdir(path);

    /*
    ** Execute program.
    */ 

    if (execvp(cmd[0], (char **)cmd)) {
      perror("mika: Runtime.exec(): execvp:");
    }
    
    woempa(9, "Execution of child process failed...\n");

    exit(0);
  }

  woempa(9, "child process has pid %d\n", pid);

  return pid;

}

/**
 * The rc_hashtable is used to store process exit codes which are read as a
 * by-product of reading the process's stdout or stderr. This is a bit of an
 * Ugly Hack (tm), but it solves the problem that we have only one chance to
 * read the exit code before the kernel throws it away.
 */
static w_hashtable rc_hashtable;

/**
 * Check to see if this pid is already in the rc_hashtable, and if so return
 * the stored exit code associated with it; the stored exit code is deleted.
 * If no entry was found, the dummy value 10001 is returned.
 */
static w_int check_rc_hashtable(w_int pid) {
  w_int rc;

  if (rc_hashtable) {
    ht_lock(rc_hashtable);
    rc = ht_read_no_lock(rc_hashtable, pid);
    ht_erase_no_lock(rc_hashtable, pid);
    ht_unlock(rc_hashtable);

    return rc;

  }

  return 10001;
}

/**
 * Store the association of a pid with an exit code in the hashtable.
 * If the hashtable does not yet exist, it will be created (with the
 * special value '10001' as its null value).
 */
static void store_rc_hashtable(w_int pid, w_int rc) {
  if (!rc_hashtable) {
    rc_hashtable = ht_create(W_Realm_system, "hashtable: process return codes", 7, NULL, NULL, 0, 10001);
  }
  ht_write(rc_hashtable, pid, rc);
}

/**
 * Remove an entry from the rc_hashtable. (May safely be called if no such
 * entry exists, or even if rc_hashtable has not yet been created).
 */
static inline void clear_rc_hashtable(w_int pid) {
  if (rc_hashtable) {
    ht_erase(rc_hashtable, pid);
  }
}

/**
 * Kill a child process. This is called by Process.finalizer(), and also
 * serves to clean up the entry in rc_hashtable.
 */
w_void host_destroy(w_int pid) {
  clear_rc_hashtable(pid);
  kill(pid, 15);
}

#define ERESTARTSYS 512

w_int host_wait(w_int pid) { 
  w_int result = 0;
  w_int status;
  w_int rc;

  rc = check_rc_hashtable(pid);
  if (rc != 10001) {

    return rc;

  }
 
  while(result == 0) {
    result = wait4(pid, &status, WNOHANG, NULL);
    x_thread_sleep(x_millis2ticks(200));
  } 

  if(result == pid) {
    if(WIFEXITED(status)) {
      rc = WEXITSTATUS(status);
      store_rc_hashtable(pid, rc);
      return rc;
    }
    else {
      return 10001;
    }
  }
  else {
    return 10001;
  }
}

// CG HACK
#define x_read read
#define x_write write

w_int host_read(w_int pid, w_int fd) {
  w_ubyte b;
  w_int   result;

  result = x_read(fd, &b, 1);
  while(result == -1 && (errno == EAGAIN || errno == EINTR)) {
    x_thread_sleep(x_millis2ticks(200));
    result = x_read(fd, &b, 1);
    
    if(result == -1 && errno == EAGAIN) {
      w_int status;
      w_int res = wait4(pid, &status, WNOHANG, NULL);
      if(res == pid) {

        if (WIFEXITED(status)) {
	  store_rc_hashtable(pid, WEXITSTATUS(status));
	}
      
        return -1;
      }
    }
  }
  /*
  if(result == -1){
    return EXECUTION_ERROR;
  }
  else */ 
  if(result == 0) {
    return -1;
  }
  else {
    return b;
  }
}

w_void host_write(w_int pid, w_int fd, w_int b) {
  w_ubyte c = b;
  x_write(fd, &c, 1);
}

w_void host_close(w_int pid, w_int fd) {
  close(fd);
}

w_int host_available(w_int pid, w_int fd){
  w_int arg=0;
  w_int res = ioctl(fd, FIONREAD, &arg);
  if(res == -1) {
    woempa(9, "Error in Available 'ioctl' failed\n");
    return EXECUTION_ERROR;
  }
  woempa(6, "Available bytes %x\n",arg);
  return arg;  		
}

w_int host_read_Array(w_int pid, w_int fd, char* bytes, w_int len) {
  w_int result;

  result = x_read(fd, bytes, (size_t)len);
  while(result == -1 && errno == EAGAIN) {
    x_thread_sleep(x_millis2ticks(200));
    result = x_read(fd, bytes, (size_t)len);
  
    if(result == -1 && errno == EAGAIN) {
      w_int status;
      w_int res = wait4(pid, &status, WNOHANG, NULL);
      if(res == pid) {
        
        if (WIFEXITED(status)) {
	  store_rc_hashtable(pid, WEXITSTATUS(status));
	}
      
        return -1;
      }
    }
  }
  /*
  if(result == -1) {
    return EXECUTION_ERROR;
  }

  else */
  if(result == 0) {
    return -1;
  }
  else {
    return result;
  }
}

