/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: StreamTokenizer.java 3799 2003-04-04 07:56:22Z dumon $
*/

package java.io;

public class StreamTokenizer {

  public int ttype;
  public double nval;
  public String sval;

  public static final int TT_EOF = -1;      //values specified according the java language specification
  public static final int TT_EOL= '\n';     //values specified according the java language specification
  public static final int TT_NUMBER = -2;   //values specified according the java language specification
  public static final int TT_WORD = -3;     //values specified according the java language specification

  private Type[] types;
  private Type   defaultType = new Type();
  
  private boolean cppstyle = false; //slashslash comments allowed
  private boolean cstyle   = false;	//slashstar comments allowed	
  private boolean eolSign  = false; //end_of_line taken into account
  private boolean tolower  = false; //translate chars to lowerCase ...
  private int lineNr = 1;
  private boolean pushback = false;
  private boolean eof = false;

  private PushbackReader buffer;

  public StreamTokenizer(Reader reader) {
    buffer = new PushbackReader(new BufferedReader(reader));
    
    types = new Type[256];
    for(int i=0; i < 256; i++) {
      types[i] = new Type();
    }
    commentChar('/');
    parseNumbers();
    quoteChar('\'');
    quoteChar('"');
    whitespaceChars(0, ' ');
    wordChars('a', 'z');
    wordChars('A', 'Z');
    wordChars(0xA0, 0xFF);
  }
/**
** @deprecated
*/
  public StreamTokenizer(InputStream is) {
    this(new InputStreamReader(is));
  }

  public void commentChar(int ch) {
    if(0 <= ch && ch <= 255) types[ch].comment = true;
  }
 
  public void eolIsSignificant(boolean eolSignificant) {
    eolSign = eolSignificant;
  }

  public int lineno() {
    return lineNr;
  }

  public void lowerCaseMode(boolean lower) {
    tolower = lower;
  }

  private int readChar() throws IOException {
    if(eof) return -1;
    int ch = buffer.read();
    return ch;
  }

  private void unreadChar(int ch) throws IOException {
    if(ch == -1) {
      eof = true;
    }
    else {
      buffer.unread(ch);
    }
  }

  public int nextToken() throws IOException {

    boolean possible_c = false;
    boolean possible_cpp = false;

    /*
    ** If pushback is not set, get on to the next token.
    ** otherwise return with the previous result.
    */

    if(!pushback) {

      /*
      ** Clear sval and read the next character
      */

      sval = "";
      int ch = readChar();

      /*
      ** Check if it might be the beginning of a c or a c++ comment.
      */
      if(ch == '/') {
        int ch2 = readChar();
        if(ch2 == '*') possible_c = true;
        if(ch2 == '/') possible_cpp = true;
        unreadChar(ch2);
      }
      
      /*
      ** Determine the type of the read character and handle it.
      */

      Type t = getType(ch);

      if(t.quote) {
        handleQuote(ch);
      }
      else if(t.white) {
        return handleWhite(ch);
      }
      else if(t.digit) {
        handleDigit(ch);
      }
      else if(t.word) {
        handleWord(ch);
      }
      else if(possible_cpp && cppstyle) {
        handleCPPComment(ch);
      }
      else if(possible_c && cstyle) {
        handleCComment(ch);
        nextToken();
      }
      else if(t.comment) {
        handleComment(ch);
      }
      else {

        /*
        ** None of the above. Set sval to null and
        ** set ttype to the character that was read.
        */
        if(ch < 256 || ch < 0){
          sval = null;
          ttype = ch;
        }
        else {
          sval = ""+((char)ch);
          ttype = TT_WORD;
        }
      }
    }
    else {

      /*
      ** Pushback was switched on, turn it off.
      */
      
      pushback = false;
    }

    return ttype;
  }
 
  private int handleWhite(int ch) throws IOException {

    /*
    ** Skip whitespace
    */

    do {

      /* 
      ** Check if we reached the end of the line
      */

      if(ch == '\r' || ch == '\n') {

        lineNr++;

        /*
        ** if the character is \r, check if the next one is \n
        */

        if(ch == '\r') {
          int ch2 = readChar();
          if(ch2 != '\n') unreadChar(ch2);
        }

        /*
        ** Stop if EOL's are significant
        */
        
        if(eolSign) {
          ttype = TT_EOL;
          sval = null;
          return ttype;
        }
      }
      
      ch = readChar();

      /*
      ** Keep doing this as long as we don't reach EOF or the 
      ** character is no longer whitespace.
      */
      
    } while(getType(ch).white && ch != -1);
    
    unreadChar(ch);

    /*
    ** Parse the next token since whitespace is discarded.
    */
    
    return nextToken();
  }
  
  private void handleWord(int ch) throws IOException {
  
    sval = "";

    /*
    ** Keep reading as long as they are word or digit
    ** characters.
    */
    
    while(getType(ch).word || getType(ch).digit) {
      sval += (char)ch;
      ch = readChar();
    }
    
    unreadChar(ch);
    ttype = TT_WORD;

    /*
    ** Convert to lowercase if needed.
    */
    
    if(tolower) sval = sval.toLowerCase();
  }
  
  private void handleDigit(int ch) throws IOException {

    /*
    ** Counter for the number of dots. Only one dot is
    ** allowed in a number. First of all, check if the
    ** first character read is a dot.
    */
    
    int dots = (ch == '.' ? 1 : 0);

    sval = "";
    ttype = TT_NUMBER;

    /*
    ** Keep on going as long as we get digits and
    ** there no more than 1 dot.
    */

    while(getType(ch).digit && dots < 2) {
      
      sval += (char)ch;
      
      ch = readChar();

      /*
      ** Check if we read a dot.
      */
      
      if (ch == '.') dots++;

      /*
      ** A minus sign is only allowed at the beginning of
      ** a number.
      */
      
      if (ch == '-') break;
    } 

    unreadChar(ch);

    /*
    ** Try to parse the number.
    */
    
    try {
      nval = Double.parseDouble(sval);
    }
    catch (NumberFormatException x) {

      /*
      ** Something went wrong.
      */
      
      if (sval.equals("-")) {

        /*
        ** There's only a minus sign...
        */
        
        ttype = '-';
      }
      else if (sval.equals(".")) {

        /*
        ** There's only a dot...
        */
        
        nval = 0.0;
      }
      else {

        /*
        ** The result is '-.'
        */
        
        nval = -0.0;
      }		
    }
  }
  
  private void handleComment(int ch) throws IOException {

    /*
    ** Skip to the end of the line.
    */
  
    while(ch != '\r' && ch != '\n' && ch != -1) {

      ch = readChar();

      /*
      ** Check if we are at the end of the line.
      */
      

    }

    if(ch != -1){
      unreadChar(ch);
    }

    /*
    ** Since comments are skipped, get on to the next token.
    */

    nextToken();
  }
  
  private void handleCComment(int ch) throws IOException {

    /*
    ** Read the *
    */
    
    ch = readChar();

    /*
    ** Keep on going until we reach a * followed by a /
    */
    
    while(true) {
      ch = readChar();

      if(ch == '*') {

        /*
        ** It's a *, check if the next character is a /
        */

        int ch2 = readChar();
        if(ch2 == '/') {

          /*
          ** It's a slash, the end of the comment has been reached.
          */

          break;
        }
        else {

          /*
          ** It's not a /, keep on looking..
          */

          unreadChar(ch2);
        }
      }
      else if(ch == -1) {

        /*
        ** If we've reached the end of the file, also stop.
        */

        break;
      }
      else if(ch == '\r' || ch == '\n') {

        /*
        ** We've reached the end of a line. C comments keep on going as
        ** long as we've don't reached a * and a /
        */

        lineNr++;

        if(ch == '\r') {

          /*
          ** Check for \r\n
          */

          int ch2 = readChar();
          if(ch2 != '\n') unreadChar(ch2);
        }
      }
    }

  }
  
  private void handleCPPComment(int ch) throws IOException {
   
    /*
    ** Read the second /
    */
    
    ch = readChar();

    /*
    ** Handle it as a normal comment.
    */
    
    handleComment(ch);
  }
  
  private void handleQuote(int ch) throws IOException {

    /*
    ** ttype is set to the quote character.
    */
    
    ttype = ch;
    sval = "";

    ch = readChar();

    /*
    ** Keep on reading until we find the other quote char or have
    ** reached EOL or EOF.
    */
    
    while(ch != ttype && ch != -1 && ch != '\n' && ch !='\r') {
      if(ch == '\\') {

        /*
        ** The character has to be escaped.
        */

        ch = readChar();

        switch(ch) {
          case 'n':  ch = '\n';  break;
          case 'r':  ch = '\r';  break;
          case 't':  ch = '\t';  break;
          default:
        }
      }
      sval += (char)ch;
      ch = readChar();
    }

    /*
    ** If we read EOL, push it back.
    */
    
    if(ch == '\r' || ch == '\n') unreadChar(ch);
  }

  public void ordinaryChar(int ch) {
    if(0 <= ch && ch <= 255) {
      types[ch].comment = false;
      types[ch].quote = false;
      types[ch].digit = false;
      types[ch].word = false;
      types[ch].white = false;
    }
  }
  
  public void ordinaryChars(int lo, int hi) {
    if(lo < 0) lo = 0;
    if(hi > 255) hi = 255;
    for(int i=lo; i<=hi; i++) ordinaryChar(i);
  }
  
  public void parseNumbers() {
    for (int ch = '0'; ch <= '9'; ch++) {
      types[ch].digit = true;
    }
    types['.'].digit = true;
    types['-'].digit = true;
  }
  
  public void pushBack(){
    pushback=true;
  }

  public void quoteChar(int ch) {
    if(0 <= ch && ch <= 255) types[ch].quote = true;
  }
  
  public void resetSyntax() {
    ordinaryChars(0, 255);
  }
  
  public void slashSlashComments(boolean cpp) {
    cppstyle = cpp;
  }

  public void slashStarComments(boolean cpp) {
    cstyle = cpp;
  }

  public String toString(){
    StringBuffer buf = 	new StringBuffer(64);
    buf.append("StreamTokenizer: line = ");
    buf.append("" + lineNr);
    buf.append(" last token :'");
    switch(ttype){
      case TT_WORD:
        buf.append(sval);
        break;
      case TT_NUMBER:
        buf.append(nval);
        break;
      case TT_EOF:
        buf.append("EOF");
        break;
      case TT_EOL:
        buf.append("EOL");
        break;
      default:
        buf.append((char)ttype);
    }
    buf.append('\'');
    return buf.toString();	
  }
  
  public void whitespaceChars(int lo, int hi) {
    if(0 <= lo && lo <= 255 && 0 <= hi && hi <= 255) {
      for(int i=lo; i<=hi; i++) {
        types[i].white = true;
      }
    }
  }
  
  public void wordChars(int lo, int hi) {
    if(0 <= lo && lo <= 255 && 0 <= hi && hi <= 255) {
      for(int i=lo; i<=hi; i++) {
        types[i].word = true;
      }
    }
  }

  private Type getType(int ch) {
    if(0 <= ch && ch <= 255) {
      return types[ch];
    }
    else {
      return defaultType;
    }
  }

  private class Type {
    public boolean comment = false;
    public boolean quote;
    public boolean digit;
    public boolean word;
    public boolean white;
  } 

}
