/**************************************************************************
* Copyright (c) 2001, 2002, 2003 by Acunia N.V. All rights reserved.      *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Philips site 5, box 3       info@acunia.com                           *
*   3001 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/*
** $Id: Runtime.java 3874 2003-04-28 10:25:52Z dbuytaert $
*/

package java.lang;

import java.io.IOException;
import java.io.OutputStream;
import java.io.InputStream;
import java.io.*;
import java.util.Vector;
import java.util.StringTokenizer;
import wonka.vm.GarbageCollector;
import wonka.vm.Wonka;

/**
** A bunch of methods used to manipulate the runtime environment.
*/

public class Runtime {

  private static boolean debugVal; // value of the debug property (if debugSet)
  private static boolean debugSet; // true iff debug property has been read

  /**
  ** The single unique instance of Runtime.
  */
  
  private static Runtime theRuntime;

  /**
  ** The single unique instance of our colleague the Finalizer.
  */
  
  private static wonka.vm.Finalizer theFinalizer;

  /**
  ** The number of instances of FinaDonna created and not yet finalized.
  ** Sycnchronize on FinaDonna.class when modifying this.
  */
  
  private static int finalizatora;

  /**
  ** The Threads which should be started on shutdown.
  */
  
  private Vector shutdownHooks;

  /**
  ** Set to true iff a shutdown is in progress.
  */
  
  private boolean shuttingDown;

  /**
  ** Inner class used to implement runFinalization()
  */
  
  private class FinaDonna {

    public FinaDonna() {
      synchronized (FinaDonna.class) {
        ++finalizatora ;
      }
    }

    protected void finalize() {
      synchronized (FinaDonna.class) {
        --finalizatora ;
        FinaDonna.class.notifyAll();
      }
    }
  }
        
  /**
  ** Static initialiser: creates the singleton.
  */
  
  static {
    theRuntime = new Runtime();
  }

  /**
  ** Private constructor to block attempts to create a new instance.
  */
  
  private Runtime() {
  }

  /**
  * Returns if debug is on
  * The first time the method is called we read the wonka.java.debug property
  * For every subsequent call we just return the value read
  * We need this method because trying the read the debug property in the 
  * constructor causes a seg fault
  */

  private boolean debug() {
     if (!debugSet)  {
        String debugProperty = Wonka.getProperty("wonka.java.debug", "false");
        debugVal = debugProperty.equals("true");
        debugSet = true;
     }
     return debugVal;
  }

  private void permissionCheck(String permission) {
    if (wonka.vm.SecurityConfiguration.USE_ACCESS_CONTROLLER) {
      java.security.AccessController.checkPermission(new RuntimePermission(permission));
    }
    else if (wonka.vm.SecurityConfiguration.USE_SECURITY_MANAGER) {
      SecurityManager sm = System.getSecurityManager();
      if (sm != null) {
        sm.checkPermission(new RuntimePermission(permission));
      }
    }
  }

  /**
  ** Add a Thread to the list of threads to be started when we decide
  ** to shut down the system.
  */
  
  public void addShutdownHook(Thread t) {
    permissionCheck("shutdownHooks");

    synchronized(this) {
      if (shuttingDown) {
        throw new IllegalStateException();
      }

      if (shutdownHooks == null) {
        shutdownHooks = new Vector();
      }

      if (shutdownHooks.contains(t)) {
        throw new IllegalArgumentException();
      }

      shutdownHooks.add(t);
    }
  }

  /**
  ** Remove a Thread from the list of threads to be started when we decide
  ** to shut down the system.
  */
  
  public boolean removeShutdownHook(Thread t) {
    permissionCheck("shutdownHooks");

    synchronized(this) {
      if (shutdownHooks == null) {

        return false;

      }

      if (shuttingDown) {
        throw new IllegalStateException();
      }

      return shutdownHooks.remove(t);
    }
  }

  /*
  ** Method which should be called to run all the shutdown hooks.
  ** The LIFO behaviour is not mandated by the specification.
  */
  
  private void runShutdownHooks() {
    synchronized(this) {
      if (shuttingDown || shutdownHooks == null) {
        return;
      }

      shuttingDown = true;
    }

    int n = shutdownHooks.size();
    while (n > 0) {
      Thread t = (Thread)shutdownHooks.elementAt(--n);
      if (debug()) {
        System.out.println("Runtime: starting shutdown hook "+t);
      }
      t.start();
    }

    n = shutdownHooks.size();
    while (n > 0) {
      Thread t = (Thread)shutdownHooks.elementAt(n-1);
      try {
        t.join();
        --n;
      }
      catch (InterruptedException ie) {}
      if (debug()) {
        System.out.println("Runtime: joined shutdown hook "+t);
      }
      shutdownHooks.removeElementAt(n);
    }
    shutdownHooks = null;

    if (debug()) {
      System.out.println("Runtime: all shutdown hooks have completed");
    }
  }

  /**
  ** Return the solitary instance of Runtime.
  */
  
  public static Runtime getRuntime() {
    return theRuntime;
  }

  /**
  ** Native method called by exit() after performing all checks.
  */
  
  private native void exit0(int status);

  /**
  ** Exit the VM, after running all shutdown hooks.
  */
  
  public void exit(int status) throws SecurityException {
    permissionCheck("VMexit");

    if (debug()) {
      System.err.println("java.lang.Runtime/exit called with status " + status + ", running shutdown hooks.");
    }
    runShutdownHooks();
    if (debug()) {
      System.err.println("Finished running shutdown hooks, goodnight.");
    }
    
    try {
      System.in.close();
      System.out.close();
      System.err.close();
    }
    catch (Throwable t) {}

    exit0(status);
  }
    
  public void halt(int status) throws SecurityException {
    permissionCheck("VMexit");
    System.err.println("java.lang.Runtime/halt called with status " + status + ", skip running shutdown hooks.");
    exit0(status);
  }
    
  /**
  **  Not allowed ...
  **  @remark always throws a SecurityException
  */
  
  public Process exec(String command) throws IOException, SecurityException {
     return exec(command, null);
  }

  /**
  **  Not allowed ...
  **  @remark always throws a SecurityException
  */
  
  public Process exec(String command, String envp[]) throws IOException, SecurityException {
    return exec(command, envp, null);
  }

  public Process exec(String command, String envp[], File path) throws IOException, SecurityException {
    StringTokenizer st = new StringTokenizer(command);
    String cmd[] = new String[st.countTokens()];
    int i = 0;
    while(st.hasMoreTokens()) {
      cmd[i++] = st.nextToken();
    }
    return exec(cmd, envp, path);
  }

  /**
  **  Not allowed ...
  **  @remark always throws a SecurityException
  */
  
  public Process exec(String cmdarray[]) throws IOException, SecurityException{
    return exec(cmdarray, null, null);
  }

  /**
  **  Not allowed ...
  **  @remark always throws a SecurityException
  */
  
  public Process exec(String cmdarray[], String envp[]) throws IOException, SecurityException{
    return exec(cmdarray, envp, null);
  }

  public Process exec(String cmdarray[], String envp[], File path) throws IOException, SecurityException{
    return exec0(cmdarray, envp, (path != null ? path.getAbsolutePath() : null));
    // throw new SecurityException("execution of commands is not allowed");
  }

  /*
  ** Go native to execute the process.
  */
  
  private native Process exec0(String cmdarray[], String envp[], String path) throws IOException;

  /**
  **  Not allowed ...
  **  @remark always throws a SecurityException
  */
  
  public void load(String filename) throws SecurityException, UnsatisfiedLinkError{
       	throw new SecurityException("load is not allowed");
  }

  /**
  ** Tell the garbage collector to do some work.
  */
  
  public void gc(){
    GarbageCollector theGC = GarbageCollector.getInstance();
    theGC.request(Integer.MAX_VALUE);
  }

  /**
  ** Create an instance of FinaDonna and then throw it away.
  */
  private void generateFinaDonna() {
    new FinaDonna();
  }

  /**
  ** Ask the garbage collector to run finalizers.  Actually what we really
  ** do is to create an instance of FinaDonna and then make it unreachable,
  ** kick the Garbage Collector, and wait for the finalizer to be run.
  */
  
  public void runFinalization(){

    GarbageCollector theGC = GarbageCollector.getInstance();
    synchronized(FinaDonna.class) {
      generateFinaDonna();
      try {
        while (finalizatora  > 0) {
          theGC.kick();
          FinaDonna.class.wait(1000);
        }
      }
      catch (InterruptedException ie) {}
    }

  }

  /**
  **  @status not implemented
  */
  
  public void traceInstructions(boolean on){}
  
  /**
  ** 	@status not implemented
  */
  
  public void traceMethodCalls(boolean on){}


  /**
  **	@depricated
  **    @remark always throws a runtime exception
  */
  
  public InputStream getLocalizedInputStream(InputStream in){
   	// throw new RuntimeException("Depricated Method");
    return new BufferedInputStream(in);
  }
  
  /**
  **	@depricated
  **    @remark always throws a runtime exception
  */
  
  public OutputStream getLocalizedOutputStream(OutputStream out){
   	// throw new RuntimeException("Depricated Method");
    return out;
  }

  public native long totalMemory();
  public native long freeMemory();

  private static native ClassLoader getCallingClassLoader();
  private static native ClassLoader getCallingCallingClassLoader();
  
  public void loadLibrary(String libname) throws SecurityException, UnsatisfiedLinkError {

    SecurityManager sm = System.getSecurityManager();

    if (sm != null) {
//      sm.checkLink(new RuntimePermission(libname));
    }

    String path = null;
   
    /*
    ** First try to look it up with the findLibrary method of ClassLoader.
    */
    
    ClassLoader cl = getCallingClassLoader();
    if(cl == null) cl = getCallingCallingClassLoader();
    
    if(cl != null) path = cl.findLibrary(libname);
    if(cl != null && path == null) path = cl.findLibrary("lib" + libname + ".so");

    if(path != null) {

      /*
      ** Found it with the classloader.
      */
      
      loadLibrary0(null, path);
    }
    else {

      /*
      ** No luck -> get the path from java.library.path and look it up.
      */
      
      path = System.getProperty("java.library.path", null);
      loadLibrary0(libname, path);
    }

  }
    
  private native void loadLibrary0(String libname, String libpath) throws UnsatisfiedLinkError;
}

