/**************************************************************************
* Copyright (c) 2001, 2002, 2003 by Acunia N.V. All rights reserved.      *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Philips site 5, box 3       info@acunia.com                           *
*   3001 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: Thread.java 3451 2003-02-06 14:25:34Z gray $
*/

package java.lang;

import java.util.Iterator;
import java.util.HashMap;

public class Thread implements Runnable {

  /*
  ** Note: this class is initialized "by hand" before the VM is fully
  ** initialized.  Consequently it must not have a static initializer.
  ** (It can have static variables, and even constant initial values
  ** for those variables, but nothing fancier and certainly no static{}
  ** clause.)
  */
  
  public final static int MIN_PRIORITY  =  1;
  public final static int MAX_PRIORITY  = 10;
  public final static int NORM_PRIORITY =  5;

  /**
   ** name       is the name of this Thread.
   */
  private String name;

  /**
   ** started    is true iff start() has been called on this thread.
   **            N.B. Remains true when the thread dies! (see 'stopped' below).
   */
  private boolean started;

  /**
   ** stopped    is true iff the thread has been started and subsequently died
   **            (either as a result of stop()/destroy() or of natural causes).
   */
  private boolean stopped;

  /**
   ** runObject  is either this Thread (if the current thread was created
   **             using a subclass of Thread) or the Runnable object used
   **             to create the thread.
   */
  private Runnable runObject;

  /**
   ** parent     is the ThreadGroup of which this thread is a member.
   */
  private ThreadGroup parent;

  /**
   ** context_classloader is the ClassLoader of last resort to be used
   ** by this Thread.
   */
  private ClassLoader context_classloader;

  /**
   ** threadLocals  maps instances of ThreadLocal (or its subclasses,
   **            including InheritableThreadLocal) onto the corresponding
   **            values for this Thread.
   */
  private HashMap threadLocals;

  /**
   ** thrown     used to store a pending exception 
   */
  private Throwable thrown; 

  private ClassLoader cl;

  private static void permissionCheck(String permission) {
    if (wonka.vm.SecurityConfiguration.USE_ACCESS_CONTROLLER) {
      java.security.AccessController.checkPermission(new RuntimePermission(permission));
    }
    else if (wonka.vm.SecurityConfiguration.USE_SECURITY_MANAGER) {
      SecurityManager sm = System.getSecurityManager();
      if (sm != null) {
        sm.checkPermission(new RuntimePermission(permission));
      }
    }
  }

  /**
   ** Thread(ThreadGroup group, Runnable runObject, String name
   **   --> name as specified, or "Thread-<n>" if name==null; 
   **       runObject as specified;
   **       parent group as specified, or if this is null then 
   **       sm.getThreadGroup() or parent of current thread.
   */
  public Thread(ThreadGroup group, Runnable runObject, String name) 
    throws SecurityException,
    IllegalThreadStateException 
  {
    ThreadGroup parent;
    SecurityManager sm = System.getSecurityManager();

    if (group == null) {
      if (sm == null) {
        parent = Thread.currentThread().getThreadGroup();
      }
      else {
        parent = sm.getThreadGroup();
      }
    }
    else parent = group;

    this.parent = parent;

    if (parent.isDestroyed()) {
      throw new IllegalThreadStateException();
    }

    if (sm != null) {
      sm.checkAccess(parent);
    }

    this.runObject = runObject;

    inheritThreadLocals(Thread.currentThread());
    context_classloader = currentThread() == null ? null : currentThread().getContextClassLoader();

    create(parent, name, runObject);
  }

  /**
   ** Thread constructor (Runnable runObject, String name)
   **   --> Thread(ThreadGroup,Runnable,String) with null ThreadGroup
   */
  public Thread(Runnable runObject, String name) {
    this((ThreadGroup)null,runObject,name);
  }

  /**
   ** Thread(ThreadGroup group,Runnable runObject)
   **   --> Thread(ThreadGroup,Runnable,String) with null String
   */
  public Thread(ThreadGroup group, Runnable runObject) 
    throws SecurityException,
    IllegalThreadStateException 
  {
    this(group,runObject,(String)null);
  }

  /**
   ** Thread(Runnable)
   **   --> Thread(ThreadGroup,Runnable,String) with null ThreadGroup, String
   */
  public Thread(Runnable runObject) {
    this((ThreadGroup)null,runObject,(String)null);
  }

  /**
   ** Thread(ThreadGroup group,String name)
   **   --> name as specified, or "Thread-<n>" if name==null; 
   **       Thread is own runObject;
   **       parent group as specified, or if this is null then 
   **       sm.getThreadGroup() or parent of current thread.
   */
  public Thread(ThreadGroup group, String name) 
    throws SecurityException,
    IllegalThreadStateException 
  {
    ThreadGroup parent;
    SecurityManager sm = System.getSecurityManager();

    if (group == null) {
      if (sm == null) {
        parent = Thread.currentThread().getThreadGroup();
      }
      else {
        parent = sm.getThreadGroup();
      }
    }
    else parent = group;

    this.parent = parent;

    if (parent.isDestroyed()) {
      throw new IllegalThreadStateException();
    }

    if (sm != null) {
      sm.checkAccess(parent);
    }

    this.runObject = this;

    inheritThreadLocals(Thread.currentThread());
    context_classloader = currentThread() == null ? null : currentThread().getContextClassLoader();

    create(parent, name, null);
  }

  /**
   ** Thread constructor (String) --> name as specified, or "Thread-<n>"
   ** if name==null; Thread is own runObject, parent is sm.getThreadGroup() 
   ** or parent of current thread.
   */
  public Thread(String name) {
    this((ThreadGroup)null,name);
  }

  /**
   ** Thread constructor with no arguments --> Thread(String) with null argument
   */
  public Thread() {
    this((ThreadGroup)null,(String)null);
  }

  /**
   ** create(ThreadGroup, String, Runnable) creates the native w_Thread and
   ** osthread structures, sets up the mapping osthread->w_Thread in the
   ** thread hashtable, and calls ThreadGroup.registerThread(this).
   */
  private native void create(ThreadGroup group, String name, Runnable runnable);

  /**
   ** activeCount() returns the number of threads executing in the
   ** current ThreadGroup and its children.  How pointless.
   */
  public static int activeCount() {
    return currentThread().getThreadGroup().activeCount();
  }

  /**
   ** run() : 
   ** If the runObject of this Thread is the Thread itself, does nothing.
   ** Otherwise, call method run() of runObject.
   */
  public void run() {
    if(runObject!=this) runObject.run();
  }

  /**
   ** checkAccess() checkes whether the SecurityManager (if any) objects
   ** to our performing priviledged actions on this thread.
   */
  public final void checkAccess() 
    throws SecurityException
  {
    if (wonka.vm.SecurityConfiguration.USE_ACCESS_CONTROLLER) {
      if (parent.getParent() == null) {
        java.security.AccessController.checkPermission(new RuntimePermission("modifyThread"));
      }
    }
    else if (wonka.vm.SecurityConfiguration.USE_SECURITY_MANAGER) {
      SecurityManager sm = System.getSecurityManager();
      if (sm != null) {
        sm.checkAccess(this);
      }
    }
  }

  /**
   ** countStackFrames() should count the number of stack frames in 
   ** this thread: the thread must be suspended.  NOT YET IMPLEMENTED.
   */
  public native int countStackFrames();

  /**
   ** _run() is the method which is used to define the initial frame
   ** when create() is called.  When a thread is started using start(),
   ** the result is to execute run() and then (when _run() terminates)
   ** to send a termination request to the ThreadGroup manager.
   */
  void _run() {
    try {
      runObject.run();
    } catch (Throwable t) {
      if (t instanceof ThreadDeath) {
      }
      else {
        parent.uncaughtException(this,t);
      }
    }
    stopped = true;
  }

  /**
   ** destroy() is supposed to destroy the thread "without performing
   ** any cleanup".  Implemented as a synonym of stop0(new ThreadDeath()).
   */
  public void destroy() {
    stop0(new ThreadDeath());
  }

  /**
   ** dumpStack() is prints a stack trace to System.err.
   */
  public static void dumpStack() {
    new Exception("Thread.dumpStack called").printStackTrace();
  }

  /**
   ** enumerate() enumerates the active threads in the current ThreadGroup
   ** and its child groups.  Another pointless synonym for a ThreadGroup method.
   */
  public static int enumerate(Thread[] threads) {
    return currentThread().getThreadGroup().enumerate(threads);
  }
 
  /**
   ** toString() creates a string which includes the thread's name, priority,
   ** and parent ThreadGroup.
   */
  public String toString() {
    return "Thread[" + name + "," + getPriority() + "," + parent.getName() +"]";
  }

  /**
   ** start() causes the Thread to be started, invoking its run() method
   ** (via a call to Thread._run()).
   */
  public void start() 
    throws IllegalThreadStateException
  {
    if (started) {
      if (stopped) {
          //System.out.println(this + " already ran, doing nothing");

        return; // silently

      }
      throw new IllegalThreadStateException(this+" already active");
    }
    //System.out.println("Setting "+this+"/started true");
    int retcode = start0();
    started = true;
    if (retcode != 0) {
      //System.out.println("Oops! Setting "+this+"/started false");
      started = false;
      throw new InternalError("Could not start "+this+", OSwald return code = "+retcode);
    }
  }

  /**
   ** Try to start the native thread.  Returns 0 on success, OSwald status on failure.
   */
  public native int start0() ;

  /**
   ** Try to join a thread.
   */
  public native void join0(long millis, int nanos);

  /**
   ** Try to sleep for a while.
   */
  public native void sleep0(long millis, int nanos);

  /**
   ** stop0() causes exception "thr" to be thrown in the thread, by sending
   ** an appropriate message to its ThreadGrouop manager.
   */
  private native void stop0(Throwable thr);

  /**
   ** stop() performs an access check and then throws a ThreadDeath in
   ** the victim thread.
   */
  public final synchronized void stop()
    throws SecurityException
  {
    checkAccess();

    stop0(new ThreadDeath());
  }

  /**
   ** stop(Throwable thr) performs an access check and then throws "thr" in
   ** the victim thread.
   */
  public final void stop(Throwable thr)
    throws SecurityException, NullPointerException
  {
    checkAccess();

    if(thr==null) {
      throw new NullPointerException();
    }

    stop0(thr);
  }


  /**
   ** suspend0() suspends a thread
   */
  private final native void suspend0();

  /**
   ** resume0() resumes a thread which was previously suspended.
   ** Note: suspension and resumption are binary, not counting
   ** (one resume0() will undo any number of suspend0()s).
   */
  private final native void resume0();

  /**
   ** suspend0() performs a checkAccess() before calling suspend0().
   */
  public final void suspend()
    throws SecurityException
  {
    checkAccess();

    suspend0();
  }



  /**
   ** resume0() performs a checkAccess() before calling resume0().
   */
  public final void resume()
    throws SecurityException
  {
    checkAccess();

    resume0();
  }



  /**
   ** getName() returns the name of the Thread
   */
  public final native String getName();

  /**
   ** setName() changes the name of the Thread.
   */
  private final native void setName0(String name);

  /**
   ** setName(String name) invokes checkAccess() before calling setName0(name).
   ** If parameter "name" is null then we set the name to "".
   */
  public final void setName(String name) 
    throws SecurityException
  {
    checkAccess();

    if(name==null) {
      setName0("");
    } else {
      setName0(name);
    }
  }

  /**
   ** getThreadGroup() returns the parent ThreadGroup.
   */
  public final ThreadGroup getThreadGroup() {
    return parent;
  }

  /**
   ** getPriority() returns the current priority of this thread.
   */
  public final native int getPriority();

  /**
   ** setPriority(int) changes the current priority of this thread.
   */
  private final native void setPriority0(int newPriority);

  /**
   ** setPriority(int newPriority) first calls checkAccess(), and then 
   ** calls setPriority0() to set the priority to the lesser of
   **   - the value requested; or
   **   - the maxPriority of the parent ThreadGroup.
   */
  public final void setPriority(int newPriority)
    throws SecurityException, IllegalArgumentException
  {
    checkAccess();

    if(newPriority<MIN_PRIORITY || newPriority>MAX_PRIORITY)
      throw new IllegalArgumentException("priority "+newPriority+" not allowed");

    if (parent!=null) {
      newPriority = Math.min(newPriority,parent.getMaxPriority());
    }
    setPriority0(newPriority);
  }


  /**
   ** isDaemon() returns the status of the Thread's daemon flag.
   */
  public final native boolean isDaemon();

  /**
   ** setDaemon0() sets the status of the Thread's daemon flag.
   ** Note: the daemon flag has no significance in Wonka.
   */
  private final native void setDaemon0(boolean on);

  /**
   ** setDaemon() first calls checkAccess() and then invokes setDaemon0()
   ** - unless the thread is currently active, in which case 
   ** IllegalThreadStateException is thrown.
   */
  public final void setDaemon(boolean on)
    throws SecurityException, IllegalThreadStateException
  {
    checkAccess();

    if(isAlive()) throw new IllegalThreadStateException(this+"");

    setDaemon0(on);
  }

  /**
   ** isAlive() returns true iff the thread is "alive" (alive, oh-oh).
   */
  public final boolean isAlive() {
    //System.out.println(this+"/started = "+started+", stopped = "+stopped+", returning "+(started && !stopped));
    return started && !stopped;
  }


  /**
   ** join() blocks the calling thread until this Thread has terminated
   ** (see the implementation of _run()).  A thread should not try to
   ** join() itself!
   */
  public final void join() throws InterruptedException {
    join(0,0);
  }

  /**
   ** join(millis) blocks the calling thread until either this Thread has 
   ** terminated or the stated number of milliseconds elapse.
   */
  public final void join(long millis) throws InterruptedException {
    join(millis,0);
  }

  /**
   ** join(millis,nanos) blocks the calling thread until either this Thread 
   ** has terminated or the stated time elapses.
   */
  public final void join(long millis, int nanos) throws InterruptedException {
    if (!isAlive()) {
      return;
    }
    join0(millis, nanos);
  }

  /**
   ** interrupt() causes the interrupt status flag to be set, and any current
   ** wait() or sleep() to be aborted.
   */
  public native synchronized void interrupt();

  /**
   ** isInterrupted() tests the state of the interrupt status flag, without
   ** affecting its value.
   */
  public native boolean isInterrupted();

  /** interrupted() tests the state of the interrupt status flag,  and if
   ** the flag was set it resets it (atomically)
   */
  public static native synchronized boolean interrupted();

  /**
   ** currentThread() returns the Thread object corresponding to the
   ** currently running thread (i.e., the calling thread).
   */
  public static native Thread currentThread();

  /**
   ** yield() causes the calling thread to "sleep" for an infinitesimal
   ** length of time (in order to give other threads at the same priority
   ** a chance to run).
   */
  public static native void yield();

  /**
   ** sleep(millis) causes the calling thread to "sleep" for
   ** the specified length of time.
   */
  public static void sleep(long millis) throws InterruptedException {
    sleep(millis, 0);
  }

  /**
   ** sleep(millis,nanos) causes the calling thread to "sleep" for
   ** the specified length of time.
   */
  public static void sleep(long millis, int nanos)
    throws InterruptedException
  {
    currentThread().sleep0(millis,nanos);
  }

  /**
   ** inheritThreadLocals(Thread t) copies those threadLocals of t
   ** which are instances of InheritableThreadLocal into the threadLocals
   ** table of the current Thread (in other words, it inherits them).
   */
  private void inheritThreadLocals(Thread t) {
    HashMap   h = t.threadLocals;

    if (h==null) return;

    Iterator i = h.keySet().iterator();

    if (threadLocals == null) {
      threadLocals = new HashMap();
    }

    while (i.hasNext()) {
      try {
        InheritableThreadLocal ihl = (InheritableThreadLocal)(i.next());
        threadLocals.put(ihl,t.threadLocals.get(ihl));
      }
      catch (ClassCastException x) {}
    }
  }

  /**
   ** setThreadLocal(ThreadLocal tl, Object obj) sets the value of tl
   ** in the current thread to obj.
   ** TODO: only allow this to be called by tl itself.
   */
  void setThreadLocal(ThreadLocal tl, Object obj) {
    if (threadLocals == null) {
      threadLocals = new HashMap();
    }

    threadLocals.put(tl,obj);
  }


  /**
   ** getThreadLocal(ThreadLocal tl) returns the value of tl in the
   ** current thread.  If no such value exists then one is created
   ** using initialValue().
   ** TODO: only allow this to be called by tl itself.
   */
  Object getThreadLocal(ThreadLocal tl) {
    if (threadLocals == null) {
      threadLocals = new HashMap();
    }

    return threadLocals.get(tl);
  }

  /**
   ** deleteThreadLocal(ThreadLocal tl) deletes the value associated with tl.
   ** TODO: only allow this to be called by tl itself.
   */
  void deleteThreadLocal(ThreadLocal tl) {
    if (threadLocals != null) {
    threadLocals.remove(tl);
    }
  }

  public ClassLoader getContextClassLoader() {
    ClassLoader caller = ClassLoader.getCallingClassLoader();
    if (caller != null && !caller.isDelegationAncestor(context_classloader)) {
      permissionCheck("getClassLoader");
    }

    if (context_classloader == null) {
      context_classloader = ClassLoader.getSystemClassLoader();
    }

    return context_classloader;
  }

  public void setContextClassLoader (ClassLoader cl) {
    ClassLoader caller = ClassLoader.getCallingClassLoader();

    if (caller != null) {
      if (context_classloader == null) {
        context_classloader = ClassLoader.getSystemClassLoader();
      }
      if (!caller.isDelegationAncestor(context_classloader)) {
        permissionCheck("setContextClassLoader");
      }
    }

    context_classloader = cl;
  }

}

