package java.net;

import java.util.Enumeration;
import java.util.Vector;

public final class NetworkInterface
{
   /**
    * The interface name
    */
   private String name;
 
   /**
    * Currently we only support 1 address per interface
    * The JDK javadoc states that multiple addresses should be supported
    */
   private String ipAddress;
   
   /**
    * Required for output in toString()
    */
   private int index;
   
   /**
    * Gets the network interface with the specified name
    */
   public static NetworkInterface getByName(String name) throws SocketException
   {
       if (name==null)
           throw new NullPointerException();
       NetworkInterface[] interfaces = getNetworkInterfacesNative();
       for (int i=0; i<interfaces.length; i++)
       {
           if (interfaces[i].getName().equals(name))
               return interfaces[i];
       }
       throw new NullPointerException("No network interface exists with name " + name);
   }
   
   /**
    * Gets the network interface with the specified address
    */
   public static NetworkInterface getByInetAddress(InetAddress addr) throws SocketException
   {
       if (addr==null)
           throw new NullPointerException();
       String addressString = addr.getHostAddress();
       NetworkInterface[] interfaces = getNetworkInterfacesNative();
       for (int i=0; i<interfaces.length; i++)
       {
           if (interfaces[i].ipAddress.equals(addressString))
               return interfaces[i];
       }
       throw new NullPointerException("No network interface exists with ip Address " + addressString);
   }
   
   /**
    * Gets all the network interfaces 
    */
   public static Enumeration getNetworkInterfaces() throws SocketException
   {
       NetworkInterface[] interfaces = getNetworkInterfacesNative();
       Vector v = new Vector();
       for (int i=0; i<interfaces.length; i++)
       {
           v.add(interfaces[i]);
       }
       return v.elements();
   }
   
   private static native NetworkInterface[] getNetworkInterfacesNative();
   
   /**
    * Returns the interface display name
    */
   public String getDisplayName()
   {
       return name;
   }
   
   /**
    * Returns the interface name
    */
   public String getName()
   {
       return name;
   }
   
   /**
    * Returns the interface addresses
    */
   public Enumeration getInetAddresses()
   {
       Vector v = new Vector();
       try
       {
          checkConnect(ipAddress);
          v.add(InetAddress.getByName(ipAddress));
       }
       catch (UnknownHostException e) {}
       return v.elements();
   }
   
   /**
    * Check security permissions to see if we can return the ip address
    */
   private void checkConnect(String address)
   {
       if (wonka.vm.SecurityConfiguration.USE_ACCESS_CONTROLLER) 
           java.security.AccessController.checkPermission(new SocketPermission(address,"resolve"));
       else if (wonka.vm.SecurityConfiguration.USE_SECURITY_MANAGER)
       {
           SecurityManager sm = System.getSecurityManager();
           if (sm != null)
               sm.checkConnect(address, -1);
       }
   }
   
   /**
    * Outputs the interface details as a string
    */
   public String toString()
   {
       return "name:" + name + " (" + getDisplayName() + ") index: " + index + " addresses:\n/" + ipAddress + ";";
   }
   
   /**
    * Check if two interfaces are the same
    */
   public boolean equals(Object obj)
   {
       NetworkInterface o = (NetworkInterface) obj;
       if (o.getName().equals(name) && o.ipAddress.equals(ipAddress))
           return true;
       else 
           return false;
   }
   
   /**
    * Returns the hash code
    */
   public int hashCode()
   {
       return name.hashCode() + ipAddress.hashCode();
   }
}
