package java.security;

import java.lang.reflect.Method;
import java.util.Random;

/**
 * An implementation of SecureRandom specifically for the
 * light-weight API, JDK 1.0, and the J2ME. Random generation is 
 * based on the traditional SHA1 with counter. Calling setSeed
 * will always increase the entropy of the hash.
 */
public class SecureRandom extends java.util.Random
{
  private static  SecureRandom rand;

// CG 20040623 Use reflection to get SHA1Digest from BouncyCastle
  private static Class digest_class;
  private static Method update_method;
  private static Method doFinal_method;
  private static Method getDigestSize_method;

  static {
    String digest_class_name = System.getProperty("wonka.security.securerandom.digest", "org.bouncycastle.crypto.digests.SHA1Digest");
    try {
      digest_class = Class.forName(digest_class_name, false, ClassLoader.getSystemClassLoader());
      update_method = digest_class.getMethod("update", new Class[]{byte[].class, int.class, int.class});
      doFinal_method = digest_class.getDeclaredMethod("doFinal", new Class[]{byte[].class, int.class});
      getDigestSize_method = digest_class.getDeclaredMethod("getDigestSize", new Class[0]);

      rand = new SecureRandom();
    }
    catch (ClassNotFoundException cnfe) {
      System.err.println("SecureRandom: cannot find digest class '" + digest_class_name + "', SecureRandom will not work");
    }
    catch (NoSuchMethodException nsme) {
      System.err.println("SecureRandom: digest class '" + digest_class_name + "' lacks " + (update_method == null ? "update([BII)" : doFinal_method == null ? "doFinal([BI)" : "getDigestSize()") + ", SecureRandom will not work");
    }
  }

  private Object digest;

  private byte[] seed;

  private long        counter = 1;
//    private SHA1Digest  digest = new SHA1Digest();
  private byte[]      state; // = new byte[digest.getDigestSize()];

  private void init() {
    try {
      digest = digest_class.newInstance();
    }
    catch (Exception e) {
      System.err.println("SecureRandom: attempt to create instance of " + digest_class + " threw exception, SecureRandom will not work");
      e.printStackTrace();
    }
    try {
      Integer size = (Integer)getDigestSize_method.invoke(digest, new Object[0]);
      state = new byte[size.intValue()];
    }
    catch (Exception e) {
      System.err.println("SecureRandom: attempt to invoke getDigestSize on " + digest + " threw exception, SecureRandom will not work");
      e.printStackTrace();
    }
  }

// public constructors
  public SecureRandom() {
    super(0);
    init();
    setSeed(System.currentTimeMillis());
  }

  public SecureRandom(byte[] inSeed) {
    super(0);
    init();
    setSeed(inSeed);
  }

	// protected constructors
	// protected SecureRandom(SecureRandomSpi srs, Provider provider);

	// public class methods
	public static SecureRandom getInstance(String algorithm)
	{
		return new SecureRandom();
	}

	public static SecureRandom getInstance(String algorithm, String provider)
	{
		return new SecureRandom();
	}

	public static byte[] getSeed(
        int numBytes)
	{
		byte[] rv = new byte[numBytes];

		rand.setSeed(System.currentTimeMillis());
		rand.nextBytes(rv);

		return rv;
	}

	// public instance methods
	public byte[] generateSeed(
        int numBytes)
	{
		byte[] rv = new byte[numBytes];
	
		nextBytes(rv);	

		return rv;
	}

	// public final Provider getProvider();
  public void setSeed(byte[] inSeed) {
  //        digest.update(inSeed, 0, inSeed.length);
    Object[] update_params = new Object[3];
    update_params[0] = inSeed;
    update_params[1] = new Integer(0);
    update_params[2] = new Integer(inSeed.length);
    try {
      update_method.invoke(digest, update_params);
    }
    catch (Exception e) {
      System.err.println("SecureRandom: attempt to invoke update on " + digest + " threw exception");
      e.printStackTrace();
    }
  }
		
	// public methods overriding random
  public void nextBytes(byte[] bytes) {
    int     stateOff = 0;

    Object[] doFinal_params = new Object[2];
    Object[] update_params = new Object[3];

    //  digest.doFinal(state, 0);
    doFinal_params[0] = state;
    doFinal_params[1] = new Integer(0);
    try {
      doFinal_method.invoke(digest, doFinal_params);
    }
    catch (Exception e) {
      System.err.println("SecureRandom: attempt to invoke doFinal on " + digest + " threw exception");
      e.printStackTrace();
    }

    for (int i = 0; i != bytes.length; i++) {
      if (stateOff == state.length) {
        byte[]  b = longToBytes(counter++);

        // digest.update(b, 0, b.length);
        // digest.update(state, 0, state.length);
        // digest.doFinal(state, 0);
        try {
          update_params[0] = b;
          update_params[1] = new Integer(0);
          update_params[2] = new Integer(b.length);
          update_method.invoke(digest, update_params);
          update_params[0] = state;
          update_params[1] = new Integer(0);
          update_params[2] = new Integer(state.length);
          update_method.invoke(digest, update_params);
          doFinal_params[0] = state;
          doFinal_params[1] = new Integer(0);
          doFinal_method.invoke(digest, doFinal_params);
        }
        catch (Exception e) {
          System.err.println("SecureRandom: attempt to invoke update/doFinal on " + digest + " threw exception");
          e.printStackTrace();
        }
        stateOff = 0;
      }
      bytes[i] = state[stateOff++];
    }

    byte[]  b = longToBytes(counter++);

    // digest.update(b, 0, b.length);
    // digest.update(state, 0, state.length);
    try {
      update_params[0] = b;
      update_params[1] = new Integer(0);
      update_params[2] = new Integer(b.length);
      update_method.invoke(digest, update_params);
      update_params[0] = state;
      update_params[1] = new Integer(0);
      update_params[2] = new Integer(state.length);
      update_method.invoke(digest, update_params);
    }
    catch (Exception e) {
      System.err.println("SecureRandom: attempt to invoke update/doFinal on " + digest + " threw exception");
      e.printStackTrace();
    }
  }

	public void setSeed(
        long rSeed)
	{
        if (rSeed != 0)
        {
		    setSeed(longToBytes(rSeed));
        }
	}

    private byte[]  intBytes = new byte[4];

    public int nextInt()
    {
        nextBytes(intBytes);

        int result = 0;

        for (int i = 0; i < 4; i++)
        {
            result = (result << 8) + (intBytes[i] & 0xff);
        }

        return result;
    }

    protected final int next(
        int numBits)
    {
        int     size = (numBits + 7) / 8;
        byte[]  bytes = new byte[size];

        nextBytes(bytes);

        int result = 0;

        for (int i = 0; i < size; i++)
        {
            result = (result << 8) + (bytes[i] & 0xff);
        }

        return result & ((1 << numBits) - 1);
    }

    private byte[]  longBytes = new byte[8];

    private byte[] longToBytes(
        long    val)
    {
        for (int i = 0; i != 8; i++)
        {
            longBytes[i] = (byte)val;
            val >>>= 8;
        }

        return longBytes;
    }
}

