/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: BitSet.java 3195 2002-11-21 11:13:07Z ruelens $
*/

package java.util;

import java.io.*;

/*
* the current implementation of BitSet doesn't allow the BitSet to shrink
* the size will only grow !!!
* a clear will at a position larger then the size will grow the BitSet
*
*/
public class BitSet implements Cloneable, java.io.Serializable {

  private static final long serialVersionUID = 7997698588986878753L;

  private long[] bits;
  private transient int bitsInUse;

/*
*  the size of the bitset is always rounded up to the next 64 increment
*/
  private int roundup (int n) {
    return ((n+63)/64);
  }

  public BitSet() {
    bits = new long[1];
  }
  
  public BitSet(int nbits) {
    if (nbits < 0 ) throw new NegativeArraySizeException();
    bits = new long[roundup(nbits)];
  }

  public String toString() {
    if(bitsInUse == 0){
      return "{}";
    }

    long[] bits = this.bits;
    int length = bits.length;

    StringBuffer buf = new StringBuffer(3*bits.length+2);

    buf.append('{');

    for (int i=0; i < length; ++i) {
      long l = bits[i];
      int j = 0;
      while(l != 0){
        if((l & 0x1) > 0){
          buf.append((i*64 + j));
          buf.append(", ");

        }
        l = l>>>1;
        j++;
      }
    }

    buf.setLength(buf.length()-2);

    buf.append('}');
    return buf.toString();
  }

  public synchronized boolean equals(Object obj) {
    if (!(obj instanceof BitSet)) {
      return false;
    }

    BitSet that = (BitSet)obj;
    int l = bits.length;
    int k = that.bits.length;
    if(k > l){
      for(int i=0 ; i < l ; i++){
        if(bits[i] != that.bits[i]){
          return false;
        }
      }
      for(int j=l ; j < k ; j++){
        if(that.bits[j] != 0){
          return false;
        }
      }
    }
    else {
      for(int i=0 ; i < k ; i++){
        if(bits[i] != that.bits[i]){
          return false;
        }
      }
      for(int j=k ; j < l ; j++){
        if(bits[j] != 0){
          return false;
        }
      }
    }

    return true;
  }
  
  public synchronized int hashCode() {
    long h = 1234;

    for (int i = bits.length-1 ; i >= 0; i--) {
      h ^= bits[i] * (i+1);
    }

    return (int)((h>>32)^h);

  }
  
  public synchronized Object clone() {
    try {
      BitSet clone = (BitSet)super.clone();
      clone.bits = (long[])bits.clone();
      return clone;
    }
    catch(CloneNotSupportedException cne){
      return null;
    }
  }

  
  public boolean get(int bitIndex) {
    if(bitIndex < 0){
      throw new IndexOutOfBoundsException();
    }
    if(bitsInUse < bitIndex){
      return false;
    }
    return (((bits[bitIndex/64])>>(bitIndex%64)) & 0x01) > 0;
  }
  
  private void grow() {
    if(bitsInUse > bits.length * 64){
      long[] nbits = new long[roundup(bitsInUse)];
      System.arraycopy(bits, 0, nbits, 0, bits.length);
      bits = nbits;
    }
  }

  public synchronized void set(int bitIndex) {
    if(bitIndex < 0){
      throw new IndexOutOfBoundsException();
    }
    if (bitIndex>=bitsInUse) {
       bitsInUse = bitIndex+1;
       grow();
    }
    int i = bitIndex / 64;
    bits[i] = bits[i] | (1L<<(bitIndex%64));
  }

  public synchronized void clear(int bitIndex) {
    if(bitIndex < 0){
      throw new IndexOutOfBoundsException();
    }
    if (bitsInUse > bitIndex) {
      int i = bitIndex / 64;
      bits[i] = bits[i] & ((1L<<(bitIndex%64)) ^ -1);
    }
    checkBitsInUse();
  }
  
  public void and(BitSet bitset) {
    int k = bitset.bits.length;
    long[] bits = this.bits;
    int l = bits.length;

    if(k < l){
      for(int j=k ; j < l ; j++){
        bits[j] = 0L;
      }
      l = k;
    }
    long[] others = bitset.bits;
    for(int i=0 ; i < l ; i++){
      bits[i] = bits[i] & others[i];
    }
    checkBitsInUse();
  }

  public void or(BitSet bitset) {
    int k = bitset.bits.length;
    long[] bits = this.bits;
    int l = bits.length;

    if(k < l){
      l = k;
    }
    long[] others = bitset.bits;
    for(int i=0 ; i < l ; i++){
      bits[i] = bits[i] | others[i];
    }
    checkBitsInUse();
  }

  public synchronized void xor(BitSet bitset) {
    int k = bitset.bits.length;
    long[] bits = this.bits;
    int l = bits.length;

    if(k < l){
      l = k;
    }
    long[] others = bitset.bits;
    for(int i=0 ; i < l ; i++){
      bits[i] = bits[i] ^ others[i];
    }
    checkBitsInUse();
  }

  public int size() {
    return bits.length * 64;
  }

/*
* this method is added in JDK 1.2
*/
  public int length() {
    return bitsInUse;
  }	

  public void andNot(BitSet bs) {
    long[] others = bs.bits;
    int k = others.length;
    long[] bits = this.bits;
    int l = bits.length;

    if(k < l){
      l = k;
    }
    for(int i=0 ; i < l ; i++){
      bits[i] = bits[i] & (others[i] ^ -1);
    }
    checkBitsInUse();
  }

  private void checkBitsInUse(){
    long[] bits = this.bits;
    int l = bits.length - 1;

    if(l >= 0){
      while(bits[l] == 0L){
        if(--l < 0){
          bitsInUse = 0;
          return;
        }
      }

      long value = bits[l]>>>1;
      int i = 1;

      while(value != 0){
        value = (value)>>>1;
        i++;
      }
      bitsInUse = l * 64 + i;
    }
  }

  private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
    in.defaultReadObject();
    checkBitsInUse();
  }
}
