/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/**
 * $Id: HashMap.java 4186 2004-07-07 14:00:04Z ruelens $
 */

package java.util;

import java.io.*;

public class HashMap extends AbstractMap implements Map, Cloneable, Serializable {

  private static final long serialVersionUID = 362498820763181265L;

  transient int capacity;
  transient int occupancy;
  private float loadFactor;
  private int threshold;

  final static Object nullKey = new Object();

  private final static float DEFAULT_LOADFACTOR = 0.75f;
  private final static int DEFAULT_CAPACITY = 16;

  transient Object[] key;
  transient Object[] value;
  transient int modCount;

  public HashMap() {
    this(DEFAULT_CAPACITY, DEFAULT_LOADFACTOR);
  }

  public HashMap(int initialCapacity) {
    this(initialCapacity, DEFAULT_LOADFACTOR);
  }

  public HashMap(int initialCapacity, float loadFactor) {
    if ( initialCapacity < 0 || loadFactor <= 0.0f ) {
    	throw new IllegalArgumentException("HashMap needs positive numbers");
    }
    this.capacity = (initialCapacity < 5 ? 5 :initialCapacity);
    this.key = new Object[capacity];
    this.value = new Object[capacity];
    this.loadFactor = (loadFactor > 1.0f ? DEFAULT_LOADFACTOR : loadFactor );
  }

  public HashMap (Map t) {
    this(DEFAULT_CAPACITY, DEFAULT_LOADFACTOR);
    Iterator it = t.entrySet().iterator();
    Map.Entry me;
    while (it.hasNext()) {
  	  me = (Map.Entry) it.next();
  	  put(me.getKey(),me.getValue());
    }
  }

  public int size() {
    return this.occupancy;
  }

  public boolean isEmpty() {
    return (this.occupancy == 0);
  }

  public boolean containsValue(Object value) {
    int nextSlot = firstBusySlot(0);
    Object[] val = this.value;
    if(value == null){
      while(nextSlot>=0) {
        if(val[nextSlot]==null) {
          return true;
        }
        nextSlot = firstBusySlot(nextSlot+1);
      }
    }
    else {
      while(nextSlot>=0) {
        if(value.equals(val[nextSlot])) {
          return true;
        }
        nextSlot = firstBusySlot(nextSlot+1);
      }
    }
    return false;
  }

  public boolean containsKey(Object key) {
    int nextSlot = firstBusySlot(0);
    if (key == null) {
      key = nullKey;
    }
    Object[] k = this.key;
    while(nextSlot>=0) {
      if(key.equals(k[nextSlot])) {
        return true;
      }
      nextSlot = firstBusySlot(nextSlot+1);
    }
    return false;
  } 

  public Object get(Object key)  {
    if(key == null) {
      key = nullKey;
    }

    int cap = this.capacity;
    int hash = key.hashCode() % cap;
    Object[] k = this.key;

    do {
      if(hash < 0){
        hash += cap;
      }
      Object o = k[hash];
      if(o == null) {
        return null;
      } else if(key.equals(o)) {
        return this.value[hash];
      } else hash--;
    } while(true);
  }          

  public Object put(Object key, Object newvalue)  {
    if(key == null) {
      key = nullKey;
    }

    int cap = this.capacity;
    int hash = key.hashCode() % cap;
    Object[] k = this.key;

    do {
      if(hash < 0){
        hash += cap;
      }
      Object o = k[hash];
      if(o == null) {

        k[hash] = key;
        this.value[hash] = newvalue;
        ++this.occupancy;
        if(occupancy >= threshold){
          resize();
        }
        modCount++;
        return null;
      } 
      else if(key.equals(o)) {
        Object[] value = this.value;
        Object oldvalue = value[hash];
        value[hash] = newvalue;
        return oldvalue;
      }
      hash--;
    } while(true);
  }

  public Object remove(Object key)  {
    if(key == null) {
      key = nullKey;
    }

    int cap = this.capacity;
    int hash = key.hashCode() % cap;
    Object[] k = this.key;

    do {
      if(hash < 0){
        hash += cap;
      }
      Object o = k[hash];
      if(o == null) {
        return null; 
      } else {
        if(key.equals(o)) {
          Object oldvalue = this.value[hash];
          deleteSlot(hash);
          modCount++;
          return oldvalue;
        }
      }
      --hash;
    } while(true);
  }

  public void putAll(Map t) {
    super.putAll(t);
  }

  public void clear() {
    if (occupancy > 0) {
      key = new Object[capacity];
      value = new Object[capacity];
      occupancy = 0;
      modCount++;
    }
  }

  public synchronized Object clone() {
    Object that = null;
    try {
      that = super.clone();
      HashMap that_hashmap = (HashMap)that;

      that_hashmap.key = (Object[])key.clone();
      that_hashmap.value = (Object[])value.clone();
    }
    catch (CloneNotSupportedException cnse) {}
    return that;
  } 

  public Set keySet() {
    return new HashMapKeySet();
  }

  public Collection values() {
    return new HashMapCollection();
  }
              
  public Set entrySet() {
    return new HashMapSet();
  }

  private int firstBusySlot(int i) {
    int j;
    int capacity = this.capacity;
    if (i<0 || i>=capacity) {
      return -1;
    }

    Object[] key = this.key;

    for(j=i;j<capacity;++j)
      if (key[j] != null)  {
        return j;
      }
    return -1;
  } 

  private void resize() {
    int oldsize = capacity;
    int newsize = oldsize * 2 + 1;

    Object[] oldkeys = this.key;
    Object[] oldvalues = this.value;
    Object[] keys = new Object[newsize];
    Object[] values = new Object[newsize];
    this.key = keys;
    this.value = values;

    this.capacity = newsize;
    this.threshold = (int)(newsize*loadFactor);

    for (int oldindex=0;oldindex<oldsize;++oldindex) {
      Object k = oldkeys[oldindex];
      if (k != null) {
        int hash = k.hashCode() % newsize;

        do {
          if(hash < 0) hash += newsize;
          if(keys[hash] == null) {
            keys[hash] = k;
            values[hash] = oldvalues[oldindex];
            break;
          } 
          hash--;
        } while(true);
      }
    }
  } 

  private void deleteSlot(int slotIndex) {
    int vacant, current, home, distance1, distance2;
    boolean happy;

    --occupancy;
    current = slotIndex;
    Object[] key = this.key;
    int capacity = this.capacity;

    while(true) {
    // R1
      key[current] = null;
      value[current] = null;
      vacant = current;

    // R2
      happy = true;
      while(happy) {
        current = current==0 ? capacity-1 : current-1;

    // R3
        if (key[current] == null) {

          return; // normal termination

        }

        home = key[current].hashCode() % capacity;
        if(home < 0){
          home += capacity;
        }
        distance1 = current<=home ? home - current : capacity + home - current;
        distance2 = current<=vacant ? vacant - current : capacity + vacant - current;

        happy = distance1<distance2;
      // back to R2
      }

    // R4
      key[vacant] = key[current];
      value[vacant] = value[current];

    // repeat from R1 ...
    }
  }

  protected class HashMapKeyIterator implements Iterator {
    int pos=firstBusySlot(0); //points to the next free slot
    int oldpos=-1; // points to the current slot
    int mc=modCount;
    int status=0;

    public  boolean hasNext() {
 		  return  firstBusySlot(pos)!=-1;
    }

    public Object next() {
      if (modCount != mc) throw new ConcurrentModificationException("don't change the set please");
      status = 1;
      if (firstBusySlot(pos)==-1) throw new NoSuchElementException("no elements left");
		  oldpos = pos;
      pos = firstBusySlot(pos+1);			
		  return (key[oldpos]== nullKey ? null : key[oldpos]);                 	
    }

    public  void remove() {
	    if (modCount != mc) throw new ConcurrentModificationException("don't change the set please");
 	    if (status != 1 ) throw new IllegalStateException("do a next() operation before remove()");
  	  HashMap.this.remove(key[oldpos]);
      mc++;
      status=-1;
    }
  }

  protected class HashMapValueIterator implements Iterator {
  	int pos=firstBusySlot(0); //points to the next free slot
  	int oldpos=-1; // points to the current slot
  	int mc=modCount;
  	int status;

    public  boolean hasNext() {
 		  return  firstBusySlot(pos)!=-1;
    }

    public Object next() {
      if (modCount != mc) throw new ConcurrentModificationException("don't change the set please");
      status = 1;
      if (firstBusySlot(pos)==-1) throw new NoSuchElementException("no elements left");
		  oldpos = pos;
      pos = firstBusySlot(pos+1);			
		  return value[oldpos];                 	
    }

    public void remove() {
	    if (modCount != mc) throw new ConcurrentModificationException("don't change the set please");
 	    if (status != 1 ) throw new IllegalStateException("do a next() operation before remove()");
  	  HashMap.this.remove(key[oldpos]);
      mc++;
      status=-1;
    }
  }

  protected class HashMapIterator implements Iterator {
  	int pos=firstBusySlot(0); //points to the next free slot
  	int oldpos=-1; // points to the current slot
  	int mc=modCount;
  	int status;

    public  boolean hasNext() {
 		  return  firstBusySlot(pos)!=-1;
    }

    public Object next() {
      if (modCount != mc) throw new ConcurrentModificationException("don't change the set please");
      status = 1;
      if (firstBusySlot(pos)==-1) throw new NoSuchElementException("no elements left");
		  oldpos = pos;
      pos = firstBusySlot(pos+1);			
		  return new Entry((key[oldpos]==nullKey ? null : key[oldpos]),value[oldpos]);                 	
    }

    public void remove() {
	    if (modCount != mc) throw new ConcurrentModificationException("don't change the set please");
 	    if (status != 1 ) throw new IllegalStateException("do a next() operation before remove()");
  	  HashMap.this.remove(key[oldpos]);
  	  mc++;
  	  pos = firstBusySlot(pos);
      status=-1;
    }
  }

  protected class HashMapCollection extends AbstractCollection {

    public Iterator iterator() {
      return new HashMapValueIterator();
    }

    public int size() {
      return occupancy;
    }
  }

  protected class HashMapKeySet extends AbstractSet {
    public Iterator iterator() {
      return new HashMapKeyIterator();
    }

    public int size() {
      return occupancy;
    }
  }

  protected class HashMapSet extends AbstractSet {
    public Iterator iterator() {
      return new HashMapIterator();
    }

    public int size() {
      return occupancy;
    }
  }


  class Entry implements Map.Entry {

    private Object key;
    private Object value;

    Entry (Object key, Object value) {
      this.key = key;
      this.value = value;
    }

    public Object getKey() {
      return this.key;
    }

    public Object getValue() {
      return this.value;
    }

    public Object setValue(Object o) throws UnsupportedOperationException, ClassCastException, IllegalArgumentException, NullPointerException {
      return HashMap.this.put(key,o);
    }

    public boolean equals (Object o) {
      boolean answer = false;
      if (o instanceof HashMap.Entry) {
        HashMap.Entry e2 = (HashMap.Entry)o;
        answer = (key == null ? e2.getKey()==null : key.equals(e2.getKey())) &&
                 (value==null ?  e2.getValue()==null : value.equals(e2.getValue()));
      }
      return answer;
    }

    public int hashCode() {
      return (key == null ? 0 : key.hashCode()) ^ (value==null ? 0 : value.hashCode());
    }

    public String toString() {
      String answer = "HashMap Entry: "+key+"="+value;
      return answer;
    }
  }

  private void readObject (ObjectInputStream s) throws IOException, ClassNotFoundException {
    int local_oc;
    s.defaultReadObject();
    this.capacity=s.readInt();
    local_oc=s.readInt();
    this.key = new Object[this.capacity];
    this.value = new Object[this.capacity];
    Object key,value;
    for (int i=0;i<local_oc;i++) {
      key=s.readObject();
      value=s.readObject();
      this.put(key,value);
    }
  }

  private void writeObject (ObjectOutputStream s) throws IOException {
    s.defaultWriteObject();
    s.writeInt (capacity);
    s.writeInt (occupancy);
    int nextSlot=firstBusySlot(0);
    for (int i=0;i<occupancy;i++) {
      s.writeObject((key[nextSlot]==nullKey ? null :key[nextSlot]));
      s.writeObject(value[nextSlot]);
      nextSlot=firstBusySlot(nextSlot+1);
    }               
  }

}
