/**************************************************************************
* Copyright (c) 2001, 2002, 2003 by Acunia N.V. All rights reserved.      *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Philips site 5, box 3       info@acunia.com                           *
*   3001 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/*
** $Id: Init.java 4161 2004-05-14 13:44:19Z dumon $
*/

package wonka.vm;

import java.io.File;
import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.Locale;
import java.util.Properties;
import java.util.TimeZone;
import java.util.Vector;
import java.util.jar.*;
import java.util.StringTokenizer;
import java.util.NoSuchElementException;

/**
 ** Class Init is a utility class used to launch the garbage collection,
 ** finaliser, and heartbeat threads, and to run the starting class or
 ** jarfile.
 */

final class Init {

  private static boolean debug = false;

  private static Method invoke_method;
  private static Object[] invoke_args;
  private static boolean launched;
  private static URLClassLoader application_class_loader;
  private static String [] start_class_args;

  private static void debug(String s) {
    if (debug) {
      System.out.println(s);
    }
  }

  private static String args2string(String[] args) {
    if (args == null || args.length == 0) {
      return "";
    }

    StringBuffer sb = new StringBuffer(args[0]);
    for (int i = 1; i < args.length; ++i) {
      sb.append(" ");
      sb.append(args[i]);
    }

    return sb.toString();
  }

  /**
   ** Take one token from the argument list and process it as a system property
   ** (begins with -D) or a normal argument.  Normal arguments are accumulated
   ** in temp_args.
   */
  private static void argumentOrSysProp(String this_arg, Vector temp_args) {
    if (this_arg.startsWith("-D")) {
      Properties sysprops = System.getProperties();
      String key, value;
      int equals_at = this_arg.indexOf('=');
      if (equals_at < 0) {
        key = this_arg.substring(2);
        debug("Init: Deleting system property '"+key+"'");
        sysprops.remove(key);
      }
      else if (equals_at == 0) {
        System.err.println("Init: Ignoring nameless property in "+this_arg);
      }
      else if (equals_at == this_arg.length()-1) {
        key = this_arg.substring(2,equals_at);
        debug("Init: Deleting system property '"+key+"'");
        sysprops.remove(key);
      }
      else {
        key = this_arg.substring(2, equals_at);
        value = this_arg.substring(equals_at + 1);
        debug("Init: Setting system property '"+key+"' to '"+value+"'");
        sysprops.put(key,value);
      }
    }
    else {
      debug("Init: passing argument"+this_arg);
      temp_args.addElement(this_arg);
    }
  }

  /**
   ** Take all the arguments in temp_args and make them into an array of
   ** String which can be passed to the starting class's main() method.
   */
  private static String[] marshallArgs(Vector temp_args) {
    String[] start_class_args = new String[temp_args.size()];

    for (int i = 0; i < temp_args.size(); ++i) {
      start_class_args[i] = (String)temp_args.elementAt(i);
    }

    return start_class_args;
  }

  /**
   ** Get the name of the Main-Class in a jar file, with the path to the
   ** jar file prepended. Return null if no mainifest or no Main-Class
   ** attribute found.
   ** If a Class-Path attribute is found, its value is parsed and every
   ** element found is added to the URL list of the Application Class Loader
   ** (after prepending the path to the jar file).
   */
  private static String getJarStartClassName(String[] args) {
    String start_class_name = null;
    if(args.length < 2){
      System.err.println("Init: -jar option is used but no JarFile is mentioned");

      return null;

    }

    try {
      File file = new File(args[1]);
      JarFile jf = new JarFile(file);
      Manifest man = jf.getManifest();
      if(man == null){
        System.err.println("Init: JarFile "+args[1]+" has no Manifest");

        return null;

      }
      start_class_name = man.getMainAttributes().getValue("Main-Class");
      if(start_class_name == null){
        System.err.println("Init: no Main-Class attribute in Manifest of JarFile "+args[1]);

        return null;

      }
      String jarpath;
      int jarpathend = args[1].lastIndexOf('/');
      if (jarpathend < 0) {
        jarpath = "";
      }
      else {
        jarpath = args[1].substring(0, jarpathend + 1);
      }
      URL url = new URL("jar:"+file.toURL()+"!/");
      Method m = URLClassLoader.class.getDeclaredMethod("addURL",  new Class[]{url.getClass()});
      m.setAccessible(true);
      m.invoke(application_class_loader,new Object[]{url});
      String jarclasspath = man.getMainAttributes().getValue("Class-Path");
      if (jarclasspath != null) {
        debug("Jarfile " + file + " has Class-Path " + jarclasspath);
        StringTokenizer toks = new StringTokenizer(jarclasspath);
        try {
          while (true) {
            file = new File(jarpath + toks.nextToken());
            url = new URL("jar:"+file.toURL()+"!/");
            debug("  Appending " + url + " to application class path");
            m.invoke(application_class_loader,new Object[]{url});
          }
        }
        catch (NoSuchElementException nsee) {
          debug("Finished with jarfile Class-Path");
        }
      }
    }
    catch(Exception e){
      System.err.println("Init: failed to load jar "+args[1]+" due to "+e);
      e.printStackTrace();

      return null;

    }

    return start_class_name;
  }

  /**
   ** Get the name of the class mentioned in the wonka.default.commandline
   ** property.
   */
  private static String getSysPropStartClassName(String start_command) {
    StringTokenizer st = new StringTokenizer(start_command);

    return st.nextToken();
  }

  /**
   ** Sift the array raw_args, using -D arguments to modify system properties
   ** and returning an array which contains the remaining arguments. Both 
   ** raw_args and the result returned may be zero-length, but they may not be null.
   */
  private static String[] processArgsArray(String[] raw_args, int offset) {
    Vector temp_args = new Vector();

    for (int i = offset; i < raw_args.length; ++i) {
      argumentOrSysProp(raw_args[i],temp_args);
    }

    return marshallArgs(temp_args);
  }

  /**
   ** Parse args_string, using -D tokens to modify system properties
   ** and returning an array which contains the remaining arguments.  
   ** Both args_string and the result returned may be zero-length, 
   ** but they may not be null.
   */
  private static String[] processArgsString(String args_string) {
    Vector temp_args = new Vector();
    StringTokenizer st = new StringTokenizer(args_string);

    while (st.hasMoreTokens()) {
      String this_arg = st.nextToken();
      argumentOrSysProp(this_arg,temp_args);
    }
    temp_args.removeElementAt(0);

    return marshallArgs(temp_args);
  }

  /**
   ** This function is invoked directly when the VM is created.  First we 
   ** parse the command line to extract the name of the starting class and
   ** any changes to system properties; the remaining arguments will be passed
   ** on to the starting class when it has been loaded.  Then we try to load
   ** starting class and to find its main(String[]) method.  If this succeeds
   ** then we first install the SecurityMAnager if required and launch some 
   ** housekeeping threads and then invoke the main(String[]) method.  Lastly,
   ** if the main(String[]) method of the starting class terminates abnormally
   ** we pass the exception thrown to the UncaughtExceptionHandler of the
   ** system thread group.
   */
  private static void main(String[] args) {
    String debugProperty = Wonka.getProperty("wonka.java.debug","false");
    if (debugProperty.equals("true")) {
        debug = true;
    }

    String start_class_name = null;
    Class start_class = null;
    application_class_loader = (URLClassLoader)ClassLoader.getSystemClassLoader();

    if(args != null && args.length > 0 && args[0] != null && args[0].equals("-jar")) {
      debug("Init: '-jar' is used");
      start_class_name = getJarStartClassName(args);
      start_class_args = processArgsArray(args, 2);
      debug("Init: remaining args : " + start_class_args);
    }
    else if (args != null && args.length > 0 && args[0] != null) {
      start_class_name = args[0];
      start_class_args = new String[args.length-1];
      start_class_args = processArgsArray(args, 1);
      debug("Init: args : " + start_class_args);
    }
    else {
      String start_command = Wonka.getProperty("wonka.default.commandline");
      if (start_command != null && start_command.length() != 0){
        start_class_name = getSysPropStartClassName(start_command);
        start_class_args = processArgsString(start_command);
      }
      debug("Init: default args : " + start_class_args);
    }

    if (start_class_name == null) {
      System.err.println("Init: no start class named.  Game over.");
      System.exit(1);
    }

    invoke_args = new Object[1];
    invoke_args[0] = start_class_args;

    try {
      start_class = Class.forName(start_class_name, false, application_class_loader);
      debug("Init: start_class = "+start_class);
    } catch (ClassNotFoundException e) {
      System.err.println("Init: no such class as "+start_class_name+" in class path!");
    }

    if (start_class == null) {
      System.err.println("Init: no start class found.  Game over.");
      System.exit(1);
    }

    try {
      invoke_method = start_class.getMethod("main",new Class[]{Class.forName("[Ljava.lang.String;")});

//      new java.awt.Panel();

      debug("Init: will invoke " + invoke_method);
      invoke_method.setAccessible(true);
    } catch (SecurityException e) {
      System.err.println("Init: unable to invoke " + invoke_method + ": " + e);
      e.printStackTrace();
    } catch (ClassNotFoundException e) {
      System.err.println("Init: unable to invoke " + invoke_method + ": " + e);
      e.printStackTrace();
    } catch (NoSuchMethodException e) {
      System.err.println("Init: unable to invoke " + invoke_method + ": " + e);
      e.printStackTrace();
    }

    if (invoke_method == null) {
      System.err.println("Init: no start method found. Game over.");
      System.exit(1);
    }

  // Install the default SecurityManager if required
    if (SecurityConfiguration.SET_SECURITY_MANAGER) {
      debug("Init: installing SecurityManager");
      if (SecurityConfiguration.USE_ACCESS_CONTROLLER) {
        debug("Init: note that Wonka libraries will not use this SecurityManager, but will call AccessController directly."); 
      }
      else if (!SecurityConfiguration.USE_SECURITY_MANAGER) {
        debug("Init: note that Wonka libraries will not use this SecurityManager (security checks are disabled)."); 
      }
      System.setSecurityManager(new SecurityManager());
    }
  // Start up the Garbage Collector
    debug("Init: starting Garbage Collector");
    GarbageCollector gc = GarbageCollector.getInstance();
  // Start up the Finalizer
    debug("Init: starting Finalizer");
    Finalizer f = Finalizer.getInstance();
  // Start up the Heartbeat
    debug("Init: starting Heartbeat");
    Heartbeat h = Heartbeat.getInstance();
  // Set the default timezone to the default default
    TimeZone.setDefault(null);

    try {
      debug("Init: starting "+invoke_method+" ...");
      invoke_method.invoke(null,invoke_args);
      launched = true;
    } catch (IllegalAccessException e) {
      System.err.println("Init: error invoking "+invoke_method+"! got "+e);
      e.printStackTrace();
    } catch (InvocationTargetException e) {
      Thread this_thread = Thread.currentThread();
      try {
        this_thread.getThreadGroup().uncaughtException(this_thread,e.getTargetException());
      }
      catch (Throwable t) {
        t.printStackTrace();
      }
    }
    finally {
      if (!launched) {
        System.exit(1);
      }
    }
  }

}

