/**************************************************************************
* Copyright  (c) 2001 by Acunia N.V. All rights reserved.                 *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Vanden Tymplestraat 35      info@acunia.com                           *
*   3000 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: hashtable.c 4226 2005-10-01 19:48:17Z chris $
*/

#include <string.h>

#include "fifo.h"
#include "hashtable.h"  
#include "list.h"
#include "ts-mem.h"
#include "realms.h"

/*
** If the AUTO_SHRINK_HASHTABLES is defined then hashtables will shrink
** when the occupancy is too low, as well as growing when it is too high.
** Only enable this if it will solve a memory problem: most of the system
** hashtables are better left to grow to their maximum size.
** TODO: consider making this definable per hashtable.
*/
// #define AUTO_SHRINK_HASHTABLES

/* 
** We implement "open addressing with linear probing", as described in
** Algorithms L (p.518) and R (p. 527) of Knuth's _Art of Computer
** Programming_, Chapter 6.4.
**
** In his discussion Knuth assumes that the hashing function h(K) acts
** directly on the key K.  In the case of java/lang/Hashtable we are
** given a function w_word (*hash)(w_word) which generates a 32-bit signed 
** integer result, but that isn't really what we want: h(K) should
** deliver a result in the range 0 to M-1, where M is the size of the
** hashtable.  So in our implementation,
**  1) if CACHE_HASCODES is defined, we store the value of (*hash)(K) in
**     the hashtable together with the key K and the associated information.
**  2) our function h(K) is equal to (*hash)(K) modulo M.
*/

/*
** ht_hash calculates the value of (hashcode MOD hashtable->currentsize),
** where MOD is unsigned integer modulus.  This is where the fun begins:
** C soesn't have such an operator, so we have to fake it.  We start by
** observing that if a/b evaluates as q and a%b evaluates as r, then C
** guarantees that a = b*q + r.  It follows that a - b*q = r, so we first
** divide b into a (discarding the remainder), then multiply the resulting
** quotient by b, and subtract the result from a to get r.  What a mess.
*/
inline static w_int ht_hash(w_hashtable hashtable, w_word hashcode) {
  w_int modulus = hashtable->currentsize;
  w_int result;

  result = hashcode - (hashcode/modulus)*modulus;

  return result;
}


/*
** Then the probe sequence is h(K), h(K)-1,...,0,M-1,M-2,...h(K)+1 and is obtained
** by calling ht_probe successively with sequence = 0, 1, 2, ...
*/

w_int ht_probe(w_hashtable hashtable, w_word hashcode, w_int sequence) {
  w_int result = ht_hash(hashtable,hashcode-sequence);

  if(result<0) {
    wabort(ABORT_WONKA,"hashcode = 0x%08x, result = %d\n",hashcode,result);
  }

//  while (result<0) {
//    result += hashtable->currentsize;
//  }

  return result;
}

/*
** For deletion, we need to examine the 'next' slot.
** ht_reprobe(hashtable,current) gives the value which follows 'current' 
** in the probe sequence
*/

static inline w_int ht_reprobe(w_hashtable hashtable, w_int current) {
  return current ? current-1 : hashtable->currentsize-1;
}

/*
** Also for deletion, we need to be able to measure the "distance" from
** one slot to another in the probing sequence: ht_distance gives the
** number of times we would have to call 'from = ht_reprobe(hashtable,from)'
** to make 'from' into 'to' ...
*/

static inline w_int ht_distance(w_hashtable hashtable, w_int from, w_int to) {
  return from>=to ? from-to : hashtable->currentsize + from-to;
}

/*
** Then the question "does r lie cyclically between i and j" becomes
**   ht_distance(hashtable,i,r) < ht_distance(hashtable,i,j).
*/

/****************************************************************************/

/*
**
** recalculateThresholds(w_hashtable) calculates a new lowthreshold
** and highthreshold from the current table size;  these thresholds 
** are used by ht_check_size to decide whether a resizing is required. 
** Currently the thresholds are hard-coded to 25% and 75% of currentsize.
*/

void
recalculateThresholds(w_hashtable h) {
  w_int maxhighthreshold = h->currentsize - h->currentsize/16;
  if(maxhighthreshold>=h->currentsize) maxhighthreshold = h->currentsize-1;

  h->lowthreshold = h->currentsize/4;

  if(h->lowthreshold==0) h->lowthreshold = 1;

  h->highthreshold = h->currentsize - h->currentsize/4;

  if(h->highthreshold>maxhighthreshold) h->highthreshold = maxhighthreshold;
}


/*
************************************************************************
*/

w_word
dummy_hash (w_word a) {

  return a;

}

w_boolean
dummy_equal (w_word a, w_word b) {

  return a==b;

}

static inline w_boolean
ht_match(w_hashtable hashtable, int idx, w_word h, w_word k) {

  return 
#ifdef CACHE_HASHCODES
         (!hashtable->hashcodes || hashtable->hashcodes[idx]==h) &&
#endif 
         (*hashtable->equal)(hashtable->keys[idx],k);

}

static inline w_boolean
ht_nullkey(w_hashtable hashtable, w_word key) {

  return key==hashtable->nullkey;

}

w_boolean
ht_occupied(w_hashtable hashtable, w_int idx) {

  return !ht_nullkey(hashtable, hashtable->keys[idx]);

}

/*
************************************************************************
*/

w_hashtable 
_ht_create (const char *f, int l,
  w_realm realm,
  char *   label,
  w_size  initialsize,
  w_word (*hash)(w_word),             
  w_boolean(*equal)(w_word,w_word), 
  w_word nullkey, w_word nullvalue
) {
  w_hashtable hashtable;
  w_size newindex;

  hashtable = allocClearedMem_realm(realm, sizeof(w_Hashtable));
  hashtable->realm = realm;
  woempa(1,"creating %s at %p, size = %d\n",label,hashtable,initialsize);

  hashtable->label  = label;
  hashtable->currentsize = initialsize;
  hashtable->occupancy = 0;
  x_monitor_create(&hashtable->monitor);

  if (equal) {
    woempa(1,"user-defined equality function at %p\n",equal);
    hashtable->equal = equal;
  }
  else {
    hashtable->equal = dummy_equal;
  }

  hashtable->hash = hash ? hash : dummy_hash;
#ifdef CACHE_HASHCODES
  if (hash) {
    hashtable->hashcodes = allocClearedMem_realm(realm, initialsize * sizeof(w_word));
  }
  else {
    hashtable->hashcodes = NULL;
  }
#endif

  hashtable->nullkey   = nullkey;
  hashtable->nullvalue = nullvalue;

  hashtable->keys = allocClearedMem_realm(realm, initialsize * sizeof(w_word));

  for(newindex=0;newindex<initialsize;++newindex) {
    hashtable->keys[newindex] = hashtable->nullkey;
  }

  recalculateThresholds(hashtable);

  return hashtable;
}

void
ht_resize(w_hashtable hashtable, w_size newsize) 
{
  w_size  oldsize = hashtable->currentsize;
  w_word *oldkeys = hashtable->keys;
  w_word *newkeys;
  w_word *oldvalues = hashtable->values;
  w_word *newvalues = NULL;
#ifdef CACHE_HASHCODES
  w_word *oldhashcodes = hashtable->hashcodes;
  w_word *newhashcodes = NULL;
#endif
  w_size  sequence;
  w_size  oldindex;
  w_size  newindex;
  w_word  key;

  woempa(1,"(re)allocating hashtable %s @ %p, new size %d\n",hashtable->label,hashtable,newsize);

  newkeys = allocClearedMem_realm(hashtable->realm, newsize * sizeof(w_word));
  if (!newkeys) {
    woempa(9, "No memory available for newkeys of %s\n", hashtable->label);

    return;

  }

  if (oldvalues) {
    newvalues = allocClearedMem_realm(hashtable->realm, newsize * sizeof(w_word));
    if (!newvalues) {
      woempa(9, "No memory available for newvalues of %s\n", hashtable->label);
      releaseMem(newkeys);

      return;

    }

  }
#ifdef CACHE_HASHCODES
  if (oldhashcodes) {
    newhashcodes = allocClearedMem_realm(hashtable->realm, newsize * sizeof(w_word));
    if (!newhashcodes) {
      woempa(9, "No memory available for newhashcodes of %s\n", hashtable->label);
      releaseMem(newkeys);
      releaseMem(newvalues);

      return;

    }
  }
#endif

  hashtable->currentsize = newsize;
  hashtable->keys = newkeys;
  hashtable->values = newvalues;
#ifdef CACHE_HASHCODES
  hashtable->hashcodes = newhashcodes;
#endif

  for(newindex=0;newindex<newsize;++newindex) {
    newkeys[newindex] = hashtable->nullkey;
  }

  if(oldsize && newsize) {
  /*
  ** Rehash old contents into new arrays.
  ** Looks long-winded, but we don't want to have the tests for
  ** the existence of oldvalues and oldhashcodes in the main loop.
  */

#ifdef CACHE_HASHCODES
    if (oldvalues && oldhashcodes) {
      for(oldindex=0;oldindex<oldsize;++oldindex) {
        key = oldkeys[oldindex];
        if (key != hashtable->nullkey) { 
          w_word value = oldvalues[oldindex];
          w_word hc    = oldhashcodes[oldindex];

          for (sequence=0; sequence<newsize; ++sequence) {
            newindex = ht_probe(hashtable, hc, (signed)sequence);
            if (!ht_occupied(hashtable, (signed)newindex)) {
              newkeys[newindex]      = key;
              newvalues[newindex]    = value;
              newhashcodes[newindex] = hc;
              break;
            }
          }
        }
      }
    }
    else 
#endif
    if (oldvalues) {
      for (oldindex=0;oldindex<oldsize;++oldindex) {
        key = oldkeys[oldindex];
        if (key != hashtable->nullkey) { 
          w_word value = oldvalues[oldindex];
          w_word hc    = hashtable->hash(key);

          for (sequence=0; sequence<newsize; ++sequence) {
            newindex = ht_probe(hashtable, hc, (signed)sequence);
            if (!ht_occupied(hashtable, (signed)newindex)) {
              newkeys[newindex]      = key;
              newvalues[newindex]    = value;
              break;
            }
          }
        }
      }
    }
#ifdef CACHE_HASHCODES
    else if (oldhashcodes) {
      for (oldindex=0; oldindex<oldsize; ++oldindex) {
        key = oldkeys[oldindex];
        if (key != hashtable->nullkey) { 
          w_word hc = oldhashcodes[oldindex];

          for (sequence=0; sequence<newsize; ++sequence) {
            newindex = ht_probe(hashtable, hc, (signed)sequence);
            if (!ht_occupied(hashtable, (signed)newindex)) {
              newkeys[newindex]      = key;
              newhashcodes[newindex] = hc;
              break;
            }
          }
        }
      }
    }
#endif
    else {
      for (oldindex=0; oldindex<oldsize; ++oldindex) {
        key = oldkeys[oldindex];
        if (key != hashtable->nullkey) { 
          for (sequence=0; sequence<newsize; ++sequence) {
            w_word hc    = hashtable->hash(key);
            newindex = ht_probe(hashtable, hc, (signed)sequence);
            if (!ht_occupied(hashtable, (signed)newindex)) {
              newkeys[newindex]      = key;
              break;
            }
          }
        }
      }
    }
  }

  if (oldkeys) {
    releaseMem(oldkeys);
  }
  if (oldvalues) {
    releaseMem(oldvalues);
  }
#ifdef CACHE_HASHCODES
  if (oldhashcodes) {
    releaseMem(oldhashcodes);
  }
#endif

  recalculateThresholds(hashtable);

  woempa(1,"hashtable keys at %p\n",hashtable->keys);
  woempa(1,"hashtable values at %p\n",hashtable->values);
#ifdef CACHE_HASHCODES
  woempa(1,"hashtable hashcodes at %p\n",hashtable->hashcodes);
#endif

}

void
ht_destroy(w_hashtable theHashtable) {

  x_monitor_delete(&theHashtable->monitor);
  if (theHashtable->keys) {
    releaseMem(theHashtable->keys);
  }
  if (theHashtable->values) {
    releaseMem(theHashtable->values);
  }
#ifdef CACHE_HASHCODES
  if (theHashtable->hashcodes) {
    releaseMem(theHashtable->hashcodes);
  }
#endif
  releaseMem(theHashtable);
}

w_size
ht_check_size(w_hashtable hashtable, w_int direction) {
  w_int       oldsize;
  w_int       newsize;

  woempa(1,"hashtable %p size now is %d, occupancy %d, thresholds %d %d\n",hashtable,hashtable->currentsize,hashtable->occupancy,hashtable->lowthreshold,hashtable->highthreshold);
  oldsize = hashtable->currentsize;
  newsize = oldsize;

  if( (hashtable->occupancy>=hashtable->highthreshold && direction>=0)
   || (hashtable->occupancy<=hashtable->lowthreshold  && direction<=0)
    ) {
    newsize = hashtable->occupancy * 2;

    if(newsize<hashtable->occupancy+2) newsize=hashtable->occupancy+2;
    if((newsize&1)==0) ++newsize;
    if((newsize%3)==0) newsize += 2;
    if((newsize%5)==0) newsize += 4;

    if(newsize==oldsize) newsize = 0;
    else {
      woempa(1,"recommend resizing to %d\n",newsize);
    }

    return newsize;

  }
  else return 0;

}


/*
** ht_rehash is a no-op: see Theorem P on p. 530 of Knuth, op. cit..
*/

w_hashtable 
ht_rehash(w_hashtable oldtable) {

  return oldtable;

}

static inline void set_value(w_hashtable hashtable, w_int idx, w_word key, w_word value) {
  if (hashtable->values) {
    hashtable->values[idx] = value;
  }
  else if (value != key) {
    hashtable->values = allocClearedMem_realm(hashtable->realm, hashtable->currentsize * sizeof(w_word));
    hashtable->values[idx] = value;
  }
}


void
ht_insert(w_hashtable hashtable, w_int idx, w_word key, w_word value)
{
  woempa(1, "hashtable %p index %d key %p value %p\n", hashtable, idx, (char *)key, (char *)value);
  /*
  ** Protect ourselves against barbarous practices of the user
  */
  if(ht_nullkey(hashtable, key) 
     || ht_occupied(hashtable, idx) 
     || hashtable->currentsize-hashtable->occupancy<=1) {

    woempa(9,"attempt to corrupt hashtable thwarted!\n");
    return;
  }

  hashtable->keys[idx] = key;
  set_value(hashtable, idx, key, value);
#ifdef CACHE_HASHCODES
  if (hashtable->hashcodes) {
    hashtable->hashcodes[idx] = (*hashtable->hash)(key);
  }
#endif
  hashtable->occupancy += 1;
}

void
ht_delete(w_hashtable hashtable, w_int idx) {
  w_int     vacant;
  w_int     current;
  w_int     home;
  w_boolean happy;

  woempa(1, "hashtable %s @ %p: delete slot %d\n",hashtable->label,hashtable,idx);
  /*
  ** The following is a summary of Knuth's Algorithm R.
  **
  ** Step R1 : Mark TABLE[i] empty, and set j equal to i.
  ** Step R2 : decrease i, modulo M
  ** Step R3 : if TABLE[i] is empty, the algorithm terminates.
  **           Otherwise set r equal to h(KEY[i]), the original hash 
  **           address of KEY[i]).  If i<=r<j (modulo M), go back to R2.
  ** Step R4 : Set TABLE[j] equal to TABLE[i], and return to step R1.
  **
  ** In Knuth's therminology, TABLE[j] is our vacant and TABLE[i] is
  ** our current; TABLE[r] corresponds to home, the "home" slot
  ** of the entry at current. 
  */

  hashtable->occupancy -= 1;
  current = idx;

  while(1) {
  /*
  ** R1
  */
    woempa(1,"mark slot %d as empty\n",current);
    hashtable->keys[current] = hashtable->nullkey;
    vacant = current;
  /*
  ** R2
  */
    happy = WONKA_TRUE;

    while(happy) {
      current = ht_reprobe(hashtable,current);
      woempa(1,"Next slot is %d\n",current);
  /*
  ** R3
  */
      if(ht_nullkey(hashtable, hashtable->keys[current])) {
        woempa(1,"slot %d is empty, algorithm terminated normally\n",current);

        return;

      }

#ifdef CACHE_HASHCODES
      if (hashtable->hashcodes) {
        home = ht_probe(hashtable,hashtable->hashcodes[current],0); 
      }
      else {
        home = ht_probe(hashtable,hashtable->keys[current],0); 
      }
#else
      home = ht_probe(hashtable,hashtable->hash(hashtable->keys[current]),0); 
#endif
      woempa(1,"  whereas in slot %d it would be %d slots from home\n",vacant,ht_distance(hashtable,home,vacant));
      if( ht_distance(hashtable,home,current) 
        < ht_distance(hashtable,home,vacant)) {
        woempa(1,"so it can stay where it is\n");
      } else {
        woempa(1,"so it needs to move ...\n");
        happy = WONKA_FALSE;
      }
    }
  /*
  ** R4
  */
    woempa(1,"copying from slot %d to slot %d\n",current,vacant);
    hashtable->keys[vacant]      = hashtable->keys[current];
    if (hashtable->values) {
      hashtable->values[vacant]    = hashtable->values[current];
    }
#ifdef CACHE_HASHCODES
    if (hashtable->hashcodes) {
      hashtable->hashcodes[vacant] = hashtable->hashcodes[current];
    }
#endif
  }
}

/*
***************************************************************************
*/

w_word /* previous value or hashtable->nullvalue */
ht_write_no_lock(w_hashtable hashtable, w_word key, 
w_word newvalue
) {
  w_int  seq;
  w_int  maxseq = hashtable->currentsize;
  w_word result = hashtable->nullvalue;
  w_word hashcode = (*hashtable->hash)(key);
  w_size newsize;
  w_int idx;

  for(seq=0;seq<maxseq;++seq) {                                                    
    idx=ht_probe(hashtable,hashcode,seq);                            
    if(ht_occupied(hashtable,idx)) {                                        
      if(ht_match(hashtable, idx, hashcode, key)) {                            
#ifdef HASHTEST
      woempa(1,"%p: MATCH with %d collisions\n",hashtable,seq);
#endif
        if (hashtable->values) {
          result = hashtable->values[idx];
        }
        else {
          result = key;
        }
        set_value(hashtable, idx, key, newvalue);
        break;                                              
      }                                                   
    }                                                    
    else {                                              
#ifdef HASHTEST
      woempa(1,"INSERT with %d collisions\n",seq);
#endif
      hashtable->keys[idx] = key;
      set_value(hashtable, idx, key, newvalue);
#ifdef CACHE_HASHCODES
      if (hashtable->hashcodes) {
        hashtable->hashcodes[idx] = hashcode;
      }
#endif
      hashtable->occupancy += 1;
      newsize = ht_check_size(hashtable,1);
      if (newsize) {
        ht_resize(hashtable, newsize);
      }
      break;                                                            
    }                                                                  
  }                                                                   

  return result;

}


w_word /* value found or hashtable->nullvalue */
ht_read_no_lock(w_hashtable hashtable, w_word key) {
  w_int  seq;
  w_int  maxseq;
  w_word result;
  w_word hashcode;
  w_int  idx;

  maxseq = hashtable->currentsize;
  result = hashtable->nullvalue;
#ifdef CACHE_HASHCODES
  hashcode = hashtable->hashcodes ? (*hashtable->hash)(key) : key;
#else
  hashcode = (*hashtable->hash)(key);
#endif

  for(seq=0;seq<maxseq;++seq) {

    idx=ht_probe(hashtable,hashcode,seq);
    if(ht_occupied(hashtable,idx)) {
      if(ht_match(hashtable, idx, hashcode, key)) {
#ifdef HASHTEST
      woempa(1,"%p: MATCH with %d collisions\n",hashtable,seq);
#endif
        result = hashtable->values ? hashtable->values[idx] : hashtable->keys[idx];
        break;
      }
    }
    else {
      break;
    }
  }
 
  return result;

}

w_word /* key found or hashtable->nullkey */
ht_findkey_no_lock(w_hashtable hashtable, w_word key) {
  w_int  seq;
  w_int  maxseq = hashtable->currentsize;
  w_word result = hashtable->nullkey;
  w_word hashcode = (*hashtable->hash)(key);
  w_int  idx;

  for(seq=0;seq<maxseq;++seq) {
    idx=ht_probe(hashtable,hashcode,seq);
    if(ht_occupied(hashtable,idx)) {
      if(ht_match(hashtable, idx, hashcode, key)) {
#ifdef HASHTEST
      woempa(1,"%p: MATCH with %d collisions\n",hashtable,seq);
#endif
        result = hashtable->keys[idx];
        break;
      }
    }
    else {
      break;
    }
  }
 
  return result;

}

w_word /* value found or NULL */
ht_erase_no_lock(w_hashtable hashtable, w_word key) {
  w_int  seq;
  w_int  maxseq = hashtable->currentsize;
  w_word result = hashtable->nullvalue;
  w_word hashcode = (*hashtable->hash)(key);
  w_int  idx;
#ifdef AUTO_SHRINK_HASHTABLES
  w_size newsize;
#endif

  for(seq=0;seq<maxseq;++seq) {

    idx=ht_probe(hashtable,hashcode,seq);
    if(ht_occupied(hashtable,idx)) {
      if(ht_match(hashtable, idx, hashcode, key)) {
        result = hashtable->values ? hashtable->values[idx] : key;
        ht_delete(hashtable,idx);
#ifdef AUTO_SHRINK_HASHTABLES
        newsize = ht_check_size(hashtable,-1);
        if (newsize) {
          ht_resize(hashtable, newsize);
        }
#endif
        break;
      }
    }
    else {
      break;
    }
  }

  return result;
 

}


w_int
ht_every(w_hashtable hashtable, void (*fun)(w_word key,w_word value)) {
  w_int count = 0;
  w_int i;

  ht_lock(hashtable);

  if (hashtable->values) {
    for(i=0;i<hashtable->currentsize;++i) {
      if(ht_occupied(hashtable,i)) {
        fun(hashtable->keys[i], hashtable->values[i]);
        ++count;
      }
    }
  }
  else {
    for(i=0;i<hashtable->currentsize;++i) {
      if(ht_occupied(hashtable,i)) {
        fun(hashtable->keys[i], hashtable->keys[i]);
        ++count;
      }
    }
  }

  ht_unlock(hashtable);

  return count;
}

w_fifo
ht_list_keys_no_lock(w_hashtable hashtable) {
  w_fifo result;
  w_size leafsize;
  w_int i;

  leafsize = 63;
  while (leafsize < (w_size)hashtable->occupancy) {
    leafsize = leafsize * 2 - 1;
  }
  result = allocFifo(hashtable->realm, leafsize);

  if (result == NULL) {

    return NULL;

  }

  ht_lock(hashtable);

  for(i=0;i<hashtable->currentsize;++i)
  {
    if(ht_occupied(hashtable,i)) {
      if (putFifo((void*)hashtable->keys[i],result) < 0) {
        wabort(ABORT_WONKA, "I don't believe it!  We sized fifo %p to have a leafsize of %d, and after %d putFifo's it overflows!\n", result, leafsize, i);
      }
    }
  }

  ht_unlock(hashtable);

  return result;
}

w_fifo
ht_list_values_no_lock(w_hashtable hashtable) {
  w_fifo result;
  w_size leafsize;
  w_int i;

  if (!hashtable->values) {

    return ht_list_keys(hashtable);

  }

  leafsize = 63;
  while (leafsize < (w_size)hashtable->occupancy) {
    leafsize = leafsize * 2 - 1;
  }
  result = allocFifo(hashtable->realm, leafsize);
  if (result == NULL) {

    return NULL;

  }


  ht_lock(hashtable);

  for(i=0;i<hashtable->currentsize;++i)
  {
    if(ht_occupied(hashtable,i)) {
      if (putFifo((void*)hashtable->values[i],result) < 0) {
        wabort(ABORT_WONKA, "I don't believe it!  We sized fifo %p to have a leafsize of %d, and after %d putFifo's it overflows!\n", result, leafsize, i);
      }
    }
  }

  ht_unlock(hashtable);

  return result;
}

w_int
ht_iterate_no_lock(w_hashtable hashtable, void (*fun)(w_word key,w_word value, void *arg1, void *arg2), void *arg1, void *arg2) {
  w_int count = 0;
  w_int i;

  if (hashtable->values) {
    for(i=0;i<hashtable->currentsize;++i) {
      if(ht_occupied(hashtable,i)) {
        fun(hashtable->keys[i], hashtable->values[i], arg1, arg2);
        ++count;
      }
    }
  }
  else {
    for(i=0;i<hashtable->currentsize;++i) {
      if(ht_occupied(hashtable,i)) {
        fun(hashtable->keys[i], hashtable->keys[i], arg1, arg2);
        ++count;
      }
    }
  }

  return count;

}


w_word /* old value of key or hashtable->nullkey */
ht_register(w_hashtable hashtable, w_word key) 
{
  w_int seq;
  w_int maxseq = hashtable->currentsize;
  w_word result = hashtable->nullkey;
  w_word hashcode = (*hashtable->hash)(key);
  w_int  idx;
  w_size newsize;
  w_int  counter;

  ht_lock(hashtable);
  for(seq=0;seq<maxseq;++seq) { 
    idx=ht_probe(hashtable,hashcode,seq);                            
    if(ht_occupied(hashtable,idx)) {                                        
      if(ht_match(hashtable, idx, hashcode, key)) {                            
#ifdef HASHTEST
      woempa(1,"%p: MATCH with %d collisions\n",hashtable,seq);
#endif
        result = hashtable->keys[idx];
        counter = (w_int)(hashtable->values[idx]);
        if(counter<HT_BIGNUM)
          hashtable->values[idx] = (w_word)(counter+1);

        break;                                              
      }                                                   
    }                                                    
    else {                                              
#ifdef HASHTEST
      woempa(1,"INSERT with %d collisions\n",seq);
#endif
      hashtable->keys[idx] = key;
      if (!hashtable->values) {
        hashtable->values = allocClearedMem_realm(hashtable->realm, hashtable->currentsize * sizeof(w_word));
      }
      hashtable->values[idx] = 1;
#ifdef CACHE_HASHCODES
      if (hashtable->hashcodes) {
        hashtable->hashcodes[idx] = hashcode;
      }
#endif
      hashtable->occupancy += 1;
      newsize = ht_check_size(hashtable,1);
      if (newsize) {
        ht_resize(hashtable, newsize);
      }
      break;                                                            
    }                                                                  
  }                                                                   
  ht_unlock(hashtable);                                                   

  return result;

}

w_word /* value of matched key or hashtable->nullkey */
ht_deregister(w_hashtable hashtable, w_word key) 
{
  w_int seq;
  w_int maxseq = hashtable->currentsize;
  w_word result = hashtable->nullvalue;
  w_word hashcode = (*hashtable->hash)(key);
  w_int  idx;
  w_int  counter;
#ifdef AUTO_SHRINK_HASHTABLES
  w_size newsize;
#endif

  ht_lock(hashtable);
  for(seq=0;seq<maxseq;++seq) {
    idx=ht_probe(hashtable,hashcode,seq);                            
    if(ht_occupied(hashtable,idx)) {                                        
      if(ht_match(hashtable, idx, hashcode, key)) { 
        counter = hashtable->values[idx];
        if(counter<HT_BIGNUM) {
          counter -= 1;
          if(counter==0) {
            result = hashtable->keys[idx];
            ht_delete(hashtable,idx);
#ifdef AUTO_SHRINK_HASHTABLES
            newsize = ht_check_size(hashtable,-1);
            if (newsize) {
              ht_resize(hashtable, newsize);
            }
#endif
          }
          else {
            hashtable->values[idx] = (w_word)counter; 
          }
          break;                                              
        }
      }                                                   
    }  
    else { 
      wabort(ABORT_WONKA,"Manure occured: did not find key 0x%08x\n",key);
    }
  }                                                                   
  ht_unlock(hashtable);                                                   

  return result;

}

 
w_word
ht_unique(w_hashtable hashtable, w_word try, w_word value, w_word min, w_word max) {
  w_boolean  found = WONKA_FALSE;
  w_word      result;

  result = try;
  if(result < min) {
    result=max;
  }
  if(result > max || result == 0) {
    result=min;
  }

  ht_lock(hashtable);

  while(found==WONKA_FALSE) {
    if (ht_read_no_lock(hashtable, result) == 0) {
      ht_write_no_lock(hashtable, result, value);
      found = WONKA_TRUE;
    }
    else {
      ++result;
      if(result > max || result == 0) {
        result = min;
      }
    }
  }

  ht_unlock(hashtable);

  return result;

}

/*
    private static long initial_hash = 0xe12398c6d9ae3b8aL;  // initial values
    private static long hash_true    = 0x851dcaa2656c6af4L;  // are arbitrary
    private static long hash_false   = 0x1af84a6b589285f7L;  // 64-bit rands
*/

#define INITIAL_BUZ 0xe12398c6d9ae3b8aLL 

// 0:255
static w_hash mixmaster[] = {
    /* 000 */ 0x4476081a7043a46fLL, 0x45768b8a6e7eac19LL, 0xebd556c1cf055952LL, 0x72ed2da1bf010101LL, 0x3ff2030b128e8a64LL,
    /* 005 */ 0xcbc330238adcfef2LL, 0x737807fe42e20c6cLL, 0x74dabaedb1095c58LL, 0x968f065c65361d67LL, 0xd3f4018ac7a4b199LL,
    /* 010 */ 0x954b389b52f24df2LL, 0x2f97a9d8d0549327LL, 0xb9bea2b49a3b180fLL, 0xaf2f42536b21f2ebLL, 0x85d991663cff1325LL,
    /* 015 */ 0xb9e1260207b575b9LL, 0xf3ea88398a23b7e2LL, 0xfaf8c83ffbd9091dLL, 0x4274fe90834dbdf9LL, 0x3f20b157b68d6313LL,
    /* 020 */ 0x68b48972b6d06b93LL, 0x694837b6eba548afLL, 0xeecb51d1acc917c9LL, 0xf1c633f02dffbcfaLL, 0xa6549ec9d301f3b5LL,
    /* 025 */ 0x451dc944f1663592LL, 0x446d6acef6ce9e4fLL, 0x1c8a5b3013206f02LL, 0x5908ca36f2dc50f7LL, 0x4fd55d3f3e880a87LL,
    /* 030 */ 0xa03a8dbeabbf065dLL, 0x3ccbbe078fabcb6dLL, 0x1da53a259116f2d0LL, 0xfb27a96fcb9af152LL, 0x50aba242e85aec09LL,
    /* 035 */ 0x24d4e414fc4fc987LL, 0x83971844a9ce535eLL, 0xc26a3fdeb849398eLL, 0xc2380d044d2e70d8LL, 0xab418aa8ae19b18fLL,
    /* 040 */ 0xd95b6b9247d5ebeaLL, 0x8b3b2171fdc60511LL, 0xe15cd0ae3fcc44afLL, 0x5a4e27f914a68f17LL, 0x377bd28ca09aafdcLL,
    /* 045 */ 0xbbeb9828594a3294LL, 0x7c8df263ae1de1b9LL, 0xba0a48a5fd1c1dd0LL, 0x57cc1b8818b98ee6LL, 0x8c570975d357dabcLL,
    /* 050 */ 0x76bdcd6f2e8826aaLL, 0x529b15b6ec4055f1LL, 0x9147c7a54c34f8a9LL, 0x2f96a7728170e402LL, 0xe46602f455eca72eLL,
    /* 055 */ 0x22834c4dd1bde03fLL, 0x2644cf5a25e368ffLL, 0x907c6de90b120f4aLL, 0xadfe8ba99028f728LL, 0xa85199ae14df0433LL,
    /* 060 */ 0x2d749b946dd3601eLL, 0x76e35457aa052772LL, 0x90410bf6e427f736LL, 0x536ad04d13e35041LL, 0x8cc0d76769b76914LL,
    /* 065 */ 0xae0249f6e3b3c01cLL, 0x1bdfd075307d6fafLL, 0xd8e04f70c221deccLL, 0x4ab23622a4281a5dLL, 0x37a5613da2fcaba7LL,
    /* 070 */ 0x19a56203666d4a9fLL, 0x158ffab502c4be93LL, 0x0bee714e332ecb2fLL, 0x69b71a59f6f74ab0LL, 0x0fc7fc622f1dfe8fLL,
    /* 075 */ 0x513966de7152a6f9LL, 0xc16fae9cc2ea9be7LL, 0xb66f0ac586c1899eLL, 0x11e124aee3bdefd7LL, 0x86cf5a577512901bLL,
    /* 080 */ 0x33f33ba6994a1fbdLL, 0xde6c4d1d3d47ff0dLL, 0x6a99220dc6f78e66LL, 0x2dc06ca93e2d25d2LL, 0x96413b520134d573LL,
    /* 085 */ 0xb4715ce8e1023afaLL, 0xe6a75900c8c66c0aLL, 0x6448f13ad54c12edLL, 0xb9057c28cf6689f0LL, 0xf4023daf67f7677aLL,
    /* 090 */ 0x877c2650767b9867LL, 0xb7ea587dcd5b2341LL, 0xc048cf111733f9bcLL, 0x112012c15bc867bfLL, 0xc95f52b1d9418811LL,
    /* 095 */ 0xa47e624ee7499083LL, 0x26928606df9b12e8LL, 0x5d020462ec3e0928LL, 0x8bbde651f6d08914LL, 0xd5db83db758e524aLL,
    /* 100 */ 0x3105e355c000f455LL, 0xdd7fe1b81a786c79LL, 0x1f3a818c8e012db1LL, 0xd902de819d7b42faLL, 0x4200e63325cda5f0LL,
    /* 105 */ 0x0e919cdc5fba9220LL, 0x5360dd54605a11e1LL, 0xa3182d0e6cb23e6cLL, 0x13ee462c1b483b87LL, 0x1b1b6087b997ee22LL,
    /* 110 */ 0x81c36d0b877f7362LL, 0xc24879932c1768d4LL, 0x1faa756e1673f9adLL, 0x61651b24d11fe93dLL, 0x30fe3d9304e1cde4LL,
    /* 115 */ 0x7be867c750747250LL, 0x973e52c7005b5db6LL, 0x75d6b699bbaf4817LL, 0x25d2a9e97379e196LL, 0xe65fb599aca98701LL,
    /* 120 */ 0x6ac27960d24bde84LL, 0xdfacc04c9fabbcb6LL, 0xa46cd07f4a97882bLL, 0x652031d8e59a1fd8LL, 0x1185bd967ec7ce10LL,
    /* 125 */ 0xfc9bd84c6780f244LL, 0x0a0c59872f61b3ffLL, 0x63885727a1c71c95LL, 0x5e88b4390b2d765cLL, 0xf0005ccaf988514dLL,
    /* 130 */ 0x474e44280a98e840LL, 0x32de151c1411bc42LL, 0x2c4b86d5aa4482c2LL, 0xccd93deb2d9d47daLL, 0x3743236ff128a622LL,
    /* 135 */ 0x42ed2f2635ba5647LL, 0x99c74afd18962dbdLL, 0x2d663bb870f6d242LL, 0x7912033bc7635d81LL, 0xb442862f43753680LL,
    /* 140 */ 0x94b1a5400aeaab4cLL, 0x5ce285fe810f2220LL, 0xe8a7dbe565d9c0b1LL, 0x219131af78356c94LL, 0x7b3a80d130f27e2fLL,
    /* 145 */ 0xbaa5d2859d16b440LL, 0x821cfb6935771070LL, 0xf68cfb6ee9bc2336LL, 0x18244132e935d2fdLL, 0x2ed0bda1f4720cffLL,
    /* 150 */ 0x4ed48cdf6975173cLL, 0xfd37a7a2520e2405LL, 0x82c102b2a9e73ce2LL, 0xadac6517062623a7LL, 0x5a1294d318e26104LL,
    /* 155 */ 0xea84fe65c0e4f061LL, 0x4f96f8a9464cfee9LL, 0x9831dff8ccdc534aLL, 0x4ca927cd0f192a14LL, 0x030900b294b71649LL,
    /* 160 */ 0x644b263b9aeb0675LL, 0xa601d4e34647e040LL, 0x34d897eb397f1004LL, 0xa6101c37f4ec8dfcLL, 0xc29d2a8bbfd0006bLL,
    /* 165 */ 0xc6b07df8c5b4ed0fLL, 0xce1b7d92ba6bccbeLL, 0xfa2f99442e03fe1bLL, 0xd8863e4c16f0b363LL, 0x033b2cccc3392942LL,
    /* 170 */ 0x757dc33522d6cf9cLL, 0xf07b1ff6ce55fec5LL, 0x1569e75f09b40463LL, 0xfa33fa08f14a310bLL, 0x6eb79aa27bbcf76bLL,
    /* 175 */ 0x157061207c249602LL, 0x25e5a71fc4e99555LL, 0x5df1fe93de625355LL, 0x235b56090c1aa55dLL, 0xe51068613eaced91LL,
    /* 180 */ 0x45bd47b893b9ff1eLL, 0x6595e1798d381f2dLL, 0xc9b5848cbcdb5ba8LL, 0x65985146ff7792bcLL, 0x4ab4a17bf05a19a0LL,
    /* 185 */ 0xfd94f4ca560ffb0cLL, 0xcf9bad581a68fa68LL, 0x92b4f0b502b1ce1aLL, 0xbcbec0769a610474LL, 0x8dbd31ded1a0fecbLL,
    /* 190 */ 0xdd1f5ed9f90e8533LL, 0x61c1e6a523f84d95LL, 0xf24475f383c110c4LL, 0xdb2dffa66f90588dLL, 0xac06d88e9ee04455LL,
    /* 195 */ 0xa215fc47c40504baLL, 0x86d7caebfee93369LL, 0x9eaec31985804099LL, 0x0fba2214abe5d01bLL, 0x5a32975a4b3865d6LL,
    /* 200 */ 0x8cceebc98a5c108fLL, 0x7e12c4589654f2dcLL, 0xa49ad49fb0d19772LL, 0x3d142dd9c406152bLL, 0x9f13589e7be2b8a5LL,
    /* 205 */ 0x5e8dbac1892967adLL, 0xcc23b93a6308e597LL, 0x1ef35f5fe874e16aLL, 0x63ae9cc08d2e274fLL, 0x5bbabee56007fc05LL,
    /* 210 */ 0xabfd72994230fc39LL, 0x9d71a13a99144de1LL, 0xd9daf5aa8dcc89b3LL, 0xe145ec0514161bfdLL, 0x143befc2498cd270LL,
    /* 215 */ 0xa8e192557dbbd9f8LL, 0xcbeda2445628d7d0LL, 0x997f0a93205d9ea4LL, 0x01014a97f214ebfaLL, 0x70c026ffd1ebedafLL,
    /* 220 */ 0xf8737b1b3237002fLL, 0x8afcbef3147e6e5eLL, 0x0e1bb0684483ebd3LL, 0x4cbad70ae9b05aa6LL, 0xd4a31f523517c363LL,
    /* 225 */ 0xdb0f057ae8e9e8a2LL, 0x400894a919d89df6LL, 0x6a626a9b62defab3LL, 0xf907fd7e14f4e201LL, 0xe10e4a5657c48f3fLL,
    /* 230 */ 0xb17f9f54b8e6e5dcLL, 0x6b9e69045fa6d27aLL, 0x8b74b6a41dc3078eLL, 0x027954d45ca367f9LL, 0xd07207b8fdcbb7ccLL,
    /* 235 */ 0xf397c47d2f36414bLL, 0x05e4e8b11d3a034fLL, 0x36adb3f7122d654fLL, 0x607d9540eb336078LL, 0xb639118e3a8b9600LL,
    /* 240 */ 0xd0a406770b5f1484LL, 0x3cbee8213ccfb7c6LL, 0x467967bb2ff89cf1LL, 0xb115fe29609919a6LL, 0xba740e6ffa83287eLL,
    /* 245 */ 0xb4e51be9b694b7cdLL, 0xc9a081c677df5aeaLL, 0x2e1fbcd8944508ccLL, 0xf626e7895581fbb8LL, 0x3ce6e9b5728a05cbLL,
    /* 250 */ 0x46e87f2664a31712LL, 0x8c1dc526c2f6acfaLL, 0x7b4826726e560b10LL, 0x2966e0099d8d7ce1LL, 0xbb0dd5240d2b2adeLL, 
    /* 255 */ 0x0d527cc60bbaa936LL
};

/*
** Rotate 64 bit hash value by count in either direction (a barrel roll)
*/

inline static w_hash rotate(w_hash value, w_int count) {

  return ((value << count) | (value >> (64 - count)));

}

// [CG 20000710] This may be lightly more efficient when count is known to be 1
inline static w_hash rotate1(w_hash value) {

  return ((value << 1) | (value >> (63)));

}

/*
** Calculate the 'Buz' hashcode of a given key that is less than 65 bytes long,
** starting from an initial hash value. This function is used to calculate longer
** keys.
**
** Reference to this hash code algorithm can be found on http://www.serve.net/buzz
*/

inline static w_hash buzHash_64(w_hash initial, w_byte *key, w_int length) {

  w_hash buz = initial;
  w_int i;
    
  for (i = 0; i < length; i++) {
//    buz = rotate(buz, 1) ^ mixmaster[*key++];
    buz = rotate1(buz) ^ mixmaster[*key++];
  }
  
  return buz;
  
}

/*
** Implementation of buzHash with a key that is unlimited in length.
*/

w_hash buzHash(w_byte *key, w_int length) {

  w_hash buz = INITIAL_BUZ;
  w_int chunksize;
  
  do {
    chunksize = (length > 64) ? 64 : length;
    buz = buzHash_64(buz, key, chunksize);
    key += chunksize;
    length -= chunksize;
  } while (length);
  
  return buz;
  
}

/*
** Implementation of functions needed for cstring hashtables, uses buzHash.
*/

w_word cstring_hash (w_word a) {
  return buzHash((w_ubyte *)a, (w_int)strlen((char *)a));
}

w_boolean cstring_equal (w_word a, w_word b) {
  return (strcmp((char *)a, (char *)b) == 0);
}

/*
** -----------------------------------------------------------------------
** ht2k_ functions: these differ from the ht_ functions in that the key
** consists of two w_words, not one.  For example the key could be a 
** w_long or w_double, or (more probably) it could be two independent
** w_word's, so that ht2k becomes an implementation of a sparse 2d matrix.
** -----------------------------------------------------------------------
*/

/*
** ht2k_hash calculates the value of (hashcode MOD hashtable->currentsize),
** where MOD is unsigned integer modulus.  See ht_hash().
*/
inline static w_int ht2k_hash(w_hashtable2k hashtable, w_word hashcode) {
  w_int modulus = hashtable->currentsize;
  w_int result;

  result = hashcode - (hashcode/modulus)*modulus;

  return result;
}


/*
** Then the probe sequence is h(K), h(K)-1,...,0,M-1,M-2,...h(K)+1 and is obtained
** by calling ht_2kprobe successively with sequence = 0, 1, 2, ...
*/

w_int ht2k_probe(w_hashtable2k hashtable, w_word hashcode, w_int sequence) {
  w_int result = ht2k_hash(hashtable,hashcode-sequence);

    if(result<0) {
      wabort(ABORT_WONKA,"hashcode = 0x%08x, result = %d\n",hashcode,result);
    }
  while (result<0) result += hashtable->currentsize;

#ifdef HASHTEST
  woempa(3, "hashtable %p hashcode %08x modulus %d sequence %d : result =%d\n", hashtable, hashcode, hashtable->currentsize, sequence, result);
#endif

  return result;
}

/*
** For deletion, we need to examine the 'next' slot.
** ht2k_reprobe(hashtable,current) gives the value which follows 'current' 
** in the probe sequence
*/

static inline w_int ht2k_reprobe(w_hashtable2k hashtable, w_int current) {
  return current ? current-1 : hashtable->currentsize-1;
}

/*
** Also for deletion, we need to be able to measure the "distance" from
** one slot to another in the probing sequence: ht2k_distance gives the
** number of times we would have to call 'from = ht2k_reprobe(hashtable,from)'
** to make 'from' into 'to' ...
*/

static inline w_int ht2k_distance(w_hashtable2k hashtable, w_int from, w_int to) {
  return from>=to ? from-to : hashtable->currentsize + from-to;
}

/*
** Then the question "does r lie cyclically between i and j" becomes
**   ht2k_distance(hashtable,i,r) < ht2k_distance(hashtable,i,j).
*/

/****************************************************************************/

/*
**
** recalculateThresholds(w_hashtable2k) calculates a new lowthreshold
** and highthreshold from the current table size;  these thresholds 
** are used by ht2k_check_size to decide whether a resizing is required. 
** It is here that the 93% occupancy limit is enforced (provided of 
** course that ht2k_check_size() is called at the appropriate moments).
*/

void
recalculateThresholds2k(w_hashtable2k h) {
  w_int maxhighthreshold = h->currentsize - h->currentsize/16;
  if(maxhighthreshold>=h->currentsize) maxhighthreshold = h->currentsize-1;

  h->lowthreshold = h->currentsize/4;

  if(h->lowthreshold==0) h->lowthreshold = 1;

  h->highthreshold = h->currentsize - h->currentsize/4;

  if(h->highthreshold>maxhighthreshold) h->highthreshold = maxhighthreshold;
}


/*
************************************************************************
*/

static inline w_word hash2k(w_word key1, w_word key2) {
/*
  w_long temp = key1;

  temp = (temp << 32) | key2;

  return (w_word)(temp / 1234567);
*/
  return key1 ^ key2;
}

static inline w_boolean
ht2k_match(w_hashtable2k hashtable, int idx, w_word h, w_word k1, w_word k2) {
  woempa(1, "%s: index %d holds 0x%08x/0x%08x, looking for  0x%08x/0x%08x\n", hashtable->label, idx, hashtable->keys1[idx], hashtable->keys2[idx], k1, k2);

  return (hashtable->keys1[idx] == k1) && (hashtable->keys2[idx] == k2);

}

static inline w_boolean
ht2k_nullkey(w_word key1, w_word key2) {

  return key1 == 0 && key2 == 0;

}

w_boolean
ht2k_occupied(w_hashtable2k hashtable, w_int idx) {

  return !ht2k_nullkey(hashtable->keys1[idx], hashtable->keys2[idx]);

}

/*
************************************************************************
*/

w_hashtable2k 
_ht2k_create (const char *f, int l,
  w_realm realm,
  char *   label,
  w_size  initialsize
) {
  w_hashtable2k hashtable;

  hashtable = allocClearedMem_realm(realm, sizeof(w_Hashtable2k));
  hashtable->realm = realm;
  woempa(1,"creating %s at %p, size = %d\n",label,hashtable,initialsize);

  hashtable->label  = label;
  hashtable->currentsize = initialsize;
  hashtable->occupancy = 0;
  x_monitor_create(&hashtable->monitor);

  hashtable->keys1 = allocClearedMem_realm(realm, initialsize * sizeof(w_word));
  hashtable->keys2 = allocClearedMem_realm(realm, initialsize * sizeof(w_word));
  hashtable->values = allocClearedMem_realm(realm, initialsize * sizeof(w_word));

  recalculateThresholds2k(hashtable);

  return hashtable;
}

void
ht2k_resize(w_hashtable2k hashtable, w_size newsize) 
{
  w_size  oldsize = hashtable->currentsize;
  w_word *oldkeys1 = hashtable->keys1;
  w_word *oldkeys2 = hashtable->keys2;
  w_word *oldvalues = hashtable->values;
  w_word *newkeys1;
  w_word *newkeys2;
  w_word *newvalues;
  w_size  sequence;
  w_size  oldindex;
  w_size  newindex;
  w_word  key1;
  w_word  key2;

  woempa(1,"(re)allocating hashtable %s @ %p, new size %d\n",hashtable->label,hashtable,newsize);

  newkeys1 = allocClearedMem_realm(hashtable->realm, newsize * sizeof(w_word));
  if (!newkeys1) {
    woempa(9, "No memory available for newkeys1 of %s\n", hashtable->label);

    return;

  }

  newkeys2 = allocClearedMem_realm(hashtable->realm, newsize * sizeof(w_word));
  if (!newkeys2) {
    woempa(9, "No memory available for newkeys2 of %s\n", hashtable->label);
    releaseMem(newkeys1);

    return;
  }

  newvalues = allocClearedMem_realm(hashtable->realm, newsize * sizeof(w_word));
  if (!newvalues) {
    woempa(9, "No memory available for newvalues of %s\n", hashtable->label);
    releaseMem(newkeys2);
    releaseMem(newkeys1);

    return;
  }

  hashtable->currentsize = newsize;
  hashtable->keys1 = newkeys1;
  hashtable->keys2 = newkeys2;
  hashtable->values = newvalues;

  for(newindex=0;newindex<newsize;++newindex) {
    newkeys1[newindex] = 0;
    newkeys2[newindex] = 0;
  }

  if(oldsize && newsize) {
  /*
  ** rehash old contents into new arrays
  */

    for(oldindex=0;oldindex<oldsize;++oldindex) {
      key1 = oldkeys1[oldindex];
      key2 = oldkeys2[oldindex];
      if(!ht2k_nullkey(key1, key2)) { 
        w_word value = oldvalues[oldindex];
        w_word hc    = hash2k(key1, key2);

        for(sequence=0;sequence<newsize;++sequence) {
          newindex = ht2k_probe(hashtable,hc, (signed)sequence);
          if(!ht2k_occupied(hashtable, (signed)newindex)) {
            newkeys1[newindex]     = key1;
            newkeys2[newindex]     = key2;
            newvalues[newindex]    = value;
            break;
          }
        }
      }
    }
  }

  if (oldkeys1) {
    releaseMem(oldkeys1);
  }
  if (oldkeys2) {
    releaseMem(oldkeys2);
  }
  if (oldvalues) {
    releaseMem(oldvalues);
  }

  recalculateThresholds2k(hashtable);

  woempa(1,"hashtable keys1 at %p\n",hashtable->keys1);
  woempa(1,"hashtable keys2 at %p\n",hashtable->keys2);
  woempa(1,"hashtable values at %p\n",hashtable->values);
#ifdef CACHE_HASHCODES
  woempa(1,"hashtable hashcodes at %p\n",hashtable->hashcodes);
#endif

}

void
ht2k_destroy(w_hashtable2k theHashtable) {

  x_monitor_delete(&theHashtable->monitor);
  if (theHashtable->keys1) {
    releaseMem(theHashtable->keys1);
  }
  if (theHashtable->keys2) {
    releaseMem(theHashtable->keys2);
  }
  if (theHashtable->values) {
    releaseMem(theHashtable->values);
  }
  releaseMem(theHashtable);
}

w_size
ht2k_check_size(w_hashtable2k hashtable, w_int direction) {
  w_int       oldsize;
  w_int       newsize;

  woempa(1,"hashtable %p size now is %d, occupancy %d, thresholds %d %d\n",hashtable,hashtable->currentsize,hashtable->occupancy,hashtable->lowthreshold,hashtable->highthreshold);
  oldsize = hashtable->currentsize;
  newsize = oldsize;

  if( (hashtable->occupancy>=hashtable->highthreshold && direction>=0)
   || (hashtable->occupancy<=hashtable->lowthreshold  && direction<=0)
    ) {
    newsize = hashtable->occupancy * 2;

    if(newsize<hashtable->occupancy+2) newsize=hashtable->occupancy+2;
    if((newsize&1)==0) ++newsize;
    if((newsize%3)==0) newsize += 2;
    if((newsize%5)==0) newsize += 4;

    if(newsize==oldsize) newsize = 0;
    else {
      woempa(1,"recommend resizing to %d\n",newsize);
    }

    return newsize;

  }
  else return 0;

}


/*
** ht2k_rehash is a no-op: see Theorem P on p. 530 of Knuth, op. cit..
*/

w_hashtable2k 
ht2k_rehash(w_hashtable2k oldtable) {

  return oldtable;

}


void
ht2k_insert(w_hashtable2k hashtable, w_int idx, w_word key1, w_word key2, w_word value)
{
  woempa(1, "hashtable %p index %d key %p %p value %p\n", hashtable, idx, (char *)key1, (char *)key2, (char *)value);
  /*
  ** Protect ourselves against barbarous practices of the user
  */
  if(ht2k_nullkey(key1, key2) 
     || ht2k_occupied(hashtable, idx) 
     || hashtable->currentsize-hashtable->occupancy<=1) {

    woempa(1,"attempt to corrupt hashtable thwarted!\n");
    return;
  }

  hashtable->keys1[idx] = key1;
  hashtable->keys2[idx] = key2;
  hashtable->values[idx] = value;
  hashtable->occupancy += 1;
}

void
ht2k_delete(w_hashtable2k hashtable, w_int idx) {
  w_int     vacant;
  w_int     current;
  w_int     home;
  w_boolean happy;

  woempa(1, "hashtable %s @ %p: delete slot %d\n",hashtable->label,hashtable,idx);
  /*
  ** The following is a summary of Knuth's Algorithm R.
  **
  ** Step R1 : Mark TABLE[i] empty, and set j equal to i.
  ** Step R2 : decrease i, modulo M
  ** Step R3 : if TABLE[i] is empty, the algorithm terminates.
  **           Otherwise set r equal to h(KEY[i]), the original hash 
  **           address of KEY[i]).  If i<=r<j (modulo M), go back to R2.
  ** Step R4 : Set TABLE[j] equal to TABLE[i], and return to step R1.
  **
  ** In Knuth's therminology, TABLE[j] is our vacant and TABLE[i] is
  ** our current; TABLE[r] corresponds to home, the "home" slot
  ** of the entry at current. 
  */

  hashtable->occupancy -= 1;
  current = idx;

  while(1) {
  /*
  ** R1
  */
    woempa(1,"mark slot %d as empty\n",current);
    hashtable->keys1[current] = 0;
    hashtable->keys2[current] = 0;
    vacant = current;
  /*
  ** R2
  */
    happy = WONKA_TRUE;

    while(happy) {
      current = ht2k_reprobe(hashtable,current);
      woempa(1,"Next slot is %d\n",current);
  /*
  ** R3
  */
      if(ht2k_nullkey(hashtable->keys1[current], hashtable->keys2[current])) {
        woempa(1,"slot %d is empty, algorithm terminated normally\n",current);

        return;

      }

      home = ht2k_probe(hashtable,hash2k(hashtable->keys1[current], hashtable->keys2[current]), 0); 
      woempa(1,"  whereas in slot %d it would be %d slots from home\n",vacant,ht2k_distance(hashtable,home,vacant));
      if( ht2k_distance(hashtable,home,current) 
        < ht2k_distance(hashtable,home,vacant)) {
        woempa(1,"so it can stay where it is\n");
      } else {
        woempa(1,"so it needs to move ...\n");
        happy = WONKA_FALSE;
      }
    }
  /*
  ** R4
  */
    woempa(1,"copying from slot %d to slot %d\n",current,vacant);
    hashtable->keys1[vacant]      = hashtable->keys1[current];
    hashtable->keys2[vacant]      = hashtable->keys2[current];
    hashtable->values[vacant]    = hashtable->values[current];
  }
}

/*
***************************************************************************
*/

w_boolean
ht2k_write_no_lock(w_hashtable2k hashtable, w_word key1, w_word key2, w_word newvalue) {
  w_int  seq;
  w_int  maxseq = hashtable->currentsize;
  w_word result = 0;
  w_word hashcode = hash2k(key1, key2);
  w_size newsize;
  w_int idx;

  for(seq=0;seq<maxseq;++seq) {                                                    
    idx=ht2k_probe(hashtable,hashcode,seq);                            
    if(ht2k_occupied(hashtable,idx)) {                                        
      if(ht2k_match(hashtable, idx, hashcode, key1, key2)) {                            
      woempa(1,"%s: REPLACE slot %d with %d collisions\n",hashtable->label,idx,seq);
        result = WONKA_TRUE;
        hashtable->values[idx] = newvalue;                           
        break;                                              
      }                                                   
    }                                                    
    else {                                              
      woempa(1,"%s: INSERT slot %d with %d collisions\n",hashtable->label,idx,seq);
      hashtable->keys1[idx] = key1;
      hashtable->keys2[idx] = key2;
      hashtable->values[idx] = newvalue;
      hashtable->occupancy += 1;
      newsize = ht2k_check_size(hashtable,1);
      if (newsize) {
        ht2k_resize(hashtable, newsize);
      }
      break;                                                            
    }                                                                  
  }                                                                   

  return result;

}


w_word /* value found or 0 */
ht2k_read_no_lock(w_hashtable2k hashtable, w_word key1, w_word key2) {
  w_int  seq;
  w_int  maxseq;
  w_word result;
  w_word hashcode;
  w_int  idx;

  maxseq = hashtable->currentsize;
  result = 0;
  hashcode = hash2k(key1, key2);

  for(seq=0;seq<maxseq;++seq) {

    idx=ht2k_probe(hashtable,hashcode,seq);
    if(ht2k_occupied(hashtable,idx)) {
      if(ht2k_match(hashtable, idx, hashcode, key1, key2)) {
        woempa(1,"%s: MATCH slot %d with %d collisions\n",hashtable->label,idx,seq);
        result = hashtable->values[idx];
        break;
      }
    }
    else {
      break;
    }
  }
 
  return result;

}

w_boolean
ht2k_erase_no_lock(w_hashtable2k hashtable, w_word key1, w_word key2) {
  w_int  seq;
  w_int  maxseq = hashtable->currentsize;
  w_boolean result = WONKA_FALSE;
  w_word hashcode = hash2k(key1, key2);
  w_int  idx;
#ifdef AUTO_SHRINK_HASHTABLES
  w_size newsize;
#endif

  for(seq=0;seq<maxseq;++seq) {

    idx=ht2k_probe(hashtable,hashcode,seq);
    if(ht2k_occupied(hashtable,idx)) {
      if(ht2k_match(hashtable, idx, hashcode, key1, key2)) {
        result = WONKA_TRUE;
        ht2k_delete(hashtable,idx);
#ifdef AUTO_SHRINK_HASHTABLES
        newsize = ht2k_check_size(hashtable,-1);
        if (newsize) {
          ht2k_resize(hashtable, newsize);
        }
#endif
        break;
      }
    }
    else {
      break;
    }
  }

  return result;
 

}


w_int
ht2k_every(w_hashtable2k hashtable, void (*fun)(w_word key1, w_word key2, w_word value)) {
  w_int count = 0;
  w_int i;

ht2k_lock(hashtable);

  for(i=0;i<hashtable->currentsize;++i)
  {
    if(ht2k_occupied(hashtable,i)) {
      fun(hashtable->keys1[i], hashtable->keys2[i], hashtable->values[i]);
      ++count;
    }
  }

ht2k_unlock(hashtable);

  return count;
}

w_fifo
ht2k_list_values(w_hashtable2k hashtable) {
  w_fifo result;
  w_size leafsize;
  w_int i;

  leafsize = 63;
  while (leafsize < (w_size)hashtable->occupancy) {
    leafsize = leafsize * 2 - 1;
  }
  result = allocFifo(hashtable->realm, leafsize);
  if (result == NULL) {

    return NULL;

  }


  ht2k_lock(hashtable);

  for(i=0;i<hashtable->currentsize;++i)
  {
    if(ht2k_occupied(hashtable,i)) {
      if (putFifo((void*)hashtable->values[i],result) <0) {
        wabort(ABORT_WONKA, "I don't believe it!  We sized fifo %p to have a leafsize of %d, and after %d putFifo's it overflows!\n", result, leafsize, i);
      }
    }
  }

  ht2k_unlock(hashtable);

  return result;
}

w_int
ht2k_iterate_no_lock(w_hashtable2k hashtable, void (*fun)(w_word key1, w_word key2, w_word value, void *arg1, void *arg2), void *arg1, void *arg2) {
  w_int count = 0;
  w_int i;

  for(i=0;i<hashtable->currentsize;++i)
  {
    if(ht2k_occupied(hashtable,i)) {
      fun(hashtable->keys1[i], hashtable->keys2[i], hashtable->values[i], arg1, arg2);
      ++count;
    }
  }

  return count;

}


 
