/**************************************************************************
* Copyright (c) 2001, 2003 by Acunia N.V. All rights reserved.            *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Philips site 5, box 3       info@acunia.com                           *
*   3001 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: comms.c 3590 2003-03-04 17:31:29Z gray $
*/

#include <stdio.h>
#include <unistd.h>
#include "comms.h"
#include "sio.h"
#include "ts-mem.h"
#include "realms.h"
#include "oswald.h"
#include "wonka.h"
#include "driver_byteserial.h"

/*
** Currently we have 4 hard-coded comm ports, 3 hard-coded
** file descriptors, and one wonkaterm.
** Should be made smarter.
** For each comm port we have a w_Sio structure, a w_Device,
** and two termios buffers; one to store the state when
** Wonka was started up (so we can restore it on exit),
** the other to store (and manipulate) the current state.
** For each file descriptor we have a w_Wfd structure and
** a w_Device, for each wonkaterm we have a w_Wt structure 
** and a w_Device.
*/

static struct termios sio0_oldtermios;
static struct termios sio0_newtermios;
static struct termios sio1_oldtermios;
static struct termios sio1_newtermios;
static struct termios sio2_oldtermios;
static struct termios sio2_newtermios;
static struct termios sio3_oldtermios;
static struct termios sio3_newtermios;

/*
** The pathnames for the four serial ports
*/
char * pathname[4];

typedef struct w_Control_Sio {
  struct pollfd pollfd;
/*
** clock speed (bits/sec)
*/
  w_int      bitrate;
/*
** Receive buffer: size, start, point where next char will be read/written,
** number of bytes currently used.
*/
  w_int      rxBufSize;
  w_ubyte    *rxBuffer;
  w_ubyte    *rxReadPtr;
  w_ubyte    *rxWritePtr;
  w_int      rxBytesInBuf;
/*
** Value of rxBytesInBuf at which DATA_AVAILABLE will be signalled
*/
  w_int      rxThreshold;
/*
** Counting semaphores used to synchronize producers and consumers of
** received bytes, sent bytes, and i/o events.
*/
  x_Sem     rxSem;    // count = bytes in buffer
  x_Sem     txSem;    // count = bytes buffered in producer
  x_Sem     evtSem;   // count = outstanding events
/*
** Which flags (DTR, RTS, DSR, CTS, CD, RI) are supported by this UART
*/
  w_short   validflags;
/*
** Values of the IOFLAGS for this sio
**   oldflags = already reported via ioevt
**   newflags = latest available values
**   thisflag: only one bit is set, indicating which bit of newflags
**             caused ioevt to return control to its caller last time.
**   thisvalue: either zero or equal to thisflag, depending on whether
**              the value of the flag identified by thisflag was 0 or 1.
*/
  w_short   oldflags;
  w_short   newflags;
  w_short   thisflag;
  w_short   thisvalue;
/*
** Other flags used internally for congestion control etc.
*/
  w_word     miscflags;
/*
** Field to store the result of a 'wait_for_event'
*/
  w_word     event_data;
/*
** Pointers to the termios structure as it was before we grabbed the sio,
** and as it is now
*/
  w_termios  oldtermios;
  w_termios  newtermios;
/*
** The name of the actual device e.g. '/dev/ttyS0'
*/
  char       *devicename;
  
} w_Control_Sio;

typedef w_Control_Sio *w_control_sio;

/*
** Somewhat ad-hoc function to set the path of an sio device
*/
void sio_set_path(int n, char *s) {
  if (n >= 0 && n < 4) {
    char *buf = allocMem_system(strlen(s)+1);
    strcpy(buf, s);
    pathname[n] = buf;
  }
  else {
    woempa(9, "sio devnum %d out of range\n", n);
  }
}

/*
** Prototypes 
*/

w_void sio_initDevice(w_device device); 
w_void sio_termDevice(w_device device); 
w_device sio_open(w_device device, w_driver_perm mode);
w_void sio_close(w_device device);
w_driver_status sio_read(w_device device, w_ubyte *bytes, w_int length, w_int *lread, x_sleep timeout);
w_driver_status sio_write(w_device device, w_ubyte *bytes, w_int length, w_int *lwritn, x_sleep timeout);
w_driver_status sio_seek(w_device device, w_int offset, x_sleep timeout);
w_driver_status sio_set(w_device device, w_driver_ioctl command, w_word param, x_sleep timeout);
w_driver_status sio_query(w_device device, w_driver_ioctl query, w_word *reply, x_sleep timeout);

/*
** sio_init sets the port state to default values, and saves
** the current state so that it can be restored later.
*/

w_void sio_initDevice(w_device device) {
  w_control_sio c;
  device->control = allocClearedMem_system(sizeof(w_Control_Sio));
  c = device->control;

  c->validflags = IOFLAG_DTR|IOFLAG_RTS|IOFLAG_DSR|IOFLAG_CTS|IOFLAG_CD|IOFLAG_RI;
  switch(device->familyMember) {
    case 0: c->bitrate = SIO0_DEFAULT_BITRATE;
            c->devicename = pathname[0];
            c->oldtermios = &sio0_oldtermios;
            c->newtermios = &sio0_newtermios;
            break;
    case 1: c->bitrate = SIO1_DEFAULT_BITRATE;
            c->devicename = pathname[1];
            c->oldtermios = &sio1_oldtermios;
            c->newtermios = &sio1_newtermios;
            break;
    case 2: c->bitrate = SIO2_DEFAULT_BITRATE;
            c->devicename = pathname[2];
            c->oldtermios = &sio2_oldtermios;
            c->newtermios = &sio2_newtermios;
            break;
    case 3: c->bitrate = SIO3_DEFAULT_BITRATE;
            c->devicename = pathname[3];
            c->oldtermios = &sio3_oldtermios;
            c->newtermios = &sio3_newtermios;
            break;
    default:
  }
            
  // atexit(sio_term);
}

w_void sio_termDevice(w_device device) {
  w_control_sio sio = device->control;
  
  if (sio->pollfd.fd>=0) {
    tcsetattr(sio->pollfd.fd, TCSAFLUSH, sio->oldtermios);
  }
  sio_close(device);

  releaseMem(device->control);
}


w_device sio_open(w_device device, w_driver_perm mode) {
  w_control_sio sio = device->control;
  
  if(device->usage == 0) {
    
    sio->miscflags = mode = wdp_read|wdp_write;
  
//  if (isNotSet(sio->miscflags, wdp_read|wdp_write)) {
// Somehow this if always returns false... Not so good...
  
      sio->pollfd.fd = open(sio->devicename,O_RDWR);
      sio->pollfd.events = POLLIN|POLLOUT;
      
      /*
      ** If a tty, save current (pre-wonka) settings in oldtermios.  (If not a tty,
      ** this call returns nonzero and we skip the rest of the tty-related stuff).
      */
      if (tcgetattr(sio->pollfd.fd,sio->oldtermios)==0) {
        /*
        ** Copy oldtermios to newtermios and make some changes:
        ** echo off, canonical mode off, extended i/p proc off, signal chars off
        ** no SIGINT on BREAK, CR>NL off, i/p parity off, strip8 off, CTS/RTS off
        ** size 8 bits, no parity check, no output processing
        ** read is blocking with no timeout
        */
        w_memcpy(sio->newtermios,sio->oldtermios,sizeof(struct termios));
        sio->newtermios->c_lflag &= ~(ECHO | ICANON | IEXTEN | ISIG);
        sio->newtermios->c_iflag &= ~(BRKINT | ICRNL | INPCK | ISTRIP | IXON);
        sio->newtermios->c_cflag &= ~(CSIZE | PARENB);
        sio->newtermios->c_cflag |= CREAD | CS8 | CLOCAL;
        sio->newtermios->c_oflag &= ~(OPOST);
        sio->newtermios->c_cc[VMIN] = 1;
        sio->newtermios->c_cc[VTIME] = 0;
        cfsetispeed(sio->newtermios,(w_size)sio->bitrate);
        cfsetospeed(sio->newtermios,(w_size)sio->bitrate);
        tcsetattr(sio->pollfd.fd,TCSAFLUSH,sio->newtermios);
      }
  
      sio->oldflags = 0;
      sio->newflags = 0;
      x_sem_create(&sio->rxSem,0);
      x_sem_create(&sio->txSem,1);
      x_sem_create(&sio->evtSem,0);
//  }

//  if (mode & wdp_read) {
 
      if (sio->rxBuffer) releaseMem(sio->rxBuffer);
      if (sio->rxBufSize==0) {
        sio->rxBufSize = sio->bitrate/RX_BUF_SIZE_DIVISOR;
      }
      sio->rxBuffer = allocMem(W_Thread_system, (w_size)sio->rxBufSize);
      sio->rxReadPtr = sio->rxBuffer;
      sio->rxWritePtr = sio->rxBuffer;
      sio->rxBytesInBuf = 0;
//  }
    
  }

  device->usage += 1;
  device->driver->usage += 1;
  
  return device;
}


w_void sio_close(w_device device) {
  w_control_sio sio = device->control;

  if(device->usage == 1) {
  
    if (sio->miscflags & (wdp_read | wdp_write)) {
      sio->miscflags &= ~(wdp_read | wdp_write);
      if (sio->pollfd.fd>=0) close(sio->pollfd.fd);

      sio->rxBufSize = 0;
      if (sio->rxBuffer) {
        releaseMem(sio->rxBuffer);
        sio->rxBuffer = NULL;
      }
    }
  
    x_sem_delete(&sio->rxSem);
    x_sem_delete(&sio->txSem);
    x_sem_delete(&sio->evtSem);
    
  }

  device->usage -= 1;
  device->driver->usage -= 1;
}


w_driver_status sio_read(w_device device, w_ubyte *bytes, w_int length, w_int *lread, x_sleep timeout) {
  w_control_sio sio = device->control;

  *lread = read(sio->pollfd.fd, bytes, (w_word)length);

  return wds_success;
  
}


w_driver_status sio_write(w_device device, w_ubyte *bytes, w_int length, w_int *lwritn, x_sleep timeout) {
  w_control_sio sio = device->control;

  *lwritn = write(sio->pollfd.fd, bytes, (w_word)length);

  return wds_success;
}


w_driver_status sio_seek(w_device device, w_int offset, x_sleep timeout) {
  return wds_not_implemented;
}


w_driver_status sio_set(w_device device, w_driver_ioctl command, w_word param, x_sleep timeout) {
  return wds_no_such_command;
}


w_driver_status sio_query(w_device device, w_driver_ioctl query, w_word *reply, x_sleep timeout) {
  return wds_no_such_command;
}


w_Driver_ByteSerial sio_driver = {
  "serial-I/O-driver-linux_v0.1",
  "sio",
  0,
  NULL,
  sio_initDevice,
  sio_termDevice,  
  sio_open,
  sio_close,  
  sio_read,
  sio_write,
  sio_seek,
  sio_set,
  sio_query
};

