/**************************************************************************
* Copyright (c) 2001, 2002, 2003 by Acunia N.V. All rights reserved.      *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Philips-site 5 box 3        info@acunia.com                           *
*   3001 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: debug.c 4294 2006-05-10 11:35:19Z chris $
*/

#include <string.h>
#include <stdio.h>

#include "loading.h"
#include "oswald.h"
#include "ts-mem.h"
#include "realms.h"
#include "threads.h"
#include "methods.h"
#include "clazz.h"
#include "hashtable.h"

#include "vfs.h"
#include "vfs_fcntl.h"

#include "checks.h"

#include <time.h>

#ifndef FALSE
#define FALSE 0
#endif

#ifndef TRUE
#define TRUE 1
#endif

extern w_clazz clazzClassLoader;

const char *abortMessages[] = {
  "Information",
  "ABORTING thread",
  "ABORTING group",
  "ABORTING root group",
  "ABORT Wonka",
};

x_Mutex woempaMutex;

#ifndef VERSION_STRING
#define VERSION_STRING "<unknown>"
#endif

const char *banner = VERSION_STRING " (Developers' version)"
                     " was compiled on " __DATE__ 
                     " at " __TIME__ ".\r\n";

void (*puts_vector)(char *s);

void PutString(char *s) {
  if (puts_vector) {
    (*puts_vector)(s);
  }
}

char hexDigit(int n) {
  return n<10 ? n+'0' : n+'a'-10;
}

void PutHex (unsigned int Value, unsigned int Width) {
  char buf[9];
  int i, x;

  x = Value;
  for (i=7;i>=0;--i) {
    buf[i] = hexDigit(x&0xf);
    x >>= 4;
  }
  buf[8] = 0;
  PutString(Width>8?buf:buf+8-Width);
}

void PutDec (unsigned int Value, unsigned int Width) {
  char buf[12];
  int i, x;
  char sign;

  x = Value;
  if (x<0) {
    sign = '-';
    x = -x;
  }
  else sign = 0;

  for (i=10;i>=0;--i) {
    buf[i] = '0'+(x%10);
    x /= 10;
  }
  buf[11] = 0;
  PutString(Width>11?buf:buf+11-Width);
}

/*
** We use a hashtable to store the woempa trigger level associated with
** each sourcefile: the key is a C string, and if the string contains a
** '/' character then the hashcode and equality functions ignore everything
** to the left of the last '/' in the string.  (In other words, the whole
** path doesn't have to match, just the 'basename').  If a file is not
** found in the hashtable then its trigger level is assumed to be
** woempa_default_trigger_level.
**
** There is one special case" file "hashtable.c" has its own trigger level
** outside of the hashtable, to avoid a deadlock when woempa is called
** from inside a hashtable function.
*/

static w_hashtable woempa_hashtable = NULL;
static w_int woempa_default_trigger_level = 1;

static inline char *
get_basename(const char *filename) {
  char *basename = (char *)filename + strlen(filename);
  while (basename-filename > 0 && *--basename != '/');

  return basename;
}

w_word
filename_hash(w_word nameword) {
  char *filename = (char*)nameword;
  char *basename = filename + strlen(filename);
  int   hashcode = 0x654321;
  char  one_byte;

  basename = get_basename(filename);

  while ((one_byte = *basename++)) {
    hashcode = (hashcode * 255) ^ one_byte;
  }

  return hashcode;
}

w_boolean
filenames_equal(w_word nameword1, w_word nameword2) {
  char *filename1 = (char*)nameword1;
  char *filename2 = (char*)nameword2;
  char *basename1;
  char *basename2;

  basename1 = get_basename(filename1);
  basename2 = get_basename(filename2);

  return strcmp(basename1,basename2) == 0;
}

static int hashtable_trigger_level = 1;

void _setTriggerLevel(const char *file, int trigger) {
  if (strcmp(get_basename(file),"hashtable.c") == 0) {
    hashtable_trigger_level = trigger;
  }
  else {

    if (woempa_hashtable == NULL) {
      woempa_hashtable = ht_create(W_Realm_system, (char*)"hashtable:woempa", 31, filename_hash, filenames_equal,0,0);
    }

  ht_write(woempa_hashtable,(w_word)file,(w_word)trigger);
  }
}

void setAllTriggerLevel(int trigger) {
  if (woempa_hashtable) {
    ht_destroy(woempa_hashtable);
    woempa_hashtable = NULL;
  }
  hashtable_trigger_level = trigger;
  woempa_default_trigger_level = trigger;
}

#define BUFSIZE        8192
static char woempa_buffer[BUFSIZE];

/*
** A small woempa aware function that can be called from the gdb debugger...
*/

void wprintf(const char *fmt, ...) {

  va_list ap;

  va_start(ap, fmt);
  x_vsnprintf(woempa_buffer, BUFSIZE, fmt , ap);
  va_end(ap);
#if defined(LINUX) || defined(NETBSD)
  write(1, woempa_buffer, strlen(woempa_buffer));
#else
  PutString(woempa_buffer);
  PutString((char *)"\15");
#endif

}

extern char *woempa_dump_file;
extern int  woempa_stderr;
extern int  woempa_bytecodecount;
extern int  woempa_bytecodetrigger;

void pclazz(w_clazz clazz) {
  woempa(9, "CLAZZ = %k\n", clazz);
}

void _woempa(const char *file, const char *function, int line, int level, const char *fmt, ... ) {

  va_list ap;
  int trigger = 0;
  w_size bufsize = BUFSIZE;
  static int od = 0;

  if (woempa_bytecodetrigger > 0 && 
      woempa_bytecodecount > woempa_bytecodetrigger)
    woempa_default_trigger_level = 2;

  if(woempa_dump_file) {
    if (! od) {
      /*         name     <--- flags --------------------------->  <------ modes -------------->  */
      od = open(woempa_dump_file, (VFS_O_WRONLY | VFS_O_CREAT | VFS_O_TRUNC), (VFS_S_IRUSR | VFS_S_IWUSR | VFS_S_IXUSR | VFS_S_IRGRP | VFS_S_IROTH));
    }
  }
  
  if (strcmp(get_basename(file),"hashtable.c") == 0) {
    trigger = hashtable_trigger_level;
  }
  else if (woempa_hashtable) {
    trigger = (int)ht_read(woempa_hashtable,(w_word)file);
  }
  else {
    trigger = woempa_default_trigger_level;
  }

  if (trigger < 1 || trigger > 9) {
    trigger = woempa_default_trigger_level;
  }
  
  if (trigger <= level && x_mutex_lock(&woempaMutex, x_eternal)==0) {
      
    w_thread jthread = currentWonkaThread;
    x_frame  jframe  = jthread ? jthread->top : NULL;
    w_method jmeth   = jframe  ? jframe->method : NULL;
    
    x_snprintf(woempa_buffer, bufsize, 
#ifdef WOEMPA_JTHREAD_FORMAT
       WOEMPA_JTHREAD_FORMAT " "
#endif
#ifdef WOEMPA_LEVEL_FORMAT
       WOEMPA_LEVEL_FORMAT " "
#endif
#ifdef WOEMPA_BYTECODECOUNT_FORMAT
       WOEMPA_BYTECODECOUNT_FORMAT " "
#endif
#ifdef WOEMPA_METHOD_FORMAT
       WOEMPA_METHOD_FORMAT " "
#endif
#ifdef WOEMPA_FUNCTION_FORMAT
       WOEMPA_FUNCTION_FORMAT " "
#endif
       "%s",
#ifdef WOEMPA_JTHREAD_FORMAT
      jthread,
#endif
#ifdef WOEMPA_LEVEL_FORMAT
      level,
#endif
#ifdef WOEMPA_BYTECODECOUNT_FORMAT
      woempa_bytecodecount,
#endif
#ifdef WOEMPA_METHOD_FORMAT
      jmeth ? jmeth->spec.declaring_clazz : NULL,
      jmeth ? jmeth->spec.name : NULL,
      jmeth ? jframe->current - jmeth->exec.code : 0,
#endif
#ifdef WOEMPA_FUNCTION_FORMAT
      function,
      line,
#endif
      ": "
    );

    va_start (ap, fmt);
    (void)x_vsnprintf(woempa_buffer + strlen(woempa_buffer), bufsize - strlen(woempa_buffer), fmt , ap);
    va_end (ap);
   
#ifdef OS_NONE
    if(woempa_stderr) {
      PutString(woempa_buffer);
      PutString((char *)"\15");
    }
#else
#ifndef ECOS
    if(od) { 
      (void)write(od, woempa_buffer, strlen(woempa_buffer));
    }
#endif //ECOS
 
   
    if(woempa_stderr) {
      if (level >= 7 || (woempa_bytecodecount > woempa_bytecodetrigger && woempa_bytecodetrigger > 0)) {
        (void)write(2, woempa_buffer, strlen(woempa_buffer));
      }
    }
#endif

    (void)x_mutex_unlock(&woempaMutex);

  }

}

void _wabort(const char *function, int line, int scope, const char *fmt, ... ) {
  x_thread kthread;
  w_thread thread;
  va_list ap;
  w_size bufsize = BUFSIZE;

  (void)x_mutex_lock(&woempaMutex, x_eternal);
#ifdef OS_NONE
  (void)write(1, (char *)"\15",1);
#endif
  x_snprintf(woempa_buffer, bufsize, "\n%s: %s %4d: ", abortMessages[scope], function, line);

  va_start (ap, fmt);
  (void)x_vsnprintf(woempa_buffer + strlen(woempa_buffer), bufsize - strlen(woempa_buffer), fmt , ap);
  va_end (ap);

  (void)write(1, woempa_buffer, strlen(woempa_buffer));
#ifdef OS_NONE
  (void)write(1, (char *)"\15",1);
#endif
  kthread = x_thread_current();
/*
** Replace with thread_hashtable lookup for now
  thread = (w_thread)kthread->wx_user_data;
*/
  thread = (w_thread)ht_read(thread_hashtable, (w_word)kthread);
  if ((*(w_word*)thread->label)!=(*(w_word*)"thread")){
    woempa(9,"currentWonkaThread called at %s:%d from non-wonka thread (label = %s)\n", function, line, thread->label);
    thread = NULL;
  }

  {
    w_fifo thread_list = ht_list_keys(thread_hashtable);
    if (thread_list) {
      x_thread other_thread;
      while ((other_thread = getFifo(thread_list))) {
        if (other_thread != kthread) {
          x_thread_suspend(other_thread);
        }
      }
    }
  }

#ifdef DEBUG
  if (thread) {
    PutString((char*)"Failing thread: ");
    PutString(threadDescription(thread)); 
    PutString((char *)"\12");
  }
  else {
    PutString((char*)"Unable to identify failing thread, sorry mate...\n");
  }
#ifdef OS_NONE
  (void)write(1, (char *)"\15",1);
#endif
#endif

  (void)x_mutex_unlock(&woempaMutex);

  if (scope > ABORT_INFO) {
#ifdef OS_NONE
    (void)x_thread_sleep(x_millis2ticks(5000));
    x_thread_deletei(x_thread_current());
#else
    abort();
#endif
  }
  
}

static char dump_buffer[BUFSIZE];

void w_dump(const char *fmt, ... ) {
  va_list ap;
  
  va_start (ap, fmt);
  (void)x_vsnprintf(dump_buffer, BUFSIZE, fmt , ap);
  va_end (ap);

#if defined(LINUX) || defined(NETBSD)
  (void)write(2, dump_buffer, strlen(dump_buffer));
#else
  PutString(dump_buffer);
  PutString((char *)"\15");
#endif
}

w_string w_dump_name(void *xref)
{
  return ((w_thread)xref)->name;
}

void w_dump_threads(void) {
  x_dump_threads();
}

w_trace_element w_trace_get(w_thread thread, x_int element) {
  x_frame frame = thread->top;
  w_trace_element result = NULL; 
  
  while (frame && element) {
    frame = frame->previous;
    element--;
  }

  if(element == 0 && frame && frame->method) {
    result = x_mem_calloc(sizeof(w_Trace_Element));
    x_mem_tag_set(result, 0x40);
    result->method = frame->method;
    result->pc = frame->current - result->method->exec.code;
    if (isNotSet(result->method->flags, ACC_NATIVE)) {
      result->line = code2line(result->method, result->method->exec.code + result->pc);
    }
  }

  return result;
}

typedef struct w_IEntry * w_ientry;

typedef struct w_IEntry {
  w_ientry next;
  w_word hash;
  w_trace_element trace;
  w_int hits;
} w_IEntry;

static w_ientry * ht;
static const w_size num_buckets = 5000;
static w_int num_ientry;

static w_ientry find(w_ientry new) {

  w_size i = new->hash % num_buckets;
  w_ientry ex;
  
  for (ex = ht[i]; ex; ex = ex->next) {
    if (ex->trace->method == new->trace->method && ex->trace->pc == new->trace->pc) {
      return ex;
    }
  }

  return NULL;

}

static void add(w_ientry new) {

  w_size i = new->hash % num_buckets;
  w_ientry ex;

  for (ex = ht[i]; ex; ex = ex->next) {
    if (ex->next == NULL) {
      ex->next = new;
      break;
    }
  }

  if (ex == NULL) {
    ht[i] = new;
  }

  num_ientry += 1;
}

static w_int ientry_sort_hits(const void * v_ientry1, const void * v_ientry2) {

  w_ientry ientry1 = *(w_ientry *)v_ientry1;
  w_ientry ientry2 = *(w_ientry *)v_ientry2;

  if(!ientry1) return -1;
  if(!ientry2) return 1;
  
  return (ientry1->hits == ientry2->hits ? 0 : (ientry1->hits < ientry2->hits ? 1 : -1));
}

static x_boolean traces_callback(void * mem, void * arg) {

  w_instance instance;
  w_object   object;
  w_ientry   ientry;
  w_ientry   ex;

  ientry = x_mem_alloc(sizeof(w_IEntry));
  if (! ientry) {
    return FALSE;
  }
  ientry->trace = (w_trace_element)mem;
  ientry->next = NULL;
  ientry->hits = 1;
  ientry->hash = (x_word)ientry->trace->method + ientry->trace->pc * 3;

  ex = find(ientry);
  if (ex) {
    ex->hits += 1;
    x_mem_free(ientry);
  }
  else {
    add(ientry);
  }
  
  return TRUE;
  
}

void reportTraces(void) {
  w_size i, j = 0;
  w_ientry entry;
  w_ientry victim;
  static w_ientry * table;

  x_mem_lock(x_eternal);
  
  ht = x_mem_calloc(num_buckets * sizeof(w_ientry));
  
  x_mem_scan(x_eternal, (0x40), traces_callback, NULL);

  table = x_mem_calloc(num_ientry * sizeof(w_ientry));
  for (i = 0; i < num_buckets; i++) {
    entry = ht[i];
    while (entry) {
      table[j++] = entry;
      entry = entry->next;
    }
  }
  
  qsort(table, num_ientry, sizeof(w_ientry), ientry_sort_hits);

  w_dump("--- Start Traces --- \n");

  for (i = 0; i < num_ientry; i++) {
    entry = table[i];
    if(entry) {
      w_dump("%5d %M %d %d\n", entry->hits, entry->trace->method, entry->trace->pc, entry->trace->line);
    }
  }

  w_dump("--- Stop Traces --- \n");

  x_mem_unlock();

  for (i = 0; i < num_buckets; i++) {
    entry = ht[i];
    while (entry) {
      victim = entry;
      entry = entry->next;
      x_mem_free(victim);
    }
  }

  x_mem_free(ht);  
  x_mem_free(table);  
  
  x_mem_unlock();
}

void w_dump_trace(void *xref) {
  w_thread thread = (w_thread)xref;
  x_frame frame = thread->top;
  w_method method;
  w_int pc;

#ifdef GC_SAFE_POINTS
  if (thread->unsafe_thread) {
    w_dump("     unsafe: thread %t\n", thread->unsafe_thread);
  }
  if (thread->gc_thread) {
    w_dump("     blocked by GC: thread %t\n", thread->gc_thread);
  }
#endif

  w_dump("     local references: %d\n", (frame->auxs - frame->thread->auxs));
  
  while (frame) {
    if (frame->method) {
      method = frame->method;
      pc = frame->current - method->exec.code;

      if (isSet(method->flags, METHOD_IS_COMPILED)) {
        w_dump("       %K/%m (jitted method)\n", method->spec.declaring_clazz, method);
      }
      else if (isSet(method->flags, ACC_NATIVE)) {
        w_dump("       %K/%m (native method)\n", method->spec.declaring_clazz, method);
      }
      else {
#ifdef STORE_METHOD_DEBUG_INFO
        w_dump("       %K/%m line %d (pc %d)\n", method->spec.declaring_clazz, method, code2line(method, method->exec.code + pc), pc);
#else
        w_dump("       %K/%m pc = %d\n", method->spec.declaring_clazz, method, pc);
#endif
      }
    }
    frame = frame->previous;
  }
}

extern x_mutex   finalizer_fifo_mutex;
extern x_mutex   enqueue_fifo_mutex;
extern x_mutex   string_mutex;
extern x_monitor reclaim_listener_monitor;
extern x_monitor gc_monitor;
#if defined(AWT_XSIM) || defined(AWT_FDEV)
extern x_monitor  tree_lock;
#endif

extern w_hashtable lock_hashtable;

static char lock_buffer[256];

void lock_iterator(w_word key, w_word value, void *v1, void *v2) {
  w_instance i = (w_instance)key;
  x_monitor  m = (x_monitor)value;
  x_snprintf(lock_buffer, 254, "%j", i);
  x_dump_monitor_if_locked(lock_buffer, m);
}

void w_dump_locks(void) {
  w_dump(" Locks :\n");
  x_dump_mutex("  Finalizer fifo mutex : ", finalizer_fifo_mutex);
  x_dump_mutex("    Enqueue fifo mutex : ", enqueue_fifo_mutex);
  x_dump_mutex("          String mutex : ", string_mutex);
  x_dump_monitor("      Reclaim listener : ", reclaim_listener_monitor);
  x_dump_monitor("        Memory monitor : ", memory_monitor);
  x_dump_monitor("            GC monitor : ", gc_monitor);
#if defined(AWT_XSIM) || defined(AWT_FDEV)
  x_dump_monitor("          AWT treelock : ", tree_lock);
#endif
  w_dump("   Instance locks :\n");
  ht_iterate(lock_hashtable, lock_iterator, NULL, NULL);
  w_dump("\n");
}


static int object_size;

static x_boolean classloaders_callback(void * mem, void * arg) {

  w_instance   instance;
  w_object     object;
  w_clazz      clazz;
  w_hashtable  ht;
  x_monitor    monitor;

  object_size += x_mem_size(mem);
  object = chunk2object(mem);
  instance = object->fields;
  clazz = object->clazz;

  if(isAssignmentCompatible(clazz, clazzClassLoader)) {
    w_dump("     %j\n", instance);
    
    ht = loader2loaded_classes(instance);
    monitor = &ht->monitor;
  
    if(monitor->owner) {
      w_dump("       loaded_hashtable (0x%08x) locked", monitor);
      if(monitor->owner->xref) {
        w_dump(" by \"%w\"", ((w_thread)(monitor->owner->xref))->name);
      }
      w_dump("\n");
    }
    
    ht = loader2unloaded_classes(instance);
    monitor = &ht->monitor;
    
    if(monitor->owner) {
      w_dump("       unloaded_hashtable (0x%08x) locked", monitor);
      if(monitor->owner->xref) {
        w_dump(" by \"%p\"", ((w_thread)(monitor->owner->xref)));
        w_dump(" by \"%w\"", ((w_thread)(monitor->owner->xref))->name);
      }
      w_dump("\n");
    }

    monitor = (x_monitor)ht_read(lock_hashtable, (w_word)instance);
    
    if(monitor && monitor->owner) {
      w_dump("       instance (0x%08x) locked", monitor);
      if(monitor->owner->xref) {
        w_dump(" by \"%p\"", ((w_thread)(monitor->owner->xref)));
        w_dump(" by \"%w\"", ((w_thread)(monitor->owner->xref))->name);
      }
      w_dump("\n");
    }
  }

  return TRUE;
}

void w_dump_monitor_info(x_monitor monitor) {
  w_int i;

  for(i = 0; i < lock_hashtable->currentsize; i++) {
    if((w_word)lock_hashtable->values[i] == (w_word)monitor) {
      w_dump("%j", lock_hashtable->keys[i]);
      break;
    }
  }
#ifdef MONITOR_INFO
  if(monitor->xref) {
    char **name;
    switch(monitor->xref_type) {
      case 0: w_dump(" at %M", ((x_frame)monitor->xref)->method); break;
      case 1: w_dump(" pthread_condition "); name = backtrace_symbols((void **)&monitor->xref, 1); w_dump("%s", name[0]); free(name); break;
      case 2: w_dump(" pthread_mutex "); name = backtrace_symbols((void **)&monitor->xref, 1); w_dump("%s", name[0]); free(name); break;
    }
  }
  else {
    w_dump(" (no location)");
  }
#endif
}

void w_dump_classloaders(void) {
  w_dump(" Classloaders :\n");
  x_mem_lock(x_eternal);
  object_size = 0;
  x_mem_scan(x_eternal, OBJECT_TAG, classloaders_callback, NULL);
  x_mem_unlock();
  w_dump("\n");
}

#ifdef SHARED_HEAP
  extern x_int mem_malloc;
  extern x_int mem_new;
#warning test1
#endif

void w_dump_meminfo(void) {
  w_dump(" Memory :\n");
  w_dump("         Total : %9d\n", x_mem_total());
  w_dump("     Available : %9d\n", x_mem_avail());
  w_dump("          Used : %9d (of which %d in objects)\n", x_mem_total() - x_mem_avail(), object_size);
#ifdef SHARED_HEAP
#warning test2
  w_dump("        Malloc : %9d\n", mem_malloc);
  w_dump("   c++ Objects : %9d\n", mem_new);
#endif
  w_dump("\n");
}

#ifdef THREAD_INFO
extern x_long start_time;
#endif

void w_dump_info(void) {
  time_t t = time(NULL);
  struct tm *now = localtime(&t);
  
  w_dump("\n");
  w_dump("Current time: %04d/%02d/%02d %02d:%02d:%02d\n", now->tm_year + 1900, now->tm_mon + 1, now->tm_mday, now->tm_hour, now->tm_min, now->tm_sec);
  w_dump("\n");
  w_dump_threads();
  w_dump_locks();
  w_dump_classloaders();
  w_dump_meminfo();
#ifdef THREAD_INFO
  start_time = x_systime_get(); 
#endif 
}
