/**************************************************************************
* Copyright (c) 2001, 2002, 2003 by Acunia N.V. All rights reserved.      *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Philips site 5, box 3       info@acunia.com                           *
*   3001 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: Throwable.c 3451 2003-02-06 14:25:34Z gray $
*/

#include <string.h>

//#include "Throwable.h"
#include "core-classes.h"
#include "exception.h"
#include "fields.h"
#include "ts-mem.h"
#include "wstrings.h"
#include "heap.h"
#include "jni.h"
#include "loading.h"
#include "realms.h"
#include "descriptor.h"
#include "methods.h"

/*
** Clean up any stack trace when a Throwable is garbage-collected.
*/
void Throwable_destructor(w_instance thisThrowable) {

  x_Exr * records = getWotsitField(thisThrowable, F_Throwable_records);
  
  if (records) {
    setWotsitField(thisThrowable, F_Throwable_records, NULL);
    woempa(5, "Destroying trace record for %j\n", thisThrowable);
    releaseMem(records);
  }

}

static const w_size bufsize = 511;

/*
** Format the first line of a stack trace (Throwable class and message),
** returning the result as a java.lang.String .
*/
w_instance Throwable_startStackTrace(JNIEnv *env, w_instance thisThrowable) {

  w_thread thread = JNIEnv2w_thread(env);
  w_instance String = NULL;
  w_string string;
  w_instance Message;
  w_string message;
  char * buffer;
  w_int i = 0;

  setWotsitField(thisThrowable, F_Throwable_frame, getWotsitField(thisThrowable, F_Throwable_records));

  Message = getReferenceField(thisThrowable, F_Throwable_detailMessage);

  buffer = allocMem_system(bufsize + 1);
  if (! buffer) {
    return NULL;
  }

  if (Message) {
    message = String2string(Message);
    i = x_snprintf(buffer, bufsize, "%k : %w", instance2clazz(thisThrowable), message);
  }
  else {
    i = x_snprintf(buffer, bufsize, "%k", instance2clazz(thisThrowable));
  }

  string = cstring2String(thread, buffer, i);
  releaseMem(buffer);
  if (string) {
    woempa(9, "TRACE %w\n", string);
    String = getStringInstance(string);
    deregisterString(string);
  }

  return String;

}

/*
** Format one line of a stack trace (method[, line number], pc),
** returning the result as a java.lang.String .
*/
w_instance formatLine(w_thread thread, w_method method, w_int pc) {

  w_int      length;
  w_int      remaining;
  char *     buffer;
  char *     bufptr;
  w_string   string;
  w_instance String = NULL;
  w_clazz    clazz = method->spec.declaring_clazz;
  w_instance loader = clazz->loader;

  buffer = allocMem_system(bufsize + 1);
  bufptr = buffer;
  remaining = bufsize;

  if (buffer) {
    length = x_snprintf(buffer, remaining, "%s method %m of class ", isSet(method->flags, ACC_STATIC) ? "static" : "instance", method);
    remaining -= length;
    bufptr += length;

    if (loader && loader != systemClassLoader) {
      w_instance loader_name = getReferenceField(loader, F_ClassLoader_ownname);

      if (loader_name) {
        length = x_snprintf(bufptr, remaining, "{%w}", String2string(getReferenceField(loader, F_ClassLoader_ownname)));
      }
      else {
        length = x_snprintf(bufptr, remaining, "{%k}", loader);
      }
      remaining -= length;
      bufptr += length;
    }

    length = x_snprintf(bufptr, remaining, "%k ", clazz);
    remaining -= length;
    bufptr += length;

    if (isSet(method->flags, METHOD_IS_COMPILED)) {
      length = x_snprintf(bufptr, remaining, "(jitted method)");
    }
    else if (isSet(method->flags, ACC_NATIVE)) {
      length = x_snprintf(bufptr, remaining, "(native method)");
    }
    else {
#ifdef STORE_METHOD_DEBUG_INFO
      length = x_snprintf(bufptr, remaining, "line %d (pc %d)", code2line(method, method->exec.code+pc), pc);
#else
      length = x_snprintf(bufptr, remaining, "pc = %d", pc);
#endif
    }
    remaining -= length;
    bufptr += length;

    string = cstring2String(thread, buffer, bufptr - buffer);
    if (string) {
      String = newStringInstance(string);
      deregisterString(string);
    }
    releaseMem(buffer);
  }

  return String;
  
}

/* 
** Get the next line of a stack trace (returning it as a java.lang.String),
** and advance the `frame' field of the Throwable to the next record.
*/
w_instance Throwable_nextStackTrace(JNIEnv *env, w_instance thisThrowable) {

  w_instance String = NULL;
  x_exr current = getWotsitField(thisThrowable, F_Throwable_frame);

  if (current) {
    String = formatLine(JNIEnv2w_thread(env), current->method, current->pc);
    if (current->position > 0) {
      setWotsitField(thisThrowable, F_Throwable_frame, current + 1);
    }
    else {
      setWotsitField(thisThrowable, F_Throwable_frame, NULL);
    }
  }

  return String;

}

/*
** Strip off the top `n' levels of a stack trace (e.g., in order to remove
** the frame created by fillInStackTrace() itself ...)
*/
static void stripStackTraceLevels(w_instance thisThrowable, w_int n) {
  x_Exr * current = getWotsitField(thisThrowable, F_Throwable_records);

  while (1) {
    current->method = (current+n)->method;
    current->pc = (current+n)->pc;
    current->position = (current+n)->position;
    if (current->position == 0) {
      break;
    }
    ++current;
  }
}

/*
** Delete the current contents of a Throwable and replace it by a stack trace
** reflecting the current stack (but excluding the call to fillInStackTrace()).
*/
w_instance Throwable_fillInStackTrace(JNIEnv *env, w_instance thisThrowable) {
  x_Exr * records = getWotsitField(thisThrowable, F_Throwable_records);
  
  if (records) {
    setWotsitField(thisThrowable, F_Throwable_records, NULL);
    woempa(1, "Destroying existing trace record for %j\n", thisThrowable);
    releaseMem(records);
  }
  fillThrowable(JNIEnv2w_thread(env), thisThrowable);
  stripStackTraceLevels(thisThrowable, 1);

  return thisThrowable;
}

