/**************************************************************************
* Copyright (c) 2001, 2002, 2003 by Acunia N.V. All rights reserved.      *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Philips site 5, box 3       info@acunia.com                           *
*   3001 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/*
** $Id: reflection.c 4140 2004-04-05 08:18:48Z dumon $
**
** Utility functions for reflection related files.
**
** The implementations for the Java conversions and promotions are to be found
** in this file. See any good JVM book or JLS book for the different promotions 
** and conversions that can take place.
*/

#include "arrays.h"
#include "checks.h"
#include "clazz.h"
#include "core-classes.h"
#include "descriptor.h"
#include "exception.h"
#include "fields.h"
#include "ieee754.h"
#include "interpreter.h"
#include "jni.h"
#include "loading.h"
#include "methods.h"

/*
** For assigning fields and variables to each other. The reflection classes need to implement
** what is called widening for some operations, or when no type conversion is required, implement
** a so called 'identity conversion' or assignment. Checking which conversions are allowed and
** what operations need to be performed by the virtual machine in order to change from one type
** to another, is done through a matrix or lookup table of conversion function pointers. A NULL
** pointer means that the conversion <from> to <to> type is not allowed, a non NULL pointer means
** that the conversion is allowed and the function pointed to will perform the assignment.
** 
*/

/*
** Two 'identity conversion assignment' functions, for 32 bit data
** and for 64 bit data. Pretty straightforward...
*/

static void i_32(void *S, void *T) {

  *((w_word*)T) = *((w_word*)S);

}

static void i_64(void *S, void *T) {

  ((w_word*)T)[WORD_LSW] = ((w_word*)S)[WORD_LSW];
  ((w_word*)T)[WORD_MSW] = ((w_word*)S)[WORD_MSW];

}

/*
** The different conversion routine <from>__<to>. Note that after testing,
** some of these can be collapsed together for compacter code.
*/

static void C__I(void *S, void *T) {

  /*
  ** No sign extension takes place...
  */

  *((w_word*)T) = (*((w_word*)S) & 0x0000ffff);

}

static void C__J(void *S, void *T) {

  /*
  ** No sign extension takes place...
  */
  
  ((w_word*)T)[WORD_LSW] = (*((w_word*)S) & 0x0000ffff);
  ((w_word*)T)[WORD_MSW] = 0;

}

static void C__F(w_word S[], w_word T[]) {
  w_float * f = (w_float*)T;

  f[0] = int32_to_float32((w_int) S[0]);
}

static void C__D(w_word S[], w_word T[]) {

  w_u64 d;
  
  d.d64 = int32_to_float64((w_int) S[0]);

  ((w_word*)T)[WORD_LSW] = d.words[WORD_LSW];
  ((w_word*)T)[WORD_MSW] = d.words[WORD_MSW];

}

static void B__S(void *S, void *T) {

  *(w_int*)T = *(w_sbyte*)S;

}

static void B__I(void *S, void *T) {

  *(w_int*)T = *(w_sbyte*)S;

}

static void B__J(void *S, void *T) {

  w_u64 l;
  
  l.s64 = *((w_sbyte*)S);

  ((w_word*)T)[WORD_LSW] = l.words[WORD_LSW];
  ((w_word*)T)[WORD_MSW] = l.words[WORD_MSW];

}

static void B__F(w_word S[], w_word T[]) {
  w_float * f = (w_float*)T;
  f[0] = int32_to_float32((w_int) S[0]);
}

static void B__D(w_word S[], w_word T[]) {

  w_u64 d;
  
  d.d64 = int32_to_float64((w_int) S[0]);

  ((w_word*)T)[WORD_LSW] = d.words[WORD_LSW];
  ((w_word*)T)[WORD_MSW] = d.words[WORD_MSW];

}

static void S__I(void *S, void *T) {

  *((w_word*)T) = (*((w_word*)S) & 0x8000) ? (*((w_word*)S) | 0xffff0000) : (*((w_word*)S) & 0x00007fff);

}

static void S__J(void *S, void *T) {

  w_u64 l;
  
  l.s64 = *((w_int*)S);

  ((w_word*)T)[WORD_LSW] = l.words[WORD_LSW];
  ((w_word*)T)[WORD_MSW] = l.words[WORD_MSW];

}

static void S__F(w_word S[], w_word T[]) {
  w_float* f = (w_float*)T;
  f[0] = int32_to_float32((w_int) S[0]);
}

static void S__D(w_word S[], w_word T[]) {

  w_u64 d;
  
  d.d64 = int32_to_float64((w_int) S[0]);

  ((w_word*)T)[WORD_LSW] = d.words[WORD_LSW];
  ((w_word*)T)[WORD_MSW] = d.words[WORD_MSW];

}

static void I__J(void *S, void *T) {

  w_u64 l;
  
  l.s64 = (w_int)*((w_word*)S);

  ((w_word*)T)[WORD_LSW] = l.words[WORD_LSW];
  ((w_word*)T)[WORD_MSW] = l.words[WORD_MSW];

}

static void I__F(w_word S[], w_word T[]) {

  w_float* f = (w_float*)T;

  *f = int32_to_float32((w_int) S[0]);
  
}

static void I__D(w_word S[], w_word T[]) {

  w_u64 d;
  
  d.d64 = int32_to_float64((w_int) S[0]);

  ((w_word*)T)[WORD_LSW] = d.words[WORD_LSW];
  ((w_word*)T)[WORD_MSW] = d.words[WORD_MSW];

}

static void J__F(w_word S[], w_word T[]) {

  w_long l = WORDS2LONG(S[1], S[0]);
  w_float* f = (w_float*)T;

  f[0] = int64_to_float32(l);
}

static void J__D(void *S, void *T) {

  w_u64 j;

  j.words[WORD_LSW] = ((w_word*)S)[WORD_LSW];
  j.words[WORD_MSW] = ((w_word*)S)[WORD_MSW];

  j.d64 = int64_to_float64(j.s64);

  ((w_word*)T)[WORD_LSW] = j.words[WORD_LSW];
  ((w_word*)T)[WORD_MSW] = j.words[WORD_MSW];

}

static void F__D(w_word S[], w_word T[]) {
  w_float* p = (w_float*)S;
  w_u64 d;
  
  d.d64 = float32_to_float64(p[0]);

  ((w_word*)T)[WORD_LSW] = d.words[WORD_LSW];
  ((w_word*)T)[WORD_MSW] = d.words[WORD_MSW];

}

typedef void (*w_widener)(void *S, void *T);


static w_widener widener_matrix[8][8] = {
  /*   +--->  to short int   long  byte  char  float doubl bool */
  /*   |                                                        */
  /* from                                                       */
  /* short  */ { i_32, S__I, S__J, NULL, NULL, S__F, S__D, NULL },
  /* int    */ { NULL, i_32, I__J, NULL, NULL, I__F, I__D, NULL },
  /* long   */ { NULL, NULL, i_64, NULL, NULL, J__F, J__D, NULL },
  /* byte   */ { B__S, B__I, B__J, i_32, NULL, B__F, B__D, NULL },
  /* char   */ { NULL, C__I, C__J, NULL, i_32, C__F, C__D, NULL },
  /* float  */ { NULL, NULL, NULL, NULL, NULL, i_32, F__D, NULL },
  /* double */ { NULL, NULL, NULL, NULL, NULL, NULL, i_64, NULL },
  /* bool   */ { NULL, NULL, NULL, NULL, NULL, NULL, NULL, i_32 },
};

/*
** Possibly convert and assign one primitive value (from) to another (to). If the widening
** conversion is not permitted, this function returns WONKA_FALSE, otherwise WONKA_TRUE indicates success.
*/

w_boolean widen(w_clazz from_clazz, void *F_data, w_clazz to_clazz, void *T_data) {

  w_widener widener;
  w_int f;
  w_int t;

  if (clazzIsPrimitive(from_clazz) && clazzIsPrimitive(to_clazz)) {
    f = from_clazz->type & 0x0f;
    t = to_clazz->type & 0x0f;

    /*
    ** In any case, zero out the memory of the 'T_data' array.
    */

#ifdef PACK_BYTE_FIELDS
    if ((to_clazz == clazz_boolean) || (to_clazz == clazz_byte)) {
      ((char*)T_data)[0] = 0;
    }
    else {
      ((w_word*)T_data)[0] = 0;
      if (isSet(to_clazz->type, VM_TYPE_TWO_CELL)) {
        ((w_word*)T_data)[1] = 0;
      }
    }
#else
    ((w_word*)T_data)[0] = 0;
    if (isSet(to_clazz->type, VM_TYPE_TWO_CELL)) {
      ((w_word*)T_data)[1] = 0;
    }
#endif

    // TODO, the following if should become a runtime check
    if ((f <= 8) && (t <= 8)) {
      woempa(1, "Trying to widen and assign type '%k' to type '%k'...\n", from_clazz, to_clazz);
      widener = widener_matrix[f - 1][t - 1];
      if (widener) {
        (*widener)(F_data, T_data);
        return WONKA_TRUE;
      }
    }
  }

  /*
  ** An error occured. Since the 'f' and 't' indexes could be bogus, we just print the indexes and not the
  ** primitive2name element since we could segfault...
  */
  
  woempa(1, "... impossible to perform widening.\n");
  
  return WONKA_FALSE;
  
}

w_clazz getWrappedValue(w_instance value, void **data) {
  w_clazz value_clazz = instance2clazz(value);

  woempa(1, "Trying to unwrap %j.\n", value);
  if (value_clazz == clazzShort) {
    *(w_word**)data = wordFieldPointer(value, F_Short_value);

    return clazz_short;

  }
  if (value_clazz == clazzInteger) {
    *(w_word**)data = wordFieldPointer(value, F_Integer_value);

    return clazz_int;

  }
  if (value_clazz == clazzLong) {
    *(w_word**)data = wordFieldPointer(value, F_Long_value);

    return clazz_long;

  }
  if (value_clazz == clazzByte) {
#ifdef PACK_BYTE_FIELDS
    *(char**)data = byteFieldPointer(value, F_Byte_value);
#else
    *(w_word**)data = byteFieldPointer(value, F_Byte_value);
#endif

    return clazz_byte;

  }
  if (value_clazz == clazzCharacter) {
    *(w_word**)data = wordFieldPointer(value, F_Character_value);

    return clazz_char;

  }
  if (value_clazz == clazzFloat) {
    *(w_word**)data = wordFieldPointer(value, F_Float_value);

    return clazz_float;

  }
  if (value_clazz == clazzDouble) {
    *(w_word**)data = wordFieldPointer(value, F_Double_value);

    return clazz_double;

  }
  if (value_clazz == clazzBoolean) {
#ifdef PACK_BYTE_FIELDS
    *(char**)data = byteFieldPointer(value, F_Boolean_value);
#else
    *(w_word**)data = byteFieldPointer(value, F_Boolean_value);
#endif

    return clazz_boolean;

  }

  return NULL;
}

/*
** Create a wrapper instance for a certain primitive type, returns NULL when the type is
** not a primitive type.
*/

w_instance createWrapperInstance(w_thread thread, w_clazz clazz, w_int *slot) {
  w_clazz wrapper_clazz;

  switch (clazz->type & 0x0f) {
    case VM_TYPE_SHORT:
      wrapper_clazz = clazzShort;
      *slot = F_Short_value;
      break;

    case VM_TYPE_INT:
      wrapper_clazz = clazzInteger;
      *slot = F_Integer_value;
      break;

    case VM_TYPE_LONG:
      wrapper_clazz = clazzLong;
      *slot = F_Long_value;
      break;

    case VM_TYPE_BYTE:
      wrapper_clazz = clazzByte;
      *slot = F_Byte_value;
      break;

    case VM_TYPE_CHAR:
      wrapper_clazz = clazzCharacter;
      *slot = F_Character_value;
      break;

    case VM_TYPE_FLOAT:
      wrapper_clazz = clazzFloat;
      *slot = F_Float_value;
      break;

    case VM_TYPE_DOUBLE:
      wrapper_clazz = clazzDouble;
      *slot = F_Double_value;
      break;

    case VM_TYPE_BOOLEAN:
      wrapper_clazz = clazzBoolean;
      *slot = F_Boolean_value;
      break;

    default:
      wrapper_clazz = NULL;
  }
  
  if (clazz) {
    return allocInstance(thread, wrapper_clazz);
  }

  wabort(ABORT_WONKA, "Wrong primitive VM_TYPE %d\n", clazz->type);

  return NULL;
  
}

x_frame invoke(JNIEnv *env, w_method method, w_instance This, w_instance Arguments) {

  w_thread thread = JNIEnv2w_thread(env);
  w_instance *arguments;
  w_clazz F_clazz;
  w_word *F_data;
  w_clazz T_clazz;
  w_word  T_data[2];
  w_int numArgs;
  w_int  spec_nargs;
  w_int i;
  x_frame frame;
  w_instance itex;
  w_instance exception;

  woempa(1, "(REFLECTION) Invoking %M, %d argument words.\n", method, method->exec.arg_i);

  frame = pushFrame(thread, method);

  numArgs = instance2Array_length(Arguments);
  spec_nargs = 0;
  if (method->spec.arg_types) {
    while (method->spec.arg_types[spec_nargs]) {
      woempa(1, "Method %w arg[%d] is %w\n", method->spec.name, spec_nargs, method->spec.arg_types[spec_nargs]->dotified);
      ++spec_nargs;
    }
  }

  if (numArgs != spec_nargs) {
    woempa(7, "Array contains %d args, method %w has %d args\n", numArgs, method->spec.name, spec_nargs);
    throwException(thread, clazzIllegalArgumentException, NULL);
  }
  else {
    classMustBeInitialized(method->spec.declaring_clazz);
    if (! isSet(method->flags, ACC_STATIC)) {
      frame->top[0].c.w = (w_word) This;
      frame->top[0].s = stack_trace;
      frame->top += 1;
    }
    arguments = instance2Array_instance(Arguments);
    for (i = 0; i < numArgs; i++) {
      if (mustBeLoaded(&method->spec.arg_types[i]) == CLASS_LOADING_FAILED) {
        break;
      }

      T_clazz = method->spec.arg_types[i];
      if (isSet(T_clazz->flags, CLAZZ_IS_PRIMITIVE)) {
        F_clazz = getWrappedValue(arguments[i], (void**)&F_data);

        if (F_clazz == NULL || !widen(F_clazz, F_data, T_clazz, T_data)) {
          if (F_clazz == NULL) {
            woempa(7, "Argument %d is not a wrapper (%k)!\n", i, instance2clazz(arguments[i]));
          }
          else {
            woempa(7, "Widening is impossible!\n");
          }
          throwException(thread, clazzIllegalArgumentException, NULL);
          break;
        }

        if (isSet(T_clazz->type, VM_TYPE_TWO_CELL)) {
          frame->top[0].c.w = T_data[0];
          frame->top[1].c.w = T_data[1];
          frame->top += 2;
        }
        else {
          frame->top[0].c.w = T_data[0];
          frame->top += 1;
        }
      }
      else {
        woempa(1, "Argument %d: target class is %k, argument is %j\n", i, T_clazz, arguments[i]);
        if (arguments[i] == NULL || isAssignmentCompatible(instance2clazz(arguments[i]), T_clazz)) {
          frame->top[0].c.w = (w_word) arguments[i];
          frame->top[0].s = stack_trace;
          frame->top += 1;
        }
        else {
          woempa(7, "%j is not assignment compatible with %k.\n", arguments[i], T_clazz);
          throwException(thread, clazzIllegalArgumentException, NULL);
          break;
        }
      }
    }

    /*
    ** If no exception is pending, all the arguments were OK and they are in the frame, let's run it now...
    */

    if (! exceptionThrown(thread)) {
      callMethod(frame, method);
      if (exceptionThrown(thread)) {
        // First clear the exception or allocating an instance won't work!
        exception = thread->exception;
        addLocalReference(thread, exception);
        clearException(thread);
        itex = allocInstance(thread, clazzInvocationTargetException);
        woempa(9, "(REFLECTION) Invocation threw an exception, wrapping it into an InvocationTargetException %p\n", itex);
        setReferenceField(itex, exception, F_InvocationTargetException_target);
        throwExceptionInstance(thread, itex);
        removeLocalReference(thread, exception);
      }
    }
  }

  return frame;

}

void voidProxyMethodCode(JNIEnv *env, w_instance thisProxy, ...) {
  w_thread thread = JNIEnv2w_thread(env);
  x_frame  current_frame = thread->top;
  x_frame  new_frame;
  w_instance handler;
  w_clazz  target_clazz;
  w_method target_method;
  w_method current_method;
  w_int    numArgs;
  w_int    i;
  w_int    length;
  w_instance arguments;

  classMustBeInitialized(clazzProxy);
  current_method = current_frame->method;
  woempa(7, "Method : %M\n", current_method);
  woempa(7, "thisProxy : %J\n", thisProxy);
  woempa(7, "F_Proxy_h : %p\n", F_Proxy_h);
  handler = getReferenceField(thisProxy, F_Proxy_h);
  woempa(7, "handler : %p\n", handler);
  target_clazz = instance2clazz(handler);
  woempa(7, "Target class : %K\n", target_clazz);
  target_method = (*env)->GetMethodID(env, clazz2Class(target_clazz), "invoke", "(Ljava/lang/Object;Ljava/lang/reflect/Method;[Ljava/lang/Object;)Ljava/lang/Object;");
  woempa(7, "Target method : %M\n", target_method);
  woempa(7, "Current method : %M\n", current_method);

  length = 0;
  arguments = NULL;
  if (current_method->spec.arg_types) {
    while (current_method->spec.arg_types[length]) {
      woempa(7, "Method %w arg[%d] is %k\n", current_method->spec.name, length, current_method->spec.arg_types[length]);
      ++length;
    }
    arguments = allocArrayInstance(thread, clazzArrayOf_Object, 1, &length);
    if (!arguments) {

      return;

    }
    for (i = 0; i < length; ++i) {
      woempa(7, "Method %w arg[%d] is %s %k\n", current_method->spec.name, i, clazzIsPrimitive(current_method->spec.arg_types[i]) ? "primitive" : "reference", current_method->spec.arg_types[i]);
    }
  }

// TODO: copy & wrap arguments from stack

  new_frame = pushFrame(thread, target_method);

  numArgs = 3;

  new_frame->top[0].c.w = thisProxy[F_Proxy_h];
  new_frame->top[0].s = stack_trace;
  new_frame->top += 1;
  new_frame->top[0].c.w = (w_word)thisProxy;
  new_frame->top[0].s = stack_trace;
  new_frame->top += 1;
  new_frame->top[0].c.w = (w_word)current_method;
  new_frame->top[0].s = stack_trace;
  new_frame->top += 1;
  new_frame->top[0].c.w = (w_word)arguments;
  new_frame->top[0].s = stack_trace;
  new_frame->top += 1;
//  callMethod(new_frame, method_InvocationHandler_invoke);
  if (exceptionThrown(thread)) {
/* TODO: adapt this stuff
    // First clear the exception or allocating an instance won't work!
    exception = thread->exception;
    addLocalReference(thread, exception);
    clearException(thread);
    itex = allocInstance(thread, clazzInvocationTargetException);
    woempa(9, "(REFLECTION) Invocation threw an exception, wrapping it into an InvocationTargetException %p\n", itex);
    setReferenceField(itex, exception, F_InvocationTargetException_target);
    throwExceptionInstance(thread, itex);
    removeLocalReference(thread, exception);
*/
  }
// TODO: pop frame!
}

w_word singleProxyMethodCode(JNIEnv *env, w_instance thisProxy, ...) {
  w_thread thread = JNIEnv2w_thread(env);
  w_method method = thread->top->method;
  x_frame  frame;
  woempa(7, "%J : %M\n", thisProxy, method);

  return 0;
}

w_long doubleProxyMethodCode(JNIEnv *env, w_instance thisProxy, ...) {
  w_thread thread = JNIEnv2w_thread(env);
  w_method method = thread->top->method;
  x_frame  frame;
  woempa(7, "%J : %M\n", thisProxy, method);

  return 0LL;
}


