/**************************************************************************
* Copyright (c) 2001, 2003, 2003 by Acunia N.V. All rights reserved.      *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Philips site 5, box 3       info@acunia.com                           *
*   3001 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: repository.c 3451 2003-02-06 14:25:34Z gray $
*/

#include <string.h>

#include "wonka.h"
#include "zip.h"
#include "repository.h"
#include "ts-mem.h"
#include "list.h"
#include "wstrings.h"
#include "threads.h"
#include "hashtable.h"
#include "core-classes.h"
#include "realms.h"
#include "clazz.h"
#include "arrays.h"
#include "loading.h"

/*
** The repository types as character strings.
*/

const char *RepositoryType2char[] = {
  "Unknown",
  "Directory",
  "Unknown1",
  "Unknown2",
  "Unknown3",
  "Raw data",
  "Manifest file",
  "Signature file",
};

/*
** The master repository, the sentinel for the linked list of repositories.
*/

w_repository W_Repository_system;

/*
** Remove a single entry from a repository.
*/

void removeRepositoryEntry(w_repository rep, w_rentry entry) {

  woempa(9, "Removing '%w' from repository 0x%08x.\n", entry->name, (w_word)rep);
  
  ht_erase(rep->hashtable, (w_word)entry->name);
  deregisterString(entry->name);
  if (entry->u_data) {
    releaseMem(entry->u_data);
  }
  if (entry->c_data) {
    releaseMem(entry->c_data);
  }
  releaseMem(entry);

}

static void loadSingleZip(w_repository rep, w_byte *data, w_size size) {

  z_zipFile zipFile;
  z_zipEntry ze;
  w_int type;
  w_rentry re;
  char *buffer;
  w_size namelength;
  w_thread thread = currentWonkaThread;
  
  woempa(1, "Processing zip file, data at %p.\n", data);
    
  zipFile = parseZipFile(thread, data);
  ze = zipFile->entries;
  do {
    type = zipEntry2RepositoryType(ze);
    if (type != RT_DIRECTORY) {
      if (uncompressZipEntry(ze)) {
        re = allocClearedMem(thread, sizeof(w_Rentry));
        namelength = strlen(ze->name);
        buffer = allocClearedMem(thread, namelength + 1);
        w_memcpy(buffer, ze->name, namelength);
        re->name = cstring2String(thread, buffer, namelength);
        woempa(1, "ZE->NAME = '%s', entry name is '%w'\n", ze->name, re->name);
        woempa(1, "Decompressed %s size %d bytes.\n", buffer, ze->u_size);
        releaseMem(buffer);
        re->u_data = ze->u_data;
        ze->u_data = NULL;
        re->u_size = ze->u_size;
        re->u_sum = ze->d_crc;
        re->flags = setRepositoryType(re->flags, zipEntry2RepositoryType(ze));
        ht_write(rep->hashtable, (w_word)re->name, (w_word)re);
	woempa(1, "Name of unzipped file is '%w'.\n", re->name);
      }
      else {
        if (ze->u_size != 0) {
          woempa(9, "Entry '%s' is in error flags = 0x%08x.\n", ze->name, ze->flags);
        }
      }
    }
    ze = ze->next;
  } while (ze != zipFile->entries);

  releaseZipFile(zipFile);

}

/*
static void dumpOneEntry(w_word key, w_word value) {
  w_string name = (w_string) key;
  w_rentry re = (w_rentry) value;
  woempa(9,"Entry %w -> %p\n", name, re);
}

static void dumpEntries(w_repository rep) {
  woempa(7,"Dumping entries of %s\n", rep->hashtable->label);
  ht_every(rep->hashtable, dumpOneEntry);
}
*/

void findReprocessEntries(w_word name_word, w_word entry_word, void *arg1, void *arg2) {

  w_int *rpindex = arg2;
  w_rentry *reprocess = arg1;
  w_rentry entry = (w_rentry)entry_word;
  w_int type = getRepositoryType(entry->flags);
  
  woempa(1, "Checking '%w' (reprocess index = %d)...\n", NM(entry), *rpindex);
  
  if (type == RT_JAR || type == RT_ZIP) {
    reprocess[*rpindex] = entry;
    *rpindex += 1;
    woempa(1, "--> needs reprocessing.\n");
  }

}

/*
** Load a repository with a given ZIP data stream. This does not decompress the entries
** in the JAR file. It merely extracts the information out of the stream and puts it
** into the repository. 
**
** When during the unpacking, the repository loader encounters an entry that ends with
** ".zip", it will recursively unpack that zip file.
**
** The following actions are performed:
**
** 2) A manifest file will be processed and the corresponding information will be
**    put in the repository.
** 3) Any signature information will be processed and attached to the appropriate class
**    type entries from step 1.
** 4) Other types (AU, ELF, RAW, ...) but NOT the directory types, will be put in the
**    repository.
** 5) An appropriate hash table for quickly locating entries will be build.
*/

static void loadRepository(w_repository rep, w_byte *data, w_size size) {

  w_rentry *reprocess;
  w_int rpindex;
  w_int i;
  w_thread thread = currentWonkaThread;

  zipInit();

  /*
  ** Load the repository with the ZIP or JAR that was given to us...
  */
  
  loadSingleZip(rep, data, size);

  /*
  ** ... now check if there were other zip or jar files and if yes, keep on processing
  ** these until all loaded...
  */
  
  do {
    rpindex = 0;
    reprocess = allocMem(thread, rep->hashtable->occupancy * sizeof(w_rentry));
    ht_iterate_no_lock(rep->hashtable, findReprocessEntries, reprocess, &rpindex);
    woempa(9, "rpindex = %d\n", rpindex);
    if (rpindex) {
      for (i = 0; i < rpindex; i++) {
        /*
        ** loadSingleZip() the entry and afterwards remove it from the hashtable...
        */
        woempa(9, "Reprocessing entry '%w'.\n", NM(reprocess[i]));
        loadSingleZip(rep, reprocess[i]->u_data, reprocess[i]->u_size);
        removeRepositoryEntry(rep, reprocess[i]);
      }
    }
    releaseMem(reprocess);
  } while (rpindex != 0);

/*
  dumpEntries(rep);
*/
  zipEnd();

}

/*
** Create a repository from a stream and load all entries found in the 
** data stream into the repository.
*/

w_repository createRepository(w_thread thread, w_byte *data, w_size size) {

  w_repository rep = allocMem(thread, sizeof(w_Repository));

  rep->hashtable = ht_create(thread2realm(thread), (char*)"hashtable:repository", 51, NULL, NULL, 0, 0);   
  loadRepository(rep, data, size);
  
  return rep;
}

/*
** Return resources used by one entry.
*/
void deleteRepositoryEntry(w_repository rep, w_string name) {
  w_rentry entry;

  ht_lock(rep->hashtable);
  woempa(1, "Destroying entry '%w' of %s\n", name, rep->hashtable->label);

  entry = (w_rentry)ht_read_no_lock(rep->hashtable, (w_word)name);
  if (entry) {
    deregisterString(entry->name);
    if (entry->u_data) {
      releaseMem(entry->u_data);
    }
    if (entry->c_data) {
      releaseMem(entry->c_data);
    }
    ht_erase_no_lock(rep->hashtable, (w_word)name);
  }
  else {
    woempa(9, "Asked to destry non-existent entry '%w' in %s!\n", name, rep->hashtable->label);
  }
  ht_unlock(rep->hashtable);
}

static void destroyEntry(w_word name_word, w_word entry_word, void *arg1, void *arg2) {
  w_rentry entry = (w_rentry)entry_word;
  woempa(1, "Destroying entry '%w' at %p\n", (w_string)name_word, entry);
  deregisterString(entry->name);
  if (entry->u_data) {
    releaseMem(entry->u_data);
  }
  if (entry->c_data) {
    releaseMem(entry->c_data);
  }
  releaseMem(entry);
}

/*
** Destroy a repository, releasing all memory used by its entries.
*/
void destroyRepository(w_repository rep) {
  ht_iterate_no_lock(rep->hashtable, destroyEntry, NULL, NULL);
  ht_destroy(rep->hashtable);
}

