/**************************************************************************
* Copyright (c) 2001, 2002, 2003 by Acunia N.V. All rights reserved.      *
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Philips-site 5, bus 3       info@acunia.com                           *
*   3001 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/


/*
** $Id: strings.c 4286 2005-11-12 21:43:08Z chris $
*/

#include <string.h>
#include <stdio.h>

#include "clazz.h"
#include "core-classes.h"
#include "fifo.h"
#include "hashtable.h"
#include "heap.h"
#include "ieee754.h"
#include "ts-mem.h"
#include "realms.h"
#include "wstrings.h"
#include "threads.h"
#include "descriptor.h"

#define STRING_HASHTABLE_SIZE                         719

w_hashtable string_hashtable;

x_mutex string_mutex;

/// The string `<NULL>'
w_string string_NULL;

/// The string `<EMPTY>'
w_string string_EMPTY;

/// The string `<BOGUS>'
w_string string_BOGUS;

static w_word hashString(w_string string);

static inline w_string allocString(w_size length, w_boolean is_latin1) {

  w_string string;
  
#ifdef BYTEWISELATIN
  string = allocMem_system(offsetof(w_String, contents) + length * (is_latin1 ? sizeof(w_byte) : sizeof(w_char)));
  string->refcount = 0;
  string->instance = NULL;
  string->length_and_flags = length + (is_latin1 ? STRING_IS_LATIN1 : 0);
#else
  string = allocMem_system(offsetof(w_String, chars) + length*sizeof(w_char));
  string->refcount = 0;
  string->instance = NULL;
  string->length = length;
#endif

  return string;

}

static inline void deallocString(w_string s) {
  woempa(1, "Releasing string `%w' (%p)\n", s, s);
  releaseMem(s);
}

/*
** Unpack a w_string into an array of w_char's.  Caller must supply a
** suitably sized buffer (i.e. space for at least source->numChars). 
** Returns the number of w_char's which were found.
*/

w_size w_string2chars(w_string source, w_char *destination) {
  w_size length = string_length(source);

#ifdef BYTEWISELATIN
  if (string_is_latin1(source)) {
    w_size i;
    w_char *dst = destination;
    w_ubyte *src = source->contents.bytes;;
    for (i = 0; i < length; ++i) {
      *dst++ = *src++;
    }
  }
  else {
    w_memcpy(destination, source->contents.chars, sizeof(w_char) * length);
  }
#else
  w_memcpy(destination, source->chars, sizeof(w_char) * length);
#endif

  return length;

}

/*
** Encode a unicode character into UTF-8; the resulting 1 to three bytes
** are written to **destination, and *destination is updated accordingly.
**
** Note that it it is the caller's responsiblility to ensure that sufficient
** space is available after **destination.
*/

void putw_char(w_char ch, w_ubyte **destination) {

  if (ch == 0) {
    *(*destination)++ = 0xC0;
    *(*destination)++ = 0x80;
  }
  else if( ch < 0x80) {
    *(*destination)++ = ch;
  }
  else if(ch < 0x800) {
    *(*destination)++ = 0xC0+(ch >> 6);
    *(*destination)++ = 0x80+(ch & 63);
  }
  else {
    *(*destination)++ = 0xE0 + (ch >> 12);
    *(*destination)++ = 0x80 + ((ch >> 6) & 63);
    *(*destination)++ = 0x80 + (ch & 63);
  }

}

w_int compareStrings(w_string s1, w_string s2) {

  w_size i;
  w_int result = 0; 

  if (string_length(s1) == 0) {
    result -= string_length(s2);
  }  
  else if (string_length(s2) == 0) {
    result = string_length(s1);
  }
  else if (s1 != s2) {
    for (i = 0; i < string_length(s1); i++) {
      if (string_char(s1, i) != string_char(s2, i)) {
        result = string_char(s1, i) - string_char(s2, i);
        break;
      }
      if ((i + 1) >= string_length(s1) || (i + 1) >= string_length(s2)) {
        result = string_length(s1) - string_length(s2);
        break;
      }
    }
  }

  return result;

}


/*
** Convert a Wonka string into an array of UTF8 w_byte's. The caller
** is responsible for releasing the memory afterwards.
** After this function returns, 'count' will be set to the number of 
** bytes that where allocated in the buffer returned.
**
*/

inline static int char_is_singlet(w_word c) {

  if ((w_char)c == 0) {
    return 0;
  }
  else if ((w_char)c & 0xff80) {
    return 0;
  }
  else {
    return 1;
  }

}

inline static int char_is_duplet(w_word c) {

  if ((w_char)c & 0xf800) {
    return 0;
  }
  else {
    return 1;
  }

}

inline static void char2duplet(w_byte *buffer, w_word thechar) {

  w_char c = (w_char)thechar;
  
  buffer[0] = (w_byte)(c>>6) | 0xC0;
  buffer[1] = (w_byte)(c&63) | 0x80;
  
}

inline static void char2triplet(w_byte *buffer, w_word thechar) {

  w_char c = (w_char)thechar;
  
  buffer[0] = (w_byte)(c>>12) | 0xE0;
  buffer[1] = (w_byte)((c>>6)&63) | 0x80;
  buffer[2] = (w_byte)(c&63) | 0x80;
}

w_byte *string2UTF8(w_string string, w_int *count) {

  w_byte *buffer;
  w_byte *handle;
  w_byte *result;
  w_size i;
  w_size l;

  l = 0;

  /*
  ** Allocate a worst case buffer, 3 bytes for each unicode char and
  ** 2 bytes for the length and one for the terminating null byte.
  */

  handle = allocMem_system(3 + (string_length(string) * (3 * sizeof (w_byte))));
  if (!handle) {

    return NULL;

  }

  buffer = handle + 2;

  for (i = 0; i < string_length(string); i++) {
    w_char ch = string_char(string, i);
    if (char_is_singlet(ch)) {
      *buffer = (w_byte)ch;
      l += 1;
      buffer += 1;
    }
    else if (char_is_duplet(ch)) {
      char2duplet(buffer, ch);
      l += 2;
      buffer += 2;
    }
    else {
      char2triplet(buffer, ch);
      l += 3;
      buffer += 3;
    }
  }
  handle[0] = (w_byte)(l >> 8);
  handle[1] = (w_byte)(l);

  result = reallocMem(handle, l + 3);
  result[l + 2] = 0;
  *count = l + 2;

  return result;

}


w_byte* chars2UTF8(w_char *chars, w_int length, w_int *count) {

  w_byte *buffer;
  w_byte *handle;
  w_byte *result;
  w_int  i;
  w_size l;

  l = 0;

  /*
  ** Allocate a worst case buffer, 3 bytes for each unicode char and
  ** 2 bytes for the length and one for the terminating null byte.
  */

  handle = allocMem_system(3 + (length * 3 * sizeof (w_byte)));
  if (!handle) {

    return NULL;

  }

  buffer = handle + 2;

  result = handle;

  for (i = 0; i < length ; i++) {
    w_char ch = chars[i];
    if (char_is_singlet(ch)) {
      *buffer = (w_byte)ch;
      l += 1;
      buffer += 1;
    }
    else if (char_is_duplet(ch)) {
      char2duplet(buffer, ch);
      l += 2;
      buffer += 2;
    }
    else {
      char2triplet(buffer, ch);
      l += 3;
      buffer += 3;
    }
  }
  handle[0] = (w_byte)(l >> 8);
  handle[1] = (w_byte)(l);

  result = reallocMem(handle, l + 3);
  result[l + 2] = 0;
  *count = l + 2;

  return result;

}

/*
** Convert a UTF8 string into a Wonka string. The string returned is a registered
** string. The UTF8 string probably comes from a class file. We first create some
** small support functions.
*/

inline static w_int is_singlet(w_word a) {

  if (! (a & 0x80)) {
    return 1;
  }

  return 0;
  
}

inline static w_int is_duplet(w_word a, w_word b) {

  if (a & 0xc0) {
    if (! (a & 0x20)) {
      if (b & 0x80) {
        if (! (b & 0x40)) {
          return 1;
        }
      }
    }
  }

  return 0;
  
}

inline static w_int is_triplet(w_word a, w_word b, w_word c) {

  if (a & 0xe0) {
    if (! (a & 0x10)) {
      if (b & 0x80) {
        if (! (b & 0x40)) {
          if (c & 0x80) {
            if (! (c & 0x40)) {
              return 1;
            }
          }
        }
      }
    }
  }

  return 0;
  
}

inline static w_char duplet2char(w_word a, w_word b) {

  w_char result = (w_char)(((a & 0x1f) << 6) + (b & 0x3f));
  
  return result;
  
}

inline static w_char triplet2char(w_word a, w_word b, w_word c) {

  w_char result = (w_char)(((a & 0x0f) << 12) + ((b & 0x3f) << 6) + (c & 0x3f));

  return result;
  
}

/*
** Now the function to create the Wonka string from a UTF8 string of bytes. The
** length argument passed is the number of UTF8 bytes. Note that the UTF8 bytes
** string is not expected to be NULL terminated, so the full length should be processed.
*/

w_string utf2String(w_thread thread, const char *utf8string, w_size utf8length) {
  w_string result;
  w_realm realm = thread2realm(thread);
  w_char *chars;
  w_string string;
  w_byte x;
  w_byte y;
  w_byte z;
  w_size idx = 0;
  w_size i = 0;
  w_boolean is_latin1 = WONKA_TRUE;

  /*
  ** Allocate a temporary buffer, we allocate it too be worst case size. This
  ** could be too large, but it will be copied over later to a buffer of a 
  ** correct size in the string itself. The worst case is when all UTF8 bytes
  ** are singlets, so we allocate 1 extra as the length.
  */
  
  if (utf8length) {
    chars = allocMem_realm(realm, (utf8length + 1) * sizeof(w_char));
    while (i < utf8length) { // [CG 20000410] WAS: <=
      x = (w_byte)utf8string[i++];

      if (is_singlet(x)) {
        chars[idx++] = (w_char)x;
      }
      else {
// [CG 20000410] add:
         if (i >= utf8length) {
           woempa(9, "Incomplete UTF8 sequence 0x%02x\n", x);
           break;
         }
// [CG 20000410] end
        y = (w_byte)utf8string[i++];
        if (is_duplet(x, y)) {
          w_char c = duplet2char(x, y);
          is_latin1 &= (c < 256);
          chars[idx++] = c;
        }
        else {
// [CG 20000410] add:
           if (i >= utf8length) {
             woempa(9, "Incomplete UTF8 sequence 0x%02x 0x%02x\n", x, y);
             break;
           }
// [CG 20000410] end
          z = (w_byte)utf8string[i++];
          if (is_triplet(x, y, z)) {
            is_latin1 = WONKA_FALSE;
            chars[idx++] = triplet2char(x, y, z);
          }
          else {
            woempa(9, "Unknown UTF8 characters 0x%02x 0x%02x 0x%02x\n", x, y, z);
          }
        }
      }
    }

#ifdef BYTEWISELATIN
    string = allocString(idx, is_latin1);
    if (is_latin1) {
      w_size j;
      w_char *src = chars;
      w_ubyte *dst = string->contents.bytes;
      for (j = 0; j < idx; ++j) {
        *dst++ = *src++;
      }
    }
    else {
      w_memcpy(string->contents.chars, chars, sizeof(w_char) * idx);
    }
#else
    string = allocString(idx, WONKA_FALSE);
    w_memcpy(string->chars, chars, sizeof(w_char) * idx);
#endif
    releaseMem(chars);
    hashString(string);
  
    woempa(1, "Created string '%w' (%p).\n", string, string);

    result = registerString(string);  
    if (result != string) {
      woempa(1, "String '%w' (%p) already existed in pool\n", result, result);
    }
  }
  else {
    result = cstring2String(W_Thread_system, "", 0);
    woempa(1, "Created empty string (%p)\n", result);
  }

  if(result == NULL) {
    w_dump("Ajajaj, not good !\n");
    abort();
  }

  return result;
  
}

w_char* utf2chars(w_thread thread, const char *utf8string,  w_int* length) {
  w_byte x;
  w_byte y;
  w_byte z;
  w_size idx = 0;
  w_int i = 0;
  w_int utf8length = *length;
  w_char* chars  = (w_char*) allocMem(thread, utf8length * sizeof (w_char));

  if(chars == NULL){
    return NULL;
  }

  while (i < utf8length) {
    x = (w_byte)utf8string[i++];
    if (is_singlet(x)) {
      chars[idx++] = (w_char)x;
    }
    else {
      if (i >= utf8length) {
        woempa(9, "Incomplete UTF8 sequence 0x%02x\n", x);
        break;
      }
      y = (w_byte)utf8string[i++];
      if (is_duplet(x, y)) {
        chars[idx++] = duplet2char(x, y);
      }
      else {
        if (i >= utf8length) {
          woempa(9, "Incomplete UTF8 sequence 0x%02x 0x%02x\n", x, y);
          break;
        }
        z = (w_byte)utf8string[i++];
        if (is_triplet(x, y, z)) {
          chars[idx++] = triplet2char(x, y, z);
        }
        else {
          woempa(9, "Unknown UTF8 characters 0x%02x 0x%02x 0x%02x\n", x, y, z);
        }
      }
    }
  }
  *length = idx;
  return chars;
}

/*
** Create a registered Wonka string from a unicode string. The unicode string
** is just an array of w_chars and the length is passed as a separate argument.
*/
w_string unicode2String(w_thread thread, w_char *chars, w_size length) {
  w_boolean is_latin1 = WONKA_TRUE;
  w_string string;
  w_string result;
  w_size i;
  
  if (length) {
    for (i = 0; i < length && is_latin1; ++i) {
      is_latin1 &= (chars[i] < 256);
    }
#ifdef BYTEWISELATIN
    string = allocString(length, is_latin1);
    if (is_latin1) {
      w_char *src = chars;
      w_ubyte *dst = string->contents.bytes;
      for (i = 0; i < length; ++i) {
        *dst++ = *src++;
      }
    }
    else {
      w_memcpy(string->contents.chars, chars, sizeof(w_char) * length);
    }
#else
    string = allocString(length, WONKA_FALSE);
    w_memcpy(string->chars, chars, sizeof(w_char) * length);
#endif
    hashString(string);
    woempa(1, "Created string '%w' (%p).\n", string, string);

    result = registerString(string);  
    if (result != string) {
      woempa(1, "String '%w' (%p) already existed in pool\n", result, result);
    }
  }
  else {
    result = cstring2String(W_Thread_system, "", 0);
    woempa(1, "Created empty string (%p)\n", result);
  }

  if(result == NULL) {
    w_dump("Zut alors ! \n");
    abort();
  }

  return result;  
  
}

/*
** Create an unregistered Wonka string from a unicode string. The unicode string
** is just an array of w_chars and the length is passed as a separate argument.
*/
w_string unicode2w_string(w_char *chars, w_size length) {

  w_string string;
  
  if (length) {
#ifdef BYTEWISELATIN
    w_boolean is_latin1;
    w_size i;

    is_latin1 = WONKA_TRUE;
    for (i = 0; i < length && is_latin1; ++i) {
      is_latin1 &= (chars[i] < 256);
    }
    string = allocString(length, is_latin1);
    if (is_latin1) {
      w_char *src = chars;
      w_ubyte *dst = string->contents.bytes;
      for (i = 0; i < length; ++i) {
        *dst++ = *src++;
      }
    }
    else {
      w_memcpy(string->contents.chars, chars, sizeof(w_char) * length);
    }
#else
    string = allocString(length, WONKA_FALSE);
    w_memcpy(string->chars, chars, sizeof(w_char) * length);
#endif
    hashString(string);
  }
  else {
    string = cstring2String(W_Thread_system, "", 0);
    woempa(1, "Empty string\n");
  }

  return string;  
  
}

/*
** Create a Wonka string from a C string. The length should NOT take into
** account the '\0' character of the C string. I.e. one can use strlen()
** to calculate the length when calling this function.
*/

w_string cstring2String(w_thread thread, const char *cstring, w_size length) {

  w_string string;
  w_string result;
  w_size i;
  char *src = (char*)cstring;

#ifdef BYTEWISELATIN
  string = allocString(length, WONKA_TRUE);
#else
  string = allocString(length, WONKA_FALSE);
#endif

  if (length) {
#ifdef BYTEWISELATIN
    w_ubyte *dst = string->contents.bytes;
#else
    w_char *dst = string->chars;
#endif
    for (i = 0; i < length; i++) {
      *dst++ = *src++;
    }
  }
  else {
#ifdef BYTEWISELATIN
    string->length_and_flags = STRING_IS_LATIN1;
#else
    string->length = 0;
#endif
  }

  hashString(string);
  woempa(1, "Created string '%w' (%p).\n", string, string);
  
  result = registerString(string);  
  if (result != string) {
    woempa(1, "String '%w' (%p) already existed in pool\n", result, result);
  }

  if(result == NULL) {
    w_dump("Incredible bad mojo 2 !!\n");
    abort();
  }
  
  return result;
}

/*
** The function that attaches a hash value to a w_string.
** From the JDK 1.2 documentation. Notice that the hashcode calculation
** has changed from JDK 1.1 to JDK 1.2 !
*/

w_word hashString(w_string string) {

  w_int hash = 0;
  w_size i;
   
  for (i=0;i<(w_size)(string_length(string));++i) {
    hash = hash*31 + string_char(string, i);
  }

#ifdef BYTEWISELATIN
  string->wonka_hash = (w_word)buzHash(string->contents.bytes, (string_is_latin1(string) ? sizeof(w_ubyte) : sizeof(w_char))*(string_length(string)));
#else
  string->wonka_hash = (w_word)buzHash((w_byte*)string->chars,((w_int)sizeof(w_char))*((w_int)string_length(string)));
#endif

  return hash;
  
}

void releaseString(w_string string) {
      woempa(1, "About to release (%p) %w\n", string, string);
      deallocString(string);
}    

/*
** The hash function for the string hashtable. It should return
** the hash value of the string that should be allready available.
*/

w_word ht_stringHash(w_word string_word) {

  w_string string = (w_string)string_word;

  return string->wonka_hash;
  
}

/*
** Register a Wonka string. The result returned is the string that should be
** used for subsequent operations; the argument string that is passed could
** be the same as the result string, if it was not registered yet, or the
** pointer to the allready registered string if it was a duplicate. If there
** was a duplicate, the passed argument string is released.
*/

w_string registerString(w_string string) {
 
  w_string result;

#ifdef DEBUG
//  if ((string_hashtable->occupancy%10000)==0) dumpStrings();
#endif

  woempa(1, "Registering %w at %p\n", string, string);
#ifdef NEWSTRINGREG
  ht_lock(string_hashtable);
  result = (w_string)ht_read_no_lock(string_hashtable, (w_word)string);
  if (result) {
    woempa(1, "Found `%w' at %p, refcount %d \n", result, result, result->refcount);
    if (result != string) {
      releaseString(string);
    }
  }
  else {
    woempa(1, "Didn't find it ... %w is new\n", string);
    result = string;
    ht_write_no_lock(string_hashtable, (w_word)result, (w_word)result);
  }
  ++result->refcount;
  woempa(1, "Refcount now %d\n", result->refcount);
  ht_unlock(string_hashtable);
#else
  result = (w_string)ht_register(string_hashtable, (w_word)string);
  if (result) {
    woempa(1, "Found it at %p ... %w\n", result,result);

    /*
    ** If the found string isn't the same one that was passed to us, 
    ** we can release it by making it garbage...
    */

    if (result != string) {
      releaseString(string);
    }
  }
  else {

    woempa(1,"Adding... %w\n", string);
    result = string;
  }
#endif

  if(result == NULL) {
    w_dump("Incredible bad mojo !!\n");
    abort();
  }

  return result;
  
}

/*
** Deregister a string. When we see that the registered count drops to 0, 
** we release the w_String structure.
**
** [CG 20000413]: modified to use ht_deregister
*/

void deregisterString(w_string string) {
#ifdef NEWSTRINGREG
  woempa(1, "Deregistering %w\n", string);
  ht_lock(string_hashtable);
  if ((--string->refcount) == 0) {
    ht_erase_no_lock(string_hashtable, (w_word)string);
    woempa(1,"releasing %w (%p)\n", string, string);
    releaseString(string);
  }
  ht_unlock(string_hashtable);
#else
  w_string garbage = NULL;

  woempa(1, "Deregistering %w\n", string);
  garbage = (w_string)ht_deregister(string_hashtable, (w_word)string);
  if (garbage) {
    woempa(1,"releasing %w (%p)\n",garbage,garbage);
    releaseString(garbage);
  }
#endif
}

/*
** The compare function for the string hashtable.
*/

w_boolean ht_stringCompare(w_word string1_word, w_word string2_word) {

  w_string string1 = (w_string)string1_word;
  w_string string2 = (w_string)string2_word;
  w_size   length1 = string_length(string1);
  w_size   length2 = string_length(string2);
  w_boolean result = WONKA_FALSE;

  /*
  ** Do a quick check first, when the two memory addresses match, the strings match.
  ** Otherwise, try some other checks, starting with the cheapest comparisons first and
  ** ending with a memory compare over the unicode strings.
  */

  woempa(1, "Comparing '%w' (%p) with '%w' (%p)\n", string1, string1, string2, string2);
  if (string1 == string2) {
    woempa(1, "Identical, return TRUE\n");
    result = WONKA_TRUE;
  }
  else if (length1 == 0 && length2 == 0) {
    woempa(1, "Both empty, return TRUE\n");
    result = WONKA_TRUE;
  }
  else {
    w_boolean is_latin1 = string_is_latin1(string1);

    woempa(1, "String1 is %slatin-1\n", is_latin1 ? "" : "not ");

    if (string_is_latin1(string2) == is_latin1 && length1 == length2) {
      woempa(1, "String2 is %slatin-1, both have length %d, hashcodes are 0x%08x/0x%08x\n", is_latin1 ? "" : "not ", length1, string1->wonka_hash, string2->wonka_hash);
      if (string1->wonka_hash == string2->wonka_hash) {
#ifdef BYTEWISELATIN
        woempa(1, "Comparing %d bytes starting at %p/%p\n", length1 * (is_latin1 ? 1 : 2), string1->contents.bytes, string2->contents.bytes);
        if (memcmp(string1->contents.bytes, string2->contents.bytes, (is_latin1 ? sizeof(w_ubyte) : sizeof(w_char)) * length1) == 0) {
          result = WONKA_TRUE;
        }
#else
        woempa(1, "Comparing %d bytes starting at %p/%p\n", length1 * 2, string1->chars, string2->chars);
        if (memcmp(string1->chars, string2->chars, sizeof(w_char) * length1) == 0) {
          result = WONKA_TRUE;
        }
#endif
      }
    }
  }

  woempa(1, "Compared (%p) %w and (%p) %w: they are %sequal.\n", string1, string1, string2, string2, result ? "" : "NOT ");
  
  return result;

}

char * print_string(char * buffer, int * remain, void * s, w_int width, w_int prec, w_flags flags) {

  w_string string = s;
  w_size padchars;
  w_size length;
  w_size i;

  if (*remain < 1) {

    return buffer;

  }

  if (s) {
    length = string_length(string);
    padchars = 0;
    if (width) {
      if (length > (w_size)width) {
        length = width;
      }
      else {
        padchars = width - length;
      }
    }

    if (flags & FMT_ALIGN_LEFT) {
      while (padchars-- && *remain > 0) {
        *buffer++ = ' ';
        *remain -= 1;
      }
    }
    
    for (i = 0; *remain > 0 && i < length; ++i) {
      w_char ch = string_char(string, i);

      switch(ch/32) {
      case 1:
      case 2:
      case 3:
      case 5:
      case 6:
      case 7:
        *buffer++ = ch;
        break;

      default:
        *buffer++ = '?';
      }
      *remain -= 1;
    }

    if (!(flags & FMT_ALIGN_LEFT)) {
      while (padchars-- && *remain > 0) {
        *buffer++ = ' ';
        *remain -= 1;
      }
    }

  }
  else {
    *buffer++ = '<';
    *buffer++ = 'n';
    *buffer++ = 'u';
    *buffer++ = 'l';
    *buffer++ = 'l';
    *buffer++ = '>';
    *remain -= 6;
  }
              
  return buffer;
                
}
                
void startStrings() {

  /*
  ** Allocate the hashtable in which the strings will be stored.
  */

  string_hashtable = ht_create(W_Realm_system, (char *)"hashtable:strings", STRING_HASHTABLE_SIZE, ht_stringHash, ht_stringCompare, 0, 0);
  woempa(1, "created string_hashtable at %p\n", string_hashtable);

  /*
  ** Create the mutex used to prevent race problems
  */
  string_mutex = (x_mutex)allocMem_system(sizeof(x_Mutex));
  x_mutex_create(string_mutex);

  /*
  ** Set up the special strings
  */

  string_NULL = cstring2String(W_Thread_system,"<NULL>",6);
  string_EMPTY = cstring2String(W_Thread_system,"<EMPTY>",7);
  string_BOGUS = cstring2String(W_Thread_system,"<BOGUS>",7);

  woempa(1, "Initialised string tables and static strings.\n");

}

w_int stringAdd(w_char *destination, w_string source) {

#ifdef BYTEWISELATIN
  if (string_is_latin1(source)) {
    w_size i;
    w_size l = string_length(source);
    w_char *dst = destination;
    w_ubyte *src = source->contents.bytes;

    for (i = 0; i < l; ++i) {
      *dst++ = *src++;
    }
  }
  else {
    w_memcpy(destination, source->contents.chars, string_length(source) * sizeof(w_char));
  }
#else
  w_memcpy(destination, source->chars, string_length(source) * sizeof(w_char));
#endif

  return string_length(source);

}

w_string dots2slashes(w_thread thread, w_string string) {
  w_string result = string;
  w_size length = string_length(string);
  w_size i;

  woempa(1, "Converting string '%w'.\n", result);

  if (length) {
#ifdef BYTEWISELATIN
    w_boolean is_latin1 = string_is_latin1(string);

    result = allocString(length, is_latin1);
    if (is_latin1) {
      w_ubyte ch;

      for (i = 0; i < length; ++i) {
        ch = string->contents.bytes[i];
        if (ch == '.') {
          ch = '/';
        }
        else if (ch == '/') {
          woempa(9,"Wow!  Found a slush char, converted it to a DEL.\n");
          ch = 127;
        }
        result->contents.bytes[i] = ch;
      }
    }
    else {
      w_char ch;

      for (i = 0; i < length; ++i) {
        ch = string->contents.chars[i];
        if (ch == '.') {
          ch = '/';
        }
        else if (ch == '/') {
          woempa(9,"Wow!  Found a slysh char, converted it to a DEL.\n");
          ch = 127;
        }
        result->contents.chars[i] = ch;
      }
    }
#else
    w_char *src = string->chars;
    w_char *dst;

    result = allocString(string_length(string), WONKA_FALSE);
    dst = result->chars;
    for (i = 0; i < length; i++) {
      if (*src == '.') {
        *dst++ = '/';
      }
      else if (*src == '/') {
        woempa(9,"Wow!  Found a slxsh char, converted it to a DEL.\n");
        *dst++ = 127;
      }
      else {
        *dst++ = *src;
      }
      src += 1;
    }
#endif
    hashString(result);
    woempa(1, "Created string '%w' (%p).\n", result, result);
  }
  else {
    woempa(1, "Empty string, returning it unchanged\n");
  }
  
  result = registerString(result);  

  return result;

}

w_string slashes2dots(w_thread thread, w_string string) {
  w_string result = string;
  w_size length = string_length(string);
  w_size i;

  woempa(1, "Converting string '%w'.\n", result);

  if (length) {
#ifdef BYTEWISELATIN
    w_boolean is_latin1 = string_is_latin1(string);

    result = allocString(length, is_latin1);
    if (is_latin1) {
      w_ubyte ch;

      for (i = 0; i < length; ++i) {
        ch = string->contents.bytes[i];
        if (ch == '/') {
          ch = '.';
        }
        else if (ch == 127) {
          woempa(9,"Wow!  Found a DEL char, converted it to a slesh.\n");
          ch = '/';
        }
        result->contents.bytes[i] = ch;
      }
    }
    else {
      w_char ch;

      for (i = 0; i < length; ++i) {
        ch = string->contents.chars[i];
        if (ch == '/') {
          ch = '.';
        }
        else if (ch == 127) {
          woempa(9,"Wow!  Found a DEL char, converted it to a slush.\n");
          ch = '/';
        }
        result->contents.chars[i] = ch;
      }
    }
#else
    w_char *src = string->chars;
    w_char *dst;

    result = allocString(string_length(string), WONKA_FALSE);
    dst = result->chars;
    for (i = 0; i < length; i++) {
      if (*src == '/') {
        *dst++ = '.';
      }
      else if (*src == 127) {
        woempa(9,"Wow!  Found a DEL char, converted it to a slosh.\n");
        *dst++ = '/';
      }
      else {
        *dst++ = *src;
      }
      src += 1;
    }
#endif
    hashString(result);
    woempa(1, "Created string '%w' (%p).\n", result, result);
  }
  else {
    woempa(1, "Empty string, returning it unchanged\n");
  }
  
  result = registerString(result);  

  return result;

}

w_instance getStringInstance(w_string s) {
  w_thread thread = currentWonkaThread;
  w_instance instance;
    
  x_mutex_lock(string_mutex, x_eternal);
  instance = s->instance;
  if (! instance) {
    x_mutex_unlock(string_mutex);
    instance = allocInstance(thread, clazzString);
    if (instance) {
      s = registerString(s);
      x_mutex_lock(string_mutex, x_eternal);
      if (s->instance) {
        // CG 20040511 This case wasn't tested for : if some other thread
        // allocated a canonical instance then we would just overwite it.
        // This was probably harmless, apart from leaking a little memory.
        woempa(1, "allocated superfluous instance %p of String for '%w', unGREENING so GC will deal with it\n", instance, s);
        instance = s->instance;
      }
      else {
        woempa(1, "allocated instance %p of String for '%w'\n", instance, s);
        s->instance = instance;
        setWotsitField(instance, F_String_wotsit, s);
      }
      x_mutex_unlock(string_mutex);
    }
  }
  else {
    newReferenceTo(instance);
    pushLocalReference(thread->top, instance);
    x_mutex_unlock(string_mutex);
  }

  return instance;

}

w_instance newStringInstance(w_string s) {
  w_instance instance;
  
  instance = allocInstance(currentWonkaThread, clazzString);
  if (instance) {
    x_mutex_lock(string_mutex, x_eternal);
    setWotsitField(instance, F_String_wotsit, registerString(s));
    woempa(1, "allocated instance %p of String for '%w'\n", instance, s);
    x_mutex_unlock(string_mutex);
  }
  return instance;
}

#ifdef DEBUG

/*
** Routines to print out statistics about the strings and string table.
*/

void qsort(void *base, size_t nmemb, size_t size, int (*compar)(const void *, const void *));

/*
** The total number of Unicode characters we have allocated.
*/

static w_int total_length = 0;

int string_sort(const void *v_string1, const void *v_string2) {

  w_string string1 = *(w_string *)v_string1;
  w_string string2 = *(w_string *)v_string2;
  w_int result;

  /*
  ** We compare on the basis of the memory pointer of the two strings so that strings
  ** that are allocated together end up together in the list.
  */

  result = (w_int)string1 - (w_int)string2;
  
  return result;
  
}

void string_iterate(w_word string_word, w_word hash, void * v_array, void * v_indx) {

  w_string string = (w_string)string_word;
  w_string *array = (w_string *)v_array;
  w_int *indx = (w_int *)v_indx;
  
  total_length += string_length(string);
  array[*indx] = string;
  *indx += 1;
  
}

void dumpSingleString(w_int i, w_string string) {

  w_int desc_registered = 0;
  
  woempa(9, "|%5d | %4d |  %4d |   %3d  | 0x%08x |'%w'\n", i,
    ht_read(string_hashtable,(w_word)string), desc_registered, string_length(string), string->wonka_hash, 
    string);

}


void dumpStrings(void) {

  w_string *array;
  w_int indx = 0;
  w_int i;
  
  array = allocMem_system(string_hashtable->occupancy * sizeof(w_string *));

  ht_iterate(string_hashtable, string_iterate, array, &indx);  

  qsort(array, (unsigned)(indx), sizeof(w_string), string_sort);
  
  woempa(9, "+------+------+-------+--------+------------+-------------------------+\n");
  woempa(9, "|  #   |  Reg |  Desc | Length |    Hash    |    First registration   | String\n");
  woempa(9, "+------+------+-------+--------+------------+-------------------------+-------------------------------------------------------------\n");
  for (i = 0; i < indx; i++) {
    dumpSingleString(i, array[i]);
  }
  woempa(9, "+------+------+-------+--------+------------+-------------------------+-------------------------------------------------------------\n");
  woempa(9, "|       Total number of bytes for Unicode characters: %6d\n", total_length * sizeof(w_char));
  woempa(9, "| Number of bytes for w_String (%d bytes) structures: %6d\n", sizeof(w_String), indx * sizeof(w_String));
  woempa(9, "+------+------+-------+--------+------------+-------------------------+-------------------------------------------------------------\n");
  

  releaseMem(array);

}
#endif

