/**************************************************************************
* Copyright (c) 2001, 2002, 2002, 2003 by Acunia N.V. All rights reserved.*
*                                                                         *
* This software is copyrighted by and is the sole property of Acunia N.V. *
* and its licensors, if any. All rights, title, ownership, or other       *
* interests in the software remain the property of Acunia N.V. and its    *
* licensors, if any.                                                      *
*                                                                         *
* This software may only be used in accordance with the corresponding     *
* license agreement. Any unauthorized use, duplication, transmission,     *
*  distribution or disclosure of this software is expressly forbidden.    *
*                                                                         *
* This Copyright notice may not be removed or modified without prior      *
* written consent of Acunia N.V.                                          *
*                                                                         *
* Acunia N.V. reserves the right to modify this software without notice.  *
*                                                                         *
*   Acunia N.V.                                                           *
*   Philips site 5, box 3       info@acunia.com                           *
*   3001 Leuven                 http://www.acunia.com                     *
*   Belgium - EUROPE                                                      *
**************************************************************************/

/*
** $Id: threads.c 4294 2006-05-10 11:35:19Z chris $
*/

#include <string.h>

#include "arrays.h"
#include "clazz.h"
#include "core-classes.h"
#include "descriptor.h"
#include "exception.h"
#include "fields.h"
#include "hashtable.h"
#include "heap.h"
#include "interpreter.h"
#include "jni.h"
#include "ts-mem.h"
#include "methods.h"
#include "locks.h"
#include "loading.h"
#include "realms.h"
#include "wstrings.h"
#include "threads.h"
#include "list.h"
#include "oswald.h"

w_boolean haveWonkaThreads = WONKA_FALSE;

w_boolean system_init_thread_started;

/*
** The maximum number of Java frames on a stack is determined by the average
** number of bytes consumed by a call and is determined in the threads.h header
** file. The 'margin_frames' variable substracts from this determined maximum
** number of frames for a safety margin; so setting it rather low, risks stack
** corruption but increases the stack depth you can have, setting it higher 
** increases safety margin at the cost of the maximum number of calls you can make.
*/

static const w_int margin_frames = 60;

w_hashtable thread_hashtable;

w_int nondaemon_thread_count;

/*
** Statics and globals
*/

const unsigned char state2liveness[] = {
  0, /* 0: Unstarted      */
  1, /* 1: Ready          */
  1, /* 2: Locking        */
  1, /* 3: Waiting        */
  0, /* 4: Not used       */
  1, /* 5: Suspended      */
  0, /* 6: Not used       */
  0, /* 7: Dying          */
  0, /* 8: Dead           */
  0, /* 9: It's dead, Jim */
};

/*
** The internal Wonka representation of the system group and thread
** and the static slots for java/lang/Thread and java/lang/ThreadGroup.
*/

w_instance I_Thread_sysInit;
w_instance I_ThreadGroup_system;

w_thread W_Thread_sysInit;
w_thread W_Thread_system;
w_group  W_ThreadGroup_system;

jclass   class_Thread;
jclass   class_ThreadGroup;
jclass   class_Runtime;

jmethodID run_method;
jmethodID registerThread_method;
jmethodID deregisterThread_method;

static x_Mutex idLock;

/*
** Hashtable relating host OS thread control block to our w_Thread structure.
**    key:   host OS thread id, 
**    value: address of our w_Thread structure
*/

w_hashtable thread_hashtable;

#define THREAD_HASHTABLE_SIZE 51

/*
** Hashtable relating host OS thread control block to our w_Group structure.
**    key:   host OS thread id, 
**    value: address of our w_Thread structure
*/

w_hashtable threadgroup_hashtable;

#define THREADGROUP_HASHTABLE_SIZE 11

static const char *unborn_thread_report(x_thread);
static const char *dying_thread_report(x_thread);

/*
** Allocate and clear out the necessary fields for a new thread structure.
*/

void attachFirstFrame(w_thread new) {

  new->Frame.previous = NULL;
  new->Frame.method = NULL;
//  new->Frame.clazz = NULL;
  new->exception = NULL;
  new->Frame.thread = new;
  new->Frame.top = new->Frame.Slots;
  new->Frame.auxs = new->auxs;
  new->top = & new->Frame;
  new->maxframes = 100; // Bootstrap value
  new->numframes = 100; // Bootstrap value

}

//  Java priority -------->    0   1   2   3   4   5   6   7   8   9  10
static w_int java2kprio[] = { 50, 48, 46, 44, 42, 40, 38, 36, 34, 32, 30};

w_int priority_j2k(w_int java_prio, w_int trim) {

  java_prio = (java_prio > 10) ? 10 : java_prio;
  java_prio = (java_prio <  0) ?  0 : java_prio;

  return java2kprio[java_prio];
    
}

w_thread createThread(w_thread parentthread, w_instance Thread, w_group parentgroup, w_string name, w_size stacksize, w_size locals) {

  w_instance Name;
//  x_queue queue;
  w_thread newthread;
  
  newthread = allocClearedMem(parentthread, sizeof(w_Thread));

  if (!newthread) {
    woempa(9, "Unable to allocate w_Thread for %w\n", name);

    return NULL;

  }

  newthread->auxs = allocClearedMem_system(sizeof(x_Slot) * locals);
  if (!newthread->auxs) {
    woempa(9, "Unable to allocate %d x_Slots for %w\n", locals, name);
    releaseMem(newthread);

    return NULL;
  }
  
  attachFirstFrame(newthread);
  newthread->maxauxs = & newthread->auxs[locals];
  newthread->natenv = &w_JNINativeInterface;
  newthread->label = (char *)"thread";
  newthread->name = name;
  newthread->group = parentgroup;
  newthread->Thread = Thread;
  newthread->kthread = allocClearedMem(newthread, sizeof(x_Thread));
  if (!newthread->kthread) {
    woempa(9, "Unable to allocate x_Thread for %w\n", name);
    releaseMem(newthread->auxs);
    releaseMem(newthread);

    return NULL;

  }

  newthread->kthread->xref = newthread;
  newthread->ksize = stacksize;
  newthread->maxframes = (stacksize / bytes_per_call) - margin_frames;
  newthread->numframes = newthread->maxframes;

#ifdef GC_SAFE_POINTS
  newthread->unsafe_thread = NULL;
  newthread->gc_thread = NULL;
#endif

  if (parentthread->jpriority > 0) {
    newthread->jpriority = parentthread->jpriority;
  }
  else {
// HACK for GM-FT only ...
    woempa(9,"WARNING! Thread `%w' has parent `%w', which is a native thread\n", NM(newthread), NM(parentthread));
    woempa(9," ^ ^ ^   Setting thread priority to 5 ...\n");
    woempa(9," | | | Fix this!\n");
    newthread->jpriority = 5;
//
  }
  newthread->kpriority = priority_j2k(newthread->jpriority,0);
  newthread->isDaemon = parentthread->isDaemon;

  newthread->run_method = run_method;
  newthread->top->top[0].c.w = (w_word) Thread;
  newthread->top->top[0].s = stack_trace;
  newthread->top->top += 1;
  newthread->state = wt_unstarted;
  setWotsitField(Thread, F_Thread_wotsit,  newthread);
  Name = newStringInstance(name);
  setReferenceField(Thread, Name, F_Thread_name);
  removeLocalReference(newthread, Name);

  return newthread;

}

void terminateThread(w_thread thread) {

  void * result;
  x_status status;

  status = x_thread_join(thread->kthread, &result, 100);
  if (status == xs_success || status == xs_no_instance) {
    status = x_thread_delete(thread->kthread);
    if (status == xs_success) {
      thread->kthread->xref = NULL;
      thread->kthread->report = dying_thread_report;
      if (thread == W_ThreadGroup_system->manager) {
        woempa(9,"This is %w, so I won't releaseMem memory that wasn't allocMem'd ...\n", thread->name);
      }
      else {
        releaseMem(thread->kthread);  
        if (thread->kstack) {
          releaseMem(thread->kstack);
        }  
        if (thread->auxs) {
          releaseMem(thread->auxs);
        }  
      }
    }
    else {
      woempa(9, "Thread delete status %s\n", x_status2char(status));
      abort();
    }
  }
  else {
    woempa(9, "Join status %s\n", x_status2char(status));
    abort();
  }

  releaseMem(thread);
}

void addThreadToGroup(w_thread thread, w_group group) {

  JNIEnv  *env = w_thread2JNIEnv(currentWonkaThread);
  w_realm  realm;

  realm = group->realm;

  /*
  ** Do the accounting of the thread use
  */

  if (!thread->isDaemon) {
    nondaemon_thread_count += 1;
    woempa(1, "Adding non-daemon thread %w, total now %d\n", thread->name, nondaemon_thread_count);
  }
  realm->thread_usage += 1;

  (*env)->CallNonvirtualVoidMethod(env, group->ThreadGroup, class_ThreadGroup, registerThread_method, thread->Thread);

}

void removeThreadFromGroup(w_thread thread, w_group group) {

  JNIEnv  *env = w_thread2JNIEnv(currentWonkaThread);
  w_realm   realm;

  realm = group->realm;

  if (!thread->isDaemon) {
    --nondaemon_thread_count;
    woempa(1, "Removed non-daemon thread %w, total now %d\n", thread->name, nondaemon_thread_count);
  }
  realm->thread_usage -= 1;

  if (thread->group != group) {
    woempa(9,"Fascinating... we should remove thread %p from group %p, but %p->group is %p\n",thread,group,thread,thread->group);
  }


  (*env)->CallNonvirtualVoidMethod(env, group->ThreadGroup, class_ThreadGroup, deregisterThread_method, thread->Thread);

}

void requestThreadTermination(w_thread thread) {
  w_boolean succeeded = WONKA_FALSE;
  w_message message;
  x_status  status;

  while (!succeeded) {
    message = (w_message)allocMem_system(sizeof(w_Message));
    if (message) {
      message->request = MESSAGE_STOP_THREAD;
      message->sender = thread;
      message->addressee = thread;
      message->bounced = 0;
      message->content = NULL;

      status = x_queue_send(thread->group->box, (void*)message, x_millis2ticks(50));
      if (status == xs_success) {
        succeeded = WONKA_TRUE;
      }
      else {
        x_thread_sleep(1);
      }
    }
  }
}
 
w_thread _currentWonkaThread(const char *f, int l) {
  x_thread kthread;
  w_thread  thread;

  kthread = x_thread_current();

  if(!kthread)
    return NULL; /* Fix for crash on O4P. */
  
  thread = (w_thread)kthread->xref;
  
  if (!thread)
    return NULL; /* MB: Thread can't have been set up yet */
  
  if (*((w_word*)thread->label)!=*((w_word*)"thread")  
     && *((w_word*)thread->label)!=*((w_word*)"group")){
    woempa(9,"currentWonkaThread called at %s:%d from non-wonka thread %p\n", f, l, kthread);
    wabort(ABORT_WONKA, "%s (%d) No Wonka thread found for kernel thread %p!\n", f, l, kthread);
  }

  return thread;
}

static char thread_report_buffer[92];

static const char *unborn_thread_report(x_thread x) {
  return "Thread not yet started";
}

static const char *dying_thread_report(x_thread x) {
  return "Thread returning to Saturn";
}

const char *running_thread_report(x_thread x) {

  w_thread t;

  if (x) {
    t = x->xref;
    x_snprintf(thread_report_buffer, 91, "%T", t);
    return thread_report_buffer;
  }

  return "!!! Wonka thread xref is NULL !!!";

}

static char group_report_buffer[92];

static const char *group_report(x_thread x) {
  w_thread t = x->xref;
  w_group g = t->group;

  if (g->ThreadGroup && g->realm && g->realm->ThreadGroup && (w_group)getWotsitField(g->realm->ThreadGroup, F_ThreadGroup_wotsit) == g) {
  // Actually this is a Realm
    w_realm r = g->realm;

    (void)x_snprintf(group_report_buffer, 91, "Realm %-24w max priority %d memory quota %-10d used %-10d thread quota %-2d used %-2d",
        ((w_group)getWotsitField(r->ThreadGroup, F_ThreadGroup_wotsit))->name, ((w_group)getWotsitField(r->ThreadGroup, F_ThreadGroup_wotsit))->maxPriority,
        r->memory_quota,r->memory_usage,r->thread_quota,r->thread_usage);

  }
  else if (g->name) {
    (void)x_snprintf(group_report_buffer, 91, "Group %-24w max priority %d", NM(g), g->maxPriority);
  }
  else {
    (void)x_snprintf(group_report_buffer, 91, "Group (unnamed)           max priority %d", NM(g), g->maxPriority);
  }
  return group_report_buffer;
}

void createThreadGroup(w_instance ThreadGroup, w_group parent, w_string name, w_size stacksize) {

  w_group oldgroup;
  w_group new;
  w_thread debiteur;
  w_instance Name;

  /*
  ** Parent is null if we are creating the system thread group. Don't use the 
  ** parent's mutex then...
  ** The group structure itself is allocated on behalf of the parent.
  */

  if (parent) {
    x_mutex_lock(parent->mutex, x_eternal);
    debiteur = parent->manager;
  }
  else {
    debiteur = W_Thread_system;
  }

  new = allocClearedMem(debiteur, sizeof(w_Group));
  if (!new) {
    woempa(9, "Unable to allocate w_Group for %w\n", name);

    return;

  }

  new->label = (char*)"group";

  if (parent == NULL) {
    new->realm = W_Realm_system;
  } 
  else {
    new->realm = parent->realm;
    if (new->realm->ThreadGroup==NULL) {
      new->realm->ThreadGroup = ThreadGroup;
    }
  }
  
  new->manager = allocClearedMem(parent->manager, sizeof(w_Thread));
  new->manager->auxs = allocClearedMem_system(sizeof(x_Slot) * default_locals);
  
  if (!new->manager) {
    woempa(9, "Unable to allocate w_Thread for manager of %w\n", name);
    releaseMem(new);

    return;

  }

  attachFirstFrame(new->manager);
  new->manager->maxauxs = & new->manager->auxs[default_locals];

  /*
  ** From this point, allocation is on behalf of the new group...
  */

  new->manager->label = (char*)"thread:group-manager";
  new->manager->group = new;
  debiteur = new->manager;
  new->mutex = (x_mutex)allocMem(debiteur, sizeof(x_Mutex));
  if (!new->mutex) {
    woempa(9, "Unable to allocate x_mutex for %w\n", name);
    releaseMem(new->manager);
    releaseMem(new);

    return;

  }

  new->box = (x_queue)allocMem(debiteur, sizeof(x_Queue));
  if (!new->box) {
    woempa(9, "Unable to allocate box (x_Queue) for %w\n", name);
    releaseMem(new->mutex);
    releaseMem(new->manager);
    releaseMem(new);

    return;

  }

  new->boxSpace = (w_byte *)allocClearedMem(debiteur, GROUP_MANAGER_BOX_SIZE);
  if (!new->boxSpace) {
    woempa(9, "Unable to allocate boxSpace for %w\n", name);
    releaseMem(new->box);
    releaseMem(new->mutex);
    releaseMem(new->manager);
    releaseMem(new);

    return;

  }

  new->manager->natenv = &w_JNINativeInterface;
  new->manager->kthread = allocClearedMem(debiteur, sizeof(x_Thread));
  if (!new->manager->kthread) {
    woempa(9, "Unable to allocate x_Thread for manager of %w\n", name);
    releaseMem(new->boxSpace);
    releaseMem(new->box);
    releaseMem(new->mutex);
    releaseMem(new->manager);
    releaseMem(new);

    return;

  }

  new->manager->kthread->xref = new->manager;
  new->manager->kthread->report = unborn_thread_report;
  new->manager->ksize = stacksize;
#ifndef O4P
  new->manager->kstack = allocMem(debiteur, new->manager->ksize);
  if (!new->manager->kstack) {
    woempa(9, "Unable to allocate native stack for manager of %w\n", name);
    releaseMem(new->manager->kthread);
    releaseMem(new->boxSpace);
    releaseMem(new->box);
    releaseMem(new->mutex);
    releaseMem(new->manager);
    releaseMem(new);

    return;

  }
#endif

  new->manager->jpriority = GROUP_MANAGER_PRIORITY;
  new->manager->kpriority = priority_j2k(GROUP_MANAGER_PRIORITY,1);
  new->parent = parent;
  new->name = name;
  new->manager->name = name;
  new->maxPriority = parent ? parent->maxPriority : 10;
  new->ThreadGroup = ThreadGroup;
#if defined(GC_SAFE_POINTS)
  new->manager->unsafe_thread = NULL;
  new->manager->gc_thread = NULL;
#endif

  x_queue_create(new->box, new->boxSpace, GROUP_MANAGER_BOX_SIZE / sizeof(w_word)); // Oswald queue's use number of WORDS as size !!

  x_mutex_create(new->mutex);

  setWotsitField(ThreadGroup, F_ThreadGroup_wotsit, new);
  Name = newStringInstance(name);
  setReferenceField(ThreadGroup, Name, F_ThreadGroup_name);
  removeLocalReference(W_Thread_system, Name);

  if (parent) {
    x_mutex_unlock(parent->mutex);
  }

  //numGroups += 1;
  new->realm->thread_usage += 1;

  /*
  ** TODO, check quota limit...
  */

  woempa(1, "registering os thread %p as Java threadGroup instance %p (%k)\n", new->manager->kthread, new->ThreadGroup, instance2clazz(new->ThreadGroup));
  oldgroup = (w_group)ht_write(threadgroup_hashtable, (w_word)new->manager->kthread, (w_word)new);

  if (oldgroup) {
    wabort(ABORT_WONKA, "Caramba! that os thread was already registered as %p!\n", oldgroup);
  }
 
  x_thread_create(new->manager->kthread, groupManagerEntry, new, new->manager->kstack, new->manager->ksize,  new->manager->kpriority, TF_START); 

}

static void releaseMessage(w_message message) {

  releaseMem(message);
  
}

void groupManagerEntry(void* WonkaThreadGroup) {

  w_group group = WonkaThreadGroup;
  w_message message;
  x_status status;
  
  group->manager->kthread->report = group_report;
  while (1) {
    status = x_queue_receive(group->box, (void**)&message, x_eternal);
    if (status != xs_success) {
      woempa(9, "Status was '%s'\n", x_status2char(status));
      continue;
    }
    woempa(1, "Received message request %d.\n", message->request);
    switch (message->request) {

      case (MESSAGE_STOP_THREAD):
        terminateThread(message->addressee);
        releaseMessage(message);         
        break;

      default: 
        woempa(1, "Unknown message request %d.\n", message->request); break;
    }
  }

}

static w_Group systemThreadGroup;
static w_Thread systemThreadGroupManager;
static x_Mutex systemThreadGroupMutex;
static x_Queue systemThreadGroupQueue;
static x_Thread systemThreadGroupManagerKthread;
static void *systemThreadGroupBoxSpace[SYSTEM_GROUP_MANAGER_BOX_SIZE];
static x_Slot systemThreadGroupManagerAuxs[manager_locals];
#ifdef O4P
#define systemThreadGroupManagerKstack NULL
#else
static char *systemThreadGroupManagerKstack[system_stack_size];
#endif

w_ubyte *initKernel(w_ubyte *start_heap) {

  x_mutex_create(&idLock);

  /*
  ** Create our internal system thread and group representation
  */

  W_ThreadGroup_system = &systemThreadGroup;
  W_ThreadGroup_system->label = (char*)"group:system";
  W_ThreadGroup_system->parent = W_ThreadGroup_system;
  W_ThreadGroup_system->maxPriority = 10;
  W_ThreadGroup_system->realm = W_Realm_system;
  setFlag(W_ThreadGroup_system->flags, WG_SYSTEM);
  // TODO make the following dependent upon the system...
  W_ThreadGroup_system->mutex = &systemThreadGroupMutex;
  x_mutex_create(&systemThreadGroupMutex);

  W_ThreadGroup_system->box = &systemThreadGroupQueue;
  W_ThreadGroup_system->boxSpace = systemThreadGroupBoxSpace;
  x_queue_create(&systemThreadGroupQueue, &systemThreadGroupBoxSpace, SYSTEM_GROUP_MANAGER_BOX_SIZE);
  
  W_ThreadGroup_system->manager = &systemThreadGroupManager;
  W_ThreadGroup_system->manager->auxs = &systemThreadGroupManagerAuxs;
   
  attachFirstFrame(W_ThreadGroup_system->manager);
  W_ThreadGroup_system->manager->maxauxs = & W_ThreadGroup_system->manager->auxs[manager_locals];
  
  W_ThreadGroup_system->manager->natenv = &w_JNINativeInterface;
  W_ThreadGroup_system->manager->label = (char*)"thread:group-manager:system";
  W_ThreadGroup_system->manager->kthread = &systemThreadGroupManagerKthread;
  W_ThreadGroup_system->manager->kthread->xref = W_ThreadGroup_system->manager;
  W_ThreadGroup_system->manager->ksize = system_stack_size;
  W_ThreadGroup_system->manager->maxframes = (system_stack_size / bytes_per_call) - margin_frames;
  W_ThreadGroup_system->manager->numframes = W_ThreadGroup_system->manager->maxframes;
#ifndef O4P
  W_ThreadGroup_system->manager->kstack = &systemThreadGroupManagerKstack;
#endif
  W_ThreadGroup_system->manager->kpriority = SYSTEM_GROUP_MANAGER_PRIORITY;
  W_ThreadGroup_system->manager->jpriority = priority_j2k(SYSTEM_GROUP_MANAGER_PRIORITY,1);
  W_ThreadGroup_system->manager->group = W_ThreadGroup_system;
#if defined(GC_SAFE_POINTS)
  W_ThreadGroup_system->manager->unsafe_thread = NULL;
  W_ThreadGroup_system->manager->gc_thread = NULL;
#endif
  
  /*
  ** W_ThreadGroup_system->manager is also known as W_Thread_system,
  ** the debiteur of last resort ...
  */

  W_Thread_system = W_ThreadGroup_system->manager;
  
  woempa(7, "Creating %t\n", W_Thread_system);
  x_thread_create(W_ThreadGroup_system->manager->kthread, 
    startWonka, NULL, W_ThreadGroup_system->manager->kstack, 
    W_ThreadGroup_system->manager->ksize,
    W_ThreadGroup_system->manager->kpriority, TF_START);
  W_ThreadGroup_system->manager->kthread->report = group_report;
#ifdef DEBUG
  PutString((char*)"Launched system thread, returning control to kernel\r\n");
#endif
  woempa(7, "Created %t\n", W_Thread_system);

  return start_heap;

}

#define INIT_CLASS "wonka.vm.Init"

extern int woempa_bytecodecount;

static w_string string_sysThread;
static w_string string_sysThread;
static w_string string_sysThreadGroup;

static w_string string_sysThreadGroup;

void startInitialThreads(void* data) {

  w_method method = NULL;
  w_instance arglist;
  w_word dims;
  w_int    i;
  JNIEnv  *env = w_thread2JNIEnv(W_Thread_sysInit);
  jclass   class_Init;
  w_instance String;
#ifdef GC_SAFE_POINTS
  w_boolean unsafe;
#endif

  woempa(9, "********* %d instances in use after initial loading. ************\n", instance_use);

  classMustBeInitialized(clazzString);
  String = getStringInstance(string_sysThread);
  setReferenceField(I_Thread_sysInit, String, F_Thread_name);
  removeLocalReference(W_Thread_system, String);
  String = getStringInstance(string_sysThreadGroup);
  setReferenceField(I_ThreadGroup_system, String, F_ThreadGroup_name);
  removeLocalReference(W_Thread_system, String);
  classMustBeInitialized(clazzString);
  String = getStringInstance(string_sysThread);
  setReferenceField(I_Thread_sysInit, String, F_Thread_name);
  removeLocalReference(W_Thread_system, String);
  String = getStringInstance(string_sysThreadGroup);
  setReferenceField(I_ThreadGroup_system, String, F_ThreadGroup_name);
  removeLocalReference(W_Thread_system, String);
  setBooleanField(I_Thread_sysInit, F_Thread_started, WONKA_TRUE);

/*
** First we gather the command line arguments (if any: in the embedded case
** there cannot be any), and use the first as the name of an initial class
** to run and the rest as arguments to be passed to that class's <init>(String[])
** method.  The name goes in arglist[0] and the arguments go in arglists[1...].
*/

  if (command_line_argument_count > 0) {
    dims = command_line_argument_count;
    woempa(1,"Allocating array of %d String[s]\n",dims);
    arglist = allocArrayInstance(W_Thread_sysInit, clazzArrayOf_String, 1, &dims);
    removeLocalReference(W_Thread_sysInit, arglist);

    for (i = 0; i < command_line_argument_count; i++) {
      String = getStringInstance(cstring2String(W_Thread_sysInit, command_line_arguments[i], strlen(command_line_arguments[i])));
      setArrayReferenceField(arglist, String, i);
      removeLocalReference(W_Thread_sysInit, String);
      woempa(9,"args[%d] = \"%w\" bytecodecount = %d\n",i,String2string(instance2Array_instance(arglist)[i]), woempa_bytecodecount);
    }
  }
  else {
    dims = 0;
    woempa(1,"Not allocating an array of String[s]\n");
    arglist = NULL;
  }

/*
** Now we call the main(String[]) method of class Init with arglist as its 
** parameter.
*/
  woempa(1, "Invoking main([Ljava.lang.String;) of %s ...\n", INIT_CLASS);

  class_Init = (*env)->FindClass(env, INIT_CLASS);
  if (!class_Init) {
    wabort(ABORT_WONKA,"Uh oh: didn't find class " INIT_CLASS ".  Game over\n");
  }

  method = (*env)->GetStaticMethodID(env, class_Init, "main", "([Ljava/lang/String;)V");
  if (method==NULL) {
    wabort(ABORT_WONKA,"Uh oh: class " INIT_CLASS " doesn't have a method 'main(String[])V'.  Game over.\n");
  }

  (*env)->CallStaticVoidMethod(env, class_Init, method, arglist);
  if ((*env)->ExceptionCheck(env)) {
    (*env)->ExceptionDescribe(env);
  }

  W_Thread_sysInit->top = & W_Thread_sysInit->Frame;
  removeThreadFromGroup(W_Thread_sysInit, W_Thread_sysInit->group);
  W_Thread_sysInit->state = wt_dying;
#ifdef GC_SAFE_POINTS
  unsafe = enterUnsafeRegion(W_Thread_sysInit);
#endif
  ht_erase(thread_hashtable,(w_word)W_Thread_sysInit->kthread);
#ifdef GC_SAFE_POINTS
  if (!unsafe) {
    enterSafeRegion(W_Thread_sysInit);
  }
#endif
  W_Thread_sysInit->state = wt_dead;
}

void startKernel() {

  JNIEnv  *env = w_thread2JNIEnv(W_ThreadGroup_system->manager);

  nondaemon_thread_count = 0;

  thread_hashtable = ht_create(W_Realm_system, (char*)"hashtable:threads", THREAD_HASHTABLE_SIZE, NULL, NULL, 0, 0);
  woempa(7, "Created thread_hashtable at %p\n",thread_hashtable);
  threadgroup_hashtable = ht_create(W_Realm_system, (char*)"hashtable:groups", THREADGROUP_HASHTABLE_SIZE, NULL, NULL, 0, 0);
  woempa(7, "Created threadgroup_hashtable at %p\n",threadgroup_hashtable);

  I_ThreadGroup_system = allocInstance(W_ThreadGroup_system->manager, clazzRealm);
  removeLocalReference(W_ThreadGroup_system->manager, I_ThreadGroup_system);
  woempa(1,"created I_ThreadGroup_system at %p\n",I_ThreadGroup_system);
  W_ThreadGroup_system->ThreadGroup = I_ThreadGroup_system;
  setWotsitField(I_ThreadGroup_system, F_ThreadGroup_wotsit, W_ThreadGroup_system);
  string_sysThreadGroup = cstring2String(W_ThreadGroup_system->manager, "SystemThreadGroup", 17);
  W_ThreadGroup_system->name = string_sysThreadGroup;
  W_ThreadGroup_system->manager->name = string_sysThreadGroup;

  I_Thread_sysInit = allocInstance(W_ThreadGroup_system->manager, clazzThread);
  removeLocalReference(W_ThreadGroup_system->manager, I_Thread_sysInit);
  string_sysThread = cstring2String(W_ThreadGroup_system->manager, "SystemInitThread", 16);

  W_Thread_sysInit = allocClearedMem(W_ThreadGroup_system->manager, sizeof(w_Thread));
  W_Thread_sysInit->auxs = allocClearedMem_system(sizeof(x_Slot) * init_locals);

  if (!W_Thread_sysInit) {
    wabort(ABORT_WONKA, "Couldn't allocate memory for initial thread!\n");
  }

  attachFirstFrame(W_Thread_sysInit);
  W_Thread_sysInit->maxauxs = & W_Thread_sysInit->auxs[init_locals];
  W_Thread_sysInit->natenv = &w_JNINativeInterface;
  W_Thread_sysInit->label = (char*)"thread:sysInit";
  W_Thread_sysInit->flags = FRAME_BOTTOM;
  W_Thread_sysInit->name = string_sysThread;
  W_Thread_sysInit->state = wt_unstarted;
  W_Thread_sysInit->group = W_ThreadGroup_system;
  W_Thread_sysInit->isDaemon = WONKA_FALSE;
  W_Thread_sysInit->jpriority = USER_PRIORITY;
  W_Thread_sysInit->Thread = I_Thread_sysInit;

  attachFirstFrame(W_Thread_sysInit);
  W_Thread_sysInit->maxauxs = & W_Thread_sysInit->auxs[init_locals];

  class_Thread = clazz2Class(clazzThread);
  class_ThreadGroup = clazz2Class(clazzThreadGroup);
  class_Runtime = clazz2Class(clazzRuntime);
  run_method = (*env)->GetMethodID(env, class_Thread, "_run", "()V"); 
  woempa(7,"run_method is %M\n",run_method);
  registerThread_method = (*env)->GetMethodID(env, class_ThreadGroup, "registerThread", "(Ljava/lang/Thread;)V");
  woempa(7,"registerThread_method is %M\n",registerThread_method);
  deregisterThread_method = (*env)->GetMethodID(env, class_ThreadGroup, "deregisterThread", "(Ljava/lang/Thread;)V");
  woempa(7,"deregisterThread_method is %M\n",deregisterThread_method);

  addThreadToGroup(W_Thread_sysInit, W_ThreadGroup_system);

  /*
  ** TODO, check quota limit...
  */

  woempa(1, "registering os thread %p as Java threadGroup instance %p (%k)\n", W_ThreadGroup_system->manager->kthread, W_ThreadGroup_system->ThreadGroup, instance2clazz(W_ThreadGroup_system->ThreadGroup));
  ht_write(threadgroup_hashtable, (w_word)W_ThreadGroup_system->manager->kthread, (w_word)W_ThreadGroup_system);

  W_ThreadGroup_system->name = string_sysThreadGroup;
#ifndef O4P
  W_Thread_sysInit->kstack = allocMem(W_ThreadGroup_system->manager, init_stack_size);
#endif
  W_Thread_sysInit->maxframes = (init_stack_size / bytes_per_call) - margin_frames;
  W_Thread_sysInit->numframes = W_Thread_sysInit->maxframes;
  W_Thread_sysInit->ksize = init_stack_size;
#ifdef GC_SAFE_POINTS
  W_Thread_sysInit->unsafe_thread = NULL;
  W_Thread_sysInit->gc_thread = NULL;
#endif

  W_Thread_sysInit->kthread = (x_thread)allocClearedMem(W_ThreadGroup_system->manager, sizeof(x_Thread));
  W_Thread_sysInit->kthread->xref = W_Thread_sysInit;
  W_Thread_sysInit->kthread->report = unborn_thread_report;
  W_Thread_sysInit->kpriority = priority_j2k(USER_PRIORITY,0);
  W_Realm_system->ThreadGroup = I_ThreadGroup_system;
  woempa(7,"W_Realm_system @ %p super @ %p (group %p) parent @ %p quotas memory %d, threads %d\n",W_Realm_system,W_Realm_system->ThreadGroup,getWotsitField(I_ThreadGroup_system, F_ThreadGroup_wotsit),NULL,W_Realm_system->memory_quota,W_Realm_system->thread_quota);

  setWotsitField(I_Thread_sysInit, F_Thread_wotsit, W_Thread_sysInit);
  setReferenceField(I_Thread_sysInit, I_ThreadGroup_system, F_Thread_parent);
  woempa(1, "'%w' instance at %p.\n", NM(W_Thread_sysInit), I_Thread_sysInit);
  woempa(1, "W_Thread_sysInit at %p.\n", W_Thread_sysInit);
  woempa(1, "registering os thread %p as thread instance %p\n", W_Thread_sysInit->kthread, W_Thread_sysInit);
  ht_write(thread_hashtable, (w_word)W_Thread_sysInit->kthread, (w_word)W_Thread_sysInit);

  W_Thread_sysInit->kthread->xref = W_Thread_sysInit;
  x_thread_create(W_Thread_sysInit->kthread, startInitialThreads, 0, W_Thread_sysInit->kstack, W_Thread_sysInit->ksize, W_Thread_sysInit->kpriority, TF_SUSPENDED);
  W_Thread_sysInit->kthread->report = running_thread_report;

}

char * print_thread_short(char* buffer, int *remain, void *data, int w, int p, unsigned int f) {

  w_thread t = (w_thread) data;
  w_size   length;
  w_string thread_name;

  if (*remain < 1) {
    return buffer;
  }

  if (t) {
    thread_name = t->name ? t->name : string_NULL;
    length = x_snprintf(buffer, *remain, "%w", thread_name);
    *remain -= length;
  }
  else {
    length = x_snprintf(buffer, *remain, "<NULL>");
    *remain -= length;
  }

  return buffer + length;

}

char * print_thread_long(char* buffer, int *remain, void *data, int w, int p, unsigned int f) {

  w_thread t = (w_thread) data;
  w_size   length;
  w_string thread_name;
  w_method method;
  w_int pc;
  w_size locals = t->top ? t->top->auxs - t->auxs : 0;

  if (*remain < 1) {
    return buffer;
  }

  if (t) {
    thread_name = t->name ? t->name : string_NULL;

    if (! t->top || ! (method = t->top->method)) {
      length = x_snprintf(buffer, *remain, "Thread %w p%-2d, %2d locals (no stack frame available)", thread_name, t->jpriority, locals);
    }
    else {
      if (isSet(method->flags, METHOD_IS_COMPILED)) {
        length = x_snprintf(buffer, *remain, "Thread %w p%-2d, %2d locals, jitted %m", thread_name, t->jpriority, locals, method);
      }
      else if (isSet(method->flags, ACC_NATIVE)) {
        length = x_snprintf(buffer, *remain, "Thread %w p%-2d, %2d locals, native %m", thread_name, t->jpriority, locals, method);
      }
      else {
        pc = t->top->current - method->exec.code;
#ifdef STORE_METHOD_DEBUG_INFO
      length = x_snprintf(buffer, *remain, "Thread %w p%-2d, %2d locals, line %d [%05d] in %m", thread_name, t->jpriority, locals, code2line(method, t->top->current), pc, t->top->method);
#else
      length = x_snprintf(buffer, *remain, "Thread %w p%-2d, %2d locals, [%05d] in %m", thread_name, t->jpriority, locals, pc, t->top->method);
#endif
      }
    }
  }
  else {
    length = x_snprintf(buffer, *remain, "<NULL>");
    *remain -= length;
  }

  *remain -= length;

  return buffer + length;

}

char * threadDescription(w_thread t) {

  x_snprintf(thread_report_buffer, 91, "%T", t);

  return thread_report_buffer;

}

char * groupDescription(w_group g) {

  if (g->ThreadGroup && g->realm->ThreadGroup == g->ThreadGroup) {
  // Actually this is a Realm
    return realmDescription(g->realm);
  }

  x_snprintf(group_report_buffer, 63, "Group %-24w max priority %d", NM(g), g->maxPriority);

  return group_report_buffer;

}
